import org.garret.consus.*;
import java.sql.*;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Properties;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

public final class BugDB extends HttpServlet {
    static final int CATEGORY_NOT_SPECIFIED = 0;
    static final int CRASH = 1;
    static final int PROGRAM_HANGS = 2;
    static final int UI_DISPLAY = 3;
    static final int UI_BEHAVIOR = 4;
    static final int CALCULATION = 5;
    static final int ERROR_HANDLING = 6;
    static final int PERFORMANCE = 7;
    static final int LICENSING = 8;
    static final int INSTALLATION = 9;
    static final int DOCUMENTATION = 10;
    static final int ENHANCEMENT = 11;
    static final int HOW_TO_QUESTION = 12;
    static final int STATUS_NOT_SPECIFIED = 0;
    static final int STATUS_OPENED = 1;
    static final int STATUS_FIXED = 2;
    static final int STATUS_CLOSED = 3;
    static final int STATUS_PENDING_Person = 4;
    static final int STATUS_PENDING_USER = 5;
    static final int STATUS_POSTPONED = 6;
    static final int STATUS_IRREPRODUCIBLE = 7;
    static final int STATUS_WITHDRAWN = 8;
    static final int STATUS_AS_DESIGNED = 9;
    static final int PRIORITY_NOT_SPECIFIED = 0;
    static final int FIX_IMMEDIATELY = 1;
    static final int FIX_BEFORE_NEXT_BUILD_RELEASE = 2;
    static final int FIX_BEFORE_NEXT_MINOR_RELEASE = 3;
    static final int FIX_BEFORE_NEXT_MAJOR_RELEASE = 4;
    static final int FIX_IF_POSSIBLE = 5;
    static final int OPTIONAL = 6;
    static final int SEVERITY_NOT_SPECIFIED = 0;
    static final int FATAL = 1;
    static final int SERIOUS = 2;
    static final int MINOR = 3;
    static final int LAST_SEVERITY = 4;
    static final String eCATEGORY_STRING[] =
        {
	"",
	"CRASH",
	"PROGRAM_HANGS",
	"UI_DISPLAY",
	"UI_BEHAVIOR",
	"CALCULATION",
	"ERROR_HANDLING",
	"PERFORMANCE",
	"LICENSING",
	"INSTALLATION",
	"DOCUMENTATION",
	"ENHANCEMENT",
	"HOW_TO_QUESTION",
	null
    };
    static final String eSTATUS_STRING[] = {
	"",
	"OPENED",
	"FIXED",
	"CLOSED",
	"PENDING_ENGINEER",
	"PENDING_USER",
	"POSTPONED",
	"IRREPRODUCIBLE",
	"WITHDRAWN",
	"AS_DESIGNED",
	null
    };
    static final String eSEVERITY_STRING[] = {
	"",
	"FATAL",
	"SERIOUS",
	"MINOR",
	null
    };
    static final String eFIXING_PRIORITY_STRING[] = {
	"",
	"FIX_IMMEDIATELY",
	"FIX_BEFORE_NEXT_BUILD_RELEASE",
	"FIX_BEFORE_NEXT_MINOR_RELEASE",
	"FIX_BEFORE_NEXT_MAJOR_RELEASE",
	"FIX_IF_POSSIBLE",
	"OPTIONAL",
	null
    };
    Connection db;
    PersistentObjectStorage storage;
    Class self;
    Class methodParamTypes[];
    int objectCacheSizeLimit = 100000;

    static String encodeURL(String str) {
        StringBuffer buf = new StringBuffer();
        for (int i = 0, n = str.length(); i < n; i++) {
            char ch = str.charAt(i);
            if (ch == ' ') {
                buf.append("+");
            } else if (!((ch >= '0' && ch <= '9') || (ch >= 'a' || ch <= 'z') || (ch >= 'A' || ch <= 'Z'))) {
                buf.append("%");
                buf.append(((ch >> 4) & 0xF) >= 10 ? ((ch >> 4) & 0xF) + 'A' - 10 : ((ch >> 4) & 0xF) + '0');
                buf.append((ch & 0xF) >= 10 ? (ch & 0xF) + 'A' - 10 : (ch & 0xF) + '0');
            } else {
                buf.append(ch);
            }
        }
        return buf.toString();
    }

    static String encodeHTML(String str) {
        if (str == null) {
            return "";
        }
        StringBuffer buf = new StringBuffer();
        for (int i = 0, n = str.length(); i < n; i++) {
            char ch = str.charAt(i);
            switch (ch) {
                case '<':
                    buf.append("&lt;");
                    break;
                case '>':
                    buf.append("&gt;");
                    break;
                case '&':
                    buf.append("&amp;");
                    break;
                case '"':
                    buf.append("&quot;");
                    break;
                default:
                    buf.append(ch);
            }
        }
        return buf.toString();
    }

    int print(HttpConnection con, ResultSet cursor) throws Exception {
        int n = 0;
        while (cursor.next()) {
            ((Printable)storage.fetchObject(cursor)).print(con);
            n += 1;
        }
        return n;
    }

    int print(HttpConnection con, TableIterator iterator) throws Exception {
        int n = 0;
        while (iterator.next()) {
            ((Printable)iterator.get()).print(con);
            n += 1;
        }
        return n;
    }

    void print(HttpConnection con, Printable arr[]) throws IOException {
        for (int i = 0; i < arr.length; i++) {
            arr[i].print(con);
        }
    }

    static final String HTML_HEAD = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\"><HTML><HEAD>";
    static final String BODY = "<BODY BGCOLOR=\"#c0c0c0\">";
    static final String EMPTY_LIST = "<OPTION>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</OPTION></SELECT>";

    void mainMenuReference(HttpConnection con) throws IOException {
        String myself = con.get("myself");
        if (myself != null) {
            con.print("<P><HR><CENTER><A HREF=\"" + con.getURI() + "?page=userForm&myself=" + myself + "&person=" + myself +
                "\">Back to main menu</A></CENTER>");
        }
        con.print("</BODY></HTML>");
    }

    void error(HttpConnection con, String msg) throws IOException {
        con.print(HTML_HEAD + "<TITLE>BUGDB error</TITLE></HEAD><BODY><H1><FONT COLOR=\"#FF0000\">" +
            encodeHTML(msg) + "</FONT></H1>");
	mainMenuReference(con);
    }

    void message(HttpConnection con, String msg) throws IOException {
        con.print(HTML_HEAD + "<TITLE>BUGDB message</TITLE></HEAD><BODY><H1><FONT COLOR=\"#004000\">" +
            encodeHTML(msg) + "</FONT></H1>");
	mainMenuReference(con);
    }

    public void addUserForm(HttpConnection con) throws Exception {
        con.print(HTML_HEAD + "<TITLE>Enter new user</TITLE></HEAD>" + BODY + "<H2>Add user</H2>" +
            "<FORM METHOD=POST ACTION=\"" + con.getURI() + "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"addUser\">" +
            "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + con.get("myself") +
            "\"><TABLE><TR><TH ALIGN=LEFT>User name:</TH>" + "<TD><INPUT TYPE=text NAME=\"name\" SIZE=30></TD></TR>" +
            "<TR><TH ALIGN=LEFT>E-mail:</TH>" + "<TD><INPUT TYPE=text NAME=\"email\" SIZE=30></TD></TR></TABLE><P>" +
            "<INPUT TYPE=submit VALUE=\"Add\">&nbsp;" + "<INPUT TYPE=reset VALUE=\"Reset\"></FORM>");
        mainMenuReference(con);
    }

    public void addEngineerForm(HttpConnection con) throws Exception {
        con.print(HTML_HEAD + "<TITLE>Enter new engineer</TITLE></HEAD>" + BODY + "<H2>Add engineer</H2>" +
            "<FORM METHOD=POST ACTION=\"" + con.getURI() + "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"addEngineer\">" +
            "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + con.get("myself") +
            "\"><TABLE><TR><TH ALIGN=LEFT>Engineer name:</TH>" + "<TD><INPUT TYPE=text NAME=\"name\" SIZE=30></TD></TR>" +
            "<TR><TH ALIGN=LEFT>E-mail:</TH>" + "<TD><INPUT TYPE=text NAME=\"email\" SIZE=30></TD></TR></TABLE><P>" +
            "<INPUT TYPE=submit VALUE=\"Add\">&nbsp;" + "<INPUT TYPE=reset VALUE=\"Reset\"></FORM>");
        mainMenuReference(con);
    }

    public void addSoftwareForm(HttpConnection con) throws Exception {
        con.print(HTML_HEAD + "<TITLE>Enter new software product</TITLE></HEAD>" + BODY + "<H2>Add software product</H2>" +
            "<FORM METHOD=POST ACTION=\"" + con.getURI() + "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"addSoftware\">" +
            "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + con.get("myself") +
            "\"><TABLE><TR><TH ALIGN=LEFT>Software name:</TH>" + "<TD><INPUT TYPE=text NAME=\"name\" SIZE=40></TD></TR>" +
            "<TR><TH ALIGN=LEFT>Version number:</TH>" + "<TD><INPUT TYPE=text NAME=\"version\" SIZE=8></TD></TR>" +
            "<TR><TH ALIGN=LEFT>Version label:</TH>" + "<TD><INPUT TYPE=text NAME=\"label\" SIZE=20></TD></TR>" +
            "<TR><TH ALIGN=LEFT>Version comment:</TH>" + "<TD><INPUT TYPE=text NAME=\"comment\" SIZE=40></TD></TR>" +
            "</TABLE><P><INPUT TYPE=submit VALUE=\"Add\">&nbsp;" + "<INPUT TYPE=reset VALUE=\"Reset\"></FORM>");
        mainMenuReference(con);
    }

    public void selectSoftwareForm(HttpConnection con) throws Exception {
        con.print(HTML_HEAD + "<TITLE>Select software product</TITLE></HEAD>" + BODY + "<H2>Select software product</H2>" +
            "<FORM METHOD=POST ACTION=\"" + con.getURI() + "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"softwareForm\">" +
            "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + con.get("myself") + "\"><SELECT SIZE=15 NAME=\"software\">");
        TableIterator iterator = storage.getTableIterator("Software");
        if (print(con, iterator) != 0) {
            con.print("</SELECT><BR><INPUT TYPE=submit NAME=\"action\" VALUE=\"Select\">");
        } else {
            con.print(EMPTY_LIST);
        }
        iterator.close();
        con.print("</FORM>");
        mainMenuReference(con);
    }

    public void removeSoftwareForm(HttpConnection con) throws Exception {
        con.print(HTML_HEAD + "<TITLE>Remove software product</TITLE></HEAD>" + BODY + "<H2>Remove software product</H2>" +
            "<FORM METHOD=POST ACTION=\"" + con.getURI() + "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"removeSoftware\">" +
            "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + con.get("myself") + "\"><SELECT SIZE=15 NAME=\"software\">");
        TableIterator iterator = storage.getTableIterator("Software");
        if (print(con, iterator) != 0) {
            con.print("</SELECT><BR><INPUT TYPE=\"submit\" VALUE=\"Remove\">");
        } else {
            con.print(EMPTY_LIST);
        }
        iterator.close();
        con.print("</FORM>");
        mainMenuReference(con);
    }

    public void selectPersonForm(HttpConnection con) throws Exception {
        con.print(HTML_HEAD + "<TITLE>Select a person</TITLE></HEAD>" + BODY + "<H2>Select a person</H2>" +
            "<FORM METHOD=POST ACTION=\"" + con.getURI() + "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"userForm\">" +
            "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + con.get("myself") + "\"><SELECT SIZE=20 NAME=\"person\">");
        TableIterator iterator = storage.getTableIterator("Person");
        if (print(con, iterator) != 0) {
            con.print("</SELECT><BR><INPUT TYPE=submit VALUE=\"Select\">");
        } else {
            con.print(EMPTY_LIST);
        }
        iterator.close();
        con.print("</FORM>");
        mainMenuReference(con);
    }

    public void removePersonForm(HttpConnection con) throws Exception {
        con.print(HTML_HEAD + "<TITLE>Remove a person</TITLE></HEAD>" + BODY + "<H2>Remove a person</H2>" +
            "<FORM METHOD=POST ACTION=\"" + con.getURI() + "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"removePerson\">" +
            "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + con.get("myself") + "\"><SELECT SIZE=20 NAME=\"person\">");
        TableIterator iterator = storage.getTableIterator("Person");
        if (print(con, iterator) != 0) {
            con.print("</SELECT><BR><INPUT TYPE=submit VALUE=\"Remove\">");
        } else {
            con.print(EMPTY_LIST);
        }
        iterator.close();
        con.print("</FORM>");
        mainMenuReference(con);
    }

    public void selectBugForm(HttpConnection con) throws Exception {
        con.print(HTML_HEAD + "<TITLE>Select a bug</TITLE></HEAD>" + BODY + "<H2>Select a bug</H2>" +
            "<FORM METHOD=POST ACTION=\"" + con.getURI() + "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"bugForm\">" +
            "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + con.get("myself") + "\"><SELECT SIZE=15 NAME=\"bug\">");
        TableIterator iterator = storage.getTableIterator("Bug");
        if (print(con, iterator) != 0) {
            con.print("</SELECT><BR><INPUT TYPE=submit NAME=\"action\" VALUE=\"Select\">");
        } else {
            con.print(EMPTY_LIST);
        }
        iterator.close();
        con.print("</FORM>");
        mainMenuReference(con);
    }

    public void removeBugForm(HttpConnection con) throws Exception {
        con.print(HTML_HEAD + "<TITLE>Remove a bug</TITLE></HEAD>" + BODY + "<H2>Remove a bug</H2>" +
            "<FORM METHOD=POST ACTION=\"" + con.getURI() + "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"removeBug\">" +
            "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + con.get("myself") + "\"><SELECT SIZE=15 NAME=\"bug\">");
        TableIterator iterator = storage.getTableIterator("Bug");
        if (print(con, iterator) != 0) {
            con.print("</SELECT><BR><INPUT TYPE=submit VALUE=\"Remove\">");
        } else {
            con.print(EMPTY_LIST);
        }
        iterator.close();
        con.print("</FORM>");
        mainMenuReference(con);
    }

    public void changePasswordForm(HttpConnection con) throws Exception {
        con.print(HTML_HEAD + "<TITLE>Change password</TITLE></HEAD>" + BODY + "<H2>Change password</H2>" +
            "<FORM METHOD=POST ACTION=\"" + con.getURI() + "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"changePassword\">" +
            "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + con.get("myself") +
            "\"><INPUT TYPE=hidden NAME=\"person\" VALUE=\"" + con.get("person") +
            "\"><TABLE><TR><TH ALIGN=LEFT>New password:</TH>" + "<TD><INPUT TYPE=password NAME=\"password\" SIZE=20</TD></TR>" +
            "<TR><TH ALIGN=LEFT>Re-type password:</TH>" + "<TD><INPUT TYPE=password NAME=\"password2\" SIZE=20</TD></TR>" +
            "</TABLE><P><INPUT TYPE=submit VALUE=\"Change\">&nbsp;" + "<INPUT TYPE=reset VALUE=\"Reset\"></FORM>");
        mainMenuReference(con);
    }

    public void addUser(HttpConnection con) throws Exception {
        Person person = new Person();
        person.sName = con.get("name");
        TableIterator iterator = storage.getTableIterator("Person");
        if (iterator.selectByKey("sName", person.sName) != null) {
            error(con, "Person already exists");
            return;
        }
        iterator.close();
        person.sEmailAddress = con.get("email");
        person.sPassword = "";
        person.status = Person.isUser;
        person.setUsedSoftware = new Software[0];
        person.setProjects = new Software[0];
        person.setReportedBugs = new Bug[0];
        person.setAssignedBugs = new Bug[0];
        person.store(storage);
        con.addPair("person", Long.toString(person.getOid(), 10));
        userForm(con);
    }

    public void addEngineer(HttpConnection con) throws Exception {
        Person person = new Person();
        person.sName = con.get("name");
        TableIterator iterator = storage.getTableIterator("Person");
        if (iterator.selectByKey("sName", person.sName) != null) {
            error(con, "Person already exists");
            return;
        }
        iterator.close();
        person.sEmailAddress = con.get("email");
        person.sPassword = "";
        person.status = Person.isEngineer;
        person.setUsedSoftware = new Software[0];
        person.setProjects = new Software[0];
        person.setReportedBugs = new Bug[0];
        person.setAssignedBugs = new Bug[0];
        person.store(storage);
        con.addPair("person", Long.toString(person.getOid(), 10));
        userForm(con);
    }

    public void removePerson(HttpConnection con) throws Exception {
        int i;
        String personOID = con.get("person");
        if (personOID == null) {
            error(con, "No person was selected");
            return;
        }
        storage.lock();
        Person person = (Person)storage.getObjectByOid(Long.parseLong(personOID, 10));
        if (person.setReportedBugs.length > 0) {
            error(con, "It is not possible to delete person who is author of existed bug reports");
        } else {
            for (i = person.setUsedSoftware.length; --i >= 0; ) {
                person.setUsedSoftware[i].setUsers = (Person[]) AnyArray.remove(person.setUsedSoftware[i].setUsers, person);
                person.setUsedSoftware[i].store();
            }
            for (i = person.setProjects.length; --i >= 0; ) {
                person.setProjects[i].setEngineers = (Person[]) AnyArray.remove(person.setProjects[i].setEngineers, person);
                person.setProjects[i].store();
            }
            for (i = person.setAssignedBugs.length; --i >= 0; ) {
                person.setAssignedBugs[i].pAssignedTo = null;
                person.setAssignedBugs[i].store();
            }
            person.remove();
            message(con, "Person is removed");
        }
    }

    public void addSoftware(HttpConnection con) throws Exception {
        Software software = new Software();
        Version version = new Version();
        software.sName = con.get("name");
        TableIterator iterator = storage.getTableIterator("Software");
        if (iterator.selectByKey("sName", software.sName) != null) {
            error(con, "Such software product already exists");
            return;
        }
        iterator.close();
        String versionStr = con.get("version");
        try {
            int p = versionStr.indexOf('.');
            version.majorVersionNumber = Integer.parseInt(versionStr.substring(0, p), 10);
            version.minorVersionNumber = Integer.parseInt(versionStr.substring(p + 1), 10);
        } catch (Exception x) {
            error(con, "Bad version number (MAJOR.MINOR expected)");
            return;
        }
        software.setEngineers = new Person[0];
        software.setUsers = new Person[0];
        software.setBugs = new Bug[0];
        version.sComment = con.get("comment");
        version.sLabel = con.get("label");
        version.released = new Date(System.currentTimeMillis());
        version.setBugs = new Bug[0];
        software.pVersions = version;
        software.store(storage);
        con.addPair("action", "Select");
        con.addPair("software", Long.toString(software.getOid(), 10));
        softwareForm(con);
    }

    public void removeSoftware(HttpConnection con) throws Exception {
        String softOID = con.get("software");
        if (softOID == null) {
            error(con, "No software was seleceted");
            return;
        }
        storage.lock();
        Software soft = (Software)storage.getObjectByOid(Long.parseLong(softOID, 10));
        if (soft.setBugs.length != 0) {
            error(con, "Can not remove software with non-empty reported bug list");
            return;
        }
        if (soft.setUsers.length != 0) {
            error(con, "Can not remove software with non-empty user list");
            return;
        }
        if (soft.setEngineers.length != 0) {
            error(con, "Can not remove software with non-empty engineers list");
            return;
        }
        soft.remove();
        message(con, "Software product is removed");
    }

    public void removeBug(HttpConnection con) throws Exception {
        String bugOID = con.get("bug");
        if (bugOID == null) {
            error(con, "No bug was selected");
        } else {
            storage.lock();
            Bug bug = (Bug)storage.getObjectByOid(Long.parseLong(bugOID, 10));
            if (bug.pReportHistory != null || bug.pWorkArounds != null) {
                error(con, "Can not remove bug with existed reports");
                return;
            }
            for (int i = bug.setSimilarBugs.length; --i >= 0; ) {
                bug.setSimilarBugs[i].setSimilarBugs = (Bug[]) AnyArray.remove(bug.setSimilarBugs[i].setSimilarBugs, bug);
                bug.setSimilarBugs[i].store();
            }
            bug.pReportedBy.setReportedBugs = (Bug[]) AnyArray.remove(bug.pReportedBy.setReportedBugs, bug);
            bug.pReportedBy.store();
            if (bug.pAssignedTo != null) {
                bug.pAssignedTo.setAssignedBugs = (Bug[]) AnyArray.remove(bug.pAssignedTo.setAssignedBugs, bug);
                bug.pAssignedTo.store();
            }
            bug.pSoftware.setBugs = (Bug[]) AnyArray.remove(bug.pSoftware.setBugs, bug);
            bug.pSoftware.store();
            bug.pVersion.setBugs = (Bug[]) AnyArray.remove(bug.pVersion.setBugs, bug);
            bug.pVersion.store();
            bug.remove();
            message(con, "Bug is removed");
        }
    }

    public void changePassword(HttpConnection con) throws Exception {
        String password = con.get("password");
        String password2 = con.get("password2");
        if (!password.equals(password2)) {
            error(con, "Passwords are not equal");
        } else {
            storage.lock();
            Person person = (Person)storage.getObjectByOid(Long.parseLong(con.get("person"), 10));
            person.sPassword = password;
            person.store();
            message(con, "Password changed");
        }
    }

    public void updatePerson(HttpConnection con) throws Exception {
        storage.lock();
        Person person = (Person)storage.getObjectByOid(Long.parseLong(con.get("person"), 10));
        String newName = con.get("newname");
        String eMail = con.get("email");
        if (eMail != null) {
            person.sEmailAddress = eMail;
        }
        if (newName != null) {
            person.sName = newName;
        }
        person.store();
        userForm(con);
    }

    public void login(HttpConnection con) throws Exception {
        String name = con.get("name");
        TableIterator iterator = storage.getTableIterator("Person");
        Person person = (Person)iterator.selectByKey("sName", name);
        if (person == null) {
            iterator.close();
            error(con, "No such person");
            return;
        }
        iterator.close();
        if (!person.checkPassword(con.get("password"))) {
            error(con, "Incorrect password");
            return;
        }
        String oid = Long.toString(person.getOid(), 10);
        con.addPair("myself", oid);
        con.addPair("person", oid);
        userForm(con);
    }

    public void bugQueryForm(HttpConnection con) throws Exception {
        int i;
        con.print(HTML_HEAD + "<TITLE>Query to locate bug</TITLE></HEAD>" + BODY +
            "<H2>Bug query</H2><FORM METHOD=POST ACTION=\"" + con.getURI() +
            "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"bugQuery\">" + "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" +
            con.get("myself") + "\"><TABLE><TR><TH ALIGN=LEFT>Description substring:</TH>" +
            "<TD><INPUT TYPE=text NAME=\"summary\" SIZE=30></TD></TR>" + "<TR><TH ALIGN=LEFT>Category:</TH>" +
            "<TD><SELECT NAME=\"category\" SIZE=1>" + "<OPTION VALUE=0 SELECTED></OPTION>");
        for (i = 1; eCATEGORY_STRING[i] != null; i++) {
            con.print("<OPTION VALUE=" + i + ">" + eCATEGORY_STRING[i] + "</OPTION>");
        }
        con.print("</SELECT></TD></TR>" + "<TR><TH ALIGN=LEFT>Severity:</TH>" + "<TD><SELECT NAME=\"severity\" SIZE=1>" +
            "<OPTION VALUE=0 SELECTED></OPTION>");
        for (i = 1; eSEVERITY_STRING[i] != null; i++) {
            con.print("<OPTION VALUE=" + i + ">" + eSEVERITY_STRING[i] + "</OPTION>");
        }
        con.print("</SELECT></TD></TR>" + "<TR><TH ALIGN=LEFT>Fixing priority:</TH>" +
            "<TD><SELECT NAME=\"priority\" SIZE=1>" + "<OPTION VALUE=0 SELECTED></OPTION>");
        for (i = 1; eFIXING_PRIORITY_STRING[i] != null; i++) {
            con.print("<OPTION VALUE=" + i + ">" + eFIXING_PRIORITY_STRING[i] + "</OPTION>");
        }
        con.print("</SELECT></TD></TR>" + "<TR><TH ALIGN=LEFT>Platform:</TH>" +
            "<TD><INPUT TYPE=text NAME=\"platform\"></TD></TR>" + "<TR><TH ALIGN=LEFT>OS</TH>" +
            "<TD><INPUT TYPE=text NAME=\"os\"></TD></TR>" + "<TR><TH ALIGN=LEFT>Software:</TH>" +
            "<TD><INPUT TYPE=text NAME=\"software\"></TD></TR>" + "<TR><TH ALIGN=LEFT>Assigned to:</TH>" +
            "<TD><INPUT TYPE=text NAME=\"engineer\"></TD></TR>" + "<TR><TH ALIGN=LEFT>Reported by:</TH>" +
            "<TD><INPUT TYPE=text NAME=\"user\"></TD></TR>" + "<TR><TH ALIGN=LEFT>Major version number:</TH>" +
            "<TD>from <INPUT TYPE=text NAME=\"minmajor\" SIZE=4>" + " to <INPUT TYPE=text NAME=\"maxmajor\" SIZE=4></TD></TR>" +
            "<TR><TH ALIGN=LEFT>Minor version number:</TH>" + "<TD>from <INPUT TYPE=text NAME=\"minminor\" SIZE=4>" +
            " to <INPUT TYPE=text NAME=\"maxminor\" SIZE=4></TD></TR></TABLE><P>" +
            "<INPUT TYPE=submit VALUE=\"Search\">&nbsp;" + "<INPUT TYPE=reset VALUE=\"Reset\">" + "</FORM></BODY></HTML>");
    }

    public void bugQuery(HttpConnection con) throws Exception {
        Query query = new Query(db);
        String p;
        try {
            query.add("select * from Bug");
            p = con.get("software");
            if (p.length() != 0) {
                query.and("pSoftware.sName like").param(p);
            }
            int category = Integer.parseInt(con.get("category"), 10);
            if (category != 0) {
                query.and("eCategory=").param(category);
            }
            int severity = Integer.parseInt(con.get("severity"), 10);
            if (severity != 0) {
                query.and("eSeverity=").param(severity);
            }
            int priority = Integer.parseInt(con.get("priority"), 10);
            if (priority != 0) {
                query.and("eFixingPriority=").param(priority);
            }
            p = con.get("os");
            if (p.length() != 0) {
                query.and("sOperatingSystem like").param(p);
            }
            p = con.get("platform");
            if (p.length() != 0) {
                query.and("sHardwarePlatform like").param(p);
            }
            p = con.get("engineer");
            if (p.length() != 0) {
                query.and("pAssignedTo is not null and pAssignedTo.sName like").param(p);
            }
            p = con.get("user");
            if (p.length() != 0) {
                query.and("pReportedBy.sName like").param(p);
            }
            p = con.get("summary");
            if (p.length() != 0) {
                query.and("sOneLineSummary like").param(p);
            }
            p = con.get("minmajor");
            int minMajorVersionNumber = (p.length() == 0) ? 0 : Integer.parseInt(p, 10);
            p = con.get("maxmajor");
            int maxMajorVersionNumber = (p.length() == 0) ? Integer.MAX_VALUE : Integer.parseInt(p, 10);
            p = con.get("minminor");
            int minMinorVersionNumber = (p.length() == 0) ? 0 : Integer.parseInt(p, 10);
            p = con.get("maxminor");
            int maxMinorVersionNumber = (p.length() == 0) ? Integer.MAX_VALUE : Integer.parseInt(p, 10);
            if (minMajorVersionNumber != 0) {
                if (maxMajorVersionNumber != Integer.MAX_VALUE) {
                    query.and("pVersion.majorVersionNumber between").param(minMajorVersionNumber)
                        .add("and").param(maxMajorVersionNumber);
                } else {
                    query.and("pVersion.majorVersionNumber>=").param(minMajorVersionNumber);
                }
            } else if (maxMajorVersionNumber != Integer.MAX_VALUE) {
                query.and("pVersion.majorVersionNumber<=").param(maxMajorVersionNumber);
            }
            if (minMinorVersionNumber != 0) {
                if (maxMinorVersionNumber != Integer.MAX_VALUE) {
                    query.and("pVersion.minorVersionNumber between").param(minMinorVersionNumber)
                        .add("and").param(maxMinorVersionNumber);
                } else {
                    query.and("pVersion.minorVersionNumber>=").param(minMinorVersionNumber);
                }
            } else if (maxMinorVersionNumber != Integer.MAX_VALUE) {
                query.and("pVersion.minorVersionNumber<=").param(maxMinorVersionNumber);
            }
        } catch (NumberFormatException x) {
            error(con, "Invalid version number");
            return;
        }
        con.print(HTML_HEAD + "<TITLE>List of selected bugs</TITLE></HEAD>" + BODY + "<H2>Selected bugs</H2>" +
            "<FORM METHOD=POST ACTION=\"" + con.getURI() + "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"bugForm\">" +
            "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + con.get("myself") + "\"><SELECT NAME=\"bug\" SIZE=20>");
        ResultSet bugs = query.executeQuery();
        if (print(con, bugs) != 0) {
            con.print("</SELECT><BR><INPUT TYPE=submit NAME=\"action\" VALUE=\"Select\">");
        } else {
            con.print(EMPTY_LIST);
        }
        bugs.close();
        con.print("</FORM>");
        mainMenuReference(con);
    }

    public void userForm(HttpConnection con) throws Exception {
        String personOID = con.get("person");
        if (personOID == null) {
            error(con, "No person was selected");
            return;
        }
        String myself = con.get("myself");
        Person me = (Person)storage.getObjectByOid(Long.parseLong(myself, 10));
        int selfStatus = me.status;
        Person person = (Person)storage.getObjectByOid(Long.parseLong(personOID, 10));
        if (person.status == Person.isAdministrator) {
            con.print(HTML_HEAD + "<TITLE>BUGDB Administrator</TITLE>" + BODY +
                "<H2>Administrator menu</H2><FONT SIZE=\"+1\">" + "<UL><LI><A HREF=\"" + con.getURI() +
                "?page=addUserForm&myself=" + myself + "\">Add user</A>" + "<LI><A HREF=\"" + con.getURI() +
                "?page=addEngineerForm&myself=" + myself + "\">Add engineer</A>" + "<LI><A HREF=\"" + con.getURI() +
                "?page=selectPersonForm&myself=" + myself + "\">Select person</A>" + "<LI><A HREF=\"" + con.getURI() +
                "?page=removePersonForm&myself=" + myself + "\">Remove person</A></UL>" + "<UL><LI><A HREF=\"" + con.getURI() +
                "?page=addSoftwareForm&myself=" + myself + "\">Add software product</A>" + "<LI><A HREF=\"" + con.getURI() +
                "?page=selectSoftwareForm&myself=" + myself + "\">Select software product</A>" + "<LI><A HREF=\"" +
                con.getURI() + "?page=removeSoftwareForm&myself=" + myself + "\">Remove software product</A></UL>" +
                "<UL><LI><A HREF=\"" + con.getURI() + "?page=selectBugForm&myself=" + myself + "\">Select bug</A>" +
                "<LI><A HREF=\"" + con.getURI() + "?page=removeBugForm&myself=" + myself + "\">Remove bug</A></UL>");
            if (selfStatus == Person.isAdministrator) {
                con.print("<UL><LI><A HREF=\"" + con.getURI() + "?page=changePasswordForm" + "&myself=" + myself + "&person=" +
                    personOID + "\">Change password</A></UL>");
            }
            con.print("</FONT></BODY></HTML>");
            return;
        }
        con.print(HTML_HEAD + "<TITLE>" + encodeHTML(person.sName) + "</TITLE></HEAD>" + BODY + "<H2>" +
            encodeHTML(person.sName) + "</H2><FONT SIZE=\"+1\">" + "<UL><LI><A HREF=\"" + con.getURI() +
            "?page=createBugReportForm&myself=" + myself + "\">Create bug report</A>");
        if (person.sEmailAddress.length() != 0) {
            con.print("<LI><A HREF=\"mailto:" + person.sEmailAddress + "\">Send e-mail</A>" + "<LI><A HREF=\"" + con.getURI() +
                "?page=bugQueryForm&myself=" + myself + "\">Find a bug</A>");
        }
        if (person == me || selfStatus == Person.isAdministrator) {
            con.print("<LI><A HREF=\"" + con.getURI() + "?page=changePasswordForm&myself=" + myself + "&person=" + personOID +
                "\">Change password</A>");
        }
        con.print("</UL></FONT><P><TABLE><TR><TH ALIGN=LEFT>Person name:</TH>" + "<TD><FORM METHOD=POST ACTION=\"" +
            con.getURI() + "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"" +
            "updatePerson\"><INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + myself +
            "\"><INPUT TYPE=hidden NAME=\"person\" VALUE=\"" + personOID +
            "\"><INPUT TYPE=text NAME=\"newname\" SIZE=30 VALUE=\"" + encodeHTML(person.sName) +
            "\"><INPUT TYPE=submit VALUE=\"Change\"></FORM></TD></TR>" + "<TR><TH ALIGN=LEFT>E-Mail:</TH>" +
            "<TD><FORM METHOD=POST ACTION=\"" + con.getURI() + "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"" +
            "updatePerson\"><INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + myself +
            "\"><INPUT TYPE=hidden NAME=\"person\" VALUE=\"" + personOID +
            "\"><INPUT TYPE=text NAME=\"email\" SIZE=30 VALUE=\"" + person.sEmailAddress +
            "\"><INPUT TYPE=submit VALUE=\"Change\"></FORM></TD></TR>");
	TableIterator iterator = storage.getTableIterator("Software");
        if (person.status != Person.isUser) {
            con.print("<TR><TH ALIGN=LEFT>Projects:</TH>" + "<TD><FORM METHOD=POST ACTION=\"" + con.getURI() +
		      "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"softwareForm\">" + "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" +
		      myself + "\">" + "<INPUT TYPE=hidden NAME=\"person\" VALUE=\"" + personOID + "\">" +
		      "<SELECT NAME=\"software\" SIZE=1>");
	    if (person.setProjects.length != 0) {
		print(con, person.setProjects);
		con.print("</SELECT><INPUT TYPE=submit NAME=\"action\" VALUE=\"Select\">" +
			  "&nbsp;<INPUT TYPE=submit NAME=\"action\" VALUE=\"Detach\">");
            } else {
                con.print(EMPTY_LIST);
            }
            if (iterator.getNumberOfRows() != 0) {
                con.print("</FORM></TD></TR>" + "<TR><TH ALIGN=LEFT>Attach to project:</TH>" +
                    "<TD><FORM METHOD=POST ACTION=\"" + con.getURI() + "\"><INPUT TYPE=hidden NAME=\"page\" " +
                    "VALUE=\"attachToProject\"><INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + myself +
                    "\"><INPUT TYPE=hidden NAME=\"person\" VALUE=\"" + personOID + "\"><SELECT NAME=\"software\" SIZE=1>");
                print(con, iterator);
                con.print("</SELECT><INPUT TYPE=submit VALUE=\"Attach\">");
            }
            con.print("</FORM></TD></TR>" + "<TR><TH ALIGN=LEFT>Find a person:</TH>" + "<TD><FORM METHOD=POST ACTION=\"" +
                con.getURI() + "\"><INPUT TYPE=hidden NAME=\"page\" " + "VALUE=\"userForm\">" +
                "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + myself + "\">" + "<INPUT TYPE=text NAME=\"name\" SIZE=30>" +
                "<INPUT TYPE=submit VALUE=\"Find\"></FORM></TD></TR>");
        }
        con.print("<TR><TH ALIGN=LEFT>Used software:</TH>" + "<TD><FORM METHOD=POST ACTION=\"" + con.getURI() +
            "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"softwareForm\">" + "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" +
            myself + "\">" + "<INPUT TYPE=hidden NAME=\"person\" VALUE=\"" + personOID + "\">" +
            "<SELECT NAME=\"software\" SIZE=1>");
            if (person.setUsedSoftware.length != 0) {
                print(con, person.setUsedSoftware);
                con.print("</SELECT><INPUT TYPE=submit NAME=\"action\" VALUE=\"Select\">" +
                    "&nbsp;<INPUT TYPE=submit NAME=\"action\" VALUE=\"Unregister\">");
        } else {
            con.print(EMPTY_LIST);
        }
        if (iterator.getNumberOfRows() != 0) {
            iterator.beforeFirst();
            con.print("</FORM></TD></TR>" + "<TR><TH ALIGN=LEFT>Register as software user:</TH>" +
                "<TD><FORM METHOD=POST ACTION=\"" + con.getURI() + "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"registerSoftware\">" +
                "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + myself + "\">" +
                "<INPUT TYPE=hidden NAME=\"person\" VALUE=\"" + personOID + "\">" + "<SELECT NAME=\"software\" SIZE=1>");
            print(con, iterator);
            con.print("</SELECT><INPUT TYPE=submit VALUE=\"Register\">");
        }
        con.print("</FORM></TD></TR></TABLE><P><B>Reported bugs:</B><BR>" + "<FORM METHOD=POST ACTION=\"" + con.getURI() +
            "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"bugForm\">" + "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + myself +
            "\">" + "<INPUT TYPE=hidden NAME=\"person\" VALUE=\"" + personOID + "\">" + "<SELECT NAME=\"bug\" SIZE=5>");
        if (person.setReportedBugs.length != 0) {
            print(con, person.setReportedBugs);
            con.print("</SELECT><BR><INPUT TYPE=submit NAME=\"action\" VALUE=\"Select\">");
        } else {
            con.print(EMPTY_LIST);
        }
        con.print("</FORM><P>");
        if (person.status != Person.isUser) {
            con.print("<P><B>Assigned bugs:</B><BR>" + "<FORM METHOD=POST ACTION=\"" + con.getURI() +
                "\"><INPUT TYPE=hidden " + "NAME=\"page\" VALUE=\"bugForm\">" + "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" +
                myself + "\">" + "<INPUT TYPE=hidden NAME=\"person\" VALUE=\"" + personOID + "\">" +
                "<SELECT NAME=\"bug\" SIZE=5>");
                if (person.setAssignedBugs.length != 0) {
                    print(con, person.setAssignedBugs);
                    con.print("</SELECT><BR><INPUT TYPE=submit NAME=\"action\" VALUE=\"Select\">&nbsp;" +
                        "<INPUT TYPE=submit NAME=\"action\" VALUE=\"Deassign\">");
            } else {
                con.print(EMPTY_LIST);
            }
            con.print("</FORM>");
        }
        if (person == me) {
            con.print("</BODY></HTML>");
        } else {
            mainMenuReference(con);
        }
    }

    public void createBugReportForm(HttpConnection con) throws Exception {
        int i;
        con.print(HTML_HEAD + "<TITLE>Bug</TITLE></HEAD>" + BODY + "<H2>Bug</H2><FORM METHOD=POST ACTION=\"" + con.getURI() +
            "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"createBugReport\">" + "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" +
            con.get("myself") + "\"><TABLE><TH ALIGN=LEFT>Summary:</TH>" +
            "<TD><INPUT TYPE=text NAME=\"summary\" SIZE=40></TD></TR>" + "<TR><TH ALIGN=LEFT>Category:</TH>" +
            "<TD><SELECT NAME=\"category\" SIZE=1>");
            for (i = 1; eCATEGORY_STRING[i] != null; i++) {
                con.print("<OPTION VALUE=" + i + ">" + eCATEGORY_STRING[i] + "</OPTION>");
        }
        con.print("</SELECT></TD></TR><TR><TH ALIGN=LEFT>Severity:</TH>" + "<TD><SELECT NAME=\"severity\" SIZE=1>");
        for (i = 1; eSEVERITY_STRING[i] != null; i++) {
            con.print("<OPTION VALUE=" + i + ">" + eSEVERITY_STRING[i] + "</OPTION>");
        }
        con.print("</SELECT></TD></TR><TR><TH ALIGN=LEFT>Priority:</TH>" + "<TD><SELECT NAME=\"priority\" SIZE=1>");
        for (i = 1; eFIXING_PRIORITY_STRING[i] != null; i++) {
            con.print("<OPTION VALUE=" + i + ">" + eFIXING_PRIORITY_STRING[i] + "</OPTION>");
        }
        con.print("</SELECT></TD></TR><TR><TH ALIGN=LEFT>Software:</TH>" + "<TD><SELECT NAME=\"software\" SIZE=1>");
        TableIterator iterator = storage.getTableIterator("Software");
        print(con, iterator);
        iterator.close();
        con.print("</SELECT></TD></TR><TR><TH ALIGN=LEFT>Version:</TH>" + "<TD><INPUT TYPE=text NAME=\"version\"></TD></TR>" +
            "<TR><TH ALIGN=LEFT>Platform:</TH>" + "<TD><INPUT TYPE=text NAME=\"platform\"></TD></TR>" +
            "<TR><TH ALIGN=LEFT>OS:</TH>" + "<TD><INPUT TYPE=text NAME=\"os\"></TD></TR></TABLE><P>" +
            "<INPUT TYPE=submit VALUE=\"Submit\">&nbsp;" + "<INPUT TYPE=reset></FORM>");
        mainMenuReference(con);
    }

    public void createBugReport(HttpConnection con) throws Exception {
        storage.lock();
        Person person = (Person)storage.getObjectByOid(Long.parseLong(con.get("myself"), 10));
        String softOID = con.get("software");
        Software product = (Software)storage.getObjectByOid(Long.parseLong(softOID, 10));
        int majorVersion, minorVersion;
        try {
            String version = con.get("version");
            int p = version.indexOf('.');
            majorVersion = Integer.parseInt(version.substring(0, p), 10);
            minorVersion = Integer.parseInt(version.substring(p + 1), 10);
        } catch (NumberFormatException x) {
            error(con, "Bad version format");
            return;
        }
        Version version;
        product.loadComponents();
        for (version = product.pVersions; version != null; version = version.pNext) {
            if (version.majorVersionNumber == majorVersion && version.minorVersionNumber == minorVersion) {
                break;
            }
        }
        if (version == null) {
            error(con, "No such software product version");
            return;
        }
        Bug bug = new Bug();
        bug.sOneLineSummary = con.get("summary");
        bug.eCategory = Integer.parseInt(con.get("category"), 10);
        bug.eFixingPriority = Integer.parseInt(con.get("priority"), 10);
        bug.eSeverity = Integer.parseInt(con.get("severity"), 10);
        bug.sOperatingSystem = con.get("os");
        bug.sHardwarePlatform = con.get("platform");
        bug.pReportedBy = person;
        person.setReportedBugs = (Bug[]) AnyArray.append(person.setReportedBugs, bug);
        bug.pAssignedTo = null;
        bug.pSoftware = product;
        bug.setSimilarBugs = new Bug[0];
        product.setBugs = (Bug[]) AnyArray.append(product.setBugs, bug);
        bug.pVersion = version;
        version.setBugs = (Bug[]) AnyArray.append(version.setBugs, bug);
        bug.nReports = 0;
        product.store();
        version.store();
        person.store();
        con.addPair("bug", Long.toString(bug.getOid(), 10));
        con.addPair("action", "Select");
        bugForm(con);
    }

    public void bugForm(HttpConnection con) throws Exception {
        int i;
        storage.lock();
        String bugOID = con.get("bug");
        if (bugOID == null) {
            error(con, "No bug was selected");
            return;
        }
        TableIterator bugs = storage.getTableIterator("Bug");
        Bug bug = (Bug)bugs.selectByOid(Long.parseLong(bugOID, 10));
        String action = con.get("action");
        String myself = con.get("myself");
        if (action.equals("Remove")) {
            bugOID = con.get("relatedbug");
            if (bugOID == null) {
                error(con, "No related bug was selected");
                bugs.close();
                return;
            }
            Bug relatedbug = (Bug)bugs.selectByOid(Long.parseLong(bugOID, 10));
            bugs.close();
            bug.setSimilarBugs = (Bug[]) AnyArray.remove(bug.setSimilarBugs, relatedbug);
            relatedbug.setSimilarBugs = (Bug[]) AnyArray.remove(relatedbug.setSimilarBugs, bug);
            bug.store();
            relatedbug.store();
            con.addPair("action", "Select");
            bugForm(con);
            bugs.close();
            return;
        }
        if (action.equals("Deassign")) {
            bug.pAssignedTo.setAssignedBugs = (Bug[]) AnyArray.remove(bug.pAssignedTo.setAssignedBugs, bug);
            bug.pAssignedTo.store();
            bug.pAssignedTo = null;
            bug.store();
            con.addPair("person", myself);
            userForm(con);
            bugs.close();
            return;
        }
        TableIterator persons = storage.getTableIterator("Person");
        Person person = (Person)persons.selectByOid(Long.parseLong(myself, 10));
        persons.close();
        int personStatus = person.status;
        con.print(HTML_HEAD + "<TITLE>Bug in " + bug.pSoftware.sName + " v. " + bug.pVersion.getVersionString() +
            "</TITLE></HEAD>" + BODY + "<H2>Bug in " + bug.pSoftware.sName + " v. " + bug.pVersion.getVersionString() +
            "</H2><FORM METHOD=POST ACTION=\"" + con.getURI() + "\"><INPUT TYPE=hidden " +
            "NAME=\"page\" VALUE=\"updateBug\">" + "<INPUT TYPE=hidden NAME=\"bug\" VALUE=" + bugOID + ">" +
            "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + myself + "\">" + "<TABLE><TH ALIGN=LEFT>Summary:</TH>" +
            "<TD><INPUT TYPE=text NAME=\"summary\" SIZE=40 VALUE=\"" + encodeHTML(bug.sOneLineSummary) + "\"></TD></TR>" +
            "<TR><TH ALIGN=LEFT>Category:</TH>" + "<TD><SELECT NAME=\"category\" SIZE=1>" + "<OPTION SELECTED VALUE=" +
            bug.eCategory + ">" + eCATEGORY_STRING[bug.eCategory] + "</OPTION>");
        for (i = 1; eCATEGORY_STRING[i] != null; i++) {
            if (i != bug.eCategory) {
                con.print("<OPTION VALUE=" + i + ">" + eCATEGORY_STRING[i] + "</OPTION>");
            }
        }
        con.print("</SELECT></TD></TR><TR><TH ALIGN=LEFT>Severity:</TH>" + "<TD><SELECT NAME=\"severity\" SIZE=1>" +
            "<OPTION SELECTED VALUE=" + bug.eSeverity + ">" + eSEVERITY_STRING[bug.eSeverity] + "</OPTION>");
        for (i = 1; eSEVERITY_STRING[i] != null; i++) {
            if (i != bug.eSeverity) {
                con.print("<OPTION  VALUE=" + i + ">" + eSEVERITY_STRING[i] + "</OPTION>");
            }
        }
        con.print("</SELECT></TD></TR><TR><TH ALIGN=LEFT>Priority:</TH>" + "<TD><SELECT NAME=\"priority\" SIZE=1>" +
            "<OPTION SELECTED VALUE=" + bug.eFixingPriority + ">" +
            eFIXING_PRIORITY_STRING[bug.eFixingPriority] + "</OPTION>");
            for (i = 1; eFIXING_PRIORITY_STRING[i] != null; i++) {
                if (i != bug.eFixingPriority) {
                    con.print("<OPTION VALUE=" + i + ">" + eFIXING_PRIORITY_STRING[i] + "</OPTION>");
                }
        }
        con.print("</SELECT></TD></TR><TR><TH ALIGN=LEFT>Platform:</TH>" + "<TD><INPUT TYPE=text NAME=\"platform\" VALUE=\"" +
            bug.sHardwarePlatform + "\"></TD></TR>" + "<TR><TH ALIGN=LEFT>OS:</TH><TD><INPUT TYPE=text NAME=\"os\"VALUE=\"" +
            bug.sOperatingSystem + "\"></TD></TR><TR><TH ALIGN=LEFT>Assigned to:</TH><TD>");
        if (bug.pSoftware.setEngineers.length != 0) {
            bug.pSoftware.loadComponents();
            con.print("<SELECT SIZE=1 NAME=\"person\">");
            if (bug.pAssignedTo != null) {
                con.print("<OPTION SELECTED VALUE=\"" + bug.pAssignedTo.getOid() + "\">" +
                    bug.pAssignedTo.sName + "</OPTION>");
            } else {
                con.print("<OPTION SELECTED VALUE=\"\"></OPTION>");
            }
            for (i = 0; i < bug.pSoftware.setEngineers.length; i++) {
                if (bug.pSoftware.setEngineers[i] != bug.pAssignedTo) {
                    bug.pSoftware.setEngineers[i].print(con);
                }
            }
            con.print("</SELECT>");
        }
        con.print("</TD></TR><TR><TH ALIGN=LEFT>Similar with:</TH>" +
            "<TD><SELECT NAME=\"similar\" SIZE=1><OPTION SELECTED VALUE=\"\"></OPTION>");
            print(con, bugs);
        bugs.close();
        con.print("</SELECT></TD></TR></TABLE><BR>");
        if (personStatus != Person.isUser) {
            con.print("<INPUT TYPE=submit NAME=\"action\" VALUE=\"Update\">&nbsp;" + "<INPUT TYPE=reset VALUE=\"Reset\">");
        }
        con.print("</FORM><P><FORM METHOD=POST ACTION=\"" + con.getURI() + "\">" +
            "<INPUT TYPE=hidden NAME=\"page\" VALUE=\"updateReportForm\">" + "<INPUT TYPE=hidden NAME=\"bug\" VALUE=" +
            bugOID + ">" + "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + myself + "\">" +
            "<B>Report history:</B><BR><SELECT NAME=\"report\" SIZE=5>");
        if (bug.pReportHistory != null) {
            for (Report report = bug.pReportHistory; report != null; report = report.pNext) {
                report.print(con);
            }
            con.print("</SELECT><BR><INPUT TYPE=submit NAME=\"action\" VALUE=\"Select\">" +
                "&nbsp;<INPUT TYPE=submit NAME=\"action\" VALUE=\"Add\">");
                if (personStatus == Person.isAdministrator) {
                    con.print("&nbsp;<INPUT TYPE=submit NAME=\"action\" VALUE=\"Remove\">");
            }
        } else {
            con.print(EMPTY_LIST + "<BR><INPUT TYPE=submit NAME=\"action\" VALUE=\"Add\">");
        }
        con.print("</FORM><P>");
        con.print("<FORM METHOD=POST ACTION=\"" + con.getURI() + "\"><INPUT TYPE=hidden " +
            "NAME=\"page\" VALUE=\"updateWorkAroundForm\">" + "<INPUT TYPE=hidden NAME=\"bug\" VALUE=" + bugOID + ">" +
            "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + myself + "\">" +
            "<B>Work arounds:</B><BR><SELECT NAME=\"workaround\" SIZE=5>");
            if (bug.pWorkArounds != null) {
                for (Report report = bug.pWorkArounds; report != null; report = report.pNext) {
                    report.print(con);
                }
                con.print("</SELECT><BR><INPUT TYPE=submit NAME=\"action\" VALUE=\"Select\">" +
                    "&nbsp;<INPUT TYPE=submit NAME=\"action\" VALUE=\"Add\">");
                if (personStatus == Person.isAdministrator) {
                    con.print("&nbsp;<INPUT TYPE=submit NAME=\"action\" VALUE=\"Remove\">");
                }
        } else {
            con.print(EMPTY_LIST + "<BR><INPUT TYPE=submit NAME=\"action\" VALUE=\"Add\">");
        }
        con.print("</FORM><P>");
        if (bug.setSimilarBugs.length != 0) {
            con.print("<FORM METHOD=POST ACTION=\"" + con.getURI() + "\"><INPUT TYPE=hidden " +
                "NAME=\"page\" VALUE=\"bugForm\">" + "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + myself + "\">" +
                "<INPUT TYPE=hidden NAME=\"relatedbug\" VALUE=" + bugOID + ">" +
                "<B>Similar bugs:</B><BR><SELECT NAME=\"bug\" SIZE=1>");
                print(con, bug.setSimilarBugs);
            con.print("</SELECT><INPUT TYPE=submit NAME=\"action\" VALUE=\"Select\">");
            if (personStatus == Person.isAdministrator) {
                con.print("&nbsp;<INPUT TYPE=submit NAME=\"action\" VALUE=\"Remove\">");
            }
            con.print("</FORM><P>");
        }
        con.print("<FONT SIZE=\"+1\"><UL>");
        if (personStatus == Person.isUser) {
            if (bug.pAssignedTo != null) {
                con.print("<LI>Assigned to <A HREF=\"mailto:" + bug.pAssignedTo.sEmailAddress + "\">" +
                    encodeHTML(bug.pAssignedTo.sName) + "</A>");
            }
            con.print("<LI>Reported by <A HREF=\"mailto:" + bug.pReportedBy.sEmailAddress + "\">" +
                encodeHTML(bug.pReportedBy.sName) + "</A></OL></FONT>");
        } else {
            if (bug.pAssignedTo != null) {
                con.print("<LI>Assigned to <A HREF=\"" + con.getURI() + "?page=userForm&myself=" + myself + "&person=" +
                    bug.pAssignedTo.getOid() + "\">" + encodeHTML(bug.pAssignedTo.sName) + "</A>");
            }
            con.print("<LI>Reported by <A HREF=\"" + con.getURI() + "?page=userForm&myself=" + myself + "&person=" +
                bug.pReportedBy.getOid() + "\">" + encodeHTML(bug.pReportedBy.sName) + "</A></OL></FONT>");
        }
        mainMenuReference(con);
    }

    public void updateBug(HttpConnection con) throws Exception {
        storage.lock();
        String bugOID = con.get("bug");
        TableIterator bugs = storage.getTableIterator("Bug");
        Bug bug = (Bug)bugs.selectByOid(Long.parseLong(bugOID, 10));
        String similarOID = con.get("similar");
        if (similarOID != null && similarOID.length() != 0) {
            Bug similar = (Bug)bugs.selectByOid(Long.parseLong(similarOID, 10));
            if (AnyArray.find(bug.setSimilarBugs, similar) < 0) {
                bug.setSimilarBugs = (Bug[]) AnyArray.append(bug.setSimilarBugs, similar);
                similar.setSimilarBugs = (Bug[]) AnyArray.append(similar.setSimilarBugs, bug);
                similar.store();
            }
        }
        bugs.close();
        String personOID = con.get("person");
        if (personOID != null && personOID.length() != 0) {
            Person person = (Person)storage.getObjectByOid(Long.parseLong(personOID));
            if (person.status == Person.isUser) {
                error(con, "No such engineer");
                return;
            }
            if (bug.pAssignedTo != null) {
                bug.pAssignedTo.setAssignedBugs = (Bug[]) AnyArray.remove(bug.pAssignedTo.setAssignedBugs, bug);
                bug.pAssignedTo.store();
            }
            bug.pAssignedTo = person;
            person.setAssignedBugs = (Bug[]) AnyArray.append(person.setAssignedBugs, bug);
            person.store();
        }
        bug.sOneLineSummary = con.get("summary");
        bug.eCategory = Integer.parseInt(con.get("category"), 10);
        bug.eFixingPriority = Integer.parseInt(con.get("priority"), 10);
        bug.eSeverity = Integer.parseInt(con.get("severity"), 10);
        bug.sOperatingSystem = con.get("os");
        bug.sHardwarePlatform = con.get("platform");
        bug.store();
        bugForm(con);
    }

    public void addReportForm(HttpConnection con) throws Exception {
        storage.lock();
        String bugOID = con.get("bug");
        Bug bug = (Bug)storage.getObjectByOid(Long.parseLong(bugOID, 10));
        con.print(HTML_HEAD + "<TITLE>Bug report</TITLE></HEAD>" + BODY + "<H2>Bug report</H2><FORM METHOD=POST ACTION=\"" +
            con.getURI() + "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"addReport\">" +
            "<INPUT TYPE=hidden NAME=\"bug\" VALUE=" + bugOID + ">" + "<INPUT TYPE=hidden NAME=\"index\" VALUE=" +
            ++bug.nReports + "><INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + con.get("myself") +
            "\"><B>Status: &nbsp;</B><SELECT SIZE=1 NAME=\"status\">");
        for (int i = 1; eSTATUS_STRING[i] != null; i++) {
            con.print("<OPTION VALUE=" + i + ">" + eSTATUS_STRING[i] + "</OPTION>");
        }
        con.print("</SELECT><P><B>Bug description:</B><P>" + "<TEXTAREA COLS=40 ROWS=5 NAME=\"description\"></TEXTAREA><P>" +
            "<INPUT TYPE=submit VALUE=\"Add\">&nbsp;" + "<INPUT TYPE=reset VALUE=\"Reset\"></FORM>");
        bug.store();
        mainMenuReference(con);
    }

    public void addReport(HttpConnection con) throws Exception {
        storage.lock();
        String bugOID = con.get("bug");
        Bug bug = (Bug)storage.getObjectByOid(Long.parseLong(bugOID, 10));
        String myself = con.get("myself");
        Person person = (Person)storage.getObjectByOid(Long.parseLong(myself, 10));
        Report report;
        int reportId = Integer.parseInt(con.get("index"), 10);
        for (report = bug.pReportHistory; report != null && report.reportId != reportId; report = report.pNext);
        if (report == null) {
            report = new Report();
            report.pAuthor = person;
            report.sDescription = con.get("description");
            report.reportId = reportId;
            report.pNext = bug.pReportHistory;
            report.status = Integer.parseInt(con.get("status"), 10);
            report.creationDate = new Date(System.currentTimeMillis());
            bug.pReportHistory = report;
            person.store();
            bug.store();
        }
        con.addPair("action", "Select");
        bugForm(con);
    }

    public void addWorkAroundForm(HttpConnection con) throws Exception {
        storage.lock();
        String bugOID = con.get("bug");
        Bug bug = (Bug)storage.getObjectByOid(Long.parseLong(bugOID, 10));
        con.print(HTML_HEAD + "<TITLE>Work around</TITLE></HEAD>" + BODY + "<H2>Work around</H2><FORM METHOD=POST ACTION=\"" +
            con.getURI() + "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"addWorkAround\">" +
            "<INPUT TYPE=hidden NAME=\"bug\" VALUE=" + bugOID + ">" + "<INPUT TYPE=hidden NAME=\"index\" VALUE=" +
            ++bug.nReports + "><INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + con.get("myself") +
            "\"><B>Status: &nbsp;</B><SELECT SIZE=1 NAME=\"status\">");
        for (int i = 1; eSTATUS_STRING[i] != null; i++) {
            con.print("<OPTION VALUE=" + i + ">" + eSTATUS_STRING[i] + "</OPTION>");
        }
        con.print("</SELECT><P><B>Description:</B><P>" + "<TEXTAREA COLS=40 ROWS=5 NAME=\"description\"></TEXTAREA><P>" +
            "<INPUT TYPE=submit VALUE=\"Add\">&nbsp;" + "<INPUT TYPE=reset VALUE=\"Reset\"></FORM>");
        bug.store();
        mainMenuReference(con);
    }

    public void addWorkAround(HttpConnection con) throws Exception {
        storage.lock();
        String bugOID = con.get("bug");
        Bug bug = (Bug)storage.getObjectByOid(Long.parseLong(bugOID, 10));
        String myself = con.get("myself");
        Person person = (Person)storage.getObjectByOid(Long.parseLong(myself, 10));
        Report report;
        int reportId = Integer.parseInt(con.get("index"), 10);
        for (report = bug.pWorkArounds; report != null && report.reportId != reportId; report = report.pNext);
        if (report == null) {
            report = new Report();
            report.pAuthor = person;
            report.sDescription = con.get("description");
            report.reportId = reportId;
            report.pNext = bug.pWorkArounds;
            report.status = Integer.parseInt(con.get("status"), 10);
            report.creationDate = new Date(System.currentTimeMillis());
            bug.pWorkArounds = report;
            person.store();
            bug.store();
        }
        con.addPair("action", "Select");
        bugForm(con);
    }

    public void updateReportForm(HttpConnection con) throws Exception {
        String action = con.get("action");
        if (action.equals("Add")) {
            addReportForm(con);
            return;
        }
        String reportIdx = con.get("report");
        if (reportIdx == null) {
            error(con, "No report was selected");
            return;
        }
        storage.lock();
        String bugOID = con.get("bug");
        Bug bug = (Bug)storage.getObjectByOid(Long.parseLong(bugOID, 10));
        int reportId = Integer.parseInt(reportIdx, 10);
        Report prev = null, report = bug.pReportHistory;
        while (report != null && report.reportId != reportId) {
            prev = report;
            report = report.pNext;
        }
        if (report == null) {
            error(con, "No such report");
            return;
        }
        if (action.equals("Remove")) {
            if (prev == null) {
                bug.pReportHistory = report.pNext;
                bug.store();
            } else {
                prev.pNext = report.pNext;
                prev.store();
            }
            report.remove();
            con.addPair("action", "Select");
            bugForm(con);
            return;
        }
        String myself = con.get("myself");
        Person person = (Person)storage.getObjectByOid(Long.parseLong(myself, 10));
        int personStatus = person.status;
        con.print(HTML_HEAD + "<TITLE>Bug report from " + report.creationDate + "</TITLE></HEAD>" + BODY +
            "<H2>Bug report from " + report.creationDate + "</H2>" + "<FORM METHOD=POST ACTION=\"" + con.getURI() +
            "\"><INPUT TYPE=hidden " + "NAME=\"page\" VALUE=\"updateReport\">" + "<INPUT TYPE=hidden NAME=\"bug\" VALUE=" +
            bugOID + ">" + "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + myself +
            "\"><INPUT TYPE=hidden NAME=\"report\" VALUE=" + reportIdx + "><B>Created by ");
        if (personStatus == Person.isUser) {
            con.print("<A HREF=\"mailto:" + report.pAuthor.sEmailAddress + "\">" + encodeHTML(report.pAuthor.sName) + "</A>");
        } else {
            con.print("<A HREF=\"" + con.getURI() + "?page=userForm&myself=" + myself + "&person=" + report.pAuthor.getOid() +
                "\">" + encodeHTML(report.pAuthor.sName) + "</A>");
        }
        con.print("<P>Status: </B><SELECT SIZE=1 NAME=\"status\">" + "<OPTION SELECTED VALUE=" + report.status + ">" +
            eSTATUS_STRING[report.status] + "</OPTION>");
        for (int i = 1; eSTATUS_STRING[i] != null; i++) {
            if (i != report.status) {
                con.print("<OPTION VALUE=" + i + ">" + eSTATUS_STRING[i] + "</OPTION>");
            }
        }
        con.print("</SELECT><P><B>Bug description:</B><BR>" + "<TEXTAREA COLS=40 ROWS=5 NAME=\"description\">" +
            report.sDescription + "</TEXTAREA><P>");
        if (personStatus != Person.isUser) {
            con.print("<INPUT TYPE=submit VALUE=\"Update\">&nbsp;" + "<INPUT TYPE=reset VALUE=\"Reset\">");
        }
        con.print("</FORM>");
        mainMenuReference(con);
    }

    public void updateWorkAroundForm(HttpConnection con) throws Exception {
        String action = con.get("action");
        if (action.equals("Add")) {
            addWorkAroundForm(con);
            return;
        }
        String reportIdx = con.get("workaround");
        if (reportIdx == null) {
            error(con, "No report was selected");
            return;
        }
        storage.lock();
        String bugOID = con.get("bug");
        Bug bug = (Bug)storage.getObjectByOid(Long.parseLong(bugOID, 10));
        int reportId = Integer.parseInt(reportIdx, 10);
        Report prev = null, report = bug.pWorkArounds;
        while (report != null && report.reportId != reportId) {
            prev = report;
            report = report.pNext;
        }
        if (report == null) {
            error(con, "No such report");
            return;
        }
        if (action.equals("Remove")) {
            if (prev == null) {
                bug.pWorkArounds = report.pNext;
                bug.store();
            } else {
                prev.pNext = report.pNext;
                prev.store();
            }
            report.remove();
            con.addPair("action", "Select");
            bugForm(con);
            return;
        }
        String myself = con.get("myself");
        Person person = (Person)storage.getObjectByOid(Long.parseLong(myself, 10));
        int personStatus = person.status;
        con.print(HTML_HEAD + "<TITLE>Work around " + report.creationDate + "</TITLE></HEAD>" + BODY + "<H2>Work around " +
            report.creationDate + "</H2>" + "<FORM METHOD=POST ACTION=\"" + con.getURI() + "\"><INPUT TYPE=hidden " +
            "NAME=\"page\" VALUE=\"updateWorkAround\">" + "<INPUT TYPE=hidden NAME=\"bug\" VALUE=" + bugOID + ">" +
            "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + myself + "\"><INPUT TYPE=hidden NAME=\"workaround\" VALUE=" +
            reportIdx + "><B>Created by ");
        if (personStatus == Person.isUser) {
            con.print("<A HREF=\"mailto:" + report.pAuthor.sEmailAddress + "\">" + encodeHTML(report.pAuthor.sName) + "</A>");
        } else {
            con.print("<A HREF=\"" + con.getURI() + "?page=userForm&myself=" + myself + "&person=" + report.pAuthor.getOid() +
                "\">" + encodeHTML(report.pAuthor.sName) + "</A>");
        }
        con.print("<P>Status: </B><SELECT SIZE=1 NAME=\"status\">" + "<OPTION SELECTED VALUE=" + report.status + ">" +
            eSTATUS_STRING[report.status] + "</OPTION>");
        for (int i = 1; eSTATUS_STRING[i] != null; i++) {
            if (i != report.status) {
                con.print("<OPTION VALUE=" + i + ">" + eSTATUS_STRING[i] + "</OPTION>");
            }
        }
        con.print("</SELECT><P><B>Bug description:</B><BR>" + "<TEXTAREA COLS=40 ROWS=5 NAME=\"description\">" +
            report.sDescription + "</TEXTAREA><P>");
        if (personStatus != Person.isUser) {
            con.print("<INPUT TYPE=submit VALUE=\"Update\">&nbsp;" + "<INPUT TYPE=reset VALUE=\"Reset\">");
        }
        con.print("</FORM>");
        mainMenuReference(con);
    }

    public void updateReport(HttpConnection con) throws Exception {
        storage.lock();
        String bugOID = con.get("bug");
        Bug bug = (Bug)storage.getObjectByOid(Long.parseLong(bugOID, 10));
        String reportId = con.get("report");
        if (reportId == null) {
            error(con, "No report was selected");
            return;
        }
        int rid = Integer.parseInt(reportId, 10);
        Report report;
        for (report = bug.pReportHistory; report != null && report.reportId != rid; report = report.pNext);
        if (report == null) {
            error(con, "No such report");
        } else {
            report.sDescription = con.get("description");
            report.status = Integer.parseInt(con.get("status"));
            report.store();
            con.addPair("action", "Select");
            bugForm(con);
        }
    }

    public void updateWorkAround(HttpConnection con) throws Exception {
        storage.lock();
        String bugOID = con.get("bug");
        Bug bug = (Bug)storage.getObjectByOid(Long.parseLong(bugOID, 10));
        String reportId = con.get("workaround");
        if (reportId == null) {
            error(con, "No report was selected");
            return;
        }
        int rid = Integer.parseInt(reportId, 10);
        Report report;
        for (report = bug.pWorkArounds; report != null && report.reportId != rid; report = report.pNext);
        if (report == null) {
            error(con, "No such report");
        } else {
            report.sDescription = con.get("description");
            report.status = Integer.parseInt(con.get("status"));
            report.store();
            con.addPair("action", "Select");
            bugForm(con);
        }
    }

    public void attachToProject(HttpConnection con) throws Exception {
        storage.lock();
        String personOID = con.get("person");
        Person person = (Person)storage.getObjectByOid(Long.parseLong(personOID, 10));
        if (person.status == Person.isUser) {
            error(con, "No such engineer");
        } else {
            String softOID = con.get("software");
            Software product = (Software)storage.getObjectByOid(Long.parseLong(softOID, 10));
            if (AnyArray.find(product.setEngineers, person) >= 0) {
                error(con, "Engineer already attached to the project");
            } else {
                product.setEngineers = (Person[]) AnyArray.append(product.setEngineers, person);
                product.store();
                person.setProjects = (Software[]) AnyArray.append(person.setProjects, product);
                person.store();
                userForm(con);
            }
        }
    }

    public void registerSoftware(HttpConnection con) throws Exception {
        storage.lock();
        String personOID = con.get("person");
        Person person = (Person)storage.getObjectByOid(Long.parseLong(personOID, 10));
        String softOID = con.get("software");
        Software product = (Software)storage.getObjectByOid(Long.parseLong(softOID, 10));
        if (AnyArray.find(product.setUsers, person) >= 0) {
            error(con, "User already registered this software");
        } else {
            product.setUsers = (Person[]) AnyArray.append(product.setUsers, person);
            product.store();
            person.setUsedSoftware = (Software[]) AnyArray.append(person.setUsedSoftware, product);
            person.store();
            userForm(con);
        }
    }

    public void softwareForm(HttpConnection con) throws Exception {
        String softOID = con.get("software");
        if (softOID == null) {
            error(con, "No software product was selected");
            return;
        }
        String action = con.get("action");
        if (action.equals("Detach")) {
            storage.lock();
            Software product = (Software)storage.getObjectByOid(Long.parseLong(softOID, 10));
            String personOID = con.get("person");
            Person person = (Person)storage.getObjectByOid(Long.parseLong(personOID, 10));
            person.setProjects = (Software[]) AnyArray.remove(person.setProjects, product);
            person.store();
            product.setEngineers = (Person[]) AnyArray.remove(product.setEngineers, person);
            product.store();
            userForm(con);
            return;
        }
        if (action.equals("Unregister")) {
            storage.lock();
            Software product = (Software)storage.getObjectByOid(Long.parseLong(softOID, 10));
            String personOID = con.get("person");
            Person person = (Person)storage.getObjectByOid(Long.parseLong(personOID, 10));
            person.setUsedSoftware = (Software[]) AnyArray.remove(person.setUsedSoftware, product);
            person.store();
            product.setUsers = (Person[]) AnyArray.remove(product.setUsers, person);
            product.store();
            userForm(con);
            return;
        }
        Software product = (Software)storage.getObjectByOid(Long.parseLong(softOID, 10));
        product.loadComponents();
        String myself = con.get("myself");
        Person person = (Person)storage.getObjectByOid(Long.parseLong(myself, 10));
        int personStatus = person.status;
        String name = encodeHTML(product.sName);
        con.print(HTML_HEAD + "<TITLE>" + name + "</TITLE></HEAD>" + BODY + "<H2>" + name + "</H2>" +
            "<FORM METHOD=POST ACTION=\"" + con.getURI() + "\"><INPUT TYPE=hidden " +
            "NAME=\"page\" VALUE=\"updateSoftware\">" + "<INPUT TYPE=hidden NAME=\"software\" VALUE=\"" + softOID + "\">" +
            "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + myself + "\">" + "<TABLE><TR><TH ALIGN=LEFT>Product name:</TH>" +
            "<TD><INPUT TYPE=text NAME=\"newname\" VALUE=\"" + name + "\"></TD></TR>");
        if (product.pVersions != null) {
            con.print("<TR><TH ALIGN=LEFT>Current version:</TH>" + "<TD><INPUT TYPE=text NAME=\"version\" SIZE=8 VALUE=\"" +
                product.pVersions.getVersionString() + "\"></TD></TR>" + "<TR><TH ALIGN=LEFT>Current version label:</TH>" +
                "<TD><INPUT TYPE=text NAME=\"label\" SIZE=20 VALUE=\"" + encodeHTML(product.pVersions.sLabel) +
                "\"></TD></TR>" + "<TR><TH ALIGN=LEFT>Current version comment:</TH>" +
                "<TD><INPUT TYPE=text NAME=\"comment\" SIZE=40 VALUE=\"" +
                encodeHTML(product.pVersions.sComment) + "\"></TD></TR>");
        } else {
            con.print("<TR><TH ALIGN=LEFT>Current version:</TH>" + "<TD><INPUT TYPE=text NAME=\"version\" SIZE=8></TD></TR>" +
                "<TR><TH ALIGN=LEFT>Current version label:</TH>" + "<TD><INPUT TYPE=text NAME=\"label\" SIZE=20></TD></TR>" +
                "<TR><TH ALIGN=LEFT>Current version comment:</TH>" +
                "<TD><INPUT TYPE=text NAME=\"comment\" SIZE=40></TD></TR>");
        }
        con.print("</TABLE><BR>");
        if (personStatus != Person.isUser) {
            con.print("<INPUT TYPE=submit VALUE=\"Update\">&nbsp;<INPUT TYPE=reset>");
        }
        con.print("</FORM><P><TABLE><TR><TH ALIGN=LEFT>Engineers:</TH>" + "<TD><FORM METHOD=POST ACTION=\"" + con.getURI() +
            "\"><INPUT TYPE=HIDDEN NAME=\"page\" VALUE=\"userForm\">" + "<INPUT TYPE=HIDDEN NAME=\"myself\" VALUE=\"" +
            myself + "\"><SELECT NAME=\"person\" SIZE=1>");
        if (product.setEngineers.length != 0) {
            print(con, product.setEngineers);
            con.print("</SELECT>");
            if (personStatus != Person.isUser) {
                con.print("<INPUT TYPE=submit VALUE=\"Select\">");
            }
        } else {
            con.print(EMPTY_LIST);
        }
        con.print("</FORM></TD></TR><TR><TH ALIGN=LEFT>Users:</TH>" + "<TD><FORM METHOD=POST ACTION=\"" + con.getURI() +
            "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"userForm\">" + "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" +
            myself + "\"><SELECT NAME=\"person\" SIZE=1>");
        if (product.setUsers.length != 0) {
            print(con, product.setUsers);
            con.print("</SELECT>");
            if (personStatus != Person.isUser) {
                con.print("<INPUT TYPE=submit VALUE=\"Select\">");
            }
        } else {
            con.print(EMPTY_LIST);
        }
        con.print("</FORM></TD></TR><TR><TH ALIGN=LEFT>Bugs:</TH>" + "<TD><FORM METHOD=POST ACTION=\"" + con.getURI() +
            "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"bugForm\">" + "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + myself +
            "\"><SELECT NAME=\"bug\" SIZE=1>");
        if (product.setBugs.length != 0) {
            print(con, product.setBugs);
            con.print("</SELECT><INPUT TYPE=submit NAME=\"action\" VALUE=\"Select\">");
        } else {
            con.print(EMPTY_LIST);
        }
        con.print("</FORM></TD></TR><TR><TH ALIGN=LEFT>Versions:</TH><TD>" + "<FORM METHOD=POST ACTION=\"" + con.getURI() +
            "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"versionForm\">" + "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" +
            myself + "\"><INPUT TYPE=HIDDEN NAME=\"software\" VALUE=\"" + softOID + "\"><SELECT NAME=\"version\" SIZE=1>");
        if (product.pVersions != null) {
            for (Version version = product.pVersions; version != null; version = version.pNext) {
                version.print(con);
            }
            con.print("</SELECT><INPUT TYPE=submit VALUE=\"Select\">");
        } else {
            con.print(EMPTY_LIST);
        }
        con.print("</FORM></TD></TR></TABLE>");
        mainMenuReference(con);
    }

    public void updateSoftware(HttpConnection con) throws Exception {
        storage.lock();
        String softOID = con.get("software");
        Software product = (Software)storage.getObjectByOid(Long.parseLong(softOID, 10));
        product.loadComponents();
        String currentVersion = con.get("version");
        int majorVersion, minorVersion;
        try {
            int p = currentVersion.indexOf('.');
            majorVersion = Integer.parseInt(currentVersion.substring(0, p), 10);
            minorVersion = Integer.parseInt(currentVersion.substring(p + 1), 10);
        } catch (Exception x) {
            error(con, "Bad version number (MAJOR.MINOR expected)");
            return;
        }
        product.sName = con.get("newname");
        if (product.pVersions != null && product.pVersions.minorVersionNumber == minorVersion &&
            product.pVersions.majorVersionNumber == majorVersion) {
                product.pVersions.sComment = con.get("label");
                product.pVersions.sLabel = con.get("comment");
        } else if (product.pVersions != null && (product.pVersions.majorVersionNumber > majorVersion ||
            (product.pVersions.majorVersionNumber == majorVersion && product.pVersions.minorVersionNumber > minorVersion))) {
                error(con, "Version number less than of current version");
                return;
        } else {
            Version version = new Version();
            version.sLabel = con.get("label");
            version.sComment = con.get("comment");
            version.majorVersionNumber = majorVersion;
            version.minorVersionNumber = minorVersion;
            version.released = new Date(System.currentTimeMillis());
            version.pNext = product.pVersions;
            product.pVersions = version;
        }
        product.save();
        con.addPair("person", con.get("myself"));
        userForm(con);
    }

    public void versionForm(HttpConnection con) throws Exception {
        String softOID = con.get("software");
        Software product = (Software)storage.getObjectByOid(Long.parseLong(softOID, 10));
        product.loadComponents();
        String versionStr = con.get("version");
        int majorVersion, minorVersion;
        try {
            int p = versionStr.indexOf('.');
            majorVersion = Integer.parseInt(versionStr.substring(0, p), 10);
            minorVersion = Integer.parseInt(versionStr.substring(p + 1), 10);
        } catch (Exception x) {
            error(con, "Bad version number (MAJOR.MINOR expected)");
            return;
        }
        Version version;
        for (version = product.pVersions; version != null; version = version.pNext) {
            if (version.majorVersionNumber == majorVersion && version.minorVersionNumber == minorVersion) {
                break;
            }
        }
        if (version == null) {
            error(con, "No such version");
            return;
        }
        String myself = con.get("myself");
        Person person = (Person)storage.getObjectByOid(Long.parseLong(myself, 10));
        con.print(HTML_HEAD + "<TITLE>" + encodeHTML(product.sName) + " v. " + versionStr + "</TITLE></HEAD>" + BODY + "<H2>" +
            encodeHTML(product.sName) + " v. " + versionStr + "</H2>" + "<FORM METHOD=POST ACTION=\"" + con.getURI() +
            "\"><INPUT TYPE=hidden " + "NAME=\"page\" VALUE=\"updateVersion\">" +
            "<INPUT TYPE=hidden NAME=\"software\" VALUE=\"" + softOID + "\">" + "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" +
            myself + "\">" + "<INPUT TYPE=hidden NAME=\"version\" VALUE=\"" + versionStr + "\">" +
            "<TABLE><TR><TH ALIGN=LEFT>Released:</TH>" + "<TD>" + version.released + "</TD></TR>" +
            "<TR><TH ALIGN=LEFT>Label:</TH>" + "<TD><INPUT TYPE=text NAME=\"label\" SIZE=20 VALUE=\"" +
            encodeHTML(version.sLabel) + "\"></TD></TR>" + "<TR><TH ALIGN=LEFT>Comment:</TH>" +
            "<TD><INPUT TYPE=text NAME=\"comment\" SIZE=40 VALUE=\"" + encodeHTML(version.sComment) + "\"></TD></TR></TABLE>");
        if (person.status != Person.isUser) {
            con.print("<P><INPUT TYPE=submit NAME=\"action\" VALUE=\"Update\">");
            if (person.status == Person.isAdministrator) {
                con.print("&nbsp;<INPUT TYPE=submit NAME=\"action\" VALUE=\"Remove\">");
            }
            con.print("&nbsp;<INPUT TYPE=reset>");
        }
        con.print("<P></FORM><B>Bugs:</B><BR><FORM METHOD=POST ACTION=\"" + con.getURI() +
            "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"bugForm\">" + "<INPUT TYPE=hidden NAME=\"myself\" VALUE=\"" + myself +
            "\">" + "<SELECT NAME=\"bug\" SIZE=5>");
        if (version.setBugs.length != 0) {
            print(con, version.setBugs);
            con.print("</SELECT><BR><INPUT TYPE=submit NAME=\"action\" VALUE=\"Select\">");
        } else {
            con.print(EMPTY_LIST);
        }
        con.print("</FORM>");
        mainMenuReference(con);
    }

    public void updateVersion(HttpConnection con) throws Exception {
        storage.lock();
        String softOID = con.get("software");
        Software product = (Software)storage.getObjectByOid(Long.parseLong(softOID, 10));
        product.loadComponents();
        String versionStr = con.get("version");
        int majorVersion, minorVersion;
        try {
            int p = versionStr.indexOf('.');
            majorVersion = Integer.parseInt(versionStr.substring(0, p), 10);
            minorVersion = Integer.parseInt(versionStr.substring(p + 1), 10);
        } catch (Exception x) {
            error(con, "Bad version number (MAJOR.MINOR expected)");
            return;
        }
        Version version, prev = null;
        for (version = product.pVersions; version != null; prev = version, version = version.pNext) {
            if (version.majorVersionNumber == majorVersion && version.minorVersionNumber == minorVersion) {
                break;
            }
        }
        if (version == null) {
            error(con, "No such version");
            return;
        }
        if (con.get("action").equals("Remove")) {
            if (version.setBugs.length != 0) {
                error(con, "Can not remove version with non-empty bugs list");
                return;
            }
            if (prev == null) {
                product.pVersions = version.pNext;
                product.store();
            } else {
                prev.pNext = version.pNext;
                prev.store();
            }
            con.addPair("action", "Select");
            softwareForm(con);
            return;
        }
        version.sComment = con.get("comment");
        version.sLabel = con.get("label");
        version.store();
        versionForm(con);
    }

    public void loginForm(HttpConnection con) throws IOException {
        con.print(HTML_HEAD + "<TITLE>Bug Tracking Database</TITLE></HEAD>" + BODY + "<H1>Bug Tracking Database</H1>" +
            "<FORM METHOD=POST ACTION=\"" + con.getURI() + "\"><INPUT TYPE=hidden NAME=\"page\" VALUE=\"login\">" +
            "<TABLE><TR><TH ALIGN=LEFT>User:</TD>" + "<TD><INPUT TYPE=text NAME=\"name\" SIZE=32></TD></TR>" +
            "<TR><TH ALIGN=LEFT>Password:</TD>" + "<TD><INPUT TYPE=password NAME=\"password\"></TD></TR></TABLE><P>" +
            "<INPUT TYPE=submit VALUE=\"Login\"></FORM></BODY></HTML>");
    }

    public void doGet(HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException {
        doGetOrPost(request, response);
    }

    public void doPost(HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException {
        doGetOrPost(request, response);
    }

    void doGetOrPost(HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException {
        response.setContentType("text/html");
        HttpConnection con = new HttpConnection(request, response);
        String page = request.getParameter("page");
        if (page == null) {
            loginForm(con);
        } else {
            try {
                Method method = self.getMethod(page, methodParamTypes);
                Object methodParams[] = new Object[1];
                methodParams[0] = con;
                try {
                    method.invoke(this, methodParams);
                } catch (InvocationTargetException x) {
                    Throwable tx = x.getTargetException();
                    tx.printStackTrace();
                    error(con, "Exception " + tx.getClass().getName() + ": " + tx.getMessage());
                } catch (Exception x) {
                    error(con, "Exception " + x.getClass().getName() + ": " + x.getMessage());
                }
                try {
                    db.commit();
                    if (storage.getObjectCacheSize() > objectCacheSizeLimit) {
                        System.out.println("Cleaning object cache...");
                        storage.clearObjectCache();
                    }
                } catch (SQLException x) {
                    System.err.println("Failed to commit transaction");
                }
            } catch (NoSuchMethodException x) {
                error(con, "No such method: " + page);
            }
        }
        con.send();
    }

    public void init() throws ServletException {
        try {
            String driverName = System.getProperty("database.driver", "org.garret.consus.ConsusDriver");
            String databaseURL = System.getProperty("database.url", "jdbc:consus:bugdb.dbs");
            System.out.println("driverName=" + driverName + ", databaseURL=" + databaseURL);
            Class.forName(driverName);
            db = DriverManager.getConnection(databaseURL);
            Statement stmt = db.createStatement();
            if (db instanceof PersistentObjectStorage) {
                storage = (PersistentObjectStorage)db;
            } else {
                storage = new JDBCObjectRelationalAdapter(db);
            }
            if (!storage.isInitialized()) {
                DatabaseMetaData meta = db.getMetaData();
                ResultSet cursor = meta.getTypeInfo();
                String oidType = "long";
                String binType = "varbinary";
                String strType = "varchar";
                String intType = "integer";
                while (cursor.next()) {
                    String name = cursor.getString("TYPE_NAME");
                    switch (cursor.getShort("DATA_TYPE")) {
                        case Types.INTEGER:
                            intType = name;
                            break;
                        case Types.BIGINT:
                            oidType = name;
                            break;
                        case Types.VARCHAR:
                            strType = name;
                            break;
                        case Types.VARBINARY:
                            binType = name;
                            break;
                    }
                }
                stmt.executeUpdate("create domain ref " + oidType);
                stmt.executeUpdate("create domain seq " + binType);
                stmt.executeUpdate("create domain str " + strType);
                stmt.executeUpdate("create domain num " + intType);
                stmt.executeUpdate("create table Bug (RID num, TID num, sOneLineSummary str, sHardwarePlatform str, sOperatingSystem str, eCategory num, eFixingPriority num, eSeverity num, nReports num, pReportedBy ref, pAssignedTo ref, pSoftware ref, pVersion ref, pReportHistory ref, pWorkArounds ref, setSimilarBugs seq)");
                stmt.executeUpdate("create table Person (RID num, TID num, sName str primary key, sEmailAddress str, sPassword str, setUsedSoftware seq, setProjects seq, setReportedBugs seq, setAssignedBugs seq, status num)");
                stmt.executeUpdate("create table Report (RID num, TID num, sDescription str, pAuthor ref, pNext ref, creationDate date, status num, reportId num)");
                stmt.executeUpdate("create table Software (RID num, TID num, sName str primary key, setEngineers seq, setUsers  seq, pVersions ref, setBugs seq)");
                stmt.executeUpdate("create table Version (RID num, TID num, sLabel str, majorVersionNumber num, minorVersionNumber num, released date, sComment str, setBugs seq, pNext ref)");
                stmt.executeUpdate("create table TypeInfo (TID num, TableName str, ClassName str)");
                stmt.executeUpdate("create table OidTable (ID ref)");
                stmt.executeUpdate("insert into TypeInfo values (1, 'Bug', 'Bug')");
                stmt.executeUpdate("insert into TypeInfo values (2, 'Person', 'Person')");
                stmt.executeUpdate("insert into TypeInfo values (3, 'Report', 'Report')");
                stmt.executeUpdate("insert into TypeInfo values (4, 'Software', 'Software')");
                stmt.executeUpdate("insert into TypeInfo values (5, 'Version', 'Version')");
                stmt.executeUpdate("insert into OidTable values (1)");
            }
            storage.open();
            db.commit();
            self = getClass();
            methodParamTypes = new Class[1];
            methodParamTypes[0] = Class.forName("HttpConnection");
            TableIterator iterator = storage.getTableIterator("Person");
            if (iterator.getNumberOfRows() == 0) {
                System.out.println("Create administrator");
                Person person = new Person();
                person.sName = "administrator";
                person.sEmailAddress = "";
                person.sPassword = "";
                person.status = Person.isAdministrator;
                person.setUsedSoftware = new Software[0];
                person.setProjects = new Software[0];
                person.setReportedBugs = new Bug[0];
                person.setAssignedBugs = new Bug[0];
                person.store(storage);
            }
            iterator.close();
            db.commit();
            stmt.close();
        } catch (Exception x) {
            x.printStackTrace();
            throw new ServletException(x.getMessage());
        }
    }

    public void destroy() {
        try {
            db.close();
        } catch (SQLException x) { }
    }
}
