import java.awt.*;
import org.garret.consus.*;

/**
 * This class represents the drawing to be displayed by Graph Editor. It contains the figures that are kept in a linked list.
 * This class extends class <code>org.garret.consus.Persistent</code>, and therefore can be automatically
 * stored in a table by Consus.
 * @see org.garret.consus.Persistent
 * @see Figure
 */
public class Drawing extends Persistent {
    public String name;
    public Figure list;

    public String toString() {
        return name;
    }

    /**
     * Adds a figure to the drawing. The new figure goes to the head of the list. The drawing is then
     * automatically stored in the table.
     */
    public void add(Figure f) {
        f.next = list;
        list = f;
        store();
    }

    /**
     * Draws the drawing onto the screen. This is done by enumerating all the figures in the drawing and calling
     * the <code>paint</code> method from each of them.
     */
    public void paint(Graphics graphics) {
        for (Figure f = list; f != null; f = f.next) {
            f.paint(graphics);
        }
    }

    /** Creates a new drawing and assigns the name. */
    public Drawing(String title) {
        name = title;
    }

    /** Default constructor to be used by Consus. */
    public Drawing() { } // constructor for database
}
