import java.awt.*;
import org.garret.consus.*;

/**
 * The abstract class, which is the parent to the graphical primitives that are represented in a drawing. Figures are linked in a
 * one-direction list. Since this class extends <code>org.garret.consus.Persistent</code>, it can be automatically
 * stored in a database table by Consus.
 */
public abstract class Figure extends Persistent implements Cloneable {
    public Figure next;

    /** Draws the figure onto the screen . */
    abstract public void paint(Graphics graphics);

    /** Sets the bounding box of the figure. */
    abstract public void set(int x1, int y1, int x2, int y2);

    /**
     * Overwrite <code>Object.clone</code> method to make sure only the "last" figure in the list is cloned. Graph Editor will
     * clone a figure being created, before adding it to the drawing.
     */
    public Object clone() {
        try {
            Assert.that(next == null);
            return super.clone();
        } catch (CloneNotSupportedException x) {
            return null;
        }
    }
}
