import org.garret.consus.*;
import java.sql.*;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Properties;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

public final class WebSQL extends HttpServlet {
    public void init() throws ServletException {
        try {
            Class.forName("org.garret.consus.ConsusDriver");
            String database = System.getProperty("database");
            if (database == null) {
                throw new ServletException("Database was not specified, use -Ddatabase option");
            }
            db = (ConsusConnection)DriverManager.getConnection("jdbc:consus:" + database);
            metaDB = db.getMetaData();
        } catch (Exception x) {
            throw new ServletException(x.getMessage());
        }
    }

    public void destroy() {
        try {
            db.close();
        } catch (SQLException x) { }
    }

    static String encodeHTML(String str) {
        StringBuffer buf = new StringBuffer();
        for (int i = 0, n = str.length(); i < n; i++) {
            char ch = str.charAt(i);
            switch (ch) {
                case '<':
                    buf.append("&lt;");
                    break;
                case '>':
                    buf.append("&gt;");
                    break;
                case '&':
                    buf.append("&amp;");
                    break;
                case '"':
                    buf.append("&quot;");
                    break;
                default:
                    buf.append(ch);
            }
        }
        return buf.toString();
    }

    void printForm(HttpServletRequest request, PrintWriter writer) throws SQLException {
        writer.println("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\"><HTML>");
        writer.println("<HEAD><TITLE>WebSQL</TITLE></HEAD><BODY bgcolor=\"#c0c0c0\">");
        writer.println("<FORM METHOD=POST ACTION=\"" + request.getRequestURI() + "\">");
        writer.println("Tables:&nbsp;&nbsp;&nbsp;<SELECT SIZE=1 name=\"table\">");
        ResultSet cursor = metaDB.getTables(null, "%", "%", null);
        String currTable = request.getParameter("table");
        while (cursor.next()) {
            String table = cursor.getString("TABLE_NAME");
            if (table.equals(currTable)) {
                writer.print("<OPTION SELECTED VALUE=\"");
            } else {
                writer.print("<OPTION VALUE=\"");
            }
            writer.print(table);
            writer.print("\">");
            writer.print(table);
            writer.println("</OPTIONS>");
        }
        cursor.close();
        writer.println("</SELECT><P>");
        writer.println("Query:<BR><TEXTAREA COLS=50 ROWS=2 NAME=\"query\">");
        String currQuery = request.getParameter("query");
        if (currQuery != null) {
            writer.print(encodeHTML(currQuery));
        }
        writer.println("</TEXTAREA><P>");
        writer.println("<INPUT TYPE=submit NAME=\"action\" VALUE=\"Select\">&nbsp;&nbsp;&nbsp;<INPUT TYPE=submit NAME=\"action\" VALUE=\"Select All\">&nbsp;&nbsp;&nbsp;<INPUT TYPE=submit NAME=\"action\" VALUE=\"Query by example\">&nbsp;&nbsp;&nbsp;<INPUT TYPE=submit NAME=\"action\" VALUE=\"Insert\">");
        writer.println("<P><HR><P>");
    }

    void select(PrintWriter writer, String query) throws SQLException {
        int i;
        Statement stmt = db.createStatement();
        if (stmt.execute(query)) {
            ResultSet cursor = stmt.getResultSet();
            ResultSetMetaData metaTable = cursor.getMetaData();
            int nColumns = metaTable.getColumnCount();
            writer.print("<TABLE BORDER><tr>");
            for (i = 1; i <= nColumns; i++) {
                if (metaTable.getColumnType(i) != Types.ARRAY) {
                    String columnName = metaTable.getColumnLabel(i);
                    writer.print("<th>");
                    writer.print(columnName == null ? "?" : columnName);
                    writer.print("</th>");
                }
            }
            writer.println("</tr>");
            int nRows = 0;
            while (cursor.next()) {
                writer.print("<tr>");
                for (i = 1; i <= nColumns; i++) {
                    String val;
                    switch (metaTable.getColumnType(i)) {
                        case Types.BLOB:
                            val = "<BLOB " + cursor.getBlob(i).length() + " bytes>";
                            break;
                        case Types.CLOB:
                            val = "<CLOB " + cursor.getClob(i).length() + " chars>";
                            break;
                        case Types.ARRAY:
                            continue;
                        default:
                            try {
                                val = cursor.getString(i);
                            } catch (Throwable x) {
                                val = "???";
                            }
                    }
                    writer.println("<td>");
                    writer.print(encodeHTML(val));
                    writer.println("</td>");
                }
                writer.println("</tr>");
                nRows += 1;
            }
            writer.println("</table><P>Total " + nRows + " rows");
        } else {
            writer.print("Update " + stmt.getUpdateCount() + " rows");
        }
        writer.close();
        stmt.close();
    }

    public void doGet(HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException {
        try {
            PrintWriter writer = response.getWriter();
            response.setContentType("text/html");
            response.setStatus(HttpServletResponse.SC_OK);
            printForm(request, writer);
            writer.print("</FORM></BODY</HTML>");
            writer.close();
            db.commit();
        } catch (Exception x) {
            response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
        }
    }

    public void queryForm(PrintWriter writer, String table, String action) throws SQLException {
        ResultSet cursor = metaDB.getColumns(null, "%", table, "%");
        writer.println("<input type=hidden name=\"callback\"><table>");
        while (cursor.next()) {
            String name = cursor.getString("COLUMN_NAME");
            writer.print("<tr><td>");
            writer.print(name);
            writer.print("</td><td><input type=text size=30 name=\"");
            writer.print(name);
            writer.print("\"></td></tr>");
        }
        writer.print("</table><p><input type=submit name=\"action\" value=\"");
        writer.print(action);
        writer.print("\">");
        writer.close();
        cursor.close();
    }

    String queryByExample(HttpServletRequest request, PrintWriter writer) throws SQLException {
        String table = request.getParameter("table");
        if (table == null) {
            writer.print("Table was not selected");
            return null;
        }
        ResultSet cursor = metaDB.getColumns(null, "%", table, "%");
        String glue = " where ";
        StringBuffer sql = new StringBuffer();
        sql.append("select * from ");
        sql.append(table);
        while (cursor.next()) {
            String name = cursor.getString("COLUMN_NAME");
            String value = request.getParameter(name);
            if (value != null && value.trim().length() != 0) {
                sql.append(glue);
                sql.append(name);
                if (value.indexOf('%') >= 0) {
                    sql.append(" like '");
                } else {
                    sql.append("='");
                }
                sql.append(encodeHTML(value));
                sql.append('\'');
                glue = " and ";
            }
        }
        cursor.close();
        return sql.toString();
    }

    int insert(String table, HttpServletRequest request) throws SQLException {
        ResultSet cursor = metaDB.getColumns(null, "%", table, "%");
        StringBuffer columns = new StringBuffer();
        StringBuffer values = new StringBuffer();
        String sep = " (";
        while (cursor.next()) {
            String name = cursor.getString("COLUMN_NAME");
            String value = request.getParameter(name);
            if (value != null) {
                columns.append(sep);
                values.append(sep);
                sep = ",";
                columns.append(name);
                values.append('\'');
                values.append(encodeHTML(value));
                values.append('\'');
            }
        }
        cursor.close();
        columns.append(") values");
        values.append(')');
        String sql = "insert into " + table + columns.toString() + values.toString();
        Statement stmt = db.createStatement();
        return stmt.executeUpdate(sql);
    }

    public void doPost(HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException {
        PrintWriter writer = response.getWriter();
        try {
            try {
                response.setContentType("text/html");
                response.setStatus(HttpServletResponse.SC_OK);
                printForm(request, writer);
                String action = request.getParameter("action");
                if (action.equals("Select")) {
                    String query;
                    if (request.getParameter("callback") != null) {
                        query = queryByExample(request, writer);
                    } else {
                        query = request.getParameter("query");
                        if (query == null) {
                            writer.println("Empty query");
                            return;
                        }
                    }
                    if (query != null) {
                        select(writer, query);
                    }
                } else if (action.equals("Select All")) {
                    String table = request.getParameter("table");
                    if (table == null) {
                        writer.println("Table was not selected");
                    }
                    select(writer, "select * from " + table);
                } else if (action.equals("Insert")) {
                    String table = request.getParameter("table");
                    if (table == null) {
                        writer.println("Table was not selected");
                        return;
                    }
                    if (request.getParameter("callback") != null) {
                        writer.print("Insert " + insert(table, request) + " record");
                    } else {
                        queryForm(writer, table, "Insert");
                    }
                } else if (action.equals("Query by example")) {
                    String table = request.getParameter("table");
                    if (table == null) {
                        writer.println("Table was not selected");
                        return;
                    }
                    queryForm(writer, table, "Select");
                } else {
                    response.setStatus(HttpServletResponse.SC_BAD_REQUEST);
                }
                db.commit();
            } catch (Exception x) {
                try {
                    db.rollback();
                } catch (SQLException e) { }
                writer.println(x.getMessage());
            }
        } finally {
            writer.println("</FORM></BODY></HTML>");
        }
    }

    DatabaseMetaData metaDB;
    Connection db;
}
