/* Consus - Pure Java Object Relational Database Management System
 * Protocol for authentication of logins from remote clients
 * Created by Konstantin Knizhnik, Apr 2000.
 */

package org.garret.consus;

import java.util.Properties;
import java.sql.SQLException;

/** 
 * Authenticator interface 
 */
public interface ConsusAuthenticator { 
    /** 
     * Performs client authentication. Information about the user is passed through the 
     * <code>Properties</code> structure. Usually it contains "<code>user</code>" and 
     * "<code>password</code>" properties. Consus server adds the "<code>host</code>"
     * property where it stores the remote Inet address of the connected socket.
     * This structure is specified while establishing connection with the server by the
     * <code>Driver.connect(String url, java.util.Properties info)</code> method.
     * 
     * @param conxn connection with the database
     * @param userInfo user login properties
     * @return <code>true</code> if client is authenticated (connection is accepted) 
     * or <code>false</code> if connection is refused.
     */
    boolean authenticate(ConsusConnection conxn, Properties userInfo) throws SQLException;

    /**
     * Add a new user.
     * 
     * @param conxn connection with the database
     * @param userInfo user properties
     * @return <code>true</code> if user is successfully added, <code>false</code> otherwise.
     */
    boolean add(ConsusConnection conxn, Properties userInfo) throws SQLException; 

    /** 
     * Remove user.
     * 
     * @param conxn connection with the database
     * @param userInfo user properties
     * @return <code>true</code> if information about user is removed,
     * <code>false</code> if there is no such user.
     */     
    boolean remove(ConsusConnection conxn, Properties userInfo) throws SQLException; 

    /**
     * Change users properties. Updates the properties used for authentication, other than user name or ID.
     * 
     * @param conxn connection with the database
     * @param userInfo new user properties (specify user name or ID of an existing user)
     */
    boolean change(ConsusConnection conxn, Properties userInfo) throws SQLException;
}








