/* Consus - Pure Java Object Relational Database Management System
 * Statement protocol implementation
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.consus;

import java.sql.*;

/**
 * <P>This is the object used for executing a static SQL statement
 * and obtaining the results produced by it. 
 *
 * <P>Only one ResultSet per Statement can be opened at any point in
 * time. Therefore, if the reading of one ResultSet is interleaved
 * with the reading of another, each must have been generated by
 * different Statements. All statement <code>execute</code> methods implicitly
 * close a statment's current ResultSet if an open one exists.
 *
 * @see Connection#createStatement
 * @see ResultSet 
 */
public interface ConsusStatement extends Statement {      
    /**
     * Insert the object into the table. This method is equivalent to
     *     executeUpdate("insert in XXX ?"); setObject(1, obj);
     *
     * @param obj object to be inserted
     * @return reference to the created object
     */
    Ref insert(Object obj) throws SQLException;

    /**
     * Get of reference of the last selected, updated or inserted object.
     * 
     * @return reference to the last manipulated object
     */
    Ref getLastRef(); 

    /**
     * Update the object in the table.
     *
     * @param ref reference to the updated object
     * @param obj object with new values 
     */
    void update(Ref ref, Object obj) throws SQLException;

    /**
     * Get the object by it's OID
     *
     * @param ref reference to the object
     * @return reference to the fetched object
     */
    Object get(Ref ref) throws SQLException;

    /**
     * Remove the object with the specifed OID
     *
     * @param ref reference to the removed object
     */
    void remove(Ref ref) throws SQLException;
} 