/* Consus - Pure Java Object Relational Database Management System
 * Import data from another JDBC database, DBF file or from standard input
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.consus;

import java.sql.*;
import java.io.*;
import java.util.Properties;
import java.util.Hashtable;

class TableWithReferences { 
    TableWithReferences next;
    String              name;
    int                 columnType[];
};

public class Loader {  
    static String tables[] = {"TABLE"};

    static Connection connect(String login) throws Exception { 
	int p, q;
	Properties info = new Properties();
	if ((p = login.indexOf('@')) > 0) { 
	    q = login.lastIndexOf(':', p);
	    if (q >= 0) { 
		info.put("user", login.substring(0, q));
		info.put("password", login.substring(q+1, p));
	    } else { 
		info.put("user", login.substring(0, p));
	    }
	} else { 
	    p = -1;
	}
	q = login.indexOf(':', p+1);
	Connection connection;
	if (q < 0) { 
	    Class.forName("org.garret.consus.ConsusDriver");
	    connection = DriverManager.getConnection("jdbc:consus:" + login);
	} else { 
	    Class.forName(login.substring(p+1, q));
	    connection = DriverManager.getConnection(login.substring(q+1), info);
	}
	if (connection == null) { 
	    throw new IllegalArgumentException("Failed to connect to " + login);
	}
	return connection;
    }


    static void loadTXT(Connection conxn, String TXTFileName, String tableName) 
	throws Exception 
    { 
	StreamTokenizer in = 
	    new StreamTokenizer(new BufferedReader(new FileReader(TXTFileName)));
	in.resetSyntax();
	in.wordChars('a', 'z');
	in.wordChars('A', 'Z');
	in.wordChars(128 + 32, 255);
	in.wordChars('0', '9');
	in.wordChars('_', '_');
	in.wordChars('$', '$');
	in.wordChars('-', '-');
	in.wordChars('+', '+');
	in.wordChars('.', '.');
	in.whitespaceChars(0, ' ');
	in.quoteChar('"');
	in.quoteChar('\'');
	in.eolIsSignificant(true);	
	in.whitespaceChars(',', ',');
	String columnList = "";
	String parameterList = "";
	String sep = "(";
	int nColumns = 0;
	String columnName[] = new String[32];
	int i;

      scanColumnList:
	while (true) { 
	    switch (in.nextToken()) { 
	      case StreamTokenizer.TT_WORD:
		if (nColumns == columnName.length) { 
		    String[] newColumns = new String[nColumns*2];
		    System.arraycopy(columnName, 0, newColumns, 0, nColumns);
		    columnName = newColumns;
		}
		columnList += sep + in.sval;
		parameterList += sep + "?";
		columnName[nColumns++] = in.sval;
		sep = ",";
		break;
	      case StreamTokenizer.TT_EOL:
		columnList += ")";	
		parameterList += ")";
		break scanColumnList;
	      default:
		System.out.println("First line should contain list of the columns");
		conxn.close();
		return;
	    }
	}
	DatabaseMetaData meta = conxn.getMetaData();
	int columnType[] = new int[nColumns];
	for (i = 0; i < nColumns; i++) { 
	    ResultSet cursor = meta.getColumns(null, "%", tableName, columnName[i]);
	    if (!cursor.next()) { 
		System.out.println("No such column " + columnName[i] + " in table " + tableName);
		conxn.close();
		return;
	    }
	    columnType[i] = cursor.getShort("DATA_TYPE");
	    cursor.close();
	}
	PreparedStatement pstmt = conxn.prepareStatement("insert into " + tableName + columnList 
							 + " values " + parameterList);
	int nTuples = 0;
	while (in.nextToken() != StreamTokenizer.TT_EOF) { 
	    for (i = 1; i <= nColumns; i++) { 
		switch (in.ttype) { 
		  case StreamTokenizer.TT_WORD:
		  case '\'':
		  case '"':
		    switch (columnType[i-1]) { 
		      case Types.DATE:
			pstmt.setDate(i, Date.valueOf(in.sval));
			break;
		      case Types.TIME:
			pstmt.setTime(i, Time.valueOf(in.sval));
			break;
		      case Types.TIMESTAMP:
			pstmt.setTimestamp(i, Timestamp.valueOf(in.sval));
			break;
		      default:
			pstmt.setString(i, in.sval);
		    }
		    break;
		  default:
		    System.out.println("Column value expected");
		    return;
		}
		in.nextToken();
	    }
	    if (in.ttype != StreamTokenizer.TT_EOL) {
		System.out.println("Ignoring rest of line " + in.lineno());
		while (in.nextToken() != StreamTokenizer.TT_EOL
		       && in.ttype != StreamTokenizer.TT_EOF);
	    }
	    if (pstmt.executeUpdate() != 1) {
		System.out.println("Column value expected");
		return;
	    } else { 
		nTuples += 1;
	    }
	}
	System.out.println("Load " + nTuples + " records to table" + tableName);
	pstmt.close();
	conxn.close();
    }

    static int loadInt(byte b[], int p) { 
	return (b[p+3] << 24) | ((b[p+2] & 0xFF) << 16) | ((b[p+1] & 0xFF) << 8) | (b[p] & 0xFF);
    }
    static int loadWord(byte b[], int p) { 
	return ((b[p+1] & 0xFF) << 8) | (b[p] & 0xFF);
    }

    static void loadDBF(Connection conxn, String DBFFileName, String tableName) 
	throws Exception 
    {
	FileInputStream in = new FileInputStream(DBFFileName);
	int i, j, k, n;
	byte hdr[] = new byte[32];
	in.read(hdr);
	int nRecords = loadInt(hdr, 4);
	int hdrLen = loadWord(hdr, 8);
	int recLen = loadWord(hdr, 10);
	int nFields = (hdrLen - 32) >>> 5;
	byte fields[] = new byte[hdrLen - hdr.length];
	in.read(fields);
	char fieldType[] = new char[nFields];
	int  fieldSize[] = new int[nFields];
	int  fieldPrec[] = new int[nFields];
	String fieldsList = "";
	String paramsList = "";
	String createStmt = "create table " + tableName;
	String sep = "(";
	for (i = 0, j = 0; i < nFields; i += 1, j += 32) { 
	    String fieldName = new String(fields, j, 10).trim();
	    fieldSize[i] = fields[j+16] & 0xFF;
	    fieldPrec[i] = fields[j+17];
	    fieldType[i] = (char)((fields[j+11] & 0xFF) & ~('a' - 'A'));
	    String typeName = null;
	    switch (fieldType[i]) { 
	      case 'C':
		typeName = "varchar";
		break;
	      case 'D':
		typeName = "date";
		break;
	      case 'L':
		typeName = "boolean";
		break;
	      case 'N':
		typeName = (fieldPrec[i] <= 0) ? "bigint" : "double";
		break;
	      case 'I':
		typeName = "integer";
		break;
	      default:
		System.out.println("Ignored field of type '" + fieldType[i] + "'");
		continue;
	    }
	    fieldsList += sep + fieldName;
	    paramsList += sep + "?";
	    createStmt += sep + fieldName + " " + typeName;
	    if (fields[j+31] != 0) { // index exdists 
		createStmt += " using index";
		System.out.println("Will create index on " + tableName + "("
				   + fieldName + ")");
	    }
	    sep = ",";
	}
	createStmt += ")";
	paramsList += ")";
	fieldsList += ")";
	byte rec[] = new byte[recLen];
	DatabaseMetaData meta = conxn.getMetaData();
	ResultSet cursor = meta.getTables(null, "%", tableName, tables);
	if (!cursor.next()) { 
	    Statement stmt = conxn.createStatement();
	    System.out.println("Create table " + tableName);
	    stmt.executeUpdate(createStmt);
	    stmt.close();
	}
	cursor.close();
	PreparedStatement pstmt = conxn.prepareStatement("insert into " + tableName + fieldsList 
							 + " values " + paramsList);
	int nInserted = 0, nAvailable = 0;
	while (--nRecords >= 0) { 
	    in.read(rec);
	    if (rec[0] != '*') { 		
		int tmp;
		String str;
		nAvailable += 1;
		for (i = 0, j = 0, k = 1; i < nFields; i++) { 
		    switch (fieldType[i]) { 
		      case 'C':
			tmp = fieldSize[i] + ((fieldPrec[i] & 0xFF) << 8);
			pstmt.setString(++j, new String(rec, k, tmp).trim());
			k += tmp;
			break;
		      case 'N':
			str = new String(rec, k, fieldSize[i]).trim();
			if (str.length() == 0) { 
			    str = "0";
			}
			pstmt.setString(++j, str);
			k += fieldSize[i];
			break;
		      case 'L':
			tmp = rec[k++] & ~('a' - 'A');
			pstmt.setBoolean(++j, tmp == 'Y' || tmp == 'T');
			break;
		      case 'I':
			pstmt.setInt(++j, loadInt(rec, k));
			k += 4;
			break;
		      case 'D':	
			str = new String(rec, k, 8).trim();
			if (str.length() != 0) { 
			    pstmt.setDate(++j, Date.valueOf(str.substring(0, 4) + "-" + 
							    str.substring(4, 6) + "-" +
							    str.substring(6, 8)));
			} else { 
			    pstmt.setDate(++j, new Date(0));
			}
			k += 8;
			break;
		      case 'M':
			k += 10;
			break;
		      default:
			k += fieldSize[i];
		    }
		}
		nInserted += pstmt.executeUpdate();
	    }
	}
	System.out.println("Insert " + nInserted + " records of " + nAvailable + " in DBF file");
	conxn.close();
	in.close();
    }
    
    public static void main(String args[]) throws Exception { 
	int i, p;
	if (args.length < 2) { 
	    System.out.println("Consus data loader\n" + 
			       "Usage:\n  java Loader dst-login src-login [table-pattern [scheme-pattern [catalog]]]\n" + 
			       "where\n  <login> is \"user:password@driver-class:URL\"");
	    return;
	}
	Connection dst = connect(args[0]);
	if (args[1].indexOf('@') < 0 && (i = args[1].length()) > 4) { 
	    if (args[1].substring(i-4).equalsIgnoreCase(".dbf")) { 
		String tableName;
		if (args.length >= 3) { 
		    tableName = args[2];
		} else {  
		    p = args[1].lastIndexOf(File.separatorChar);
		    tableName = args[1].substring(p+1, i-4);
		}
		loadDBF(dst, args[1], tableName);
		return;
	    } 
	    if (args[1].substring(i-4).equalsIgnoreCase(".txt")) { 
		String tableName;
		if (args.length >= 3) { 
		    tableName = args[2];
		} else {  
		    p = args[1].lastIndexOf(File.separatorChar);
		    tableName = args[1].substring(p+1, i-4);
		}
		loadTXT(dst, args[1], tableName);
		return;
	    }
	}
	Connection src = connect(args[1]);
	Hashtable refMap = null;
	TableWithReferences refTable = null;
	if (src instanceof ConsusConnection && dst instanceof ConsusConnection) {
	    refMap = new Hashtable();
	}
	DatabaseMetaData dstMeta = dst.getMetaData();
	DatabaseMetaData srcMeta = src.getMetaData();
	String tableNamePattern = args.length >= 3 ? args[2] : "%";
	String schemeNamePattern = args.length >= 4 ? args[3] : "%";
	String catalogName = args.length >= 5 ? args[4] : null;
	ResultSet srcCursor = 
	    srcMeta.getTables(catalogName, schemeNamePattern, tableNamePattern, tables);
	Statement stmt = src.createStatement();

	while (srcCursor.next()) { 
	    String tableName = srcCursor.getString("TABLE_NAME");
	    if (tableName.equals("org.garret.consus.MetaTable")) { 
		continue;
	    }
	    ResultSet cursor = stmt.executeQuery("select * from " + tableName);
	    ResultSetMetaData meta = cursor.getMetaData();
	    int nColumns = meta.getColumnCount();
	    String columnName[] = new String[nColumns];
	    int columnType[] = new int[nColumns];
	    String columnList = "";
	    String parameterList = "";
	    for (i = 0; i < nColumns; i++) { 
		if (i != 0) { 
		    columnList += ",";
		    parameterList += ",";
		}
		columnList += columnName[i] = meta.getColumnName(i+1);
		columnType[i] = meta.getColumnType(i+1);
		parameterList += "?";
	    }
	    ResultSet dstCursor = dstMeta.getTables(null, "%", tableName, tables);
	    if (!dstCursor.next()) { 
		// Table doesn't exist
		String sql = "create table " + tableName;
		String sep = "(";
		for (i = 0; i < nColumns; i++) { 
		    String type;
		    switch (columnType[i]) {
		      case Types.BIT:
			type = "bit";
			break;
		      case Types.TINYINT:
			type = "tinyint";
			break;
		      case Types.SMALLINT:
			type = "smallint";
			break;
		      case Types.INTEGER:
			type = "integer";
			break;
		      case Types.BIGINT:
			type = "bigint";
			break;
		      case Types.FLOAT:
			type = "float";
			break;
		      case Types.DOUBLE:
			type = "double";
			break;
		      case Types.REAL:
			type = "real";
			break;
		      case Types.VARCHAR:
		      case Types.LONGVARCHAR:
			type = "varchar";
			break;
		      case Types.CHAR:
			type = "char";
			break;
		      case Types.REF:
			type = "reference";
			if ((p = columnName[i].indexOf('$')) >= 0) {  
			    type += " to " + columnName[i].substring(p+1);
			}
			break;
		      case Types.DATE:
			type = "date";
			break;
		      case Types.TIME:
			type = "time";
			break;
		      case Types.TIMESTAMP:
			type = "timestamp";
			break;
		      case Types.BINARY:
			type = "binary";
			break;
		      case Types.VARBINARY:
		      case Types.LONGVARBINARY:
			type = "varbinary";
			break;
		      case Types.NUMERIC:
			type = "numeric";
			break;			
		      case Types.DECIMAL:
			type = "decimal";
			break;
		      case Types.BLOB:
			type = "blob";
			break;
		      case Types.CLOB:
			type = "clob";
			break;
		      default:
			System.out.println("Failed to import column of type " + columnType[i]);
			dst.close();
			src.close();
			return;
		    }
		    sql += sep + columnName[i] + " " + type;
		    sep = ",";
		}
		sql += ")";
		Statement createStmt = dst.createStatement();
		createStmt.executeUpdate(sql);
		System.out.println("Create table " + tableName);
		ResultSet indexCursor = srcMeta.getIndexInfo(srcCursor.getString("TABLE_CAT"),
							     srcCursor.getString("TABLE_SCHEM"),
							     tableName, false, true); 
		while (indexCursor.next()) { 
		    sql = "create index " + indexCursor.getString("INDEX_NAME") 
			+ " on " + tableName + "(" + indexCursor.getString("COLUMN_NAME") + ")";
		    createStmt.executeUpdate(sql);
		    System.out.println("Create index on field " 
				       + indexCursor.getString("COLUMN_NAME"));
		}
		indexCursor.close();
		createStmt.close();		    
	    }
	    PreparedStatement pstmt = dst.prepareStatement("insert into " + tableName
							   + " (" + columnList + ") values ("
							   + parameterList + ")");		
	    int srcCount = 0, dstCount = 0;
	    boolean importReferences = false;
	    while (cursor.next()) { 
		srcCount += 1;
		for (i = 0; i < nColumns;) { 
		    String type;
		    switch (columnType[i++]) {
		      case Types.BIT:
			pstmt.setBoolean(i, cursor.getBoolean(i));
			break;
		      case Types.TINYINT:
			pstmt.setByte(i, cursor.getByte(i));
			break;
		      case Types.SMALLINT:
			pstmt.setShort(i, cursor.getShort(i));
			break;
		      case Types.INTEGER:
			pstmt.setInt(i, cursor.getInt(i));
			break;
		      case Types.BIGINT:
			pstmt.setLong(i, cursor.getLong(i));
			break;
		      case Types.FLOAT:
			pstmt.setFloat(i, cursor.getFloat(i));
			break;
		      case Types.DOUBLE:
		      case Types.REAL:
			pstmt.setDouble(i, cursor.getDouble(i));
			break;
		      case Types.LONGVARCHAR:
		      case Types.VARCHAR:
		      case Types.CHAR:
			pstmt.setString(i, cursor.getString(i));
			break;
		      case Types.REF:
			pstmt.setRef(i, cursor.getRef(i));
			importReferences = true;
			break;
		      case Types.DATE:
			pstmt.setDate(i, cursor.getDate(i));
			break;
		      case Types.TIME:
			pstmt.setTime(i, cursor.getTime(i));
			break;
		      case Types.TIMESTAMP:
			pstmt.setTimestamp(i, cursor.getTimestamp(i));
			break;
		      case Types.BINARY:
		      case Types.VARBINARY:
		      case Types.LONGVARBINARY:
			pstmt.setBytes(i, cursor.getBytes(i));
			break;
		      case Types.BLOB:
			pstmt.setBlob(i, cursor.getBlob(i));
			break;
		      case Types.CLOB:
			pstmt.setClob(i, cursor.getClob(i));
			break;
		      default:
			pstmt.setObject(i, cursor.getObject(i));
			break;
		    }
		}
		dstCount += pstmt.executeUpdate();
		if (refMap != null) { 
		    refMap.put(cursor.getRef(0), 
			       ((ConsusPreparedStatement)pstmt).getLastRef());
		}
	    }
	    if (importReferences) { 
		TableWithReferences rtab = new TableWithReferences();
		rtab.name = tableName;
		rtab.columnType = columnType;
		rtab.next = refTable;
		refTable = rtab;
	    }
	    System.out.println("Import " + dstCount + " records from table " + tableName 
			       + "(" + srcCount + " records)");
	    cursor.close();
	    pstmt.close();
	}
	    
	if (refMap != null) { 
	    Statement select = dst.createStatement();
	    while (refTable != null) { 
		ResultSet cursor = select.executeQuery("select * from " + refTable.name 
						       + " for update");
		while (cursor.next()) { 
		    boolean updated = false;
		    for (i = refTable.columnType.length; --i >= 0;) { 
			if (refTable.columnType[i] == Types.REF) {
			    Ref ref = cursor.getRef(i+1);
			    if (ref != null) { 
				cursor.updateObject(i+1, refMap.get(ref));
				updated = true;
			    }
			}
		    }
		    if (updated) { 
			cursor.updateRow();
		    }
		}
		refTable = refTable.next;
	    }
	}
	src.close();
	dst.close();
	System.out.println("Import done");
    }
}



