/* Consus - Pure Java Object Relational Database Management System
 * Persistent Object Storage error
 * Created by Konstantin Knizhnik, Apr 2000.
 */
package org.garret.consus;

/**
 * Persistent Object Storage error. All exceptions thrown by Consus during operations with
 * persistent objects are caught and the <code>POSError</code> exception is rethrown.
 * The <code>POSError</code> exception is derived from the <code>RuntimeException</code> class,
 * so that it does not need to be caught or declared by methods working with persistent objects.
 */
public class POSError extends RuntimeException { 
    protected Exception ex;

    /**
     * Get the original exception.
     * @return SQLException for persistent storage error
     */
    public Exception reason() {
	return ex;
    }

    /**
     * Checks whether unique contraint violations were the reason for the persistent object storage error.
     * @return <code>true</code> if ConsusConstraintViolation exception was thrown
     */

    public boolean isUniqueConstraintViolated() { 
	return ex instanceof ConsusDuplicateException;
    }

    /**
     * Constructs the exception.
     * @param x original exception
     */
    public POSError(Exception x) { 
	super(x.getMessage());
	ex = x;
    }
}
