/* Consus - Pure Java Object Relational Database Management System
 * Remote Consus Server
 * Created by Konstantin Knizhnik, Apr 2000.
 */
package org.garret.consus;

import org.garret.internal.consus.*;

import java.io.*;
import java.sql.*;
import java.util.Properties;

public class Server { 
    static BufferedReader in = new BufferedReader(new InputStreamReader(System.in)); 

    static String input(String prompt) {
	while (true) { 
	    try { 
		System.out.print(prompt);
		String answer = in.readLine().trim();
		if (answer.length() != 0) {
		    return answer;
		}
	    } catch (IOException x) {}
	}
    }

    public static void main(String args[]) throws Exception { 
	Properties info;
	if (args.length < 2) { 
	    System.out.println("Usage: java Server <database-name> <port> [<authenticator-class>]");
	    return;
	}
	int port = Integer.parseInt(args[1]);
	Class.forName("org.garret.consus.ConsusDriver");
	ConsusConnection connection = 
	    (ConsusConnection)DriverManager.getConnection("jdbc:consus:" + args[0]);	
	if (connection == null) { 
	    System.err.println("No connection with database " + args[0]);
	    return;
	}
	ConsusAuthenticator authenticator = null;
	if (args.length >= 3) {
	    authenticator = (ConsusAuthenticator)Class.forName(args[2]).newInstance();
	}
	ServerManager localManager = null;
	ServerManager globalManager = null;
	globalManager = new ServerManager(connection, port, false, authenticator);	
	globalManager.start();
	try { 
	    localManager = new ServerManager(connection, port, true, authenticator);	
	    localManager.start();
	} catch (Throwable x) {}

	while (true) { 
	    String answer = input("> ");
	    if (answer.equalsIgnoreCase("exit")) { 
		break;
	    } else if (answer.equalsIgnoreCase("add")) { 
		System.out.println("Adding new user");
		info = new Properties();
		info.put("user", input("User: "));
		info.put("password", input("Password: "));
		if (authenticator != null) { 
		    if (authenticator.add(connection, info)) { 
			System.out.println("User added");
		    } else { 
			System.out.println("User already exists");
		    }
		    connection.commit();
		}
	    } else if (answer.equalsIgnoreCase("remove")) { 
		System.out.println("Removing user");
		info = new Properties();
		info.put("user", input("User: "));
		if (authenticator != null) { 
		    if (authenticator.remove(connection, info)) { 
			System.out.println("User removed");
		    } else { 
			System.out.println("No such user");
		    }
		    connection.commit();
		}
	    } else if (answer.equalsIgnoreCase("change")) { 
		System.out.println("Changing user password");
		info = new Properties();
		info.put("user", input("User: "));
		info.put("password", input("Password: "));
		if (authenticator != null) { 
		    if (authenticator.change(connection, info)) { 
			System.out.println("Password changed");
		    } else { 
			System.out.println("No such user");
		    }
		    connection.commit();
		}
	    } else { 
		System.out.println("Commands: add, remove, change, exit");
	    }
	}
	globalManager.terminate();
	if (localManager != null) { 
	    localManager.terminate();
	}
	connection.close();
    }
}

