/* Consus - Pure Java Object Relational Database Management System
 * Default implementation of the authentication interface based on the user login name and password
 * Created by Konstantin Knizhnik, Apr 2000.
 */

package org.garret.consus;

import java.util.Properties;
import java.sql.*;

/**
 * Basic implementation of the authentication interface based on the user login name and password.
 */
public class StdAuthenticator implements ConsusAuthenticator { 
    /**
     * ConsusAuthenticator constructor.
     * @param conxn dastabase conenction
     * @param userInfo user properties (shoud contain at least "user" and "password" properties)
     */
    public boolean authenticate(ConsusConnection conxn, Properties userInfo) throws SQLException 
    {
	PreparedStatement stmt
	    = conxn.prepareStatement("select * from consus.UserInfo where user=? and password=?");
	stmt.setString(1, userInfo.getProperty("user", "guest"));
	stmt.setString(2, userInfo.getProperty("password", ""));
	ResultSet cursor = stmt.executeQuery();
	boolean authenticationOk = cursor.next();
	cursor.close();
	stmt.close();
	return authenticationOk;
    }

    /**
     * Add a new user.
     * 
     * @param conxn connection with the database
     * @param userInfo user properties
     * @return <code>true</code> if the user is successfully added, <code>false</code> otherwise.
     */    
    public boolean add(ConsusConnection conxn, Properties userInfo) throws SQLException 
    {
	String user = userInfo.getProperty("user");
	if (user == null) { 
	    return false;
	}
	PreparedStatement stmt
	    = conxn.prepareStatement("select * from consus.UserInfo where user=? for update");
	stmt.setString(1, user);
	ResultSet cursor = stmt.executeQuery();
	boolean success = false;
	if (!cursor.next()) { 
	    cursor.moveToInsertRow();
	    cursor.updateString("user", user);
	    cursor.updateString("password", userInfo.getProperty("password", ""));
	    cursor.insertRow();
	    success = true;
	}
	cursor.close();
	stmt.close();
	return success;
    }

    /** 
     * Remove a user.
     * 
     * @param conxn connection with the database
     * @param userInfo user properties
     * @return <code>true</code> if the information about the user is removed,
     * <code>false</code> if there is no such user.
     */     
    public boolean remove(ConsusConnection conxn, Properties userInfo) throws SQLException 
    {
	String user = userInfo.getProperty("user");
	if (user == null) { 
	    return false;
	}
	PreparedStatement stmt
	    = conxn.prepareStatement("delete from consus.UserInfo where user=?");
	stmt.setString(1, user);
	boolean success = stmt.executeUpdate() == 1;
	stmt.close();
	return success;
    }

    /**
     * Change user properties.
     * 
     * @param conxn connection with the database
     * @param userInfo users new properties (user name and other properties needed to identify
     * user should not be changed. 
     */
    public boolean change(ConsusConnection conxn, Properties userInfo) throws SQLException 
    {
	String user = userInfo.getProperty("user");
	if (user == null) { 
	    return false;
	}
	PreparedStatement stmt
	    = conxn.prepareStatement("update consus.UserInfo set password=? where user=?");
	stmt.setString(1, userInfo.getProperty("password", ""));
	stmt.setString(2, user);
	boolean success = stmt.executeUpdate() == 1;
	stmt.close();
	return success;
    }
}



