/* Consus - Pure Java Object Relational Database Management System
 * Connection pool data source.
 * Created by Konstantin Knizhnik, Mar 2001.
 */
package org.garret.consusx;

import java.io.*;
import java.sql.*;
import javax.sql.*;


public class ConsusConnectionPoolDataSource extends ConsusDataSource
                                            implements ConnectionPoolDataSource 
{ 
    public PooledConnection getPooledConnection() throws SQLException 
    { 
	return getPooledConnection(user, password);
    }
    
    public synchronized PooledConnection getPooledConnection(String username, String password) throws SQLException
    {
	if (availableConnections == null) { 
	    return new ConsusPooledConnection(this, username, password);
	} else { 
	    PooledConnection conxn = availableConnections;
	    availableConnections = availableConnections.next;
	    return conxn;
	}
    }

    public Connection getConnection(String username, String password) throws SQLException
    {
	return getPooledConnection().getConnection();
    }

    synchronized void release(ConsusPooledConnection conxn) { 
	conxn.next = availableConnections;
	availableConnections = conxn;
    }

    transient ConsusPooledConnection availableConnections;
}





