/* Consus - Pure Java Object Relational Database Management System
 * Consus pooled connection implementation.
 * Created by Konstantin Knizhnik, Mar 2001.
 */
package org.garret.consusx;
import org.garret.consus.*;
import org.garret.internal.consus.*;

import java.io.*;
import java.sql.*;
import javax.sql.*;
import javax.naming.*;
import java.util.*;

interface PhysicalConnection extends ConsusTransaction { 
    public void closePhysical() throws SQLException;
};

public class ConsusPooledConnection implements PooledConnection {

    public void addConnectionEventListener(ConnectionEventListener listener)
    {
	listeners.add(listener);
    }

    public void close() throws SQLException
    {
	con.closePhysical();
    }


    void release() throws SQLException 
    { 
	Iterator i = listeners.iterator();
	ConnectionEvent event = new ConnectionEvent(this);
	while (i.hasNext()) { 
	    ConnectionEventListener listener = 
		(ConnectionEventListener)i.next();
	    listener.connectionClosed(event);
	}
	listeners.clear();
	pool.release(this);
    }

    class LocalConnectionHandle extends org.garret.internal.consus.ConsusConnection 
				implements PhysicalConnection 
    { 
	public void close() throws SQLException 
	{ 
	    commitTransaction();
	    release();
	}

	public void closePhysical() throws SQLException 
	{ 
	    super.close();
	}

	LocalConnectionHandle(String fileName, String url, Properties info, 
			      Driver driver) throws SQLException
	{
	    super(fileName, url, info, driver);
	}
    }

    class RemoteConnectionHandle extends ClientConnection
				 implements PhysicalConnection 
    { 
	public void close() throws SQLException 
	{ 
	    commitTransaction();
	    release();
	}

	public void closePhysical() throws SQLException 
	{ 
	    super.close();
	}

	RemoteConnectionHandle(String address, String url, Properties info, 
			       Driver driver) throws SQLException
	{
	    super(address, url, info, driver);
	}
    }

    public Connection getConnection() throws SQLException
    {
	return con;
    } 
	
    public void removeConnectionEventListener(ConnectionEventListener listener) 
    {
	listeners.remove(listener);
    }

    PhysicalConnection createLocalConnection(String fileName, String url, 
					     Properties info, Driver driver) throws SQLException
    { 
	return new LocalConnectionHandle(fileName, url, info, driver);
    }

    PhysicalConnection createRemoteConnection(String address, String url, 
					      Properties info, Driver driver) throws SQLException
    { 
	return new RemoteConnectionHandle(address, url, info, driver);
    }

    ConsusPooledConnection(ConsusConnectionPoolDataSource ds, String username, String password) 
	throws SQLException
    { 
	pool = ds;
	Properties props = new Properties();
	if (username != null) { 
	    props.put("user", username);
	}
	if (password != null) { 
	    props.put("password", password);
	}
	try { 
	    if (ds.getServerName() == null) {
		Driver driver = (Driver)Class.forName("org.garret.consus.ConsusDriver").newInstance();
		String url = "jdbc:consus:" + ds.getDatabaseName();
		con = createLocalConnection(ds.getDatabaseName(),
					    url, 
					    props,
					    driver);
	    } else { 
		Driver driver = (Driver)Class.forName("org.garret.consus.ClientDriver").newInstance();
		String address = ds.getServerName() + ":" + ds.getPortNumber();
		String url = "jdbc:consus-client:" + address;
		con = createRemoteConnection(address,
					     url, 
					     props,
					     driver);
	    }
	} catch (ClassNotFoundException x) { 
	    throw new SQLException("Driver class not found");
	} catch (InstantiationException x) { 
	    throw new SQLException("Failed to create driver instance");	    
	} catch (IllegalAccessException x) { 
	    throw new SQLException("Failed to create driver instance");	    
	}
	listeners = new TreeSet();
    }

    ConsusPooledConnection         next;
    ConsusConnectionPoolDataSource pool;
    PhysicalConnection             con;
    TreeSet                        listeners;
}






