/* Consus - Pure Java Object Relational Database Management System
 * Consus transactional connection implementation.
 * Created by Konstantin Knizhnik, Mar 2001.
 */
package org.garret.consusx;

import java.io.*;
import java.sql.*;
import javax.sql.*;
import org.garret.consus.*;
import org.garret.internal.consus.*;
import java.util.Properties;
import javax.transaction.xa.*;


public class ConsusXAConnection extends ConsusPooledConnection implements XAConnection 
{     
    class LocalXAConnectionHandle extends LocalConnectionHandle
    {
        public void commit() throws SQLException 
	{ 
	    throw new SQLException("Explicit commit is not possible in distributed transaction mode");
	}

        public void rollback() throws SQLException 
        { 
	    throw new SQLException("Explicit rollback is not possible in distributed transaction mode");
	}

        public void setAutoCommit(boolean autoCommit) throws SQLException 
	{ 
	    if (autoCommit) { 
		throw new SQLException("AutoCommit is not possible in distributed transaction mode");
	    }
	}

	LocalXAConnectionHandle(String fileName, String url, Properties info, 
				Driver driver) throws SQLException
	{
	    super(fileName, url, info, driver);
	}
    }


    class RemoteXAConnectionHandle extends RemoteConnectionHandle
    {
        public void commit() throws SQLException 
	{ 
	    throw new SQLException("Explicit commit is not possible in distributed transaction mode");
	}

        public void rollback() throws SQLException 
        { 
	    throw new SQLException("Explicit rollback is not possible in distributed transaction mode");
	}

        public void setAutoCommit(boolean autoCommit) throws SQLException 
	{ 
	    if (autoCommit) { 
		throw new SQLException("AutoCommit is not possible in distributed transaction mode");
	    }
	}

	RemoteXAConnectionHandle(String address, String url, Properties info, 
				 Driver driver) throws SQLException
	{
	    super(address, url, info, driver);
	}
    }

    class ConsusXAResource implements XAResource { 
        public void commit(Xid xid, boolean onePhase) throws XAException 
	{
	    try { 
		if (onePhase) { 
		    con.commitTransaction();
		} else { 
		    con.completeTransaction();
		}
	    } catch (SQLException x) { 
		throw new XAException(XAException.XAER_RMERR);
	    }
	}  

        public void end(Xid xid, int flags) throws XAException 
	{
	}  

	public void forget(Xid xid) throws XAException 
	{
	}  

	public int getTransactionTimeout() throws XAException 
	{
	    return 0;
	}  
	
	public boolean isSameRM(XAResource xares) throws XAException 
	{
	    return xares instanceof ConsusXAResource;
	}  
	
        public int prepare(Xid xid) throws XAException 
	{
	    try { 
		con.prepareTransaction(xid.getFormatId(), 
				       xid.getGlobalTransactionId(), 
				       xid.getBranchQualifier());
		return XA_OK;
	    } catch (SQLException x) { 
		throw new XAException(XAException.XAER_RMERR);
	    }
	}  
	
	public Xid[] recover(int flag) throws XAException 
	{
	    try { 
		ConsusTransId tid = con.getTransId();
		if (tid.getGlobalTransactionId() != null) {
		    Xid[] recovered = new Xid[1];
		    recovered[0] = new ConsusXid(tid.getFormatId(), 
						 tid.getGlobalTransactionId(), 
						 tid.getBranchQualifier());
		    return recovered;
		} 
		return new Xid[0];
	    } catch (SQLException x) { 
		throw new XAException(XAException.XAER_RMERR);
	    }
	}  

	public void rollback(Xid xid) throws XAException 
	{
	    try { 
		con.rollbackTransaction();
	    } catch (SQLException x) { 
		throw new XAException(XAException.XAER_RMERR);
	    }
	}  

	public boolean setTransactionTimeout(int seconds) throws XAException 
	{
	    return false;
	}   
	
	public void start(Xid xid, int flags)  throws XAException 
	{
	}   
    }

    PhysicalConnection createLocalConnection(String fileName, String url, 
					     Properties info, Driver driver) throws SQLException
    { 
	return new LocalXAConnectionHandle(fileName, url, info, driver);
    }

    PhysicalConnection createRemoteConnection(String address, String url, 
					      Properties info, Driver driver) throws SQLException
    { 
	return new RemoteXAConnectionHandle(address, url, info, driver);
    }

    public XAResource getXAResource() throws SQLException 
    {
	if (resource == null) { 
	    resource = new ConsusXAResource();
	}
	return resource;
    }


    ConsusXAConnection(ConsusXADataSource ds, String username, String password) throws SQLException 
    {
	super(ds, username, password);
    }

    transient XAResource resource;
}



