/* Consus - Pure Java Object Relational Database Management System
 * Node for ALTER TABLE statement
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import  java.sql.SQLException;
    
class AlterTableNode extends StmtNode { 
    TableDescriptor table;
    FieldDescriptor addedList;
    FieldDescriptor modifiedList;
    FieldDescriptor droppedList;
    RenameNode      renameList;
    boolean         changedConstraints;

    int executeStmt(ConsusStatement stmt) throws SQLException { 
	FieldDescriptor alt, fd, curr, next, prev = null;
	
	stmt.db.beginTransaction(true);
	
	if ((droppedList != null || modifiedList != null || droppedList != null) 
	    && (table.superClass != null || table.nSubclasses != 0)) 
	{ 
	    throw new ConsusRuntimeError("Derived or inherited table can not be altered");
	}
	boolean changed = droppedList != null || addedList != null || changedConstraints;

	for (alt = droppedList; alt != null; alt = alt.next) {
	    fd = table.getColumn(alt.name);
	    if (fd == null) { 
		throw new ConsusRuntimeError("No field " + alt.name + " in table " + table.name);
	    }
	    for (curr = table.fields, prev = null; curr != fd; prev = curr, curr = curr.next);
	    if (prev == null) { 
		table.fields = fd.next;
	    } else { 
		prev.next = fd.next;
	    }
	    if (fd.index != 0) { 
		Btree.drop(stmt.db, fd.index);
	    }
	    table.nSelfFields -= 1;
	    table.nColumns -= 1;
	}
	for (fd = table.fields; fd != null; fd = fd.next) { 
	    if ((fd.attr & (FieldDescriptor.OnDeleteCascade|FieldDescriptor.PrimaryKey|FieldDescriptor.Unique)) != 0
		 && (fd.attr & FieldDescriptor.Indexed) == 0)
	    {
		fd.attr |= FieldDescriptor.Indexed;
	    }
	}
	for (fd = table.fields; fd != null; fd = fd.next) { 
	    fd.oldType = fd.type;
	    fd.oldOffs = fd.offs;
	    fd.oldSize = fd.size;
	}
	for (alt = modifiedList; alt != null; alt = alt.next) {
	    fd = table.getColumn(alt.name);
	    if (fd == null) { 
		throw new ConsusRuntimeError("No field " + alt.name + " in table " + table.name);
	    }
	    if ((alt.type < FieldDescriptor.tpString 
		 && fd.type < FieldDescriptor.tpString)
		|| (alt.type == fd.type 
		    && alt.type >= FieldDescriptor.tpString))
	    {
		fd.className = alt.className;
		fd.attr = alt.attr;
		if (fd.type != alt.type) { 
		    fd.type = alt.type;
		    changed = true;
		    if (fd.index != 0) { 
			Btree.drop(stmt.db, fd.index);
			fd.index = Btree.allocate(stmt.db, fd.type);
		    }
		}
	    } else { 
		throw new ConsusRuntimeError("Requested conversion is not possible");
	    }
	}
	for (fd = addedList; fd != null; fd = next) { 
	    next = fd.next;
	    fd.calculateAttributes();
	    fd.oldType = FieldDescriptor.tpUnknown;
	    if ((fd.attr & FieldDescriptor.Indexed) != 0) { 
		fd.index = Btree.allocate(stmt.db, fd.type);
	    }
	}
	for (RenameNode node = renameList; node != null; node = node.next) { 
	    fd = table.getColumn(node.oldName);
	    boolean isColumn = true;
	    if (fd == null) { 
		isColumn = false;
		for (fd = table.fields; fd != null && !node.oldName.equals(fd.longName); fd = fd.next);
		if (fd == null) { 
		    throw new ConsusRuntimeError("No such field '" + node.oldName + 
						 "' in table " + table.name);
		}
	    }
	    if (isColumn) {
		table.renameColumn(fd, node.newName);
	    }
	    fd.longName = node.newName;
	    fd.name = node.newName.substring(node.newName.lastIndexOf('.')+1);
	}
	if (changed) { 
	    int size = Record.sizeof;
	    table.columnMapping.clear();
	    for (fd = table.fields; fd != null; fd = fd.next) {
		if ((fd.attr & FieldDescriptor.ComponentOfArray) == 0) { 
		    fd.offs = size;
		    size += FieldDescriptor.sizeof[fd.type];
		}
	    }
	    table.fixedSize = size;
	    table.completeDeclaration();
	    stmt.db.reformatTable(table, table);
	}
	table.altered = true;
	table.addIndices(true);
	stmt.db.resetForeignKeyConstraints();
	stmt.db.endStatement();
	return 0;
    }	

    AlterTableNode(TableDescriptor tbl, 
		   FieldDescriptor addedFields, 
		   FieldDescriptor modifiedFields, 
		   FieldDescriptor droppedFields,
		   RenameNode renameNodes, 
		   boolean changedConstrs) 
    {  
	super(tpVoid, opAlterTable);
	table = tbl;
	addedList = addedFields;
	modifiedList = modifiedFields;
	droppedList = droppedFields;
	changedConstraints = changedConstrs;
	renameList = renameNodes;
    }
}
