/* Consus - Pure Java Object Relational Database Management System
 * B-Tree page methods implementation
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import java.sql.*;

class BtreePage { 
    static final int firstKeyOffs = 4;
    static final int keySpace = Page.pageSize - firstKeyOffs;
    static final int strKeySize = 8;    
    static final int maxItems = keySpace / 4;    
    static final int maxStrKeySize = (keySpace - strKeySize*2)/4;

    static int getnItems(Page pg) { 
	return Bytes.unpack2(pg.data, 0);
    }
    static int getSize(Page pg) { 
	return Bytes.unpack2(pg.data, 2);
    }
    static int getKeyStrOid(Page pg, int index) {
		return Bytes.unpack4(pg.data, firstKeyOffs + index*8);
    }
    static int getKeyStrSize(Page pg, int index) { 
	return Bytes.unpack2(pg.data, firstKeyOffs + index*8+4);
    }
    static int getKeyStrOffs(Page pg, int index) { 
	return Bytes.unpack2(pg.data, firstKeyOffs + index*8+6);
    }

    static int getRecord(Page pg, int index) { 
	return Bytes.unpack4(pg.data, firstKeyOffs + index*4);
    }
    
    static void setnItems(Page pg, int nItems) { 
	Bytes.pack2(pg.data, 0, (short)nItems);
    }
    static void setSize(Page pg, int size) { 
	Bytes.pack2(pg.data, 2, (short)size);
    }
    static void setKeyStrOid(Page pg, int index, int oid) { 
	Bytes.pack4(pg.data, firstKeyOffs + index*8, oid);
    }
    static void setKeyStrSize(Page pg, int index, int size) { 
	Bytes.pack2(pg.data, firstKeyOffs + index*8+4, (short)size);
    }
    static void setKeyStrOffs(Page pg, int index, int offs) { 
	Bytes.pack2(pg.data, firstKeyOffs + index*8+6, (short)offs);
    }
    static void setKeyStrChars(Page pg, int offs, char[] str) { 
	int len = str.length;
	for (int i = 0; i < len; i++) { 
	    Bytes.pack2(pg.data, firstKeyOffs + offs, (short)str[i]);
	    offs += 2;
	}
    }
    static void setRecord(Page pg, int index, int oid) { 
	Bytes.pack4(pg.data, firstKeyOffs + index*4, oid);
    }

    static boolean find(ConsusConnection db, int pageId, SearchContext sc, 
			int type, int height)
	throws SQLException
    {
	Page pg = db.getPage(pageId);
	int l = 0, n = getnItems(pg), r = n;
	int oid;
	height -= 1;
	try { 
	    if (type == FieldDescriptor.tpString) { 
		if (sc.firstKey != null) { 
		    SearchValue firstKey = sc.firstKey;
		    while (l < r)  {
			int i = (l+r) >> 1;
			if (firstKey.compareStr(pg, i) >= sc.firstKey.inclusion) {
			    l = i + 1; 
			} else { 
			    r = i;
			}
		    }
		    Assert.that(r == l); 
		}
		if (sc.lastKey != null) { 
		    SearchValue lastKey = sc.lastKey;
		    if (height == 0) { 
			while (l < n) { 
			    if (-lastKey.compareStr(pg, l) >= sc.lastKey.inclusion) { 
				return false;
			    }
			    oid = getKeyStrOid(pg, l);
			    if (sc.stmt.cursor.checkType(oid)
				&& (sc.condition == null || sc.stmt.evaluate(sc.condition, oid)))
			    {
				if (!sc.stmt.cursor.add(oid)) { 
				    return false;
				}
			    }
			    l += 1;
			}
		    } else { 
			do {
			    if (!find(db, getKeyStrOid(pg, l), sc, type, height)) {
				return false;
			    }
			    if (l == n) { 
				return true;
			    }
			} while (lastKey.compareStr(pg, l++) >= 0);
			return false;
		    }
		} else { 
		    if (height == 0) { 
			if (sc.condition != null) { 
			    while (l < n) { 
				oid = getKeyStrOid(pg, l);
				if (sc.stmt.cursor.checkType(oid)
				    && sc.stmt.evaluate(sc.condition, oid))
				{
				    if (!sc.stmt.cursor.add(oid)) { 
					return false;
				    }
				}
				l += 1;
			    }
			} else { 
			    while (l < n) { 
				oid = getKeyStrOid(pg, l);
				if (sc.stmt.cursor.checkType(oid)) { 
				    if (!sc.stmt.cursor.add(oid)) { 
					return false;
				    }
				}
				l += 1;
			    }
			}
		    } else {
			do {
			    if (!find(db, getKeyStrOid(pg, l), sc, type, height)) {
				return false;
			    }
			} while (++l <= n);
		    }
		}
	    } else { 
		if (sc.firstKey != null) {
		    while (l < r)  {
			int i = (l+r) >> 1;
			if (sc.firstKey.compare(pg, i, type) >= sc.firstKey.inclusion) {
			    l = i+1;
			} else {
			    r = i;
			}
		    }
		    Assert.that(r == l);
		}
		if (sc.lastKey != null) {
		    if (height == 0) {
			while (l < n) {
			    if (-sc.lastKey.compare(pg, l, type) >= sc.lastKey.inclusion) {
				return false;
			    }
			    oid = getRecord(pg, maxItems-1-l);
			    if (sc.stmt.cursor.checkType(oid)
				&& (sc.condition == null || sc.stmt.evaluate(sc.condition, oid)))
			    {
				if (!sc.stmt.cursor.add(oid)) {
				    return false;
				}
			    }
			    l += 1;
			}
			return true;
		    } else {
			do {
			    if (!find(db, getRecord(pg, maxItems-1-l), sc, type, height)) {
				return false;
			    }
			    if (l == n) {
				return true;
			    }
			} while (sc.lastKey.compare(pg, l++, type) >= 0);
			return false;
		    }
		} 
		if (height == 0) { 
		    if (sc.condition != null) { 
			while (l < n) { 
			    oid = getRecord(pg, maxItems-1-l); 
			    if (sc.stmt.cursor.checkType(oid) 
				&& sc.stmt.evaluate(sc.condition, oid)) 
			    {
				if (!sc.stmt.cursor.add(oid)) { 
				    return false;
				}
			    }
			    l += 1;
			}
		    } else { 
			while (l < n) { 
			    oid = getRecord(pg, maxItems-1-l);
			    if (sc.stmt.cursor.checkType(oid)) { 
				if (!sc.stmt.cursor.add(oid)) { 
				    return false;
				}
			    }
			    l += 1;
			}
		    }
		} else { 
		    do {
			if (!find(db, getRecord(pg, maxItems-1-l), sc, type, height)) {
			    return false;
			}
		    } while (++l <= n);
		}
	    }    
	} finally { 
	    db.pool.unfix(pg);
	}
	return true;
    }    


    static int allocate(ConsusConnection db, int root, int type, SearchValue ins) 
	throws SQLException
    {
	int pageId = db.allocatePage();
	Page pg = db.putPage(pageId);
	setnItems(pg, 1);
	if (type == FieldDescriptor.tpString) { 
	    int len = ins.sval.length;
	    setSize(pg, len*2);
	    setKeyStrOffs(pg, 0, keySpace - len*2);
	    setKeyStrSize(pg, 0, len);
	    setKeyStrOid(pg, 0, ins.oid);
	    setKeyStrOid(pg, 1, root); 
	    setKeyStrChars(pg, keySpace - len*2, ins.sval);
	} else { 
	    ins.pack(pg, 0, type);
	    setRecord(pg, maxItems-2, root);
	}
	db.pool.unfix(pg);
	return pageId;
    }

    static int clonePage(ConsusConnection db, int origPageId, int type, int height) 
	throws SQLException
    {
	int newPageId = db.allocatePage();
	Page newPage = db.putPage(newPageId);
	Page origPage = db.getPage(origPageId);
	int i, n = getnItems(origPage);
	System.arraycopy(origPage.data, 0, newPage.data, 0, Page.pageSize); 

	if (--height != 0) {
	    if (type == FieldDescriptor.tpString) { // page of strings
		for (i = 0; i <= n; i++) { 
		    setKeyStrOid(newPage, i, 
				 clonePage(db, getKeyStrOid(origPage, i), type, height));
		}
	    } else { 
		for (i = 0; i <= n; i++) {
		    setRecord(newPage, maxItems-i-1, 
			      clonePage(db, getRecord(origPage, maxItems-i-1), type, height));
		}
	    }
	}
	db.pool.unfix(newPage);
	db.pool.unfix(origPage);
	return newPageId;
    }


    static void memcpy(Page dst_pg, int dst_idx, Page src_pg, int src_idx, 
		       int len, int itemSize) 
    { 
	System.arraycopy(src_pg.data, firstKeyOffs + src_idx*itemSize, 
			 dst_pg.data, firstKeyOffs + dst_idx*itemSize, 
			 len*itemSize);
    }

    static int insert(ConsusConnection db, int pageId, int type, SearchValue ins, int height, 
		      boolean unique)
	throws SQLException
    {
	Page pg = db.getPage(pageId);
	int result;
	int l = 0, n = getnItems(pg), r = n;
	try { 
	    if (type == FieldDescriptor.tpString) { 	    
		while (l < r)  {
		    int i = (l+r) >> 1;
		    if (ins.compareStr(pg, i) > 0) { 
			l = i+1; 
		    } else { 
			r = i;
		    }
		}
		Assert.that(l == r);
		if (--height != 0) { 
		    result = insert(db, getKeyStrOid(pg, r), type, ins, height, unique);
		    Assert.that(result != Btree.op_not_found);
		    if (result != Btree.op_overflow) {
			return result;                              
		    }                                             
		} else if (unique && r < n && ins.compareStr(pg, r) == 0) {
		    return Btree.op_duplicate;
		}
		db.pool.unfix(pg);
		pg = db.putPage(pageId);
		return insertStrKey(db, pg, r, ins, height);
	    } else { 
		while (l < r)  {
		    int i = (l+r) >> 1;
		    if (ins.compare(pg, i, type) > 0) l = i+1; else r = i;
		}
		Assert.that(l == r);
		/* insert before e[r] */
		if (--height != 0) {
		    result = insert(db, getRecord(pg, maxItems-r-1), type, ins, height, unique);
		    Assert.that(result != Btree.op_not_found);
		    if (result != Btree.op_overflow) {
			return result;                              
		    }                                             
		    n += 1;
		} else if (unique && r < n && ins.compare(pg, r, type) == 0) { 
		    return Btree.op_duplicate;
		}
		db.pool.unfix(pg);
		pg = db.putPage(pageId);
		int itemSize = FieldDescriptor.sizeof[type];
		int max = keySpace / (4 + itemSize);
		if (n < max) {
		    memcpy(pg, r+1, pg, r, n - r, itemSize);
		    memcpy(pg, maxItems-n-1, pg, maxItems-n, n-r, 4);
		    ins.pack(pg, r, type);
		    setnItems(pg, getnItems(pg)+1);
		    return Btree.op_done;
		} else { /* page is full then divide page */
		    pageId = db.allocatePage();
  	            Page b = db.putPage(pageId);
		    Assert.that(n == max);
		    int m = max/2;
		    if (r < m) {
			memcpy(b, 0, pg, 0, r, itemSize);
			memcpy(b, r+1, pg, r, m-r-1, itemSize);
			memcpy(pg, 0, pg, m-1, max-m+1, itemSize);
			memcpy(b, maxItems-r, pg, maxItems-r, r, 4);
			ins.pack(b, r, type);
			memcpy(b, maxItems-m, pg, maxItems-m+1, m-r-1, 4);
			memcpy(pg, maxItems-max+m-1, pg, maxItems-max, max-m+1, 4);
		    } else {
			memcpy(b, 0, pg, 0, m, itemSize);
			memcpy(pg, 0, pg, m, r-m, itemSize);
			memcpy(pg, r-m+1, pg, r, max-r, itemSize);
			memcpy(b, maxItems-m, pg, maxItems-m, m, 4);
			memcpy(pg, maxItems-r+m, pg, maxItems-r, r-m, 4);
			ins.pack(pg, r-m, type);
			memcpy(pg, maxItems-max+m-1, pg, maxItems-max, max-r, 4);
		    }
		    ins.oid = pageId;
		    ins.extract(b.data, firstKeyOffs + (m-1)*itemSize, type);
		    if (height == 0) {
			setnItems(pg, max - m + 1);
			setnItems(b, m);
		    } else {
			setnItems(pg, max - m);
			setnItems(b, m - 1);
		    }
		    db.pool.unfix(b);
		    return Btree.op_overflow;
		}
	    }
	} finally { 
	    db.pool.unfix(pg);
	}
    }

    static int insertStrKey(ConsusConnection db, Page pg, int r, SearchValue ins, int height)
	throws SQLException
    {
	int nItems = getnItems(pg);
	int size = getSize(pg);
	int n = (height != 0) ? nItems + 1 : nItems;
 	// insert before e[r]
	int len = ins.sval.length;
	if (size + len*2 + (n+1)*strKeySize <= keySpace) { 
	    memcpy(pg, r+1, pg, r, n-r, strKeySize);
	    size += len*2;
	    setKeyStrOffs(pg, r, keySpace - size);
	    setKeyStrSize(pg, r, len);
	    setKeyStrOid(pg, r, ins.oid);
	    setKeyStrChars(pg, keySpace - size, ins.sval);
	    nItems += 1;
	} else { // page is full then divide page
	    int  pageId = db.allocatePage();
	    Page b = db.putPage(pageId);
	    int  moved = 0;
	    int  inserted = len*2 + strKeySize;
	    int  prevDelta = (1 << 31) + 1;

	    for (int bn = 0, i = 0; ; bn += 1) {
		int addSize, subSize;
		int j = nItems - i - 1;
		int keyLen = getKeyStrSize(pg, i);
		if (bn == r) {
		    keyLen = len;
		    inserted = 0;
		    addSize = len;
		    if (height == 0) {
			subSize = 0;
			j += 1;
		    } else { 
			subSize = getKeyStrSize(pg, i);
		    }
		} else { 
		    addSize = subSize = keyLen;
		    if (height != 0) {
			if (i + 1 != r) { 
			    subSize += getKeyStrSize(pg, i+1);
			    j -= 1;
			} else { 
			    inserted = 0;
			}
		    }
		}
		int delta = (moved + addSize*2 + (bn+1)*strKeySize) 
		    - (j*strKeySize + size - subSize*2 + inserted);
		if (delta >= -prevDelta) {
		    char[] insKey = ins.sval;
		    if (height == 0) { 
			ins.getStr(b, bn-1);
		    } else {
			Assert.that("String fits in the B-Tree page", 
				    moved + (bn+1)*strKeySize <= keySpace);
			if (bn != r) { 
			    ins.getStr(pg, i);
			    setKeyStrOid(b, bn, getKeyStrOid(pg, i));
			    size -= keyLen*2;
			    i += 1;
			} else {
			    setKeyStrOid(b, bn, ins.oid);
			}
		    }
		    nItems = compactify(pg, i);		    
		    if (bn < r || (bn == r && height == 0)) { 
			memcpy(pg, r-i+1, pg, r-i, n - r, strKeySize);
			size += len*2;
			nItems += 1;
			Assert.that("String fits in the B-Tree page", 
				    size + (n-i+1)*strKeySize <= keySpace);
			setKeyStrOffs(pg, r-i, keySpace - size);
			setKeyStrSize(pg, r-i, len);
			setKeyStrOid(pg, r-i, ins.oid);
			setKeyStrChars(pg, keySpace - size, insKey);
		    }
		    setnItems(b, bn);
		    setSize(b, moved);
		    setSize(pg, size);
		    setnItems(pg, nItems);
		    ins.oid = pageId;
		    db.pool.unfix(b);
		    return Btree.op_overflow;
		}
		moved += keyLen*2;
		prevDelta = delta;
		Assert.that("String fits in the B-Tree page", 
			    moved + (bn+1)*strKeySize <= keySpace);
		setKeyStrSize(b, bn, keyLen);
		setKeyStrOffs(b, bn, keySpace - moved);
		if (bn == r) { 
		    setKeyStrOid(b, bn, ins.oid);
		    setKeyStrChars(b, keySpace - moved, ins.sval);
		} else { 
		    setKeyStrOid(b, bn, getKeyStrOid(pg, i));
		    memcpy(b, keySpace - moved, pg, getKeyStrOffs(pg, i), keyLen*2, 1);
		    size -= keyLen*2;
		    i += 1;
		}
	    }
	}
	setnItems(pg, nItems);
	setSize(pg, size);
	return size + strKeySize*(nItems+1) < keySpace/2 
	    ? Btree.op_underflow : Btree.op_done;
    }


    static int compactify(Page pg, int m) 
	throws SQLException
    {
	int i, j, offs, len, n = getnItems(pg);
	int[] size  = new int[keySpace/2+1];
	int[] index = new int[keySpace/2+1];
	if (m == 0) { 
	    return n;
	}
	int nZeroLengthStrings = 0;
	if (m < 0) {
	    m = -m;
	    for (i = 0; i < n-m; i++) { 
		len = getKeyStrSize(pg, i);
		if (len != 0) { 
		    offs = getKeyStrOffs(pg, i) >>> 1;
		    size[offs + len] = len;
		    index[offs + len] = i;
		} else { 
		    nZeroLengthStrings += 1;
		}
	    }	
	    for (; i < n; i++) { 
		len = getKeyStrSize(pg, i);
		if (len != 0) { 
		    offs = getKeyStrOffs(pg, i) >>> 1;
		    size[offs + len] = len;
		    index[offs + len] = -1;
		}
	    }
	} else { 
	    for (i = 0; i < m; i++) { 
		len = getKeyStrSize(pg, i);
		if (len != 0) { 
		    offs = getKeyStrOffs(pg, i) >>> 1;
		    size[offs + len] = len;
		    index[offs + len] = -1;
		}
	    }
	    for (; i < n; i++) { 
		len = getKeyStrSize(pg, i);
		if (len != 0) { 
		    offs = getKeyStrOffs(pg, i) >>> 1;
		    size[offs + len] = len;
		    index[offs + len] = i - m;
		} else { 
		    nZeroLengthStrings += 1;
		}		    
		setKeyStrOid(pg, i-m, getKeyStrOid(pg, i));
		setKeyStrSize(pg, i-m, len);
	    }	
	    setKeyStrOid(pg, i-m, getKeyStrOid(pg, i));
	}
	int nItems = n -= m;
	n -= nZeroLengthStrings;
	for (offs = keySpace/2, i = offs; n != 0; i -= len) { 
	    len = size[i];
	    j = index[i];
	    if (j >= 0) {
		offs -= len;
		n -= 1;
		setKeyStrOffs(pg, j, offs*2);
		if (offs != i - len) { 
		    memcpy(pg, offs, pg, i - len, len, 2);
		}
	    }
	}
	return nItems;
    }
    
    static int removeStrKey(Page pg, int r) throws SQLException
    {
	int len = getKeyStrSize(pg, r)*2;
	int offs = getKeyStrOffs(pg, r);
	int size = getSize(pg);
	int nItems = getnItems(pg);
	if ((nItems+1)*strKeySize >= keySpace) { 
	    memcpy(pg, r, pg, r+1, nItems-r-1, strKeySize);
	} else { 
	    memcpy(pg, r, pg, r+1, nItems-r, strKeySize);
	} 
	if (len != 0) { 
	    memcpy(pg, keySpace - size + len, pg, keySpace - size, size - keySpace + offs, 1);
	    for (int i = nItems; --i >= 0; ) { 
		if (getKeyStrOffs(pg, i) < offs) { 
		    setKeyStrOffs(pg, i, getKeyStrOffs(pg, i) + len);
		}
	    }
	    setSize(pg, size -= len);
	}
	setnItems(pg, nItems - 1);
	return size + strKeySize*nItems < keySpace/2 
	    ? Btree.op_underflow : Btree.op_done;
    }
    
    static int replaceStrKey(ConsusConnection db, Page pg, int r, SearchValue ins, int height)
	throws SQLException
    {
	ins.oid = getKeyStrOid(pg, r);
	removeStrKey(pg, r);
	return insertStrKey(db, pg, r, ins, height);
    }
    
    static int handlePageUnderflow(ConsusConnection db, Page pg, int r, int type, SearchValue rem, int height)
	throws SQLException
    {
	int nItems = getnItems(pg);
	if (type == FieldDescriptor.tpString) { 
	    Page a = db.putPage(getKeyStrOid(pg, r));
	    int an = getnItems(a);
	    if (r < nItems) { // exists greater page
		Page b = db.getPage(getKeyStrOid(pg, r+1));
		int bn = getnItems(b); 
		int merged_size = (an+bn)*strKeySize + getSize(a) + getSize(b);
		if (height != 1) { 
		    merged_size += getKeyStrSize(pg, r)*2 + strKeySize*2;
		}
		if (merged_size > keySpace) {
		    // reallocation of nodes between pages a and b
		    int i, j, k;
		    db.pool.unfix(b);
		    b = db.putPage(getKeyStrOid(pg, r+1));
		    int size_a = getSize(a);
		    int size_b = getSize(b);
		    int addSize, subSize;
		    if (height != 1) {
			addSize = getKeyStrSize(pg, r);
			subSize = getKeyStrSize(b, 0);
		    } else { 
			addSize = subSize = getKeyStrSize(b, 0);
		    }
		    i = 0;
		    int prevDelta = (an*strKeySize + size_a) - (bn*strKeySize + size_b);
		    while (true) { 
			i += 1;
			int delta = ((an+i)*strKeySize + size_a + addSize*2)
			    - ((bn-i)*strKeySize + size_b - subSize*2);
			if (delta >= 0) {
			    if (delta >= -prevDelta) { 
				i -= 1;
			    }
			    break;
			}
			size_a += addSize*2;
			size_b -= subSize*2;
			prevDelta = delta;
			if (height != 1) { 
			    addSize = subSize;	
			    subSize = getKeyStrSize(b, i);
			} else { 
			    addSize = subSize = getKeyStrSize(b, i);
			}
		    }
		    int result = Btree.op_done;
		    if (i > 0) { 
			k = i;
			if (height != 1) { 
			    int len = getKeyStrSize(pg, r);
			    setSize(a, getSize(a) + len*2);
			    setKeyStrOffs(a, an, keySpace - getSize(a));
			    setKeyStrSize(a, an, len);
			    memcpy(a, getKeyStrOffs(a, an),
				   pg, getKeyStrOffs(pg, r), len*2, 1);
			    k -= 1;
			    an += 1;
			    setKeyStrOid(a, an+k, getKeyStrOid(b, k));
			    setSize(b, getSize(b) - getKeyStrSize(b, k)*2);
			}
			for (j = 0; j < k; j++) { 
			    int len = getKeyStrSize(b, j);
			    setSize(a, getSize(a) + len*2);
			    setSize(b, getSize(b) - len*2);
			    setKeyStrOffs(a, an, keySpace - getSize(a));
			    setKeyStrSize(a, an, len);
			    setKeyStrOid(a, an, getKeyStrOid(b, j));
			    memcpy(a, getKeyStrOffs(a, an),
				   b, getKeyStrOffs(b, j), len*2, 1);
			    an += 1;
			}
			rem.getStr(b, i-1);
			result = replaceStrKey(db, pg, r, rem, height);
			setnItems(a, an);
			setnItems(b, compactify(b, i));
		    }
		    db.pool.unfix(a);
		    db.pool.unfix(b);
		    return result;
		} else { // merge page b to a
		    if (height != 1) { 
			int r_len = getKeyStrSize(pg, r);
			setKeyStrSize(a, an, r_len);
			setSize(a, getSize(a) + r_len*2);
			setKeyStrOffs(a, an, keySpace - getSize(a));
			memcpy(a, getKeyStrOffs(a, an), 
			       pg, getKeyStrOffs(pg, r), r_len*2, 1);
			an += 1;
			setKeyStrOid(a, an+bn, getKeyStrOid(b, bn));
		    }
		    for (int i = 0; i < bn; i++, an++) { 
			setKeyStrSize(a, an, getKeyStrSize(b, i));
			setKeyStrOffs(a, an, getKeyStrOffs(b, i) - getSize(a));
			setKeyStrOid(a, an, getKeyStrOid(b, i));
		    }
		    setSize(a, getSize(a) + getSize(b));
		    setnItems(a, an);
		    memcpy(a, keySpace - getSize(a), b, keySpace - getSize(b), getSize(b), 1);
		    db.pool.unfix(a);
		    db.pool.unfix(b);
		    db.freePage(getKeyStrOid(pg, r+1));
		    setKeyStrOid(pg, r+1, getKeyStrOid(pg, r));
		    return removeStrKey(pg, r);
		}
	    } else { // page b is before a
		Page b = db.getPage(getKeyStrOid(pg, r-1));
		int bn = getnItems(b); 
		int merged_size = (an+bn)*strKeySize + getSize(a) + getSize(b);
		if (height != 1) { 
		    merged_size += getKeyStrSize(pg, r-1)*2 + strKeySize*2;
		}
		if (merged_size > keySpace) {
		    // reallocation of nodes between pages a and b
		    int i, j, k, len;
		    db.pool.unfix(b);
		    b = db.putPage(getKeyStrOid(pg, r-1));
		    int size_a = getSize(a);
		    int size_b = getSize(b);
		    int addSize, subSize;
		    if (height != 1) {
			addSize = getKeyStrSize(pg, r-1);
			subSize = getKeyStrSize(b, bn-1);
		    } else { 
			addSize = subSize = getKeyStrSize(b, bn-1);
		    }
		    i = 0;
		    int prevDelta = (an*strKeySize + size_a) - (bn*strKeySize + size_b);
		    while (true) { 
			i += 1;
			int delta = ((an+i)*strKeySize + size_a + addSize*2)
			    - ((bn-i)*strKeySize + size_b - subSize*2);
			if (delta >= 0) {
			    if (delta >= -prevDelta) { 
				i -= 1;
			    }
			    break;
			}
			prevDelta = delta;
			size_a += addSize*2;
			size_b -= subSize*2;
			if (height != 1) { 
			    addSize = subSize;	
			    subSize = getKeyStrSize(b, bn-i-1);
			} else { 
			    addSize = subSize = getKeyStrSize(b, bn-i-1);
			}
		    }
		    int result = Btree.op_done;
		    if (i > 0) { 
			k = i;
			Assert.that(i < bn);
			if (height != 1) { 
			    setSize(b, getSize(b) - getKeyStrSize(b, bn-k)*2);
			    memcpy(a, i, a, 0, an+1, strKeySize);
			    k -= 1;
			    setKeyStrOid(a, k, getKeyStrOid(b, bn));
			    len = getKeyStrSize(pg, r-1);
			    setKeyStrSize(a, k, len);
			    setSize(a, getSize(a) + len*2);
			    setKeyStrOffs(a, k, keySpace - getSize(a));
			    memcpy(a, getKeyStrOffs(a, k), 
				   pg, getKeyStrOffs(pg, r-1), len*2, 1);
			} else { 
			    memcpy(a, i, a, 0, an, strKeySize);
			}
			for (j = 0; j < k; j++) { 
			    len = getKeyStrSize(b, bn-k+j);
			    setSize(a, getSize(a) + len*2);
			    setSize(b, getSize(b) - len*2);
			    setKeyStrOffs(a, j, keySpace - getSize(a));
			    setKeyStrSize(a, j, len);
			    setKeyStrOid(a, j, getKeyStrOid(b, bn-k+j));
			    memcpy(a, getKeyStrOffs(a, j),
				   b, getKeyStrOffs(b, bn-k+j), len*2, 1);
			}
			an += i;
			setnItems(a, an);
			rem.getStr(b, bn-k-1);
			result = replaceStrKey(db, pg, r-1, rem, height);
			setnItems(b, compactify(b, -i));
		    }
		    db.pool.unfix(a);
		    db.pool.unfix(b);		    
		    return result;
		} else { // merge page b to a
		    if (height != 1) { 
			memcpy(a, bn + 1, a, 0, an+1, strKeySize);
			int len = getKeyStrSize(pg, r-1);
			setKeyStrSize(a, bn, len);
			setSize(a, getSize(a) + len*2);
			setKeyStrOffs(a, bn, keySpace - getSize(a));
			setKeyStrOid(a, bn, getKeyStrOid(b, bn));
			memcpy(a, getKeyStrOffs(a, bn), 
			       pg, getKeyStrOffs(pg, r-1), len*2, 1);
			an += 1;
		    } else { 
			memcpy(a, bn, a, 0, an, strKeySize);
		    }
		    for (int i = 0; i < bn; i++) { 
			setKeyStrOid(a, i, getKeyStrOid(b, i));
			setKeyStrSize(a, i, getKeyStrSize(b, i));
			setKeyStrOffs(a, i, getKeyStrOffs(b, i) - getSize(a));
		    }
		    an += bn;
		    setnItems(a, an);
		    setSize(a, getSize(a) + getSize(b));
		    memcpy(a, keySpace - getSize(a), b, keySpace - getSize(b), getSize(b), 1);
		    db.pool.unfix(a);
		    db.pool.unfix(b);
		    db.freePage(getKeyStrOid(pg, r-1));
		    return removeStrKey(pg, r-1);
		}
	    }
	} else { 
	    Page a = db.putPage(getRecord(pg, maxItems-r-1));
	    int an = getnItems(a);
	    int itemSize = FieldDescriptor.sizeof[type];
	    if (r < nItems) { // exists greater page
		Page b = db.getPage(getRecord(pg, maxItems-r-2));
		int bn = getnItems(b); 
		Assert.that(bn >= an);
		if (height != 1) { 
		    memcpy(a, an, pg, r, 1, itemSize);
		    an += 1;
		    bn += 1;
		}
		int merged_size = (an+bn)*(4 + itemSize);
		if (merged_size > keySpace) { 
		    // reallocation of nodes between pages a and b
		    int i = bn - ((an + bn) >> 1);
		    db.pool.unfix(b);
		    b = db.putPage(getRecord(pg, maxItems-r-2));
		    memcpy(a, an, b, 0, i, itemSize);
		    memcpy(b, 0, b, i, bn-i, itemSize);
		    memcpy(a, maxItems-an-i, b, maxItems-i, i, 4);
		    memcpy(b, maxItems-bn+i, b, maxItems-bn, bn-i, 4);
		    memcpy(pg, r, a, an+i-1, 1, itemSize);
		    setnItems(b, getnItems(b) - i);
		    setnItems(a, getnItems(a) + i);
		    db.pool.unfix(a);
		    db.pool.unfix(b);
		    return Btree.op_done;
		} else { // merge page b to a  
		    memcpy(a, an, b, 0, bn, itemSize);
		    memcpy(a, maxItems-an-bn, b, maxItems-bn, bn, 4);
		    db.freePage(getRecord(pg, maxItems-r-2));
		    memcpy(pg, maxItems-nItems, pg, maxItems-nItems-1, 
			   nItems - r - 1, 4);
		    memcpy(pg, r, pg, r+1, nItems - r - 1, itemSize);
		    setnItems(a, getnItems(a) + bn);
		    setnItems(pg, nItems - 1);
		    db.pool.unfix(a);
		    db.pool.unfix(b);
		    return nItems*(itemSize + 4) < keySpace/2
			? Btree.op_underflow : Btree.op_done;
		}
	    } else { // page b is before a
		Page b = db.getPage(getRecord(pg, maxItems-r));
		int bn = getnItems(b); 
		Assert.that(bn >= an);
		if (height != 1) { 
		    an += 1;
		    bn += 1;
		}
		int merged_size = (an+bn)*(4 + itemSize);
		if (merged_size > keySpace) { 
		    // reallocation of nodes between pages a and b
		    int i = bn - ((an + bn) >> 1);
		    db.pool.unfix(b);
		    b = db.putPage(getRecord(pg, maxItems-r));
		    memcpy(a, i, a, 0, an, itemSize);
		    memcpy(a, 0, b, bn-i, i, itemSize);
		    memcpy(a, maxItems-an-i, a, maxItems-an, an, 4);
		    memcpy(a, maxItems-i, b, maxItems-bn, i, 4);
		    if (height != 1) { 
			memcpy(a, i-1, pg, r-1, 1, itemSize);
		    }
		    memcpy(pg, r-1, b, bn-i-1, 1, itemSize);
		    setnItems(b, getnItems(b) - i);
		    setnItems(a, getnItems(a) + i);
		    db.pool.unfix(a);
		    db.pool.unfix(b);
		    return Btree.op_done;
		} else { // merge page b to a
		    memcpy(a, bn, a, 0, an, itemSize);
		    memcpy(a, 0, b, 0, bn, itemSize);
		    memcpy(a, maxItems-an-bn, a, maxItems-an, an, 4);
		    memcpy(a, maxItems-bn, b, maxItems-bn, bn, 4);
		    if (height != 1) { 
			memcpy(a, bn-1, pg, r-1, 1, itemSize);
		    }
		    db.freePage(getRecord(pg, maxItems-r));
		    setRecord(pg, maxItems-r, getRecord(pg, maxItems-r-1));
		    setnItems(a, getnItems(a) + bn);
		    setnItems(pg, nItems - 1);
		    db.pool.unfix(a);
		    db.pool.unfix(b);
		    return nItems*(itemSize + 4) < keySpace/2
			? Btree.op_underflow : Btree.op_done;
		}
	    }
	}
    }
   
    static int remove(ConsusConnection db, int pageId, int type, SearchValue rem, int height)
	throws SQLException
    {
	Page pg = db.getPage(pageId);
	try { 
	    int i, n = getnItems(pg), l = 0, r = n;
	    
	    if (type != FieldDescriptor.tpString) { 
		int itemSize = FieldDescriptor.sizeof[type];
		while (l < r)  {
		    i = (l+r) >> 1;
		    if (rem.compare(pg, i, type) > 0) { 
			l = i+1; 
		    } else { 
			r = i;
		    }
		}
		if (--height == 0) {
		    int oid = rem.oid;
		    while (r < n) {
			if (rem.compare(pg, r, type) == 0) {
			    if (getRecord(pg, maxItems-r-1) == oid) {
				db.pool.unfix(pg);
				pg = db.putPage(pageId);
				memcpy(pg, r, pg, r+1, n - r - 1, itemSize);
				memcpy(pg, maxItems-n+1, pg, maxItems-n, n - r - 1, 4);
				setnItems(pg, --n);
				return n*(itemSize + 4) < keySpace/2
				    ? Btree.op_underflow : Btree.op_done;
			    }
			} else {
			    break;
			}
			r += 1;
		    }
		    return Btree.op_not_found;
		}
		do { 
		    switch (remove(db, getRecord(pg, maxItems-r-1), type, rem, height)) {
		      case Btree.op_underflow: 
			db.pool.unfix(pg);
			pg = db.putPage(pageId);
			return handlePageUnderflow(db, pg, r, type, rem, height);
		      case Btree.op_done:
			return Btree.op_done;
		    } 
		} while (++r <= n);
		return Btree.op_not_found;
	    } else { 
		while (l < r)  {
		    i = (l+r) >> 1;
		    if (rem.compareStr(pg, i) > 0) { 
			l = i+1; 
		    } else { 
			r = i;
		    }
		}
		if (--height != 0) {
		    do {  
			switch (remove(db, getKeyStrOid(pg, r), type, rem, height)) {
			  case Btree.op_underflow: 
			    db.pool.unfix(pg);
			    pg = db.putPage(pageId);
			    return handlePageUnderflow(db, pg, r, type, rem, height);
			  case Btree.op_done:
			    return Btree.op_done;
			  case Btree.op_overflow:
			    db.pool.unfix(pg);
			    pg = db.putPage(pageId);
			    return insertStrKey(db, pg, r, rem, height);
			}
		    } while (++r <= n);
		} else { 
		    while (r < n) { 
			if (rem.compareStr(pg, r) == 0) { 
			    if (getKeyStrOid(pg, r) == rem.oid) { 
				db.pool.unfix(pg);
				pg = db.putPage(pageId);
				return removeStrKey(pg, r);
			    }
			} else { 
			    break;
			}
			r += 1;
		    }
		}
		return Btree.op_not_found;
	    }
	} finally { 
	    db.pool.unfix(pg);
	}
    }

    static void purge(ConsusConnection db, int pageId, int type, int height)
	throws SQLException
    {
	if (--height != 0) { 
	    Page pg = db.getPage(pageId);
	    int n = getnItems(pg)+1;
	    if (type == FieldDescriptor.tpString) { // page of strings
		while (--n >= 0) { 
		    purge(db, getKeyStrOid(pg, n), type, height);
		}
	    } else { 
		while (--n >= 0) { 
		    purge(db, getRecord(pg, maxItems-n-1), type, height);
		}
	    }
	    db.pool.unfix(pg);
	} 
	db.freePage(pageId);
    }

    static boolean traverseForward(ConsusConnection db, int treeId, ConsusStatement stmt, 
				   ExprNode condition, int type, int height)
	throws SQLException
    {
	Page pg = db.getPage(treeId);
	int oid;
	try { 
	    int i, n = getnItems(pg);
	    if (--height != 0) {
		if (type == FieldDescriptor.tpString) { // page of strings
		    for (i = 0; i <= n; i++) { 
			if (!traverseForward(db, getKeyStrOid(pg, i), stmt, condition, type, height)) 
			{ 
			    return false;
			}
		    }
		} else { 
		    for (i = 0; i <= n; i++) { 
			if (!traverseForward(db, getRecord(pg, maxItems-i-1), stmt, condition, type, height)) 
			{ 
			    return false;
			}
		    }
		}
	    } else { 
		if (type != FieldDescriptor.tpString) { // page of scalars
		    if (condition == null) { 
			for (i = 0; i < n; i++) { 
			    oid = getRecord(pg, maxItems-1-i);
			    if (stmt.cursor.checkType(oid)) { 
				if (!stmt.cursor.add(oid)) { 
				    return false;
				}
			    }
			}
		    } else { 
			for (i = 0; i < n; i++) { 
			    oid = getRecord(pg, maxItems-1-i);
			    if (stmt.cursor.checkType(oid) && stmt.evaluate(condition, oid)) {  
				if (!stmt.cursor.add(oid)) { 
				    return false;
				}
			    }
			}
		    }
		} else { // page of strings
		    if (condition == null) { 
			for (i = 0; i < n; i++) {
			    oid = getKeyStrOid(pg, i);
			    if (stmt.cursor.checkType(oid)) { 
				if (!stmt.cursor.add(oid)) { 
				    return false;
				}
			    }
			}
		    } else { 
			for (i = 0; i < n; i++) { 
			    oid = getKeyStrOid(pg, i);
			    if (stmt.cursor.checkType(oid) && stmt.evaluate(condition, oid)) {  
				if (!stmt.cursor.add(oid)) { 
				    return false;
				}
			    }
			}	
		    }
		}
	    }
	    return true;
	} finally { 
	    db.pool.unfix(pg);
	}
    }

    static boolean traverseBackward(ConsusConnection db, int pageId, ConsusStatement stmt,
				    ExprNode condition, int type, int height)
	throws SQLException
    {
	Page pg = db.getPage(pageId);
	int oid;
	try { 
	    int i, n = getnItems(pg);
	    if (--height != 0) {
		if (type == FieldDescriptor.tpString) { // page of strings
		    do {
			if (!traverseBackward(db, getKeyStrOid(pg, n), 
					      stmt, condition, type, height)) 
			{ 
			    return false;
			}
		    } while (--n >= 0);
		} else { 
		    do { 
			if (!traverseBackward(db, getRecord(pg, maxItems-n-1), 
					      stmt, condition, type, height)) 
			{ 
			    return false;
			}
		    } while (--n >= 0);
		}
	    } else { 
		if (type != FieldDescriptor.tpString) { // page of scalars
		    if (condition == null) { 
			while (--n >= 0) { 
			    oid = getRecord(pg, maxItems-1-n);
			    if (stmt.cursor.checkType(oid)) { 
				if (!stmt.cursor.add(oid)) { 
				    return false;
				}
			    }
			}
		    } else { 
			while (--n >= 0) { 
			    oid = getRecord(pg, maxItems-1-n); 
			    if (stmt.cursor.checkType(oid) && stmt.evaluate(condition, oid)) {  
				if (!stmt.cursor.add(oid)) { 
				    return false;
				}
			    }
			}
		    }
		} else { // page of strings
		    if (condition == null) { 
			while (--n >= 0) { 
			    oid = getKeyStrOid(pg, n);
			    if (stmt.cursor.checkType(oid)) { 
				if (!stmt.cursor.add(oid)) {
				    return false;
				}
			    }
			}
		    } else { 
			while (--n >= 0) { 
			    oid = getKeyStrOid(pg, n);
			    if (stmt.cursor.checkType(oid) && stmt.evaluate(condition, oid)) {  
				if (!stmt.cursor.add(oid)) { 
				    return false;
				}
			    }
			}
		    }
		}	
	    }
	} finally { 
	    db.pool.unfix(pg);
	}
	return true;
    }

}
