/* Consus - Pure Java Object Relational Database Management System
 * Represents column in select list
 * Created by Konstantin Knizhnik, Feb 2000.
 */


package org.garret.internal.consus;
import org.garret.consus.*;
import java.sql.SQLException;

class ColumnNode extends ExprNode { 
    ColumnNode      next;
    int             index;
    FieldDescriptor field;
    ExprNode        expr;
    String          name;
    int             tableDep;
    boolean         sorted;
    long            iVal;
    double          dVal;
    String          sVal;
    boolean         bVal;
    int             oidVal;
    byte[]          binVal;
    Object          blob;

    int dependsOn() { return tableDep; }

    FieldDescriptor getField() { 
	return field;
    }

    boolean equalsNode(ExprNode node) { 
	return node == this;
    }

    String dump() { 
	return "(" + opMnem[tag] + " " + index + ")";
    }

    void assign(Object obj) { 
	switch (type) { 
	  case tpBool:
	    bVal = ((Boolean)obj).booleanValue();
	    break;
	  case tpInt:
	    iVal = ((Long)obj).longValue();
	    break;
	  case tpStr:
	    sVal = (String)obj;
	    break;
	  case tpReal:
	    dVal = ((Double)obj).doubleValue();
	    break;
	  case tpRef:
	    oidVal = (int)((ConsusRef)obj).getOid();
	    break;
	  case tpBlob:
	  case tpClob:
	    blob = obj;
	    break;
	  default:
	    Assert.failed("Invalif column type: " + type);
	}
    }

    long evaluateInt(ConsusStatement stmt) throws SQLException { 
	return iVal;
    }

    double evaluateReal(ConsusStatement stmt) throws SQLException { 
	return dVal;
    }

    String evaluateStr(ConsusStatement stmt) throws SQLException { 
	return sVal;
    }

    int evaluateOid(ConsusStatement stmt) throws SQLException { 
	return oidVal;
    }
    
    boolean evaluateBool(ConsusStatement stmt) throws SQLException { 
	return bVal;
    }

    byte[] evaluateBinary(ConsusStatement stmt) throws SQLException { 
	return binVal;
    }

    final Object getObject() throws SQLException { 
	switch (type) { 
	  case tpInt:
	    return new Long(iVal);
	  case tpReal:
	    return new Double(dVal);
	  case tpStr:
	    return sVal;
	  case tpRef:
	    return new ConsusRef(oidVal);
	  case tpBool:
	    return new Boolean(bVal);
	  case tpArrayInt1:
	    return binVal;
	  case tpBlob:
	  case tpClob:
	    return blob;
	  default:
	    throw new ConsusRuntimeError("Invalid column type " + type);
	}
    }
	
    void evaluate(ConsusStatement stmt) throws SQLException { 
	switch (type) { 
	  case tpInt:
	    iVal = expr.evaluateInt(stmt);
	    break;
	  case tpReal:
	    dVal = expr.evaluateReal(stmt);
	    break;
	  case tpStr:
	    sVal = expr.evaluateStr(stmt);
	    break;
	  case tpRef:
	    oidVal = expr.evaluateOid(stmt);
	    break;
	  case tpBool:
	    bVal = expr.evaluateBool(stmt);
	    break;
	  case tpArrayInt1:
	    binVal = expr.evaluateBinary(stmt);
	    break;
	  case tpBlob:
	  case tpClob:
	    blob = expr.evaluateObj(stmt);
	    break;
	  default:
	    throw new ConsusRuntimeError("Invalid column type " + type);
	}
    }

    ColumnNode(int index) { // pseudocolumn for DatabaseMetaData
	super(tpStr, opColumn);
	this.index = index;
    }

    ColumnNode(int type, String name, int index) { 
	super(type, opColumn);
	this.name = name;
	this.index = index;
    }

    ColumnNode(int index, ExprNode expr, String alias, ColumnNode chain) { 
	super(expr.type, opColumn);
	this.expr = expr;
	this.index = index;
	if (expr instanceof LoadNode && 
	    (((LoadNode)expr).base == null || ((LoadNode)expr).base.tag == opTable)) 
	{ 
	    field = expr.getField();
	    name = field.name;
	}
	tableDep = expr.dependsOn();
	if (alias != null) { 
	    name = alias;
	}
	next = chain;
    }
}

