/* Consus - Pure Java Object Relational Database Management System
 * Node for values in UPDATE/INSERT operations
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import  java.sql.SQLException;


class ColumnValue {
    ColumnValue     next;
    FieldDescriptor field;
    ExprNode        value;

    void setValue(ExprNode expr, int pos) throws SQLException { 
	value = expr;
	if (expr.tag == ExprNode.opParameter) { 
	    switch (field.type) { 
	      case FieldDescriptor.tpBool:
		expr.type = ExprNode.tpBool;
		break;
	      case FieldDescriptor.tpInt1:
	      case FieldDescriptor.tpInt2:
	      case FieldDescriptor.tpInt4:
	      case FieldDescriptor.tpInt8:
		expr.type = ExprNode.tpInt;
		break;
	      case FieldDescriptor.tpDate:
	      case FieldDescriptor.tpTime:
	      case FieldDescriptor.tpTimestamp:
		expr.type = ExprNode.tpInt;
		break;
	      case FieldDescriptor.tpReal4:
	      case FieldDescriptor.tpReal8:
		expr.type = ExprNode.tpReal;
		break;
	      case FieldDescriptor.tpString:
		expr.type = ExprNode.tpStr;
		break;
	      case FieldDescriptor.tpArray:
		expr.type = ExprNode.tpArrayInt1;
		Assert.that(field.components[0].type == FieldDescriptor.tpInt1);
		break;
	      case FieldDescriptor.tpClob:
		expr.type = ExprNode.tpClob;
		break;
	      case FieldDescriptor.tpBlob:
		expr.type = ExprNode.tpBlob;
		break;
	      case FieldDescriptor.tpReference:
		expr.type = ExprNode.tpRef;
		break;
	      default:
		Assert.failed("Invalid field type " + field.type);
	    }
	} else { 	    
	    if ((field.type == FieldDescriptor.tpBool && expr.type == ExprNode.tpBool)
		|| (field.type >= FieldDescriptor.tpInt1 && field.type <= FieldDescriptor.tpTimestamp
		    && (expr.type == ExprNode.tpInt || expr.type == ExprNode.tpReal || expr.type == ExprNode.tpStr))
		|| (field.type == FieldDescriptor.tpString && expr.type == ExprNode.tpStr)
		|| (field.type == FieldDescriptor.tpReference && expr.type == ExprNode.tpRef)
		|| (field.type == FieldDescriptor.tpArray 
		    && field.components[0].type == FieldDescriptor.tpInt1 
		    && expr.type == ExprNode.tpArrayInt1))
	    {
		return;
	    }
	    throw new ConsusCompileError("Type of expresion is incompatible with column type", pos);
	}
    }

    void evaluate(ConsusStatement stmt) throws SQLException { 
	Object val = null;
	switch (value.type) { 
	  case ExprNode.tpBool:
	    val = new Boolean(value.evaluateBool(stmt));
	    break;
	  case ExprNode.tpInt:
	    val = new Long(value.evaluateInt(stmt));
	    break;
	  case ExprNode.tpReal:
	    val = new Double(value.evaluateReal(stmt));
	    break;
	  case ExprNode.tpStr:
	    val = value.evaluateStr(stmt);
	    break;
	  case ExprNode.tpArrayInt1:
	    val = value.evaluateBinary(stmt);
	    break;
	  case ExprNode.tpBlob:
	    val = ((ParameterNode)value).obj;
	    break;
	  case ExprNode.tpClob:
	    val = ((ParameterNode)value).obj;
	    break;
	  case ExprNode.tpRef:
	    val = new ConsusRef(value.evaluateOid(stmt));
	    break;
	  default:
	    Assert.failed("Invalid value type " + value.type);
	}
	stmt.columnValue[field.id] = val;	
	stmt.updated[field.id] = true;	
    }

    ColumnValue(FieldDescriptor fd) { 
	field = fd;
    }
}

