/* Consus - Pure Java Object Relational Database Management System
 * JSQL compiler and interpreter
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import  java.sql.*;
import  java.lang.reflect.*;
import  java.util.Hashtable;
	
class Symbol { 
    int tkn;

    Symbol(int tkn) { 
	this.tkn = tkn;
    }
}


class Binding {
    Binding next;
    String  name;
    boolean used;
    int     loopId;

    Binding(String ident, int loop, Binding chain) { 
	name = ident;
	used = false;
	next = chain;
	loopId = loop;
    }
}


class IncorrectModeException extends ConsusCompileError {
    IncorrectModeException(int p) { 
	super("Subqueries and group functions can be used only in SELECT statement", p); 
    }
}

class Compiler extends ExprNode {
    int           pos;
    char[]        buf;
    char[]        str;
    long          ivalue;
    String        svalue;
    double        fvalue;
    String        query;
    String        ident;
    int           lex;
    int           vars;
    Binding       bindings;

    SelectNode    select;
    boolean       selectTuplesMode;
    
    int           context;
    static final int CTX_COLUMNS  = 0;
    static final int CTX_WHERE    = 1;
    static final int CTX_ORDER_BY = 2;
    static final int CTX_GROUP_BY = 3;
    static final int CTX_HAVING   = 4;

    ConsusConnection db;
    TableDescriptor  table;
    String           refTableName;

    ParameterNode[]  params;
    int              nParams;

    static Hashtable symtab;
    static Hashtable domainHash;


    final static int tknIdent = 1;
    final static int tknLpar = 2;
    final static int tknRpar = 3;
    final static int tknLbr = 4;
    final static int tknRbr = 5;
    final static int tknDot = 6;
    final static int tknComma = 7;
    final static int tknPower = 8;
    final static int tknIconst = 9;
    final static int tknSconst = 10;
    final static int tknFconst = 11;
    final static int tknAdd = 12;
    final static int tknSub = 13;
    final static int tknMul = 14;
    final static int tknDiv = 15;
    final static int tknAnd = 16;
    final static int tknOr = 17;
    final static int tknNot = 18;
    final static int tknNull = 19;
    final static int tknNeg = 20;
    final static int tknEq = 21;
    final static int tknNe = 22;
    final static int tknGt = 23;
    final static int tknGe = 24;
    final static int tknLt = 25;
    final static int tknLe = 26;
    final static int tknBetween = 27;
    final static int tknEscape = 28;
    final static int tknExists = 29;
    final static int tknLike = 30;
    final static int tknIn = 31;
    final static int tknLength = 32;
    final static int tknLower = 33;
    final static int tknUpper = 34;
    final static int tknAbs = 35;
    final static int tknIs = 36;
    final static int tknInteger = 37;
    final static int tknReal = 38;
    final static int tknString = 39;
    final static int tknBlob = 40;
    final static int tknClob = 41;
    final static int tknOid = 42;
    final static int tknCast = 43;
    final static int tknCol = 44;
    final static int tknTrue = 45;
    final static int tknFalse = 46;
    final static int tknWhere = 47;
    final static int tknOrder = 48;
    final static int tknAsc = 49;
    final static int tknDesc = 50;
    final static int tknSubstr= 51;
    final static int tknSin = 52;
    final static int tknCos = 53;
    final static int tknTan = 54;
    final static int tknAsin = 55;
    final static int tknAcos = 56;
    final static int tknAtan = 57;
    final static int tknSqrt = 58;
    final static int tknLog = 59;
    final static int tknExp = 60;
    final static int tknCeil = 61;
    final static int tknFloor = 62;
    final static int tknBy = 63;
    final static int tknPrimary = 64;
    final static int tknKey = 65;
    final static int tknUsing = 66;
    final static int tknCreate = 67;
    final static int tknDrop = 68;
    final static int tknDelete = 69;
    final static int tknUpdate = 70;
    final static int tknSet = 71;
    final static int tknValues = 72;
    final static int tknTable = 73;
    final static int tknIndex = 74;
    final static int tknOn = 75;
    final static int tknInto = 76;
    final static int tknTinyint = 77;
    final static int tknSmallint = 78;
    final static int tknBit = 79;
    final static int tknBigint = 80;
    final static int tknFloat = 81;
    final static int tknBoolean = 82;
    final static int tknDouble = 83;
    final static int tknChar = 84;
    final static int tknVarchar = 85;
    final static int tknBinary = 86;
    final static int tknVarbinary = 87;
    final static int tknDate = 88;
    final static int tknTime = 89;
    final static int tknTimestamp = 90;
    final static int tknReference = 91;
    final static int tknFrom = 92;
    final static int tknParam = 93;
    final static int tknRollback = 94;
    final static int tknCommit = 95;
    final static int tknSelect = 96;
    final static int tknInsert = 97;
    final static int tknFor = 98;
    final static int tknTo = 99;
    final static int tknStart = 100;
    final static int tknFollowing = 101;
    final static int tknFirst = 102;
    final static int tknLast = 103;
    final static int tknDomain = 104;
    final static int tknAlter = 105;
    final static int tknADD = 106;
    final static int tknModify = 107;
    final static int tknAll = 108;
    final static int tknDistinct = 109;
    final static int tknSysdate = 110;
    final static int tknCount = 111;
    final static int tknMin = 112;
    final static int tknMax = 113;
    final static int tknAvg = 114;
    final static int tknSum = 115;
    final static int tknAs  = 116;
    final static int tknAny  = 117;
    final static int tknSome  = 118;
    final static int tknHaving = 119;
    final static int tknGroup  = 120;
    final static int tknJoin   = 121;
    final static int tknNatural  = 122;
    final static int tknUnion  = 123;
    final static int tknNumeric  = 124;
    final static int tknDecimal  = 125;
    final static int tknRename   = 126;
    final static int tknUnique  = 127;
    final static int tknForeign  = 128;
    final static int tknReferences = 129;
    final static int tknCascade = 130;
    final static int tknConstraint = 131;
    final static int tknInt = 132;
    final static int tknLongvarbinary = 133;
    final static int tknLongvarchar = 134;
    final static int tknThis = 135;
    final static int tknEof = 136;

    static { 
	symtab = new Hashtable(1013);
	symtab.put("abs", new Symbol(tknAbs));
	symtab.put("acos", new Symbol(tknAcos));
	symtab.put("add", new Symbol(tknADD));
	symtab.put("all", new Symbol(tknAll));
	symtab.put("alter", new Symbol(tknAlter));
	symtab.put("and", new Symbol(tknAnd));
	symtab.put("any", new Symbol(tknAny));
	symtab.put("as", new Symbol(tknAs));
	symtab.put("asc", new Symbol(tknAsc));
	symtab.put("asin", new Symbol(tknAsin));
	symtab.put("atan", new Symbol(tknAtan));
	symtab.put("avg", new Symbol(tknAvg));
	symtab.put("between", new Symbol(tknBetween));
	symtab.put("binary", new Symbol(tknBinary));
	symtab.put("bigint", new Symbol(tknBigint));
	symtab.put("bit", new Symbol(tknBit));
	symtab.put("blob", new Symbol(tknBlob));
	symtab.put("boolean", new Symbol(tknBoolean));
	symtab.put("by", new Symbol(tknBy));
	symtab.put("cascade", new Symbol(tknCascade));
	symtab.put("cast", new Symbol(tknCast));
	symtab.put("char", new Symbol(tknChar));
	symtab.put("ceil", new Symbol(tknCeil));
	symtab.put("clob", new Symbol(tknClob));
	symtab.put("commit", new Symbol(tknCommit));
	symtab.put("constraint", new Symbol(tknConstraint));
	symtab.put("cos", new Symbol(tknCos));
	symtab.put("count", new Symbol(tknCount));
	symtab.put("create", new Symbol(tknCreate));
	symtab.put("date", new Symbol(tknDate));
	symtab.put("decimal", new Symbol(tknDecimal));
	symtab.put("delete", new Symbol(tknDelete));
	symtab.put("desc", new Symbol(tknDesc));
	symtab.put("distinct", new Symbol(tknDistinct));
	symtab.put("domain", new Symbol(tknDomain));
	symtab.put("double", new Symbol(tknDouble));
	symtab.put("drop", new Symbol(tknDrop));
	symtab.put("escape", new Symbol(tknEscape));
	symtab.put("exists", new Symbol(tknExists));
	symtab.put("exp", new Symbol(tknExp));
	symtab.put("false", new Symbol(tknFalse));
	symtab.put("first", new Symbol(tknFirst));
	symtab.put("float", new Symbol(tknFloat));
	symtab.put("floor", new Symbol(tknFloor));
	symtab.put("following", new Symbol(tknFollowing));
	symtab.put("for", new Symbol(tknFor));
	symtab.put("foreign", new Symbol(tknForeign));
	symtab.put("from", new Symbol(tknFrom));
	symtab.put("group", new Symbol(tknGroup));
	symtab.put("having", new Symbol(tknHaving));
	symtab.put("in", new Symbol(tknIn));
	symtab.put("index", new Symbol(tknIndex));
	symtab.put("into", new Symbol(tknInto));
	symtab.put("insert", new Symbol(tknInsert));
	symtab.put("is", new Symbol(tknIs));
	symtab.put("int", new Symbol(tknInt));
	symtab.put("integer", new Symbol(tknInteger));
	symtab.put("join", new Symbol(tknJoin));
	symtab.put("key", new Symbol(tknKey));
	symtab.put("last", new Symbol(tknLast));
	symtab.put("length", new Symbol(tknLength));
	symtab.put("like", new Symbol(tknLike));
	symtab.put("log", new Symbol(tknLog));
	symtab.put("longint", new Symbol(tknBigint));
	symtab.put("lower", new Symbol(tknLower));
	symtab.put("max", new Symbol(tknMax));
	symtab.put("min", new Symbol(tknMin));
	symtab.put("modify", new Symbol(tknModify));
	symtab.put("natural", new Symbol(tknNatural));
	symtab.put("not", new Symbol(tknNot));
	symtab.put("null", new Symbol(tknNull));
	symtab.put("numeric", new Symbol(tknNumeric));
	symtab.put("oid", new Symbol(tknOid));
	symtab.put("on", new Symbol(tknOn));
	symtab.put("or", new Symbol(tknOr));
	symtab.put("order", new Symbol(tknOrder));
	symtab.put("primary", new Symbol(tknPrimary));
	symtab.put("real", new Symbol(tknReal));
	symtab.put("reference", new Symbol(tknReference));
	symtab.put("references", new Symbol(tknReferences));
	symtab.put("rename", new Symbol(tknRename));
	symtab.put("rollback", new Symbol(tknRollback));
	symtab.put("select", new Symbol(tknSelect));
	symtab.put("set", new Symbol(tknSet));
	symtab.put("sin", new Symbol(tknSin));
	symtab.put("smallint", new Symbol(tknSmallint));
	symtab.put("some", new Symbol(tknSome));
	symtab.put("sqrt", new Symbol(tknSqrt));
	symtab.put("start", new Symbol(tknStart));
	symtab.put("string", new Symbol(tknString));
	symtab.put("sysdate", new Symbol(tknSysdate));
	symtab.put("current_date", new Symbol(tknSysdate)); // make it compatible with Cloudscape
	symtab.put("substr", new Symbol(tknSubstr));
	symtab.put("sum", new Symbol(tknSum));
	symtab.put("table", new Symbol(tknTable));
	symtab.put("tan", new Symbol(tknTan));
	symtab.put("time", new Symbol(tknTime));
	symtab.put("this", new Symbol(tknThis));
	symtab.put("timestamp", new Symbol(tknTimestamp));
	symtab.put("tinyint", new Symbol(tknTinyint));
	symtab.put("to", new Symbol(tknTo));
	symtab.put("true", new Symbol(tknTrue));
	symtab.put("values", new Symbol(tknValues));
	symtab.put("varbinary", new Symbol(tknVarbinary));
	symtab.put("varchar", new Symbol(tknVarchar));
	symtab.put("longvarbinary", new Symbol(tknLongvarbinary));
	symtab.put("longvarchar", new Symbol(tknLongvarchar));
	symtab.put("union", new Symbol(tknUnion));
	symtab.put("unique", new Symbol(tknUnique));
	symtab.put("update", new Symbol(tknUpdate));
	symtab.put("upper", new Symbol(tknUpper));
	symtab.put("using", new Symbol(tknUsing));
	symtab.put("where", new Symbol(tknWhere));
	domainHash = new Hashtable(113);
    }

    Compiler(ConsusConnection conxn, String sql) { 
	super(0, 0); 
	query = sql;
	buf = sql.toCharArray();
	str = new char[buf.length];
	pos = 0;
	db = conxn;
    }


    final int scan() throws SQLException {
	int p = pos;
	int eol = buf.length;
	char ch = 0;
	int i, j;
	while (true) { 
	    while (p < eol && Character.isWhitespace(ch = buf[p])) { 
		p += 1;
	    }
	    if (p == eol) { 
		pos = p;
		return tknEof;
	    }
	    pos = ++p;
	    switch (ch) { 
	      case '+':
		return tknAdd;
	      case '-':
		if (p < eol && buf[p] == '-') { 
		    while (++p < eol && buf[p] != '\n');
		    if (p == eol) {
			return tknEof;
		    }
		    continue;
		} 
		return tknSub;
	      case '*':
		return tknMul;
	      case '/':
		return tknDiv;
	      case '.':
		return tknDot;
	      case ',':
		return tknComma;
	      case '(':
		return tknLpar;
	      case ')':
		return tknRpar;
	      case '[':
		return tknLbr;
	      case ']':
		return tknRbr;
	      case ':':
		return tknCol;
	      case '^':
		return tknPower;
	      case '?':
		return tknParam;
	      case '<':
		if (p < eol) { 
		    if (buf[p] == '=') { 
			pos += 1;
			return tknLe;
		    } 
		    if (buf[p] == '>') { 
			pos += 1;
			return tknNe;
		    }
		}
		return tknLt;
	      case '>':
		if (p < eol && buf[p] == '=') { 
		    pos += 1;
		    return tknGe;
		} 
		return tknGt;
	      case '=':
		return tknEq;
	      case '!':
		if (p == eol || buf[p] != '=') { 
		    throw new ConsusCompileError("Invalid token '!'", p-1);
		} 
		pos += 1;
		return tknNe;
	      case '|':
		if (p == eol || buf[p] != '|') { 
		    throw new ConsusCompileError("Invalid token '!'", p-1);
		} 
		pos += 1;
		return tknAdd;
	      case '"':
		while (true) { 
		    if (p == eol) { 
			throw new ConsusCompileError("Unexpected end of quoted identifier", p);
		    }
		    if (buf[p] == '"') { 
			ident = query.substring(pos, p);
			pos = p+1;
			return tknIdent;
		    }
		    p += 1;
		}
	      case '\'':
		i = 0; 
		while (true) { 
		    if (p == eol) { 
			throw new ConsusCompileError("Unexpected end of string constant", p);
		    }
		    if (buf[p] == '\'') { 
			if (++p == eol || buf[p] != '\'') { 
			    svalue = new String(str, 0, i);
			    pos = p;
			    return tknSconst;
			}
		    }
		    str[i++] = buf[p++];
		}
	      case '0': case '1': case '2': case '3': case '4': 
	      case '5': case '6': case '7': case '8': case '9':
		i = p - 1;
		while (p < eol && Character.isDigit(ch = buf[p])) { 
		    p += 1;
		}
		if (ch == '.' || ch == 'e' || ch == 'E') { 
		    while (++p < eol 
			   && (Character.isDigit(buf[p]) 
			       || buf[p] == 'e' || buf[p] == 'E' || buf[p] == '.' 
			       || ((ch == 'e' || ch == 'E') 
				   && (buf[p] == '-' || buf[p] == '+'))));
		    pos = p;
		    try { 
			fvalue = 
			    Double.valueOf(query.substring(i, p)).doubleValue();
		    } catch(NumberFormatException x) { 
			throw new ConsusCompileError("Bad floating point constant", i);
		    }
		    return tknFconst;
		} else { 
		    pos = p;
		    try { 
			ivalue = Long.parseLong(query.substring(i, p), 10);
		    } catch(NumberFormatException x) { 
			throw new ConsusCompileError("Bad floating point constant", i);
		    }
		    return tknIconst;
		}
	      default:
		if (Character.isLetter(ch) || ch == '$' || ch == '_') { 
		    i = p-1;
		    j = 0;
		    if (ch != '_') {  
			ch |= (char)('a'-'A');
		    }
		    str[j++] = ch;
		    while (p < eol && (Character.isLetterOrDigit(ch = buf[p]) 
				       || ch == '$' || ch == '_'))
		    {
			if (ch != '_') {  
			    ch |= (char)('a'-'A');
			}
			str[j++] = ch;
			p += 1;
		    }
		    pos = p;
		    Symbol s = (Symbol)symtab.get(new String(str, 0, j));
		    if (s == null) { 
			ident = query.substring(i, p);
			return tknIdent;
		    }
		    return s.tkn;
		} else { 
		    throw new ConsusCompileError("Invalid symbol: " + ch, p-1);
		}
	    }
	}
    }
	
    final void expect(int tkn, String what) throws SQLException { 
	int p = pos;
	if ((lex = scan()) != tkn) { 
	    throw new ConsusCompileError("Expect " + what, p);
	}
    }

    final void expected(int tkn, String what) throws SQLException { 
	if (lex != tkn) { 
	    throw new ConsusCompileError(what + " expected", pos-1);
	}
    }

    final void allocateParameters() { 
	int n = 0;
	int len = buf.length;
	for (int i = 0; i < len; i++) { 
	    if (buf[i] == '\'') { 
		while (++i < len && buf[i] != '\'');
	    } else if (buf[i] == '?') { 
		n += 1;
	    }
	}
	if (n > 0) { 
	    params = new ParameterNode[n];
	} else { 
	    params = null;
	}
    }


    final FieldDescriptor getColumn() throws SQLException {
	int p = pos;
	expect(tknIdent, "column name");
	String column = ident;
	while ((lex = scan()) == tknDot) { 
	    expect(tknIdent, "column name");
	    column += "." + ident;
	}
	FieldDescriptor fd = table.getColumn(column);
	if (fd == null) { 
	    throw new ConsusCompileError("No such column " + column + " in table " + table.name, p);
	}
	return fd;
    }

    final OrderNode orderBy() throws SQLException { 
	int p = pos;
	if (select.columns != null) { 
	    int i = -1;
	    lex = scan();
	    if (lex == tknIdent) { 
		ExprNode expr = objectField(null);
		if (expr instanceof ColumnNode) { 
		    i = ((ColumnNode)expr).index - 1;	    
		}
	    } else if (lex == tknIconst) {
		i = (int)ivalue - 1;
		lex = scan();
	    } else { 
		throw new ConsusCompileError("Column name or index expected", p);
	    }
	    if (i < 0 || i >= select.columns.length) { 
		throw new ConsusCompileError("Invalid column specifier in " + 
					     ((context == CTX_ORDER_BY) ? "order" : "group") + " by part", p);
	    }
	    boolean ascent = true;
	    switch (lex) { 
	      case tknAsc:
		lex = scan();
		break;
	      case tknDesc:
		ascent = false;
		lex = scan();
	    }
	    select.columns[i].sorted = true;
	    return new OrderNode(i, ascent);
	} 
	FieldDescriptor fd = getColumn();
	if (fd.type == FieldDescriptor.tpArray) { 
	    throw new ConsusCompileError("Can not use binary fields in order by clause", p);
	}
	boolean ascent = true;
	switch (lex) { 
	  case tknAsc:
	    lex = scan();
	    break;
	  case tknDesc:
	    ascent = false;
	    lex = scan();
	}
	return new OrderNode(db, fd, ascent);
    }
	    
    final TableDescriptor getTable() throws SQLException
    {
	int p = pos;
	expect(tknIdent, "table name");        
	String tableName = ident;
	while ((lex = scan()) == tknDot) { 
	    expect(tknIdent, "class name");
	    tableName += "." + ident;
	}
	if (table != null && tableName.equals(table.name)) { 
	    return table;
	}
	table = db.getTableByName(tableName);
	if (table == null) { 
	    throw new ConsusCompileError("Unknown table " + tableName, p);
	}
	return table;
    }

    final TableDescriptor lookupTable() throws SQLException
    {
	String tableName = ident;
	while ((lex = scan()) == tknDot) { 
	    expect(tknIdent, "class name");
	    tableName += "." + ident;
	}
	TableDescriptor t = db.getTableByName(tableName);
	if (t == null) { 
	    throw new ConsusCompileError("Unknown table " + tableName, pos-1);
	}
	return t;
    }


    final ExprNode condition() throws SQLException { 
	int p = pos;
	ExprNode node = disjunction();
	if (node == null || node.type != tpBool) {
	    throw new ConsusCompileError("Boolean expression expected", p);
	}
	return node;
    }

    final IteratorVar getIterator() throws SQLException { 
	FieldDescriptor fd = getColumn();
	if (fd.type != FieldDescriptor.tpReference 
	    && (fd.type != FieldDescriptor.tpArray 
		|| fd.components[0].type != FieldDescriptor.tpReference)) 
	{
	    throw new ConsusCompileError("Iterator field should be of references or array of reference type", pos-1);
	}
	IteratorVar var = new IteratorVar();
	var.field = fd;
	return var;
    }


    final void getReferencedPrimaryKey(FieldDescriptor fd) throws SQLException
    {
	String column;
	TableDescriptor thisTable = table;
	TableDescriptor refTable = getTable();
	int p = pos;
	if (lex == tknLpar) {
	    lex = scan();
	    if (lex == tknOid) { 
		p = pos;
		lex = scan();
	        column = "oid";
	    } else if (lex == tknIdent) { 			
		column = ident;
		while ((lex = scan()) == tknDot) { 
		    expect(tknIdent, "column name");
		    column += "." + ident;
		    p = pos;
		}
		if (refTable.getColumn(column) == null) { 
		    throw new ConsusCompileError("No such column " + column + " in table " + refTable.name, p);
		}
	    } else { 
 		throw new ConsusCompileError("Column name expected ", p-1);
	    }
	    if (lex != tknRpar) { 
		throw new ConsusCompileError("')' expected ", p-1);
	    }
	    lex = scan();
	} else { 
	    column = fd.longName;
	    if (refTable.getColumn(column) == null) { 
		throw new ConsusCompileError("No such column " + column + " in table " + refTable.name, p);
	    }
	}
	table = thisTable;
	fd.fkKeyName = column;
	fd.fkTableName = refTable.name;	
    }


    final StmtNode alterTable() throws SQLException { 
	int i;
	int p = pos;
	expect(tknTable, "TABLE");
	getTable();
	FieldDescriptor fd, fds[];
	FieldDescriptor modifiedFields = null;
	FieldDescriptor droppedFields = null;
	FieldDescriptor addedFields = null;
	boolean changedConstraints = false;
	RenameNode renameNodes = null;
	int op = lex;
	while (true) { 
	    if (op == tknADD || op == tknModify || op == tknDrop) { 
		boolean lpar = false;
		p = pos;
		lex = scan();
		switch (lex) { 
		  case tknLpar:
		    lex = scan();
		    lpar = true;
		    break;
		  case tknUnique:
		    fds = getConstraintField();
		    for (i = 0; i < fds.length; i++) { 
			fd = fds[i];
			if (op == tknDrop) { 
			    fd.attr &= ~FieldDescriptor.Unique;
			} else if (op == tknAdd) { 
			    if (fds.length == 1) { 
				fd.attr |= FieldDescriptor.Unique;
			    }
			} else { 
			    throw new ConsusCompileError("DROP UNIQUE or ADD UNIQUE expected", p);
			}
		    }
		    changedConstraints = true;
		    lex = op = scan();
		    continue;
		  case tknPrimary:
		    expect(tknKey, "KEY");
		    fds = getConstraintField();
		    for (i = 0; i < fds.length; i++) { 
			fd = fds[i];
			if (op == tknDrop) {
			    fd.attr &= ~FieldDescriptor.PrimaryKey;
			} else if (op == tknAdd) { 
			    fd.attr |= FieldDescriptor.PrimaryKey;
			} else { 
			    throw new ConsusCompileError("DROP PRIMARY KEY or ADD PRIMARY KEY expected", p);
			}
		    }
		    changedConstraints = true;		    
		    op = lex = scan();
		    continue;
		  case tknForeign:
		    expect(tknKey, "KEY");
		    fds = getConstraintField();
		    fd = fds[0];
		    if (op == tknDrop) { 
			fd.attr &= ~(FieldDescriptor.ForeignKey|FieldDescriptor.OnDeleteCascade);
			fd.refTable = null;
			fd.className = null;
		    } else { 
			fd.attr |= FieldDescriptor.ForeignKey;
			expect(tknReferences, "REFERENCES");
			getReferencedPrimaryKey(fd);
			if (lex == tknOn) { 
			    expect(tknDelete, "DELETE");
			    expect(tknCascade, "CASCADE");
			    fd.attr |= FieldDescriptor.OnDeleteCascade;
			}
		    }
		    changedConstraints = true;		    
		    op = lex = scan();
		    continue;		    
		}
		while (true) { 
		    expected(tknIdent, "field name");
		    FieldDescriptor prev;
		    switch (op) { 
		      case tknDrop:
			fd = new FieldDescriptor(table, ident);
			fd.next = droppedFields;
			droppedFields = fd;
			lex = scan();
			break;
		      case tknModify:
			fd = new FieldDescriptor(table, ident);
			setColumnType(fd);
			fd.next = modifiedFields;
			modifiedFields = fd;
			break;
		      case tknADD:
			fd = new FieldDescriptor(table, ident);
			setColumnType(fd);
			fd.next = addedFields;
			addedFields = fd;
		    }
		    if (lex != tknComma) { 
			break;
		    }
		    lex = scan();
		}
		op = lex;
		if (lpar) { 
		    expected(tknRpar, "')'");
		    op = scan();
		} 
	    } else if (op == tknRename) {
		boolean lpar = false;
		p = pos;
		lex = scan();
		if (lex == tknLpar) { 
		    lex = scan();
		    lpar = true;
		}
		while (true) { 
		    String oldColumnName;
		    if (lex == tknIdent) { 
			oldColumnName = ident;
			while ((lex = scan()) == tknDot) { 
			    expect(tknIdent, "Component name");
			    oldColumnName += "." + ident;
			}
		    } else { 
			expected(tknSconst, "Column name expected"); 
			oldColumnName = svalue;
			lex = scan();
		    } 
		    if (lex == tknAs) { 
			lex = scan();
		    }
		    String newColumnName;
		    if (lex == tknIdent) { 
			newColumnName = ident;
			while ((lex = scan()) == tknDot) { 
			    expect(tknIdent, "Component name");
			    newColumnName += "." + ident;
			}
		    } else { 
			expected(tknSconst, "Column name expected"); 
			newColumnName = svalue;
			lex = scan();
		    } 
		    renameNodes = new RenameNode(oldColumnName, newColumnName, renameNodes);
		    if (lex != tknComma) { 
			break;
		    }
		    lex = scan();
		}
		op = lex;
		if (lpar) { 
		    expected(tknRpar, "')'");
		    op = scan();
		} 
	    } else { 		
		lex = op;
		return new AlterTableNode(table, addedFields, modifiedFields, droppedFields,
					  renameNodes, changedConstraints);
	    }
	}
    }


    final ColumnNode addColumns(TableNode tabref, int nColumns, ColumnNode list, boolean join) { 
	if (join && tabref.joinTable != null) { 
	    list = addColumns(tabref.joinTable, nColumns, list, true);
	    if (list != null) { 
		nColumns = list.index;
	    }
	}
	TableDescriptor table = tabref.table;
      forAllColumns:
	for (int i = 0, n = table.nColumns; i < n; i++) { 
	    FieldDescriptor fd = table.columns[i];
	    for (JoinColumn jc = tabref.joinColumns; jc != null; jc = jc.next) { 
		if (jc.right == fd) { 
		    continue forAllColumns;
		}
	    }
	    list = new ColumnNode(++nColumns, LoadNode.create(fd, tabref), null, list);
	}
	return list;
    }

    // Select attributes
    final static int SA_SUBQUERY  = 1;
    final static int SA_EXISTS    = 2;
    final static int SA_SET_UNION = 4; 
    
    final ExprNode addJoinCondition(TableNode leftTable, FieldDescriptor leftColumn, 
				    TableNode rightTable, FieldDescriptor rightColumn, 
				    ExprNode joinCondition) throws SQLException
    {	
	ExprNode left = LoadNode.create(leftColumn, leftTable);
	ExprNode right = LoadNode.create(rightColumn, rightTable);
	int tag;
	switch (left.type) {
	  case ExprNode.tpInt:
	    tag =  opIntEq;
	    break;
	  case ExprNode.tpReal:
	    tag =  opRealEq;
	    break;
	  case ExprNode.tpBool:
	    tag =  opBoolEq;
	    break;
	  case ExprNode.tpRef:
	    tag =  opOidEq;
	    break;
	  case ExprNode.tpStr:
	    tag =  opStrEq;
	    break;
	  default:
	    throw new ConsusCompileError("Invalid type for JOIN column", pos-1);
	}
	ExprNode condition = new BinOpNode(tpBool, tag, left, right);
	if (joinCondition != null) { 
	    if (joinCondition.tag == opBoolAnd) { 
		((BinOpNode)joinCondition).right = condition = 
		    new BinOpNode(tpBool, opBoolAnd, ((BinOpNode)joinCondition).right, condition);
	    } else { 
		condition = new BinOpNode(tpBool, opBoolAnd, joinCondition, condition);
	    }
	}
	return condition;
    }
    
    final SelectNode selectExpr(int selectAttr) throws SQLException { 
	SelectNode stmt;
	if (lex == tknLpar) { 
	    lex = scan();
	    stmt = selectExpr(selectAttr);
	    expected(tknRpar, "')'");
	    lex = scan();
	} else if (lex == tknSelect) { 
	    stmt = selectStatement(selectAttr);
	} else { 
	    throw new ConsusCompileError("SELECT expected", pos-1);
	}
	return stmt;
    }
	    
    final SelectNode selectStatement(int selectAttr) throws SQLException { 
	int i;
	String tableName;
	SelectNode outer = select;
	if (outer != null && !selectTuplesMode) { 
	    throw new IncorrectModeException(pos);
	}
	int p = pos;
	lex = scan();
	select = new SelectNode(outer);
	select.selectType = SelectNode.SELECT_DEFAULT;
	if (lex == tknAll) { 
	    select.selectType = SelectNode.SELECT_ALL;
	    p = pos;
	    lex = scan();
	} else if (lex == tknDistinct) { 
	    select.selectType = SelectNode.SELECT_DISTINCT;
	    p = pos;
	    lex = scan();
	}
	boolean selectAllColumns = false;
	if (lex == tknMul) { 
	    selectAllColumns = true;
	    p = pos;
	    lex = scan();
	}
	if (lex != tknFrom || db.getContext().selectTuplesMode) {
	    selectTuplesMode = true;
	}

      selectObjects:
	if (!selectTuplesMode) { 
	    lex = scan();
	    if (lex != tknIdent) { 
		selectTuplesMode = true;
		pos = p;
		lex = scan();
		break selectObjects;
	    }
	    table = lookupTable();
	    if (lex == tknJoin || lex == tknNatural || lex == tknComma) { 
		selectTuplesMode = true;
		pos = p;
		lex = scan();
		break selectObjects;
	    }
	    select.table = table;
	    if (lex == tknWhere) { 
		context = CTX_WHERE;
		select.condition = condition();
	    }
	    if (lex == tknStart) { 
		expect(tknFrom, "FROM");
		IteratorNode iterator = new IteratorNode(table);
		p = pos;
		switch (scan()) { 
		  case tknFirst:
		    iterator.type = IteratorNode.START_FROM_FIRST;
		    break;
		  case tknLast:
		    iterator.type = IteratorNode.START_FROM_LAST;
		    break;
		  case tknParam:
		    iterator.type = IteratorNode.START_FROM_PARAM;
		    iterator.param = params[nParams++] = new ParameterNode(tpCursor);
		    break;
                  case tknIconst:
		    iterator.type = IteratorNode.START_FROM_PARAM;
		    iterator.param = new ParameterNode(tpRef);
                    iterator.param.itype = tpRef;
                    iterator.param.oid = (int)ivalue;
		    break;
		  default:
		    throw new ConsusCompileError("FIRST, LAST or ? expected", pos-1);
		}
		if ((lex = scan()) == tknFollowing) { 
		    expect(tknBy, "BY");
		    IteratorVar var = getIterator();
		    iterator.fields = var;
		    while (lex == tknComma) { 
			var = var.next = getIterator();
		    }
		}
		select.iterator = iterator;
	    }
	    if (lex == tknGroup) { 
		selectTuplesMode = true;
		pos = p;
		lex = scan();
	    }
	}
      selectTuples:
	if (selectTuplesMode) { 
	    while (lex != tknFrom) { 
		if (lex == tknEof) { 
		    throw new ConsusCompileError("FROM expected", p);
		}
		lex = scan();
	    }
	    int nTables = 0;	       
	    TableNode tableList = null;
	    boolean join = false;
	    boolean naturalJoin = false;
	    ExprNode joinCondition = null;
	    ExprNode lastJoinCondition = null;
	    int nJoinConditions = 0;
	    int nParentheses = 0;
	    while (true) {
		while ((lex = scan()) == tknLpar) { 
		    nParentheses += 1;
		}
		expected(tknIdent, "Table name");
		tableName = ident;
		while ((lex = scan()) == tknDot) { 
		    expect(tknIdent, "class name");
		    tableName += "." + ident;
		}
		table = db.getTableByName(tableName);
		if (table == null) { 
		    throw new ConsusCompileError("Unknown table " + tableName, pos-1);
		}
		tableList = new TableNode(table, ++nTables, tableList);
		if (naturalJoin) { 
		    JoinColumn joinColumns = null;
		    TableNode joinList = tableList.next;
		    do { 
			TableDescriptor leftTable = joinList.table;
			for (i = table.columns.length; --i >= 0;) { 
			    FieldDescriptor rightColumn = table.columns[i];
			    FieldDescriptor leftColumn = leftTable.getColumn(rightColumn.longName);
			    if (leftColumn != null) { 
				if (leftColumn.type != rightColumn.type) { 
				    throw new ConsusCompileError("Types of columns for NATURAL JOIN do not match", pos-1);
				}
				lastJoinCondition = addJoinCondition(tableList.next, leftColumn, 
								     tableList, rightColumn, 
								     lastJoinCondition);
				if (++nJoinConditions == 2) { 
				    joinCondition = lastJoinCondition;
				}
				joinColumns = new JoinColumn(leftColumn, rightColumn, joinColumns);
			    }
			} 			
		    } while ((joinList = joinList.joinTable) != null);
		    if (joinColumns == null) { 
			throw new ConsusCompileError("NATURAL JOIN not possible", pos-1);
		    }
		    tableList.joinColumns = joinColumns;
		    tableList.joinTable = tableList.next;
		} else if (join) { 
		    expected(tknUsing, "USING");
		    JoinColumn joinColumns = null;
		    do { 
			expect(tknIdent, "column name");
			FieldDescriptor[] columns = table.columns;
			FieldDescriptor rightColumn = table.getColumn(ident);
			TableNode joinList = tableList.next;
			FieldDescriptor leftColumn = null;
			while (joinList != null && (leftColumn = joinList.table.getColumn(ident)) == null) { 
			    joinList = joinList.joinTable;
			}
			if (leftColumn == null || rightColumn == null 
			    || leftColumn.type != rightColumn.type)
			{
			    throw new ConsusCompileError("JOIN column '" + ident 
			       + "' doesn't present in both tables or types mismatch", pos-1);
			}
			lastJoinCondition = addJoinCondition(tableList.next, leftColumn, 
							 tableList, rightColumn, 
							 lastJoinCondition);
			if (++nJoinConditions == 2) { 
			    joinCondition = lastJoinCondition;
			}
			joinColumns = new JoinColumn(leftColumn, rightColumn, joinColumns);
		    } while ((lex = scan()) == tknComma);
		    tableList.joinTable = tableList.next;
		    tableList.joinColumns = joinColumns;
		} 
		while (lex == tknRpar && nParentheses != 0) { 
		    nParentheses -= 1;
		    lex = scan();
		}
		if (lex == tknAs) { 
		    expect(tknIdent, "table alias name");
		    tableList.name = ident;
		    lex = scan();
		} else if (lex == tknIdent) { 
		    tableList.name = ident;
		    lex = scan();
		}		
		join = naturalJoin = false;
		if (lex == tknNatural) { 
		    expect(tknJoin, "JOIN");
		    naturalJoin = true;
		} else if (lex == tknJoin) { 
		    join = true;
		} else if (lex != tknComma) { 
		    break;
		}
	    }
	    if (nParentheses != 0) { 
		throw new ConsusCompileError("')' expected", pos-1);
	    }
	    select.tables = new TableNode[nTables];
	    i = nTables;
	    do { 
		select.tables[--i] = tableList;
	    } while((tableList = tableList.next) != null);
	    
	    int savePos = pos;
	    int saveLex = lex;
	    int nColumns = 0;
	    ColumnNode columnList = null;
	    
	    if ((selectAttr & SA_EXISTS) != 0) { 
		// replace "exists (select ...)" with "(select count(*) ...) <> 0"
		select.selectType = SelectNode.SELECT_COUNT;
		select.existsOptimization = true;
		columnList = new ColumnNode(1, new GroupNode(tpInt, opCountAll,
							     new IntLiteralNode(1)), null, null);
		nColumns = 1;
	    } else { 
		pos = p; // returns back to the fields list
		context = CTX_COLUMNS;
		if (selectAllColumns) { 
		    for (i = 0; i < nTables; i++) { 
			columnList = addColumns(select.tables[i], nColumns, columnList, false);
			if (columnList != null) { 
			    nColumns = columnList.index;
			}
		    }
		    lex = scan();
		} else { 
		  scan_columns:
		    do { 
			p = pos;
			if (scan() == tknIdent) { 
			    // handle T.* 
			    tableName = ident;
			    while (scan() == tknDot) { 
				if ((lex = scan()) == tknMul) { 
				    for (i = 0; i < nTables; i++) { 
					if (tableName.equals(select.tables[i].name)) { 
					    columnList = addColumns(select.tables[i], nColumns, columnList, true);
					    if (columnList != null) { 
						nColumns = columnList.index;
					    }
					    lex = scan();
					    continue scan_columns;
					}
				    }
				    throw new ConsusCompileError("Invalid table reference: " + tableName, p);
				} else if (lex == tknIdent) { 
				    tableName += "." + ident;
				} else { 
				    break;
				}
			    }
			}  
			pos = p; 
			ExprNode expr = disjunction();
			String alias = null;
			if (lex == tknAs) { 
			    expect(tknIdent, "alias name");
			    alias = ident;
			    lex = scan();
			} else if (lex == tknIdent) { 
			    alias = ident;
			    lex = scan();
			}
			columnList = new ColumnNode(++nColumns, expr, alias, columnList);
		    } while (lex == tknComma);
		}
		expected(tknFrom, "FROM");
	    }
	    select.groupColumns = select.resultColumns = select.columns 
		= new ColumnNode[nColumns];
	    i = nColumns;
	    do { 
		select.columns[--i] = columnList;
	    } while ((columnList = columnList.next) != null);
	    Assert.that(i == 0);
	    ColumnNode groups = null;
	    for (i = 0; i < nColumns; i++) { 
		ColumnNode column = select.columns[i];
		if (column.expr instanceof GroupNode) { 
		    if (groups == null) { 
			select.columns = new ColumnNode[nColumns]; 
			System.arraycopy(select.groupColumns, 0, select.columns, 0, nColumns);
		    }
		    select.columns[i] = new ColumnNode(column.index, 
						       ((GroupNode)column.expr).expr, 
						       null, null);
		    ((GroupNode)column.expr).columnNo = i;
		    column.next = groups;
		    groups = column;
		}
	    }
	    if (groups != null) { 
		select.groups = groups;
		if (select.columns.length == 1 && groups.expr.tag == opCountAll) { 
		    select.selectType = SelectNode.SELECT_COUNT;
		}
	    }
	    lex = saveLex;
	    pos = savePos;
	    if (lex == tknWhere) { 
		context = CTX_WHERE;
		select.condition = condition();
	    }
	    if (nJoinConditions > 0) { 
		if (select.condition == null) { 
		    select.condition = nJoinConditions == 1 ? lastJoinCondition : joinCondition;
		} else if (nJoinConditions == 1) { 
		    select.condition = 
			new BinOpNode(tpBool, opBoolAnd, lastJoinCondition, select.condition);
		} else { 
		    ((BinOpNode)lastJoinCondition).right 
			= new BinOpNode(tpBool, opBoolAnd, ((BinOpNode)lastJoinCondition).right,
					select.condition);
		    select.condition = joinCondition;
		}
	    }
	    if (select.condition != null) { 
		ExprNode and;
		int nConjuncts = 1;
		for (and = select.condition; and.tag == opBoolAnd; and = ((BinOpNode)and).right) { 
		    nConjuncts += 1;
		}
		if (nConjuncts > 1) { 
		    ExprNode conjuncts[] = new ExprNode[nConjuncts];
		    for (i=0, and=select.condition; and.tag == opBoolAnd; and = ((BinOpNode)and).right) {
			conjuncts[i++] = ((BinOpNode)and).left;
		    }
		    conjuncts[i] = and;
		    java.util.Arrays.sort(conjuncts);
		    and = new BinOpNode(tpBool, opBoolAnd, conjuncts[i-1], conjuncts[i]);
		    i -= 1;
		    while (--i >= 0) {
			and = new BinOpNode(tpBool, opBoolAnd, conjuncts[i], and);
		    }
		    select.condition = and;
		}
	    }
	    if (lex == tknGroup) { 
		context = CTX_GROUP_BY;
		expect(tknBy, "BY");
		OrderNode last = select.groupBy = orderBy();
		while (lex == tknComma) { 
		    last = last.next = orderBy();
		}	    
		if (lex == tknHaving) { 
		    context = CTX_HAVING;
		    select.having = condition();
		}
	    }
	    if (lex == tknStart) { 
		throw new ConsusCompileError("START FROM .. FOLLOWING BY construction is not possible for tuples", pos-1);
	    }
	}
	OrderNode last = null;
	if (lex == tknOrder) { 
	    expect(tknBy, "BY");
	    context = CTX_ORDER_BY;
	    last = select.order = orderBy();
	    while (lex == tknComma) { 
		last = last.next = orderBy();
	    }
	}
	if (lex == tknFor) { 
	    expect(tknUpdate, "UPDATE");
	    select.forUpdate = true;
	    lex = scan();
	}
	if (selectTuplesMode) {
	    if ((selectAttr & SA_SUBQUERY) != 0) { 
		if (select.resultColumns.length == 1) { 
		    select.type = select.resultColumns[0].type;
		} else { 
		    throw new ConsusCompileError("Sunquery should return the single column", p);
		}
	    }
	}
	SelectNode stmt = select;
	select = outer;
	if (lex == tknUnion) { 
	    selectAttr |= SA_SET_UNION;
	    if ((lex = scan()) == tknAll) { 
		selectAttr &= ~SA_SET_UNION;
		lex = scan();
	    }
	    SelectNode unionStmt = selectExpr(selectAttr);
	    if ((stmt.resultColumns != null 
		 && (unionStmt.resultColumns == null 
		     || stmt.resultColumns.length != unionStmt.resultColumns.length)) ||
		(stmt.resultColumns == null 
		 && (unionStmt.resultColumns != null 
		     || stmt.table.isSuperclassFor(unionStmt.table))))
	    {
		throw new ConsusCompileError("Operands of UNION operator should return objects "+
					     "of the same class or tuples with equal number of columns", pos-1);
	    }
	    unionStmt.isUnion = true;
	    unionStmt.isSetUnion = (selectAttr & SA_SET_UNION) != 0;
	    if (unionStmt.forUpdate) { 
		stmt.forUpdate = true;
	    }
	    stmt.unionWith = unionStmt;
	}
	return stmt;
    }

    final StmtNode statement() throws SQLException { 
	int p;
	ColumnValue first = null;
	FieldDescriptor fd;
	ExprNode criteria = null;
	StmtNode stmt;

	allocateParameters();

	switch (scan()) { 
	  case tknCommit:
	    stmt = new CommitNode();
	    lex = scan();
	    break;

	  case tknRollback:	    
	    stmt = new RollbackNode();
	    lex = scan();
	    break;

	  case tknAlter:
	    stmt = alterTable();
	    break;

	  case tknCreate:
	    p = pos;
	    switch (scan()) { 
	      case tknTable:
		stmt = createTable();
		break;

	      case tknIndex:
		lex = scan();
		if (lex == tknIdent) { 
		    lex = scan();
		}
		expected(tknOn, "ON");
		getTable();
		p = pos;
	        expected(tknLpar, "'('");
		fd = getColumn();
		if (fd.type == FieldDescriptor.tpArray) { 
		    throw new ConsusCompileError("Binary fields can not be indexed", p);
		}
		while (lex == tknComma) { 
		    getColumn();
		}
		expected(tknRpar, "')'");
		stmt = new AddIndexNode(table, fd);
		lex = scan();
		break;

	      case tknDomain:
		p = pos;
		lex = scan();
		if (lex != tknIdent) { 
		    throw new ConsusCompileError("Domain name expected", p);
		}
		FieldDescriptor domain = new FieldDescriptor(table, ident);
		setColumnType(domain);
		stmt = new CreateDomainNode(domain);
		break;
		
	      default:
		throw new ConsusCompileError("CREATE [TABLE, INDEX or DOMAIN] expected", p);
	    }
	    break;

	  case tknSelect:
	    p = pos;
	    try { 
		stmt = selectStatement(0); 
	    } catch(IncorrectModeException x) {
		pos = p;
		selectTuplesMode = true;
		stmt = selectStatement(0); 
	    }
	    break;
	    
	  case tknDelete:
	    expect(tknFrom, "FROM");
	    getTable();
	    if (lex == tknWhere) { 
		criteria = condition();
	    }
	    stmt = new DeleteNode(table, criteria); 
	    break;

	  case tknInsert:
	    expect(tknInto, "INTO");
	    getTable();
	    if (lex == tknLpar) {
		first = new ColumnValue(getColumn());
		ColumnValue last = first;
		while (lex == tknComma) { 
		    last = last.next = new ColumnValue(getColumn());
		}	
		expected(tknRpar, "')'");
		lex = scan();
	    } else { 
		for (int i = table.columns.length; --i >= 0;) {
		    ColumnValue cv = new ColumnValue(table.columns[i]);
		    cv.next = first;
		    first = cv;
		}
	    }
	    if (lex == tknParam) { 
		params[0] = new ParameterNode(tpRef);
		nParams = 1;
		stmt = new InsertObjectNode(table);
	    } else if (lex == tknValues) { 
		expect(tknLpar, "'('");
		p = pos;
		ColumnValue column = first;
		do { 
		    if (column == null) { 
			throw new ConsusCompileError("Too many values", p);
		    }
		    p = pos;
		    column.setValue(disjunction(), p);
		    column = column.next;
		} while (lex == tknComma);
		expected(tknRpar, "')'");
		if (column != null) { 
		    throw new ConsusCompileError("Too few values", p);
		}
		stmt = new InsertNode(table, first);
		lex = scan();
	    } else { 
		throw new ConsusCompileError("VALUES or parameter expected", pos-1);
	    }
	    break;

	  case tknUpdate:
	    getTable();
	    expected(tknSet, "SET");
	    first = null;
	    do { 
		ColumnValue column = new ColumnValue(getColumn());
		expected(tknEq, "'='");
                p = pos;
		column.setValue(disjunction(), p);	    
		column.next = first;
		first = column;
	    } while (lex == tknComma);

	    if (lex == tknWhere) { 
		criteria = condition();
	    }
	    stmt = new UpdateNode(table, criteria, first);
	    break;

	  case tknDrop:
	    p = pos;
	    switch (scan()) { 
	      case tknTable:
		getTable();
		stmt = new DropTableNode(table);
		break;

	      case tknIndex:
		getTable();
		expected(tknLpar, "(");
		fd = getColumn();
		expected(tknRpar, "')'");
		stmt = new DropIndexNode(table, fd);
		lex = scan();
		break;

	      default:
		throw new ConsusCompileError("DROP [TABLE or INDEX] expected", p);
	    }
	    break;

	  default:
	    throw new ConsusCompileError("Invalid JSQL statement", 0);
	}
	if (lex != tknEof) { 
	    throw new ConsusCompileError("JSQL syntax error", pos-1);
	}
	return stmt;
    }
    
    void setColumnType(FieldDescriptor fd) throws SQLException {
	int type;
	int p = pos;
	switch (scan()) {
	  case tknBit:
	  case tknBoolean:
	    type = FieldDescriptor.tpBool;
	    lex = scan();
	    break;
	  case tknChar:
	  case tknVarchar:
	  case tknLongvarchar:
	  case tknString:
	    type = FieldDescriptor.tpString;
	    if ((lex = scan()) == tknLpar) { 
		// skip fields length 
		expect(tknIconst, "field width");
		expect(tknRpar, "')'");
		lex = scan();
	    }
	    break;	      
	  case tknTinyint:
	    type = FieldDescriptor.tpInt1;
	    lex = scan();
	    break;	      
	  case tknSmallint:
	    type = FieldDescriptor.tpInt2;
	    lex = scan();
	    break;	      
	  case tknInteger:
	  case tknInt:
	    type = FieldDescriptor.tpInt4;
	    lex = scan();
	    break;	      
	  case tknBigint:
	    type = FieldDescriptor.tpInt8;
	    lex = scan();
	    break;
	  case tknFloat:
	    type = FieldDescriptor.tpReal4;
	    lex = scan();
	    break;
	  case tknReal:
	  case tknDouble:
	    type = FieldDescriptor.tpReal8;
	    lex = scan();
	    break;
	  case tknNumeric:
	  case tknDecimal:
	    type = FieldDescriptor.tpReal8;
	    if ((lex = scan()) == tknLpar) { 
		// skip fields length 
		expect(tknIconst, "field precision");
		if ((lex = scan()) == tknComma) { 
		    expect(tknIconst, "field scale");
		    lex = scan();
		}
		expected(tknRpar, "')'");
		lex = scan();
	    }
	    break;
	  case tknDate:
	    type = FieldDescriptor.tpDate;
	    lex = scan();
	    break;
	  case tknTime:
	    type = FieldDescriptor.tpTime;
	    lex = scan();
	    break;
	  case tknTimestamp:
	    type = FieldDescriptor.tpTimestamp;
	    lex = scan();
	    break;
	  case tknBinary:
	  case tknVarbinary:
	  case tknLongvarbinary:
	    type = FieldDescriptor.tpArray;
	    if ((lex = scan()) == tknLpar) { 
		// skip fields length 
		expect(tknIconst, "field width");
		expect(tknRpar, "')'");
		lex = scan();
	    }
	    break;
	  case tknReference:
	    type = FieldDescriptor.tpReference;
	    if ((lex = scan()) == tknTo) { 
		expect(tknIdent, "referenced table name");
		String tableName = ident;
		while ((lex = scan()) == tknDot) { 
		    expect(tknIdent, "class name");
		    tableName += "." + ident;
		}
		fd.className = tableName;
	    }
	    break;
	  case tknBlob:
	    type = FieldDescriptor.tpBlob;
	    lex = scan();
	    break;
	  case tknClob:
	    type = FieldDescriptor.tpClob;
	    lex = scan();
	    break;
	  case tknIdent:
	    FieldDescriptor domain = (FieldDescriptor)domainHash.get(ident);
	    if (domain != null) { 
		type = domain.type;
		fd.className = domain.className;
		fd.attr |= domain.attr;
		lex = scan();
		break;
	    }
	  default:
	    throw new ConsusCompileError("column type expected", p);
	}
	fd.type = type;
      constraints:while (true) { 
	    switch (lex) { 
	      case tknPrimary:
		expect(tknKey, "KEY");
		fd.attr |= FieldDescriptor.PrimaryKey;
		if (type != FieldDescriptor.tpArray) { 
		    fd.attr |= FieldDescriptor.Indexed;
		}
		if ((lex = scan()) == tknIdent) { 
		    lex = scan();
		}
		break;
	      case tknUnique:
		fd.attr |= FieldDescriptor.Unique;
		if (type != FieldDescriptor.tpArray) { 
		    fd.attr |= FieldDescriptor.Indexed;
		}
		lex = scan();
		break;
	      case tknUsing:
		expect(tknIndex, "INDEX");
		if (type == FieldDescriptor.tpArray) { 
		    throw new ConsusCompileError("Binary fields can not be indexed", p);
		}
		lex = scan();
		fd.attr |= FieldDescriptor.Indexed;
		break;
	      case tknForeign:
		expect(tknKey, "KEY");
		expect(tknReferences, "REFERENCES");
		// no break
	      case tknReferences:
		fd.attr |= FieldDescriptor.ForeignKey;
		getReferencedPrimaryKey(fd);
		if (lex == tknOn) { 
		    expect(tknDelete, "DELETE");
		    expect(tknCascade, "CASCADE");
		    fd.attr |= FieldDescriptor.Indexed|FieldDescriptor.OnDeleteCascade;
		}
		break;
	      case tknNot:
		expect(tknNull, "NULL");
		// no break
	      case tknNull:
		lex = scan();
		break;
	      default: 
		break constraints;
	    }
	}
    }	

    final FieldDescriptor[] getConstraintField() throws SQLException { 
	expect(tknLpar, "'('");
	FieldDescriptor[] fds = new FieldDescriptor[1];
	int i = 0;
	do { 
	    FieldDescriptor fd = getColumn();
	    if (i == fds.length) { 
		FieldDescriptor[] newFds = new FieldDescriptor[i+1];
		System.arraycopy(fds, 0, newFds, 0, i);
		fds = newFds;
	    } 
	    fds[i++] = fd;
	} while (lex == tknComma);
	if (lex != tknRpar) { 
	    throw new ConsusCompileError("')' expected", pos-1);
	}
	return fds;
    }

    final void parseConstraints() throws SQLException { 
	FieldDescriptor fd, fds[];
	int p = pos;
	do { 
	    int i;
	    p = pos;
	    lex = scan();
	    switch (lex) { 
	      case tknPrimary:
		expect(tknKey, "KEY");
		fds = getConstraintField();
		for (i = 0; i < fds.length; i++) { 
		    fd = fds[i];
		    fd.attr |= FieldDescriptor.PrimaryKey;
		    if (fd.type != FieldDescriptor.tpArray) { 
			fd.attr |= FieldDescriptor.Indexed;
		    }
		}
		lex = scan();
		break;
	      case tknForeign:
		expect(tknKey, "KEY");
		fds = getConstraintField();
		fd = fds[0];
		expect(tknReferences, "REFERENCES");
		getReferencedPrimaryKey(fd);
		fd.attr |= FieldDescriptor.ForeignKey;
		if (lex == tknOn) { 
		    expect(tknDelete, "DELETE");
		    expect(tknCascade, "CASCADE");
		    fd.attr |= FieldDescriptor.Indexed|FieldDescriptor.OnDeleteCascade;
		    lex = scan();
		}
		break;
	      case tknUnique:
		fds = getConstraintField();
		if (fds.length == 1) { 
		    fd = fds[0];
		    fd.attr |= FieldDescriptor.Unique;
		    if (fd.type != FieldDescriptor.tpArray) { 
			fd.attr |= FieldDescriptor.Indexed;
		    }
		}
		lex = scan();
		break;
	      default:
		throw new ConsusCompileError("Column name or constraint expected", p);
	    }
	} while (lex == tknComma);
	
	if (lex != tknEof) { 
	    throw new ConsusCompileError("Constraint specification expected", p);
	}
    }

    final StmtNode createTable() throws SQLException { 
	expect(tknIdent, "table name");
	String tableName = ident;
	while ((lex = scan()) == tknDot) { 
	    expect(tknIdent, "table name");
	    tableName += "." + ident;
	}	
	table = new TableDescriptor(db, tableName);
	table.created = true;
	expected(tknLpar, "'('");
	do { 
	    int i;
	    int p = pos;
	    FieldDescriptor fd, fds[];
	    lex = scan();
	    if (lex == tknConstraint) { 
		expect(tknIdent, "Constraint name");
		p = pos;
		lex = scan();
		if (lex == tknIdent) { 
		    throw new ConsusCompileError("Constraint specification expected", p);
		}
	    }		    
	    switch (lex) { 
	      case tknIdent:
		fd = new FieldDescriptor(table, ident);
		setColumnType(fd);
		fd.calculateAttributes();
		break;
	      case tknPrimary:
		expect(tknKey, "KEY");
		fds = getConstraintField();
		for (i = 0; i < fds.length; i++) { 
		    fd = fds[i];
		    fd.attr |= FieldDescriptor.PrimaryKey;
		    if (fd.type != FieldDescriptor.tpArray) { 
			fd.attr |= FieldDescriptor.Indexed;
		    }
		}
		lex = scan();
		break;
	      case tknForeign:
		expect(tknKey, "KEY");
		fds = getConstraintField();
		fd = fds[0];
		expect(tknReferences, "REFERENCES");
		getReferencedPrimaryKey(fd);
		fd.attr |= FieldDescriptor.ForeignKey;
		if (lex == tknOn) { 
		    expect(tknDelete, "DELETE");
		    expect(tknCascade, "CASCADE");
		    fd.attr |= FieldDescriptor.Indexed|FieldDescriptor.OnDeleteCascade;
		    lex = scan();
		}
		break;
	      case tknUnique:
		fds = getConstraintField();
		if (fds.length == 1) { 
		    fd = fds[0];
		    fd.attr |= FieldDescriptor.Unique;
		    if (fd.type != FieldDescriptor.tpArray) { 
			fd.attr |= FieldDescriptor.Indexed;
		    }
		}
		lex = scan();
		break;
	      default:
		throw new ConsusCompileError("Column name or constraint expected", p);
	    }
	} while (lex == tknComma);

	expected (tknRpar, "')'");
	lex = scan();
	table.completeDeclaration();
	return new CreateTableNode(table);
    }

    final void matchTypes(ExprNode left, ExprNode right) throws SQLException { 
	if (left.tag == opParameter) { 
	    if (right.tag == opParameter) { 
		throw new ConsusCompileError("Failed to determine the type of parameter", 
					     pos-1);
	    }
	    left.type = right.type;
	} 
	if (right.tag == opParameter) { 
	    right.type = left.type;
	} 
    }

    final void matchType(ExprNode expr, int type) throws SQLException { 
	if (expr.tag == opParameter) { 
	    expr.type = type;
	} 
    }

    final ExprNode disjunction() throws SQLException {
	ExprNode left = conjunction();
	if (lex == tknOr) { 
	    int p = pos;
	    ExprNode right = disjunction();
	    matchTypes(left, right);
	    if (left.type == tpInt && right.type == tpInt) { 
		left = new BinOpNode(tpInt, opIntOr, left, right);
	    } else if (left.type == tpBool && right.type == tpBool) {
		left = new BinOpNode(tpBool, opBoolOr, left, right);
	    } else { 
		throw new ConsusCompileError("Bad operands for OR operator", p);
	    }
	}
	return left;
    }

    final ExprNode conjunction() throws SQLException {
	ExprNode left = comparison();
	if (lex == tknAnd) { 
	    int p = pos;
	    ExprNode right = conjunction();
	    matchTypes(left, right);
	    if (left.type == tpInt && right.type == tpInt) { 
		left = new BinOpNode(tpInt, opIntAnd, left, right);
	    } else if (left.type == tpBool && right.type == tpBool) {
		left = new BinOpNode(tpBool, opBoolAnd, left, right);
	    } else { 
		throw new ConsusCompileError("Bad operands for AND operator", p);
	    }
	}
	return left;
    }

    final static ExprNode int2real(ExprNode expr) {
	if (expr.tag == opIntConst) { 
	    return new RealLiteralNode((double)((IntLiteralNode)expr).value);
	} 
	return new UnaryOpNode(tpReal, opIntToReal, expr);
    }

    final static ExprNode int2str(ExprNode expr) {
	if (expr.tag == opIntConst) { 
	    return new StrLiteralNode(Long.toString(((IntLiteralNode)expr).value, 10));
	} 
	return new UnaryOpNode(tpStr, opIntToStr, expr);
    }

    final static ExprNode real2str(ExprNode expr) {
	if (expr.tag == opRealConst) { 
	    return new StrLiteralNode(Double.toString(((RealLiteralNode)expr).value));
	} 
	return new UnaryOpNode(tpStr, opRealToStr, expr);
    }

    final void compare(ExprNode expr, BinOpNode list) throws SQLException
    {
        ExprNode selector = expr;
        for (int n = 1; list != null; n++) { 
            ExprNode elem = list.left;
            BinOpNode tail = (BinOpNode)list.right;
            int cop = opNop;
            matchTypes(expr, elem);
            if (expr.type == tpInt) { 
                if (elem.type == tpReal) { 
                    expr = new UnaryOpNode(tpReal, opIntToReal, expr);
                    cop = opRealEq;
                } else if (elem.type == tpInt) { 
                    cop = opIntEq;
                } else if (elem.type == tpRef) { 
                    cop = opOidEq;
                } else if (elem.type == tpStr) { 
                    expr = new UnaryOpNode(tpStr, opIntToStr, expr);
                    cop = opStrEq;
                }
            } else if (expr.type == tpReal) {
                if (elem.type == tpReal) { 
                    cop = opRealEq;
                } else if (elem.type == tpInt) { 
                    cop = opRealEq;
                    elem = int2real(elem);
                } else if (elem.type == tpStr) { 
                    expr = new UnaryOpNode(tpStr, opRealToStr, expr);
                    cop = opStrEq;
                }
            } else if (expr.type == tpStr) { 
                if (elem.type == tpStr) { 
                    cop = opStrEq;
                } else if (elem.type == tpInt) { 
                    cop = opStrEq;
                    elem = int2str(elem);
                } else if (elem.type == tpReal) { 
                    cop = opStrEq;
                    elem = real2str(elem);
                }
            } else if (expr.type == tpRef && (elem.type == tpRef || elem.type == tpInt)) {
                cop = opOidEq;
            } else if (expr.type == tpBool && elem.type == tpBool) {
                cop = opBoolEq;
            }
            if (cop == opNop) { 
                throw new ConsusCompileError("Expression "+n+" in right part of IN "+
                                             "operator has incompatible type", pos);
            } 
            list.type = tpBool;
            if (tail != null) { 
                list.left = new BinOpNode(tpBool, cop, expr, elem);
                list.tag = opBoolOr;
                expr = selector;
            } else { 
                list.right = elem;
                list.left = expr;
                list.tag = cop;
            }
            list = tail;
	}
    }


    final ExprNode convertToString(ExprNode node, int p) throws SQLException { 
	switch (node.type) { 
	  case tpInt:
	    return int2str(node);
	  case tpReal:
	    return real2str(node);
	  case tpStr:
	    return node;
	  default:
	    throw new ConsusCompileError("Operand of numeric or string type expected", p);
	}
    }

    final ExprNode convertFromString(ExprNode strNode, ExprNode targetNode) { 
	FieldDescriptor fd = targetNode.getField();
	if (fd != null) { 
	    try { 
		String val = ((StrLiteralNode)strNode).value;
		switch (fd.type) { 
		  case FieldDescriptor.tpBool:
		    if (val.equalsIgnoreCase("true")) { 
			return new ConstantNode(tpBool, opTrue);
		    } else if (val.equalsIgnoreCase("false")) { 
			return new ConstantNode(tpBool, opFalse);
		    }
		    break;
		  case FieldDescriptor.tpInt1:
		  case FieldDescriptor.tpInt2:
		  case FieldDescriptor.tpInt4:
		  case FieldDescriptor.tpInt8:
		    return new IntLiteralNode(Long.parseLong(val));
		  case FieldDescriptor.tpReal4:
		  case FieldDescriptor.tpReal8:
		    return new RealLiteralNode(Double.parseDouble(val));
		  case FieldDescriptor.tpDate:
		    return new IntLiteralNode(Date.valueOf(val).getTime());
		  case FieldDescriptor.tpTime:
		    return new IntLiteralNode(Time.valueOf(val).getTime());
		  case FieldDescriptor.tpTimestamp:
		    return new IntLiteralNode(Timestamp.valueOf(val).getTime());
		}
	    } catch(Exception x) {}			
	}
	return strNode;
    }


    final static int findWildcard(String str, int pos) { 
	for (int i = pos, len = str.length(); i < len; i++) { 
            char ch = str.charAt(i);
	    if (ch == '%' || ch == '_' || ch == '\'') { 
		return i;
	    }
	}
	return -1;
    }

    final ExprNode comparison() throws SQLException {
	int  leftPos = pos;
	ExprNode left, right;
	left = addition();
	int cop = lex;
	if (cop == tknEq || cop == tknNe || cop == tknGt || cop == tknGe
	    || cop == tknLe || cop == tknLt || cop == tknBetween 
	    || cop == tknLike || cop == tknNot || cop == tknIs || cop == tknIn)
	{
	    int rightPos = pos;
	    boolean not = false;
	    if (cop == tknNot) { 
		not = true;
		cop = scan();
		if (cop != tknLike && cop != tknBetween && cop != tknIn) { 
		    throw new ConsusCompileError("LIKE, BETWEEN or IN expected", 
					   rightPos);
		} 
		rightPos = pos;
	    } else if (cop == tknIs) {
		if (left.type != tpRef) { 
		    throw new ConsusCompileError("IS [NOT] NULL predicate can be applied only to references", 
					   rightPos);
		} 
		rightPos = pos;
		if ((cop = scan()) == tknNull) { 
		    left = new UnaryOpNode(tpBool, opIsNull, left);
		} else if (cop == tknNot) { 
		    rightPos = pos;
		    expect(tknNull, "NULL");
		    left = new UnaryOpNode(tpBool, opBoolNot, 
					   new UnaryOpNode(tpBool, opIsNull, left));
		} else {
                    if (left.tag == opGetOid) { 
                        return new ThisInstanceOfNode(lookupTable());
                    } else { 
                        return new InstanceOfNode(left, lookupTable());
                    } 
		} 
		lex = scan();
		return left;
	    }	
	    right = addition();
	    if (cop == tknIn) { 
		int type;
		if (right.tag == opSelect) { 
		    matchType(left, right.type);
		    if (left.type != right.type) { 
			throw new ConsusCompileError("Incompatible types of IN operator operands", 
						     rightPos);
		    }
		    switch (right.type) {
		      case tpInt:
			left = new BinOpNode(tpBool, opScanSetInt, left, right);
			break;
		      case tpBool:
			left = new BinOpNode(tpBool, opScanSetBool, left, right);
			break;
		      case tpReal:
			left = new BinOpNode(tpBool, opScanSetReal, left, right);
			break;
		      case tpStr:
			left = new BinOpNode(tpBool, opScanSetStr, left, right);
			break;
		      case tpRef:
			left = new BinOpNode(tpBool, opScanSetRef, left, right);
			break;
		      default:
			Assert.failed("Invalid subquery result type");
		    }
		} else { 
		    switch (right.type) {
		      case tpArrayBool:
			matchType(left, tpBool);
			if (left.type != tpBool) { 
			    throw new ConsusCompileError("Incompatible types of IN operator operands", 
							 rightPos);
			}
			left = new BinOpNode(tpBool, opScanArrayBool, left, right);
			break;
		      case tpArrayInt1:
			matchType(left, tpInt);
			if (left.type != tpInt) { 
			    throw new ConsusCompileError("Incompatible types of IN operator operands", 
							 rightPos);
			}
			left = new BinOpNode(tpBool, opScanArrayInt1, left, right);
			break;
		      case tpArrayInt2:
			matchType(left, tpInt);
			if (left.type != tpInt) { 
			    throw new ConsusCompileError("Incompatible types of IN operator operands", 
							 rightPos);
			}
			left = new BinOpNode(tpBool, opScanArrayInt2, left, right);
			break;
		      case tpArrayInt4:
			matchType(left, tpInt);
			if (left.type != tpInt) { 
			    throw new ConsusCompileError("Incompatible types of IN operator operands", 
							 rightPos);
			}
			left = new BinOpNode(tpBool, opScanArrayInt4, left, right);
			break;
		      case tpArrayInt8:
			matchType(left, tpInt);
			if (left.type != tpInt) { 
			    throw new ConsusCompileError("Incompatible types of IN operator operands", 
							 rightPos);
			}
			left = new BinOpNode(tpBool, opScanArrayInt8, left, right);
			break;
		      case tpArrayReal4:
			matchType(left, tpReal);
			if (left.type == tpInt) {
			    left = int2real(left);
			} else if (left.type != tpReal) { 
			    throw new ConsusCompileError("Incompatible types of IN operator operands", 
							 rightPos);
			}
			left = new BinOpNode(tpBool, opScanArrayReal4, left, right);
			break;
		      case tpArrayReal8:
			if (left.type == tpInt) {
			    left = int2real(left);
			} else if (left.type != tpReal) { 
			    throw new ConsusCompileError("Incompatible types of IN operator operands", 
							 rightPos);
			}
			left = new BinOpNode(tpBool, opScanArrayReal8, left, right);
			break;
		      case tpArrayRef:
			matchType(left, tpRef);
			if (left.type != tpRef) { 
			    throw new ConsusCompileError("Incompatible types of IN operator operands", 
							 rightPos);
			}
			left = new BinOpNode(tpBool, opScanArrayRef, left, right);
			break;
		      case tpArrayStr:
			matchType(left, tpStr);
			if (left.type != tpStr) { 
			    throw new ConsusCompileError("Incompatible types of IN operator operands", 
							 rightPos);
			}
			left = new BinOpNode(tpBool, opScanArrayStr, left, right);
			break;
		      case tpStr:
			matchType(left, tpStr);
			if (left.type != tpStr) { 
			    throw new ConsusCompileError("Left operand of IN expression hasn't string type",
							 leftPos);
			}
			left = new BinOpNode(tpBool, opInString, left, right);
			break;
		      case tpList:
			compare(left, (BinOpNode)right);
			left = right;
			break;
		      default:
			throw new ConsusCompileError("List of expressions or array expected", 
						     rightPos);
		    }
		}
	    } else if (cop == tknBetween) { 
		int andPos = pos;
		expected(tknAnd, "AND");
		matchTypes(left, right);
		ExprNode right2 = addition();
		matchTypes(left, right2);
		if (left.type == tpStr || right.type == tpStr || right2.type == tpStr) {
		    left = new TripleOpNode(opStrBetween, 
					    convertToString(left, leftPos), 
					    convertToString(right, rightPos), 
					    convertToString(right2, andPos));
		} else if (left.type == tpReal || right.type == tpReal || right2.type == tpReal) {
		    if (left.type == tpInt) { 
			left = int2real(left);
		    } else if (left.type != tpReal) { 
			throw new ConsusCompileError("operand of BETWEEN operator should be of integer, real or string type", 
					       leftPos);
		    }
		    if (right.type == tpInt) {
			right = int2real(right);
		    } else if (right.type != tpReal) { 
			throw new ConsusCompileError("operand of BETWEEN operator should be of integer, real or string type", 
					       rightPos);
		    }
		    if (right2.type == tpInt) {
			right2 = int2real(right2);
		    } else if (right2.type != tpReal) { 
			throw new ConsusCompileError("operand of BETWEEN operator should be of integer, real or string type", 
					       andPos);
		    }
		    left = new TripleOpNode(opRealBetween, left, right, right2);
		} 
		else if (left.type == tpInt 
			 && right.type == tpInt 
			 && right2.type == tpInt)
		{		    
		    left = new TripleOpNode(opIntBetween, left, right, right2);
		}
	    } else if (cop == tknLike) {  
		matchType(left, tpStr);
		matchType(right, tpStr);
		if (left.type != tpStr || right.type != tpStr) { 
		    throw new ConsusCompileError("operands of LIKE operator should be of string type", 
					   rightPos);
		}
		if (lex == tknEscape) { 
		    rightPos = pos;
		    if (scan() != tknSconst) { 
			throw new ConsusCompileError("String literal espected after ESCAPE", rightPos);
		    }
		    left = new TripleOpNode(opStrLikeEsc, left, right,
					   new StrLiteralNode(svalue));
		    lex = scan();
		} else { 
                    if (right.tag == opStrConst) {
                        String str = ((StrLiteralNode)right).value;
                        int i = findWildcard(str, 0);
                        if (i < 0) { 
                            left = new BinOpNode(tpBool, opStrEq, left, right);
                        } else if (i == 0 && str.charAt(0) == '%' 
                                   && (i = findWildcard(str, 1)) == str.length()-1
                                   && str.charAt(i) == '%') 
                        {
                            ((StrLiteralNode)right).value = str.substring(1, i);
                            left = new BinOpNode(tpBool, opInString, right, left);
                        } else {
                            left = new TripleOpNode(opStrLike, left, right, null);
                        }
                    } else {
                        left = new TripleOpNode(opStrLike, left, right, null);
                    }
		}
	    } else { 
		if (left.type == tpStr) { 
		    if (left.tag == opStrConst && right.type != tpStr) { 
			left = convertFromString(left, right);
		    }
		} else if (right.type == tpStr) { 
		    if (right.tag == opStrConst && left.type != tpStr) { 
			right = convertFromString(right, left);
		    }
		}		    
		matchTypes(left, right);
		int tag = 0;
		if (left.type == tpStr || right.type == tpStr) { 
		    left = convertToString(left, leftPos);
		    right = convertToString(right, rightPos);
		    tag = opStrEq;
		} else if (left.type == tpReal || right.type == tpReal) { 
		    if (left.type == tpInt) { 
			left = int2real(left);
		    } else if (left.type != tpReal) { 
			throw new ConsusCompileError("operands of relation operator should be of intger, real or string type", 
					       leftPos);
		    }
		    if (right.type == tpInt) { 
			right = int2real(right);
		    } else if (right.type != tpReal) { 
			throw new ConsusCompileError("operands of relation operator should be of intger, real or string type", 
					       rightPos);
		    }
		    tag = opRealEq;
		} else if (left.type == tpInt && right.type == tpInt) 
		{ 
		    tag = opIntEq;
		} 
		else if (left.type == tpRef && (right.type == tpRef || right.type == tpInt)
			 || (left.type == tpInt && right.type == tpRef)) 
		{ 
		    if (cop != tknEq && cop != tknNe) { 
			throw new ConsusCompileError("References can be checked only for equality", 
					       rightPos);
		    }
		    tag = opOidEq;
		} else if (left.type == tpBool && right.type == tpBool) { 
		    if (cop != tknEq && cop != tknNe) { 
			throw new ConsusCompileError("Boolean variables can be checked only for equality",
					       rightPos);
		    }
		    tag = opBoolEq;
		} else { 
		    throw new ConsusCompileError("operands of relation operator should be of integer, real or string type", 
					   rightPos);
		}
		if ((left.tag < opLoadBool || left.tag > opLoadStr) &&
		    (right.tag >= opLoadBool && right.tag <= opLoadStr))
		{
		    // swap operands
		    ExprNode tmp = left;
		    left = right;
		    right = tmp;
		    if (cop == tknGt || cop == tknGe) { 
			cop += 2; // replace Gt with Lt or Ge with Le
		    } else if (cop == tknLt || cop == tknLe) { 
			cop -= 2; // replace Lt with Gt or Le with Ge
		    }
		}
		left = new BinOpNode(tpBool, tag+cop-tknEq, left, right);
		if (right instanceof SubqueryElemNode) { 
		    QuantorNode quantor = ((SubqueryElemNode)right).quantor;
		    quantor.condition = left;
		    left = quantor;
		}
	    }
	    if (not) { 
		left = new UnaryOpNode(tpBool, opBoolNot, left);
	    }
	}
	return left;
    }


    final ExprNode addition() throws SQLException { 
	int leftPos = pos;
	ExprNode left = multiplication();
	while (lex == tknAdd || lex == tknSub) { 
	    int cop = lex;
	    int rightPos = pos;
	    ExprNode right = multiplication();
	    matchTypes(left, right);
	    if (left.type == tpReal || right.type == tpReal) { 
		if (left.type == tpInt) { 
		    left = int2real(left);
		} else if (left.type != tpReal) { 
		    throw new ConsusCompileError("operands of arithmetic operators should be of integer or real type", 
					   leftPos);
		}
		if (right.type == tpInt) { 
		    right = int2real(right);
		} else if (right.type != tpReal) { 
		    throw new ConsusCompileError("operands of arithmetic operator should be of integer or real type", 
					   rightPos);
		}
		left = new BinOpNode(tpReal, cop == tknAdd 
				     ? opRealAdd : opRealSub,
				     left, right);
	    } 
	    else if (left.type == tpInt && right.type == tpInt) 
	    { 
		left = new BinOpNode(tpInt, cop == tknAdd 
				     ? opIntAdd : opIntSub,
				     left, right);
	    } else if (left.type == tpStr && right.type == tpStr) { 
		if (cop == tknAdd) { 
		    left = new BinOpNode(tpStr, opStrConcat, 
					 left, right);
		} else { 
		    throw new ConsusCompileError("Operation - is not defined for strings", 
					   rightPos);
		}
	    } else { 
		throw new ConsusCompileError("operands of arithmentic operator should be of integer or real type", 
				       rightPos);
	    }
	    leftPos = rightPos;
	}
	return left;
    }


    final ExprNode multiplication() throws SQLException { 
	int leftPos = pos;
	ExprNode left = power();
	while (lex == tknMul || lex == tknDiv) { 
	    int cop = lex;
	    int rightPos = pos;
	    ExprNode right = power();
	    matchTypes(left, right);
	    if (left.type == tpReal || right.type == tpReal) { 
		if (left.type == tpInt) { 
		    left = int2real(left);
		} else if (left.type != tpReal) { 
		    throw new ConsusCompileError("operands of arithmetic operators should be of integer or real type", 
					   leftPos);
		}
		if (right.type == tpInt) { 
		    right = int2real(right);
		} else if (right.type != tpReal) { 
		    throw new ConsusCompileError("operands of arithmetic operator should be of integer or real type", 
					   rightPos);
		}
		left = new BinOpNode(tpReal, cop == tknMul 
				     ? opRealMul : opRealDiv,
				     left, right);
	    } else if (left.type == tpInt && right.type == tpInt) { 
		left = new BinOpNode(tpInt, cop == tknMul 
				     ? opIntMul : opIntDiv,
				     left, right);
	    } else { 
		throw new ConsusCompileError("operands of arithmentic operator should be of integer or real type", 
				       rightPos);
	    }
	    leftPos = rightPos;
	}
	return left;
    }


    final ExprNode power() throws SQLException { 
	int leftPos = pos;
	ExprNode left = term();
	if (lex == tknPower) { 
	    int rightPos = pos;
	    ExprNode right = power();
	    matchTypes(left, right);
	    if (left.type == tpReal || right.type == tpReal) { 
		if (left.type == tpInt) { 
		    left = int2real(left);
		} else if (left.type != tpReal) { 
		    throw new ConsusCompileError("operands of arithmetic operators should be of integer or real type", 
					   leftPos);
		}
		if (right.type == tpInt) { 
		    right = int2real(right);
		} else if (right.type != tpReal) { 
		    throw new ConsusCompileError("operands of arithmetic operator should be of integer or real type", 
					   rightPos);
		}
		left = new BinOpNode(tpReal, opRealPow, left, right);
	    } else if (left.type == tpInt && right.type == tpInt) { 
		left = new BinOpNode(tpInt, opIntPow, left, right);
	    } else { 
		throw new ConsusCompileError("operands of arithmentic operator should be of integer or real type", 
				       rightPos);
	    }
	}
	return left;
    }


    final ExprNode arrayElement(ExprNode expr) throws SQLException {
	while (true) { 
	    int p = pos;
	    int type;
	    int tag;
	    FieldDescriptor field;

	    switch (lex) {
	      case tknDot:	
		expect(tknIdent, "field name");
		expr = objectField(expr);
		break;
	      case tknLbr:
		if ((field = expr.getField()) != null) { 
		    if (field.type != FieldDescriptor.tpArray 
			&& field.type != FieldDescriptor.tpString) 
		    { 
			throw new ConsusCompileError("Index can be applied only to arrays or strings", 
						     p);
		    }
		    if (field.type == FieldDescriptor.tpArray) { 
			field = field.components[0];
		    }
		}
		switch (expr.type) { 
		  case tpArrayBool:
		    tag = opGetAtInt1;
		    type = tpBool;
		    break;
		  case tpStr:
		    tag = opStrGetAt;
		    type = tpInt;
		    break;
		  case tpArrayInt1:
		    tag = opGetAtInt1;
		    type = tpInt;
		    break;
		  case tpArrayInt2:
		    tag = opGetAtInt2;
		    type = tpInt;
		    break;
		  case tpArrayInt4:
		    tag = opGetAtInt4;
		    type = tpInt;
		    break;
		  case tpArrayInt8:
		    tag = opGetAtInt8;
		    type = tpInt;
		    break;
		  case tpArrayReal4:
		    tag = opGetAtReal4;
		    type = tpReal;
		    break;
		  case tpArrayReal8:
		    tag = opGetAtReal8;
		    type = tpReal;
		    break;
		  case tpArrayStr:
		    tag = opGetAtStr;
		    type = tpStr;
		    break;
		  case tpArrayRef:
		    tag = opGetAtInt4;
		    type = tpRef;
		    break;
		  case tpArray:
		    tag = opGetAtObj;
		    type = field.type == FieldDescriptor.tpArray 
			? tpArray : tpStructure;
		    break;
		  default: 
		    throw new ConsusCompileError("Index can be applied only to arrays or strings", 
						 p);
		}
		p = pos;
		ExprNode index = disjunction();
		expected(tknRbr, "']'");
		matchType(index, tpInt);
		if (index.type != tpInt && index.type != tpFreeVar) {
		    throw new ConsusCompileError("Index should have integer type",p);
		}
		if (tag == opStrGetAt) { 
		    expr = new StrGetAtNode(expr, index);
		} else { 
		    expr = new GetAtNode(type, tag, expr, index, field);
		}
		lex = scan();
		break;
	      default:
		return expr;
	    }	
	}
    }


    final ExprNode objectField(ExprNode base) throws SQLException
    {
	FieldDescriptor fd = null;
	int p = pos - 1;
	String column = ident;
	TableDescriptor td = table;

	if (base == null) { 
	    if (!selectTuplesMode) { 
		fd = table.getColumn(column);
	    } else { 
		while (true) { 
		    SelectNode curr, inner = null;
		    for (curr = select; curr != null; inner = curr, curr = curr.outer) { 
			int i;
			if (curr.columns != null) { 
			    for (i = curr.columns.length; --i >= 0;) { 
				if (column.equals(curr.columns[i].name)) { 
				    lex = scan();
				    if (inner != null && inner.outDep < curr.columns[i].tableDep) { 
					inner.outDep = curr.columns[i].tableDep;
				    }
				    return curr.columns[i];
				}
			    }
			}
			for (i = curr.tables.length; --i >= 0;) { 
			    TableNode tnode = curr.tables[i]; 
			    if (column.equals(tnode.name)) { 
				expect(tknDot, "."); 
				expect(tknIdent, "field name");
				while ((fd = tnode.table.getColumn(ident)) == null
				       && (tnode = tnode.joinTable) != null);
				if (fd == null) { 
				    throw new ConsusCompileError("No such column '" + ident + 
								 "' in table " + column, p);
				} else if (curr.columns != null) { 
				    for (int j = curr.columns.length; --j >= 0;) { 
					if (curr.columns[j].expr instanceof LoadNode 
                                            && ((LoadNode)curr.columns[j].expr).base != null
                                            && ((LoadNode)curr.columns[j].expr).base.tag == opTable
                                            && ((TableNode)((LoadNode)curr.columns[j].expr).base).name.equals(column)
                                            && curr.columns[j].getField() == fd)
					{
					    lex = scan();
					    if (inner != null && inner.outDep <= i) {
						inner.outDep = i+1;
					    }
					    return curr.columns[j];
					}
				    }
				}
			    } else { 
				fd = tnode.table.getColumn(column);
			    }
			    if (fd != null) { 
				if (inner != null && inner.outDep <= i) {
				    inner.outDep = i+1;
				}
				ExprNode expr = LoadNode.create(fd, tnode);
				if (context != CTX_COLUMNS && context != CTX_WHERE 
				    && curr == select) 
				{ 
				    int n = select.columns.length;
				    ColumnNode columns[] = new ColumnNode[n+1];
				    System.arraycopy(select.columns, 0, columns, 0, n);
				    expr = columns[n] = new ColumnNode(n+1, expr, null, null);
				    select.columns = columns;
				} 
				lex = scan();
				return expr;
			    } 
			}
		    }
		    if (scan() == tknDot) { 
			expect(tknIdent, "class name");
			column += "." + ident;
		    } else { 
			break;
		    }
		}
		if (fd == null) { 
		    throw new ConsusCompileError("Failed to locate column '" + column + "'", p);
		}
	    }
	} else { 
	    fd = base.getField();
	    if (base.type == tpRef) { 
		if (fd == null) { 
		    throw new ConsusCompileError("Type of referenced table is unknown", p);
		}
		if (fd.className == null) { 
		    throw new ConsusCompileError("Abstract reference field " + fd.name
						 + " can not be dereferenced", p);
		}
		base = new DerefNode(base, fd);
		td = fd.refTable;
		if (td == null) { 
		    td = db.getTableByName(fd.className);
		    if (td == null) { 
			throw new ConsusCompileError("Reference to the unknown table " + fd.className, 
						     p);
		    }
		    fd.refTable = td;
		}
		fd = td.getColumn(column);
	    } else if (base.type == tpStructure) { 
		Assert.that(fd != null && fd.type == FieldDescriptor.tpStructure);
		fd = fd.getComponent(column);
	    } else { 
		throw new ConsusCompileError("Left operand of '.' should be structure or reference", 
					     p);
	    }
	}
        while (fd != null && fd.type == FieldDescriptor.tpStructure) { 
	    expect(tknDot, ".");
	    expect(tknIdent, "field name");
	    p = pos-1;
	    column = ident;
	    fd = fd.getComponent(column);
	}	    
	if (fd == null) { 
	    throw new ConsusCompileError("No such column '" + column + "' in " + td.name, p);
	}
	lex = scan();
	return LoadNode.create(fd, base);
    }

    final ExprNode buildList() throws SQLException { 
        ExprNode expr = disjunction(); 
        return new BinOpNode(tpList, opNop, expr, lex == tknComma ? buildList() : null);
    }

    final ExprNode groupFunction(int tag) throws SQLException { 
	ColumnNode[] saveColumns = select.columns;
	int p = pos;
	int type = tpInt;
	ExprNode expr = term();
	if (tag != opCountAll && tag != opCountDis) { 
	    if (expr.type == tpReal) { 
		tag +=  opRealMinAll - opIntMinAll;
		type = tpReal;
	    } else if (expr.type != tpInt) { 
		throw new ConsusCompileError("MIN,MAX and AVG functions can be applied only to integer or real expressions", 
					     p);
	    }		
	}
	GroupNode group = new GroupNode(type, tag, expr);
	ColumnNode groups = select.groups;
	if (context == CTX_HAVING) { 
	    while (groups != null) { 
		if (groups.expr.equalsNode(group)) { 
		    select.columns = saveColumns;
		    return groups;
		}
		groups = groups.next;
	    }
	}
	return group;
    }

    final ExprNode term() throws SQLException {
        int prevLex = lex;
	int cop = scan();
	int p = pos;
	ExprNode expr;
	Binding bp;
        lex = cop;

	switch (cop) { 
	  case tknEof:
	  case tknOrder:
	    return null;
	  case tknParam:
	    lex = scan();
	    return params[nParams++] = new ParameterNode(tpVoid);
	  case tknIdent:
	    for (bp = bindings; bp != null; bp = bp.next) { 
		if (bp.name.equals(ident)) { 
		    lex = scan();
		    bp.used = true;
		    return new IndexNode(bp.loopId);
		}
	    }
	    return arrayElement(objectField(null));

	  case tknCast:
	    expect(tknLpar, "'('");
	    expr = disjunction();
	    if (lex == tknAs) { 
		lex = scan();
	    }
	    switch (lex) { 
	      case tknChar:
	      case tknVarchar:
	      case tknLongvarchar:
	      case tknString:
		switch (expr.type) { 
		  case tpInt:
		    expr = new UnaryOpNode(tpStr, opIntToStr, expr);
		    break;
		  case tpReal:
		    expr = new UnaryOpNode(tpStr, opRealToStr, expr);
		    break;
		  case tpStr:
		    break;
		  case tpBool:
		    expr = new UnaryOpNode(tpStr, opBoolToStr, expr);
		    break;
		  default:
		    throw new ConsusCompileError("Conversion can be performed only between string,numeric and boolean types", p);
		}
		break;

	      case tknInt:
	      case tknInteger:
	      case tknBigint:
	      case tknSmallint:
	      case tknTinyint:
	      case tknDate:
	      case tknTime:
	      case tknTimestamp:
		switch (expr.type) { 
		  case tpInt:
		    break;
		  case tpReal:
		    expr = new UnaryOpNode(tpInt, opRealToInt, expr);
		    break;
		  case tpStr:
		    expr = new UnaryOpNode(tpInt, opStrToInt, expr);
		    break;
		  case tpBool:
		    expr = new UnaryOpNode(tpInt, opBoolToInt, expr);
		    break;
		  default:
		    throw new ConsusCompileError("Conversion can be performed only between string,numeric and boolean types", p);
		}
		break;

	      case tknReal:
	      case tknFloat:
	      case tknDouble:
	      case tknNumeric:
	      case tknDecimal:
		switch (expr.type) { 
		  case tpInt:
		    expr = new UnaryOpNode(tpReal, opIntToReal, expr);
		    break;
		  case tpReal:
		    break;
		  case tpStr:
		    expr = new UnaryOpNode(tpReal, opStrToReal, expr);
		    break;
		  case tpBool:
		    expr = new UnaryOpNode(tpReal, opBoolToReal, expr);
		    break;
		  default:
		    throw new ConsusCompileError("Conversion can be performed only between string,numeric and boolean types", p);
		}
		break;

	      case tknBoolean:
	      case tknBit:
		switch (expr.type) { 
		  case tpInt:
		    expr = new UnaryOpNode(tpBool, opIntToBool, expr);
		    break;
		  case tpReal:
		    expr = new UnaryOpNode(tpBool, opRealToBool, expr);
		    break;
		  case tpStr:
		    expr = new UnaryOpNode(tpBool, opStrToBool, expr);
		    break;
		  case tpBool:
		    break;
		  default:
		    throw new ConsusCompileError("Conversion can be performed only between string,numeric and boolean types", p);
		}
		break;

	      default:
		throw new ConsusCompileError("Conversion can be performed only between string,numeric and boolean types", p);
	    }
	    expect(tknRpar, "')'");	    
	    lex = scan();
	    return arrayElement(expr); 

	  case tknSubstr:
	    expect(tknLpar, "'('");
	    expr = disjunction();
	    matchType(expr, tpStr);
	    if (expr.type != tpStr) {
		throw new ConsusCompileError("First operand of SUBSTR function should have STRING type", p);
	    }  
	    expected(tknComma, "','");
	    p = pos;
	    ExprNode from = disjunction();
	    matchType(from, tpInt);
	    if (from.type != tpInt) { 
		throw new ConsusCompileError("Second operand of SUBSTR function should have INTEGER type", p);
	    }
	    ExprNode length = null;
	    if (lex == tknComma) { 
		p = pos;
		length = disjunction();
		matchType(length, tpInt);
		if (length.type != tpInt) { 
		    throw new ConsusCompileError("Third operand of SUBSTR function should have INTEGER type", p);
		}
	    }
	    expr = new TripleOpNode(tpStr, opSubStr, expr, from, length);
	    expected(tknRpar, "')'");
	    lex = scan();
	    return arrayElement(expr); 

	  case tknAny:
	  case tknSome:
	  case tknAll:
	    expr = term();
	    if (!(expr instanceof SelectNode)) { 
		throw new ConsusCompileError("Subquery expected", p);
	    }
	    return new SubqueryElemNode(cop == tknAll ? opQuantorAll : opQuantorAny, expr);

	  case tknExists:
	    lex = scan();
	    if (lex == tknIdent) { 
		bindings = bp = new Binding(ident, vars++, bindings);
		if (vars >= ConsusStatement.maxIndexVars) { 
		    throw new ConsusCompileError("Too many nested EXISTS clauses", p);
		}
		p = pos;
		expect(tknCol, "':'");
		p = pos;
		expr = term();
		if (expr.type != tpBool) { 
		    throw new ConsusCompileError("Expresion in EXISTS clause should be of boolean type", 
						 p);
		}
		if (bp.used) { 
		    expr = new ExistsNode(expr, vars-1);
		}
		vars -= 1;	    
		bindings = bp.next;
	    } else if (lex == tknLpar) { 
		expr = new BinOpNode(tpBool, opIntNe, selectExpr(SA_SUBQUERY|SA_EXISTS), 
				     new IntLiteralNode(0));
	    } else if (lex == tknSelect) { 
		expr = new BinOpNode(tpBool, opIntNe, selectStatement(SA_SUBQUERY|SA_EXISTS), 
				     new IntLiteralNode(0));
	    } else { 
		throw new ConsusCompileError("Subquery or free variable name expected", p);
	    }	    
	    return expr;
	  case tknSysdate:
	    lex = scan();
	    return new ConstantNode(tpInt, opSysdate);	    
	  case tknOid:
	  case tknThis:
	    lex = scan();
	    return new ConstantNode(tpRef, opGetOid);
	  case tknFalse:
	    expr = new ConstantNode(tpBool, opFalse);
	    break;
	  case tknTrue:
	    expr = new ConstantNode(tpBool, opTrue);
	    break;
	  case tknNull:
	    expr = new ConstantNode(tpRef, opNull);
	    break;
	  case tknFirst:
	    expr = new ConstantNode(tpRef, opFirst);
	    break;
	  case tknLast:
	    expr = new ConstantNode(tpRef, opLast);
	    break;
	  case tknIconst:
	    expr = new IntLiteralNode(ivalue);
	    break;
	  case tknFconst:
	    expr = new RealLiteralNode(fvalue);
	    break;
	  case tknSconst:
	    expr = new StrLiteralNode(svalue);
	    lex = scan();
	    return arrayElement(expr); 
	  case tknSin:
	  case tknCos:
	  case tknTan:
	  case tknAsin:
	  case tknAcos:
	  case tknAtan:
	  case tknExp:
	  case tknLog:
	  case tknSqrt:
	  case tknCeil:
	  case tknFloor:
	    expr = term();
	    matchType(expr, tpReal);
	    if (expr.type == tpInt) { 
		expr = int2real(expr);
	    } else if (expr.type != tpReal) { 
		throw new ConsusCompileError("Numeric argument expected", p);
	    }
	    return new UnaryOpNode(tpReal, cop+opRealSin-tknSin, expr);
	  case tknCount:
	    if (!selectTuplesMode) { 
		throw new IncorrectModeException(pos);
	    }
	    expect(tknLpar, "'('");
	    p = pos;
	    lex = scan();
	    if (lex == tknMul) { 
		expr = new GroupNode(tpInt, opCountAll, new IntLiteralNode(1));
		p = pos;
		lex = scan();
	    } else { 
		int op = opCountDis;
		if (lex != tknDistinct) { 
		    pos = p; // restore position of pointer
		    op = opCountAll;
		}
		expr = groupFunction(op);
	    }
	    expected(tknRpar, "')'");
	    lex = scan();
	    return expr;
	  case tknMin:
	  case tknMax:
	  case tknAvg:
	  case tknSum:
	    if (!selectTuplesMode) { 
		throw new IncorrectModeException(pos);
	    } else { 
		boolean distinct = false;
		expect(tknLpar, "'('");
		p = pos;
		lex = scan();
		if (lex == tknDistinct) { 
		    distinct = true;
		} else if (lex != tknAll) { 
		    pos = p; // restore position of pointer
		}
		expr = groupFunction(cop-tknMin+(distinct ? opIntMinDis : opIntMinAll));
		expected(tknRpar, "')'");
		lex = scan();
	    }
	    return expr;

	  case tknAbs:
	    expr = term();
	    if (expr.type == tpInt) { 
		return new UnaryOpNode(tpInt, opIntAbs, expr);
	    } else if (expr.type == tpReal) { 
		return new UnaryOpNode(tpReal, opRealAbs, expr);
	    } else { 
		throw new ConsusCompileError("ABS function can be applied only to integer or real expression", 
				       p);
	    }
	  case tknLength:
	    expr = term();
	    matchType(expr, tpStr);
	    if (expr.type == tpStr) { 
		return new UnaryOpNode(tpInt, opStrLength, expr);
	    } else if (expr.type >= tpArrayBool) { 
		return new UnaryOpNode(tpInt, opLength, expr);
	    } else { 
		throw new ConsusCompileError("LENGTH function is defined only for arrays and strings",
				       p);
	    } 
	  case tknLower:
	    expr = term();
	    matchType(expr, tpStr);
	    if (expr.type == tpStr) { 
		return arrayElement(new UnaryOpNode(tpStr,opStrLower,expr));
	    }
	    throw new ConsusCompileError("LOWER function can be applied only to string argument",
				       p);
	  case tknUpper:
	    expr = term();
	    matchType(expr, tpStr);
	    if (expr.type == tpStr) { 
		return arrayElement(new UnaryOpNode(tpStr,opStrUpper,expr));
	    }
	    throw new ConsusCompileError("UPPER function can be applied only to string argument",
				       p);
	  case tknInt:
	  case tknInteger:
	    expr = term();
	    matchType(expr, tpReal);
	    switch (expr.type) { 
	      case tpReal:
		return new UnaryOpNode(tpInt, opRealToInt, expr);
	      case tpStr:
		return new UnaryOpNode(tpInt, opStrToInt, expr);
	      case tpBool:
		return new UnaryOpNode(tpInt, opBoolToInt, expr);
	      case tpInt:
		return expr;
	      default:
		throw new ConsusCompileError("INTEGER function can be applied only to expression of real,string or boolean type", 
					     p);
	    }
	  case tknReal:
	    expr = term();
	    matchType(expr, tpInt);
	    switch (expr.type) {
	      case tpInt:
		return new UnaryOpNode(tpReal, opIntToReal, expr);
	      case tpStr:
		return new UnaryOpNode(tpReal, opStrToReal, expr);
	      case tpBool:
		return new UnaryOpNode(tpReal, opBoolToReal, expr);
	      case tpReal:
		return expr;
	      default:
		throw new ConsusCompileError("REAL function can be applied only to expression of integer, srting or boolean type", 
					     p);
	    }
	  case tknString:
	    expr = term();	
	    switch (expr.type) {
	      case tpInt:
		return arrayElement(new UnaryOpNode(tpStr, opIntToStr, 
						    expr));
	      case tpReal:
		return arrayElement(new UnaryOpNode(tpStr, opRealToStr, 
						    expr));
	      case tpBool:
		return arrayElement(new UnaryOpNode(tpStr, opBoolToStr, 
						    expr));
	      case tpStr:
		return arrayElement(expr);
	      default:
		throw new ConsusCompileError("STRING function can be applied only to numeric or boolean expression", 
					     p);
	    }
	  case tknLpar:
	  {
	    expr = disjunction();
            if (lex == tknComma) { 
                expr = new BinOpNode(tpList, opNop, expr, buildList());
            } else if (prevLex == tknIn && expr.tag != opSelect) { 
                expr = new BinOpNode(tpList, opNop, expr, null);
            }
	    expected(tknRpar, "')'");
	    break;
	  }
	  case tknNot:
	    expr = comparison();
	    if (expr.type == tpInt) { 
		if (expr.tag == opIntConst) { 
		    IntLiteralNode ic = (IntLiteralNode)expr;
		    ic.value = ~ic.value;
		} else {
		    expr = new UnaryOpNode(tpInt, opIntNot, expr);
		} 
		return expr;
	    } else if (expr.type == tpBool) { 
		return new UnaryOpNode(tpBool, opBoolNot, expr);
	    } else { 
		throw new ConsusCompileError("NOT operator can be applied only to integer or boolean expressions", 
				       p);
	    }
	  case tknAdd:
	    throw new ConsusCompileError("Using of unary plus operator has no sense",
				   p);
	  case tknSub:
	    expr = term();
	    if (expr.type == tpInt) { 
		if (expr.tag == opIntConst) { 
		    IntLiteralNode ic = (IntLiteralNode)expr;
		    ic.value = -ic.value;
		} else {
		    expr = new UnaryOpNode(tpInt, opIntNeg, expr);
		} 
	    } else if (expr.type == tpReal) { 
		if (expr.tag == opRealConst) { 
		    RealLiteralNode fc = (RealLiteralNode)expr;
		    fc.value = -fc.value;
		} else {
		    expr = new UnaryOpNode(tpReal, opRealNeg, expr);
		} 
	    } else { 
		throw new ConsusCompileError("Unary minus can be applied only to numeric expressions", 
				       p);
	    }
	    return expr;
	  case tknSelect: // subquery
	    SelectNode subselect = selectStatement(SA_SUBQUERY);
	    expr = subselect;
	    if (subselect.resultColumns.length != 1) { 
		throw new ConsusCompileError("Subquery should return only one column", p);
	    }
	    return expr;
	  default:
	    throw new ConsusCompileError("operand expected", p);
	}
	lex = scan();
	return expr;
    }
}




