/* Consus - Pure Java Object Relational Database Management System
 * Abstraction of fiel uded by Consus. Encapsulate RandomAccessFile, mutifile and raid file.
 * Created by Konstantin Knizhnik, Sep 2001.
 */
package org.garret.internal.consus;
import java.io.*;
import java.lang.reflect.Method;
import java.sql.SQLException;
import org.garret.consus.ConsusIOError;

public class ConsusFile { 
    RandomAccessFile file;
    String           name;
    ConsusFile       next;

    ConsusFile(String name) throws SQLException { 
	try { 
	    this.name = name;
	    if (name.endsWith(".raid0")) { 
		file = new Raid0File(name, "rw");
	    } else if (name.endsWith(".raid1")) { 
		file = new Raid1File(name, "rw");
	    } else if (name.endsWith(".mfile")) { 
		file = new RandomAccessMultiFile(name, "rw");
	    } else { 
		file = new RandomAccessFile(name, "rw");
	    }
	} catch(IOException x) { 
	    throw new ConsusIOError("Failed to create file '" + name + "':" + x);
	}
    }

    static final long MAX_FILE_SIZE = Long.MAX_VALUE-2;

    public boolean lock() { 
	try { 
	    Class cls = file.getClass();
	    Method getChannel = cls.getMethod("getChannel", new Class[0]);
	    if (getChannel != null) { 
		Object channel = getChannel.invoke(file, new Object[0]);
		if (channel != null) { 
		    cls = channel.getClass();
		    Class[] paramType = new Class[3];
		    paramType[0] = Long.TYPE;
		    paramType[1] = Long.TYPE;
		    paramType[2] = Boolean.TYPE;
		    Method lock = cls.getMethod("tryLock", paramType);
		    if (lock != null) { 
			Object[] param = new Object[3];
			param[0] = new Long(MAX_FILE_SIZE);
			param[1] = new Long(1);
			param[2] = new Boolean(false);
			return lock.invoke(channel, param) != null;
		    }
		}
	    }
	} catch (Exception x) {}

	return true;
    }

    public void write(long pos, byte[] buf) throws SQLException 
    {
	try { 
	    file.seek(pos);
	    file.write(buf, 0, buf.length);
	} catch(IOException x) { 
	    throw new ConsusIOError("Write to the file failed: " + x);
	}
    }

    public int read(long pos, byte[] buf) throws SQLException 
    { 
	try { 
	    file.seek(pos);
	    return file.read(buf, 0, buf.length);
	} catch(IOException x) { 
	    throw new ConsusIOError("Read from the file failed: " + x);
	}
    }
	
    public void sync() throws SQLException 
    { 
	try { 	
	    if (file instanceof RandomAccessMultiFile) { 
		((RandomAccessMultiFile)file).sync();
	    } else {
		file.getFD().sync();
	    }
	} catch(IOException x) { 
	    throw new ConsusIOError("Sync failed: " + x);
	}
    }
    
    public void close() throws SQLException { 
	try { 
	    file.close();
	} catch(IOException x) {             
	    throw new ConsusIOError("Close failed: " + x);
	}
    }
}


