/* Consus - Pure Java Object Relational Database Management System
 * Interface for advanced transaction support.
 * Created by Konstantin Knizhnik, Apr 2001.
 */
package org.garret.internal.consus;

import java.sql.*;

/**
 * Interface for advanced transaction management.
 * This interface provides methods for nested transactions and two phase commit protocol
 * for distributed transactions
 */
public interface ConsusTransaction extends Connection {
    /**
     * Commit nested subtransaction. Exterior transaction will be committed when all nested
     * subtransaction were successfully committed.
     */ 
    void commitSubtransaction() throws SQLException; 

    /**
     * Rollback local transaction
     */
    void rollbackTransaction() throws SQLException;

    /**
     * Method invoked by distributed transaction meanager at the second pahse of two
     * phase commit protocol to complete local part of transaction
     */
    void completeTransaction() throws SQLException;

    /**
     * Get transaction identifier
     */
    ConsusTransId getTransId() throws SQLException;

    /**
     * Commit local transaction
     */
    void commitTransaction() throws SQLException;

    /**
     * Prepare part of distributed transaction. This method is invoked by transaction
     * manager at first phase of two phase distributed transaction commit protocol
     */
    void prepareTransaction(int fmtid, byte[] gtrid, byte[] bqual) throws SQLException;
};
