/* Consus - Pure Java Object Relational Database Management System
 * Abstract node class - common base for all nodes of internal representation of JSQL statement
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import java.sql.*;

class ExprNode implements Comparable { 
    int type;
    int tag;

    final static int tpBool          = 0;
    final static int tpInt           = 1;
    final static int tpReal          = 2;
    final static int tpFreeVar       = 3;
    final static int tpList          = 4;
    final static int tpRef           = 5;
    final static int tpStr           = 6;
    final static int tpArray         = 7;
    final static int tpArrayBool     = 8;
    final static int tpArrayInt1     = 9;
    final static int tpArrayInt2     = 10;
    final static int tpArrayInt4     = 11;
    final static int tpArrayInt8     = 12;
    final static int tpArrayReal4    = 13;
    final static int tpArrayReal8    = 14;
    final static int tpArrayStr      = 15;
    final static int tpArrayRef      = 16;
    final static int tpStructure     = 18;
    final static int tpBlob          = 19;
    final static int tpClob          = 20;
    final static int tpCursor        = 21;
    final static int tpTable         = 22;
    final static int tpVoid          = 23;

    final static int opNop     = 0;
    final static int opIntAdd  = 1;
    final static int opIntSub  = 2;
    final static int opIntMul  = 3;
    final static int opIntDiv  = 4;
    final static int opIntAnd  = 5;
    final static int opIntOr   = 6;
    final static int opIntNeg  = 7;
    final static int opIntNot  = 8;
    final static int opIntAbs  = 9;
    final static int opIntPow  = 10;
    final static int opIntEq   = 11;
    final static int opIntNe   = 12;
    final static int opIntGt   = 13;
    final static int opIntGe   = 14;
    final static int opIntLt   = 15;
    final static int opIntLe   = 16;
    final static int opIntBetween = 17;

    final static int opRealEq   = 18;
    final static int opRealNe   = 19;
    final static int opRealGt   = 20;
    final static int opRealGe   = 21;
    final static int opRealLt   = 22;
    final static int opRealLe   = 23;
    final static int opRealBetween = 24;

    final static int opStrEq   = 25;
    final static int opStrNe   = 26;
    final static int opStrGt   = 27;
    final static int opStrGe   = 28;
    final static int opStrLt   = 29;
    final static int opStrLe   = 30;
    final static int opStrBetween = 31;
    final static int opStrLike    = 32;
    final static int opStrLikeEsc = 33;

    final static int opBoolEq   = 34;
    final static int opBoolNe   = 35;
    
    final static int opOidEq   = 36;
    final static int opOidNe   = 37;
    
    final static int opRealAdd = 38;
    final static int opRealSub = 39;
    final static int opRealMul = 40;
    final static int opRealDiv = 41;
    final static int opRealNeg = 42;
    final static int opRealAbs  = 43;
    final static int opRealPow  = 44;

    final static int opIntToReal = 45;
    final static int opRealToInt = 46;
    final static int opIntToStr = 47;
    final static int opRealToStr = 48;
    final static int opStrToInt = 49;
    final static int opStrToReal = 50;

    final static int opIsNull = 51;

    final static int opStrGetAt  = 52;
    final static int opGetAtInt1 = 53;
    final static int opGetAtInt2 = 54;
    final static int opGetAtInt4 = 55;
    final static int opGetAtInt8 = 56;
    final static int opGetAtReal4 = 57;
    final static int opGetAtReal8 = 58;
    final static int opGetAtStr = 59;
    final static int opGetAtObj = 60;

    final static int opFirst     = 61;
    final static int opLast      = 62;

    final static int opSysdate   = 63;
    final static int opTable     = 64;
    final static int opColumn    = 65;

    final static int opIntToBool = 66;
    final static int opRealToBool = 67;
    final static int opStrToBool = 68;

    final static int opStrLength = 70;
    final static int opLength = 71;

    final static int opSubStr = 72;

    final static int opExists = 73;
    final static int opIndexVar = 74;

    final static int opFalse   = 75;
    final static int opTrue    = 76;
    final static int opNull    = 77;

    final static int opGetOid  = 78;

    final static int opCommit  = 79;
    final static int opRollback = 80;

    final static int opLoadBlob  = 81;
    final static int opLoadClob  = 82;

    final static int opScanArrayBool = 83;
    final static int opScanArrayChar = 84;
    final static int opScanArrayInt1 = 85;
    final static int opScanArrayInt2 = 86;
    final static int opScanArrayInt4 = 87;
    final static int opScanArrayInt8 = 88;
    final static int opScanArrayReal4 = 89;
    final static int opScanArrayReal8 = 90;
    final static int opScanArrayStr = 91;
    final static int opScanArrayRef = 92;
    final static int opInString = 93;

    final static int opParameter = 94;
    final static int opIntConst = 95;
    final static int opRealConst = 97;
    final static int opStrConst = 98;
    
    final static int opDeref = 99;

    final static int opBoolToStr = 100;
    final static int opBoolToInt = 101;
    final static int opBoolToReal = 102;

    final static int opRealSin = 103;
    final static int opRealCos = 104;
    final static int opRealTan = 105;
    final static int opRealAsin = 106;
    final static int opRealAcos = 107;
    final static int opRealAtan = 108;
    final static int opRealSqrt = 109;
    final static int opRealExp = 110;
    final static int opRealLog = 111;
    final static int opRealCeil = 112;
    final static int opRealFloor = 113;

    final static int opBoolAnd = 114;
    final static int opBoolOr  = 115;
    final static int opBoolNot = 116;

    final static int opStrLower = 117;
    final static int opStrUpper = 118;
    final static int opStrConcat = 119;

    final static int opLoadBool = 120;
    final static int opLoadChar = 121;
    final static int opLoadInt1 = 122;
    final static int opLoadInt2 = 123;
    final static int opLoadInt4 = 124;
    final static int opLoadInt8 = 125;
    final static int opLoadReal4 = 126;
    final static int opLoadReal8 = 127;
    final static int opLoadStr  = 128;
    final static int opLoadBase = 129;

    final static int opSelect = 130;
    final static int opUpdate = 131;
    final static int opDelete = 132;
    final static int opInsert = 133;
    final static int opInsertObject = 134;

    final static int opCreateTable = 135;
    final static int opCreateIndex = 136;
    final static int opDropTable = 137;
    final static int opDropIndex = 138;
    final static int opAlterTable = 139;

    final static int opLoadBin = 140;

    final static int opIntMinAll = 141;
    final static int opIntMaxAll = 142;
    final static int opIntAvgAll = 143;
    final static int opIntSumAll = 144;
    final static int opIntMinDis = 146;
    final static int opIntMaxDis = 147;
    final static int opIntAvgDis = 148;
    final static int opIntSumDis = 149;
    
    final static int opRealMinAll = 151;
    final static int opRealMaxAll = 152;
    final static int opRealAvgAll = 153;
    final static int opRealSumAll = 154;
    final static int opRealMinDis = 156;
    final static int opRealMaxDis = 157;
    final static int opRealAvgDis = 158;
    final static int opRealSumDis = 159;
    
    final static int opCountAll   = 160;
    final static int opCountDis   = 161;

    final static int opScanSetBool = 162;
    final static int opScanSetInt  = 163;
    final static int opScanSetReal = 164;
    final static int opScanSetStr = 165;
    final static int opScanSetRef = 166;
    final static int opSetNotEmpty = 167;

    final static int opQuantorAll = 168;
    final static int opQuantorAny = 169;
    final static int opSubqueryElem = 170;

    final static int opCreateDomain = 171;

    final static int opInstanceOf = 172;
    final static int opThisInstanceOf = 173;

    final static String tpMnem[] = { 
	"Bool",
	"Int",
	"Real",
	"FreeVar",
	"List",
	"Ref",
	"Str",
	"Array",
	"ArrayBool",
	"ArrayInt1",
	"ArrayInt2",
	"ArrayInt4",
	"ArrayInt8",
	"ArrayReal4",
	"ArrayReal8",
	"ArrayStr",
	"ArrayRef",
	"Structure",
	"Blob",
	"Clob",
	"Cursor",
	"Table",
	"Void"
    };

    final static String opMnem[] = {
	"Nop",
	"IntAdd",
	"IntSub",
	"IntMul",
	"IntDiv",
	"IntAnd",
	"IntOr",
	"IntNeg",
	"IntNot",
	"IntAbs",
	"IntPow",
	"IntEq",
	"IntNe",
	"IntGt",
	"IntGe",
	"IntLt",
	"IntLe",
	"IntBetween",
	"RealEq",
	"RealNe",
	"RealGt",
	"RealGe",
	"RealLt",
	"RealLe",
	"RealBetween",
	"StrEq",
	"StrNe",
	"StrGt",
	"StrGe",
	"StrLt",
	"StrLe",
	"StrBetween",
	"StrLike",
	"StrLikeEsc",
	"BoolEq",
	"BoolNe",
	"OidEq",
	"OidNe",
	"RealAdd",
	"RealSub",
	"RealMul",
	"RealDiv",
	"RealNeg",
	"RealAbs",
	"RealPow",
	"IntToReal",
	"RealToInt",
	"IntToStr",
	"RealToStr",
	"StrToInt",
	"StrToReal",
	"IsNull",
	"StrGetAt",
	"GetAtInt1",
	"GetAtInt2",
	"GetAtInt4",
	"GetAtInt8",
	"GetAtReal4",
	"GetAtReal8",
	"GetAtStr",
	"GetAtObj",
	"First",
	"Last",
	"Sysdate",
	"Table",
	"Column",
	"IntToBool",
	"RealToBool",
	"StrToBool",
	"",	
	"StrLength",
	"Length",
	"SubStr",
	"Exists",
	"IndexVar",
	"False",
	"True",
	"Null",
	"GetOid",
	"Commit",
	"Rollback",
	"LoadBlob",
	"LoadClob",
	"ScanArrayBool",
	"ScanArrayChar",
	"ScanArrayInt1",
	"ScanArrayInt2",
	"ScanArrayInt4",
	"ScanArrayInt8",
	"ScanArrayReal4",
	"ScanArrayReal8",
	"ScanArrayStr",
	"ScanArrayRef",
	"InString",
	"Parameter",
	"IntConst",
	"",
	"RealConst",
	"StrConst",
	"Deref",
	"BoolToStr",
	"BoolToInt",
	"BoolToReal",
	"RealSin",
	"RealCos",
	"RealTan",
	"RealAsin",
	"RealAcos",
	"RealAtan",
	"RealSqrt",
	"RealExp",
	"RealLog",
	"RealCeil",
	"RealFloor",
	"BoolAnd",
	"BoolOr",
	"BoolNot",
	"StrLower",
	"StrUpper",
	"StrConcat",
	"LoadBool",
	"LoadChar",
	"LoadInt1",
	"LoadInt2",
	"LoadInt4",
	"LoadInt8",
	"LoadReal4",
	"LoadReal8",
	"LoadStr",
	"LoadBase",
	"Select",
	"Update",
	"Delete",
	"Insert",
	"InsertObject",
	"CreateTable",
	"CreateIndex",
	"DropTable",
	"DropIndex",
	"AlterTable",
	"LoadBin",
	"IntMinAll",
	"IntMaxAll",
	"IntAvgAll",
	"IntSumAll",
	"",
	"IntMinDis",
	"IntMaxDis",
	"IntAvgDis",
	"IntSumDis",
	"",
	"RealMinAll",
	"RealMaxAll",
	"RealAvgAll",
	"RealSumAll",
	"",
	"RealMinDis",
	"RealMaxDis",
	"RealAvgDis",
	"RealSumDis",
	"CountAll",
	"CountDis",
	"ScanSetBool",
	"ScanSetInt",
	"ScanSetReal",
	"ScanSetStr",
	"ScanSetRef",
	"SetNotEmpty",
	"QuantorAll",
	"QuantorAny",
	"SubqueryElem",
        "CreateDomain", 
        "InstanceOf", 
        "ThisInstanceOf" 
    };


    FieldDescriptor getField() { 
	return null;
    }

    boolean evaluateBool(ConsusStatement stmt) throws SQLException { 
	throw new AbstractMethodError();
    }
    int     evaluateOid(ConsusStatement stmt) throws SQLException { 
	return (int)evaluateInt(stmt);
    }
    long    evaluateInt(ConsusStatement stmt) throws SQLException { 
	throw new AbstractMethodError();
    }
    double  evaluateReal(ConsusStatement stmt) throws SQLException { 
	System.out.println(getClass());
	throw new AbstractMethodError();
    }
    String  evaluateStr(ConsusStatement stmt) throws SQLException { 
	throw new AbstractMethodError();
    }
    Object[] evaluateSet(ConsusStatement stmt) throws SQLException { 
	throw new AbstractMethodError();
    }
    byte[]  evaluateBinary(ConsusStatement stmt) throws SQLException { 
	throw new AbstractMethodError();
    }
    Object  evaluateObj(ConsusStatement stmt) throws SQLException { 
	throw new AbstractMethodError();
    }
   int evaluateBase(ConsusStatement stmt) throws SQLException { 
	throw new AbstractMethodError();
    }
    int dependsOn() { 
	return 0;
    }
    boolean equalsNode(ExprNode node) { 
	return false;
    }
    public int compareTo(Object obj) { 
	return dependsOn() - ((ExprNode)obj).dependsOn();
    }
    String dump() { 
	return opMnem[tag];
    }
    public ExprNode(int type, int tag) {
	this.type = type;
	this.tag = tag;
    }
}

