/* Consus - Pure Java Object Relational Database Management System
 * Node for array element access operation
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import  java.sql.SQLException;

class GetAtNode extends ExprNode { 
    ExprNode        left;
    ExprNode        right;
    FieldDescriptor field;

    String dump() { 
	return "(" + opMnem[tag] + " " + left.dump() + " " + right.dump() + ")";
    }
    boolean equalsNode(ExprNode node) { 
	return node.tag == tag && left.equalsNode(((GetAtNode)node).left) 
	    && right.equalsNode(((GetAtNode)node).right);
    }

    FieldDescriptor getField() { 
	return field;
    }

    int dependsOn() { 
	return Math.max(left.dependsOn(), right.dependsOn());
    }

    final int getElemOffs(ConsusStatement stmt, int elemSize) throws SQLException 
    {
	int offs = left.evaluateBase(stmt);
	int len = Varying.getSize(stmt.rec, offs);
	byte[] saveRec = stmt.rec;
	int index = (int)right.evaluateInt(stmt);
	stmt.rec = saveRec;
	if (index >= len) { 
	    if (right.tag == opIndexVar) { 
		throw new IndexOutOfRangeError(((IndexNode)right).loopId);
	    }
	}
	if (index < 0 || index >= len) { 
	    throw new ConsusRuntimeError("Index " + index + " is out of range");
	}
	return Varying.getOffs(stmt.rec, offs) + index*elemSize;
    }

    int evaluateBase(ConsusStatement stmt) throws SQLException { 
	return getElemOffs(stmt, field.size);
    }
    
    long evaluateInt(ConsusStatement stmt) throws SQLException 
    {
	byte[] saveRec = stmt.rec;	
	long val = 0;
        int offs;
	switch (tag) { 
	  case opGetAtInt1:
            offs = getElemOffs(stmt, 1);
	    val = stmt.rec[offs];
	    break;
	  case opGetAtInt2:
	    offs = getElemOffs(stmt, 2);
	    val = Bytes.unpack2(stmt.rec, offs);
	    break;
	  case opGetAtInt4:
	    offs = getElemOffs(stmt, 4);
	    val = Bytes.unpack4(stmt.rec, offs);
	    break;
	  case opGetAtInt8:
	    offs = getElemOffs(stmt, 8);
	    val = Bytes.unpack8(stmt.rec, offs);
	    break;
	  default:
	    Assert.failed("Invalid tag " + tag);
	}
	stmt.rec = saveRec;
	return val;
    }	    

    double evaluateReal(ConsusStatement stmt) throws SQLException  
    { 
	byte[] saveRec = stmt.rec;	
	double val = 0;
        int offs;
	switch (tag) { 
	  case opGetAtReal4:
	    offs = getElemOffs(stmt, 4);
	    val = Float.intBitsToFloat(Bytes.unpack4(stmt.rec, offs));
	    break;
	  case opGetAtReal8:
	    offs = getElemOffs(stmt, 8);
	    val = Double.longBitsToDouble(Bytes.unpack8(stmt.rec, offs));
	    break;
	  default:
	    Assert.failed("Invalid tag " + tag);
	}
	stmt.rec = saveRec;
	return val;
    }	    

    boolean evaluateBool(ConsusStatement stmt) throws SQLException { 
	byte[] saveRec = stmt.rec;	
        int offs = getElemOffs(stmt, 1);
	boolean val = stmt.rec[offs] != 0;
	stmt.rec = saveRec;
	return val;
    }

    String evaluateStr(ConsusStatement stmt) throws SQLException { 
	byte[] saveRec = stmt.rec;	
	int offs = getElemOffs(stmt, Varying.sizeof);
	String val = Varying.getString(stmt.rec, offs);
	stmt.rec = saveRec;
	return val;
    }

    int evaluateOid(ConsusStatement stmt) throws SQLException { 
	byte[] saveRec = stmt.rec;	
	int offs = getElemOffs(stmt, 4);
	int val = Bytes.unpack4(stmt.rec, offs);
	stmt.rec = saveRec;
	return val;
    }

    GetAtNode(int type, int tag, ExprNode base, ExprNode index, FieldDescriptor fd) { 
	super(type, tag);
	left = base;
	right = index;
	field = fd;
    }
}
