/* Consus - Pure Java Object Relational Database Management System
 * Iterator for START from ... FOLLOWING BY construction
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import  java.sql.*;

class IteratorNode { 
    static final int START_FROM_PARAM   = 0;
    static final int START_FROM_FIRST   = 1;
    static final int START_FROM_LAST    = 2;
    
    int             type;
    ParameterNode   param;
    IteratorVar     fields;
    TableDescriptor table;

    final boolean iterate(ConsusStatement stmt, ExprNode condition, int oid) throws SQLException { 
	while (oid != 0) {
	    stmt.rec = null;
	    if (!stmt.cursor.checkType(oid)) { 
		return true;
	    }
	    if (condition == null || stmt.evaluate(condition, oid)) { 
		if (!stmt.cursor.add(oid)) { 
		    return false;
		}
	    }
	    IteratorVar var = fields;
	    if (var != null) { 
		byte[] rec = stmt.rec;
		if (rec == null) {
		    rec = stmt.db.get(oid);
		}
		if (var.next == null && var.field.type == FieldDescriptor.tpReference) { 
		    oid = Bytes.unpack4(rec, var.field.offs);
		    continue;
		} else { 
		    do { 
			if (var.field.type == FieldDescriptor.tpReference) { 
			    if (!iterate(stmt, condition, Bytes.unpack4(rec, var.field.offs))) { 
				return false;
			    }
			} else { 
			    int n = Varying.getSize(rec, var.field.offs);
			    int offs = Varying.getOffs(rec, var.field.offs);
			    while (--n >= 0) { 
				if (!iterate(stmt, condition, Bytes.unpack4(rec, offs))) { 
				    return false;
				}
				offs += 4;
			    }
			}
		    } while ((var = var.next) != null);
		}
	    } 
	    break;
	}
	return true;
    }

    void iterate(ConsusStatement stmt, ExprNode condition) throws SQLException { 
	switch (type) { 
	  case START_FROM_PARAM:
	    switch (param.itype) { 
	      case ExprNode.tpRef:
		iterate(stmt, condition, param.oid);
		break;
	      case ExprNode.tpArrayRef:
		{
		    Ref[] refs = (Ref[])param.obj;
		    for (int i = 0; i < refs.length; i++) {
			if (refs[i] != null) { 
			    if (!iterate(stmt, condition, (int)((ConsusRef)refs[i]).getOid())) { 
				return;
			    }
			}
		    }
		    break;
		}
	      case ExprNode.tpCursor:
		{
		    ConsusResultSet cursor = (ConsusResultSet)param.obj;
		    for (int i = 0; i < cursor.nSelected; i++) { 
			if (!iterate(stmt, condition, cursor.selection[i])) { 
			    return;
			}
		    }
		    break;
		}
	      default:
		Assert.failed("Invalid parameter type");
	    }
	    break;
	  case START_FROM_FIRST:
	    iterate(stmt, condition, table.firstRow);
	    break;
	  case START_FROM_LAST:
	    iterate(stmt, condition, table.lastRow);
	    break;
	  default:
	    Assert.failed("Invalid iterator type");
	}
    }

    IteratorNode(TableDescriptor tab) {
	table = tab;
    }
}
