/* Consus - Pure Java Object Relational Database Management System
 * Impelementation of compare operation for ORDER BY
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

class OrderNode { 
    OrderNode        next;
    FieldDescriptor  field;
    int              columnNo;
    boolean          ascent;
    ConsusConnection db;

    OrderNode(ConsusConnection db, FieldDescriptor field, boolean ascent) { 
	this.db = db;
	this.field = field;
	this.ascent = ascent;
    }

    OrderNode(int columnNo, boolean ascent) { 
	this.columnNo = columnNo;
	this.ascent = ascent;
    }

    OrderNode() {}

    int compareTuples(Object[] t1, Object[] t2) { 
        int diff = ((Comparable)t1[columnNo]).compareTo(t2[columnNo]);
        if (diff != 0) { 
            return ascent ? diff : -diff;
        } else if (next != null) { 
            return next.compareTuples(t1, t2);
        }
	return 0;
    }

    int compare(byte[] obj1, byte[] obj2) { 
	int diff = 0;
	OrderNode node = this;

	do { 
	    int offs = node.field.offs;
	    switch (node.field.type) { 
	      case FieldDescriptor.tpBool:
	      case FieldDescriptor.tpInt1:
		diff = obj1[offs] - obj2[offs];
		break;
	      case FieldDescriptor.tpInt2:
		diff = Bytes.unpack2(obj1, offs) - Bytes.unpack2(obj2, offs);
		break;
	      case FieldDescriptor.tpInt4:
	      case FieldDescriptor.tpReference:
	      {
		  int l = Bytes.unpack4(obj1, offs);
		  int r = Bytes.unpack4(obj2, offs);
		  diff = l < r ? -1 : l == r ? 0 : 1;
		  break;
	      }
	      case FieldDescriptor.tpDate:
	      case FieldDescriptor.tpTime:
	      case FieldDescriptor.tpTimestamp:
	      case FieldDescriptor.tpInt8:
	      {
		  long l = Bytes.unpack8(obj1, offs);
		  long r = Bytes.unpack8(obj2, offs);
		  diff = l < r ? -1 : l == r ? 0 : 1;
		  break;
	      }
	      case FieldDescriptor.tpReal4:
	      {
		  float l = Float.intBitsToFloat(Bytes.unpack4(obj1, offs));
		  float r = Float.intBitsToFloat(Bytes.unpack4(obj2, offs));
		  diff = l < r ? -1 : l == r ? 0 : 1;
		  break;
	      }
	      case FieldDescriptor.tpReal8:
	      {
		  double l = Double.longBitsToDouble(Bytes.unpack8(obj1, offs));
		  double r = Double.longBitsToDouble(Bytes.unpack8(obj2, offs));
		  diff = l < r ? -1 : l == r ? 0 : 1;
		  break;
	      }
	      case FieldDescriptor.tpString:
	      {
		  int len1 = Varying.getSize(obj1, offs);
		  int ofs1 = Varying.getOffs(obj1, offs);
		  int len2 = Varying.getSize(obj2, offs);
		  int ofs2 = Varying.getOffs(obj2, offs);
		  int len = len1 < len2 ? len1 : len2;
		  while (--len >= 0 
			 && (diff=Bytes.unpack2(obj1,ofs1) - Bytes.unpack2(obj2,ofs2)) == 0) 
		  {
		      ofs1 += 2;
		      ofs2 += 2;
		  }
		      
		  if (len < 0) {
		      diff = len1 - len2;
		  }
		  break;
	      }
	      default:
		Assert.failed("Invalid type " + field.type);
	    }
	    if (!node.ascent) { 
		diff = -diff;
	    }
	    if (diff != 0) { 
		break;
	    }
	} while ((node = node.next) != null);
	return diff;
    }
}
