/* Consus - Pure Java Object Relational Database Management System
 * QSort implementation for different element types
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

class QSort { 
    /**
     * Swaps x[a] with x[b].
     */
    final static void swap(int ref[], int x[], int a, int b) {
	int t = x[a];
	x[a] = x[b];
	x[b] = t;
	int r = ref[a];
	ref[a] = ref[b];
	ref[b] = r;
    }

    /**
     * Swaps x[a .. (a+n-1)] with x[b .. (b+n-1)].
     */
    final static void vecswap(int ref[], int x[], int a, int b, int n) {
	for (int i=0; i<n; i++, a++, b++)
	    swap(ref, x, a, b);
    }

    /**
     * Returns the index of the median of the three indexed integers.
     */
    final static int med3(int x[], int a, int b, int c) {
	return (x[a] < x[b] ?
		(x[b] < x[c] ? b : x[a] < x[c] ? c : a) :
		(x[b] > x[c] ? b : x[a] > x[c] ? c : a));
    }

    final static void sort(int[] ref, int[] x, int off, int len) { 
	if (len < 7) {
	    for (int i=off; i<len+off; i++)
		for (int j=i; j>off && x[j-1]>x[j]; j--)
		    swap(ref, x, j, j-1);
	    return;
	}

	// Choose a partition element, v
	int m = off + len/2;       // Small arrays, middle element
	if (len > 7) {
	    int l = off;
	    int n = off + len - 1;
	    if (len > 40) {        // Big arrays, pseudomedian of 9
		int s = len/8;
		l = med3(x, l,     l+s, l+2*s);
		m = med3(x, m-s,   m,   m+s);
		n = med3(x, n-2*s, n-s, n);
	    }
	    m = med3(x, l, m, n); // Mid-size, med of 3
	}
	int v = x[m];

	// Establish Invariant: v* (<v)* (>v)* v*
	int a = off, b = a, c = off + len - 1, d = c;
	while(true) {
	    while (b <= c && x[b] <= v) {
		if (x[b] == v)
		    swap(ref, x, a++, b);
		b++;
	    }
	    while (c >= b && x[c] >= v) {
		if (x[c] == v)
		    swap(ref, x, c, d--);
		c--;
	    }
	    if (b > c)
		break;
	    swap(ref, x, b++, c--);
	}

	// Swap partition elements back to middle
	int s, n = off + len;
	s = Math.min(a-off, b-a  );  vecswap(ref, x, off, b-s, s);
	s = Math.min(d-c,   n-d-1);  vecswap(ref, x, b,   n-s, s);

	// Recursively sort non-partition-elements
	if ((s = b-a) > 1)
	    sort(ref, x, off, s);
	if ((s = d-c) > 1)
	    sort(ref, x, n-s, s);
    }


    /**
     * Swaps x[a] with x[b].
     */
    final static void swap(int ref[], double x[], int a, int b) {
	double t = x[a];
	x[a] = x[b];
	x[b] = t;
	int r = ref[a];
	ref[a] = ref[b];
	ref[b] = r;
    }

    /**
     * Swaps x[a .. (a+n-1)] with x[b .. (b+n-1)].
     */
    final static void vecswap(int ref[], double x[], int a, int b, int n) {
	for (int i=0; i<n; i++, a++, b++)
	    swap(ref, x, a, b);
    }

    /**
     * Returns the index of the median of the three indexed integers.
     */
    final static int med3(double x[], int a, int b, int c) {
	return (x[a] < x[b] ?
		(x[b] < x[c] ? b : x[a] < x[c] ? c : a) :
		(x[b] > x[c] ? b : x[a] > x[c] ? c : a));
    }

    final static void sort(int[] ref, double[] x, int off, int len) { 
	if (len < 7) {
	    for (int i=off; i<len+off; i++)
		for (int j=i; j>off && x[j-1]>x[j]; j--)
		    swap(ref, x, j, j-1);
	    return;
	}

	// Choose a partition element, v
	int m = off + len/2;       // Small arrays, middle element
	if (len > 7) {
	    int l = off;
	    int n = off + len - 1;
	    if (len > 40) {        // Big arrays, pseudomedian of 9
		int s = len/8;
		l = med3(x, l,     l+s, l+2*s);
		m = med3(x, m-s,   m,   m+s);
		n = med3(x, n-2*s, n-s, n);
	    }
	    m = med3(x, l, m, n); // Mid-size, med of 3
	}
	double v = x[m];

	// Establish Invariant: v* (<v)* (>v)* v*
	int a = off, b = a, c = off + len - 1, d = c;
	while(true) {
	    while (b <= c && x[b] <= v) {
		if (x[b] == v)
		    swap(ref, x, a++, b);
		b++;
	    }
	    while (c >= b && x[c] >= v) {
		if (x[c] == v)
		    swap(ref, x, c, d--);
		c--;
	    }
	    if (b > c)
		break;
	    swap(ref, x, b++, c--);
	}

	// Swap partition elements back to middle
	int s, n = off + len;
	s = Math.min(a-off, b-a  );  vecswap(ref, x, off, b-s, s);
	s = Math.min(d-c,   n-d-1);  vecswap(ref, x, b,   n-s, s);

	// Recursively sort non-partition-elements
	if ((s = b-a) > 1)
	    sort(ref, x, off, s);
	if ((s = d-c) > 1)
	    sort(ref, x, n-s, s);
    }



    /**
     * Swaps x[a] with x[b].
     */
    final static void swap(int ref[], long x[], int a, int b) {
	long t = x[a];
	x[a] = x[b];
	x[b] = t;
	int r = ref[a];
	ref[a] = ref[b];
	ref[b] = r;
    }

    /**
     * Swaps x[a .. (a+n-1)] with x[b .. (b+n-1)].
     */
    final static void vecswap(int ref[], long x[], int a, int b, int n) {
	for (int i=0; i<n; i++, a++, b++)
	    swap(ref, x, a, b);
    }

    /**
     * Returns the index of the median of the three indexed integers.
     */
    final static int med3(long x[], int a, int b, int c) {
	return (x[a] < x[b] ?
		(x[b] < x[c] ? b : x[a] < x[c] ? c : a) :
		(x[b] > x[c] ? b : x[a] > x[c] ? c : a));
    }

    final static void sort(int[] ref, long[] x, int off, int len) { 
	if (len < 7) {
	    for (int i=off; i<len+off; i++)
		for (int j=i; j>off && x[j-1]>x[j]; j--)
		    swap(ref, x, j, j-1);
	    return;
	}

	// Choose a partition element, v
	int m = off + len/2;       // Small arrays, middle element
	if (len > 7) {
	    int l = off;
	    int n = off + len - 1;
	    if (len > 40) {        // Big arrays, pseudomedian of 9
		int s = len/8;
		l = med3(x, l,     l+s, l+2*s);
		m = med3(x, m-s,   m,   m+s);
		n = med3(x, n-2*s, n-s, n);
	    }
	    m = med3(x, l, m, n); // Mid-size, med of 3
	}
	long v = x[m];

	// Establish Invariant: v* (<v)* (>v)* v*
	int a = off, b = a, c = off + len - 1, d = c;
	while(true) {
	    while (b <= c && x[b] <= v) {
		if (x[b] == v)
		    swap(ref, x, a++, b);
		b++;
	    }
	    while (c >= b && x[c] >= v) {
		if (x[c] == v)
		    swap(ref, x, c, d--);
		c--;
	    }
	    if (b > c)
		break;
	    swap(ref, x, b++, c--);
	}

	// Swap partition elements back to middle
	int s, n = off + len;
	s = Math.min(a-off, b-a  );  vecswap(ref, x, off, b-s, s);
	s = Math.min(d-c,   n-d-1);  vecswap(ref, x, b,   n-s, s);

	// Recursively sort non-partition-elements
	if ((s = b-a) > 1)
	    sort(ref, x, off, s);
	if ((s = d-c) > 1)
	    sort(ref, x, n-s, s);
    }



    /**
     * Swaps x[a] with x[b].
     */
    final static void swap(int ref[], float x[], int a, int b) {
	float t = x[a];
	x[a] = x[b];
	x[b] = t;
	int r = ref[a];
	ref[a] = ref[b];
	ref[b] = r;
    }

    /**
     * Swaps x[a .. (a+n-1)] with x[b .. (b+n-1)].
     */
    final static void vecswap(int ref[], float x[], int a, int b, int n) {
	for (int i=0; i<n; i++, a++, b++)
	    swap(ref, x, a, b);
    }

    /**
     * Returns the index of the median of the three indexed integers.
     */
    final static int med3(float x[], int a, int b, int c) {
	return (x[a] < x[b] ?
		(x[b] < x[c] ? b : x[a] < x[c] ? c : a) :
		(x[b] > x[c] ? b : x[a] > x[c] ? c : a));
    }

    final static void sort(int[] ref, float[] x, int off, int len) { 
	if (len < 7) {
	    for (int i=off; i<len+off; i++)
		for (int j=i; j>off && x[j-1]>x[j]; j--)
		    swap(ref, x, j, j-1);
	    return;
	}

	// Choose a partition element, v
	int m = off + len/2;       // Small arrays, middle element
	if (len > 7) {
	    int l = off;
	    int n = off + len - 1;
	    if (len > 40) {        // Big arrays, pseudomedian of 9
		int s = len/8;
		l = med3(x, l,     l+s, l+2*s);
		m = med3(x, m-s,   m,   m+s);
		n = med3(x, n-2*s, n-s, n);
	    }
	    m = med3(x, l, m, n); // Mid-size, med of 3
	}
	float v = x[m];

	// Establish Invariant: v* (<v)* (>v)* v*
	int a = off, b = a, c = off + len - 1, d = c;
	while(true) {
	    while (b <= c && x[b] <= v) {
		if (x[b] == v)
		    swap(ref, x, a++, b);
		b++;
	    }
	    while (c >= b && x[c] >= v) {
		if (x[c] == v)
		    swap(ref, x, c, d--);
		c--;
	    }
	    if (b > c)
		break;
	    swap(ref, x, b++, c--);
	}

	// Swap partition elements back to middle
	int s, n = off + len;
	s = Math.min(a-off, b-a  );  vecswap(ref, x, off, b-s, s);
	s = Math.min(d-c,   n-d-1);  vecswap(ref, x, b,   n-s, s);

	// Recursively sort non-partition-elements
	if ((s = b-a) > 1)
	    sort(ref, x, off, s);
	if ((s = d-c) > 1)
	    sort(ref, x, n-s, s);
    }



    /**
     * Swaps x[a] with x[b].
     */
    final static void swap(int ref[], short x[], int a, int b) {
	short t = x[a];
	x[a] = x[b];
	x[b] = t;
	int r = ref[a];
	ref[a] = ref[b];
	ref[b] = r;
    }

    /**
     * Swaps x[a .. (a+n-1)] with x[b .. (b+n-1)].
     */
    final static void vecswap(int ref[], short x[], int a, int b, int n) {
	for (int i=0; i<n; i++, a++, b++)
	    swap(ref, x, a, b);
    }

    /**
     * Returns the index of the median of the three indexed integers.
     */
    final static int med3(short x[], int a, int b, int c) {
	return (x[a] < x[b] ?
		(x[b] < x[c] ? b : x[a] < x[c] ? c : a) :
		(x[b] > x[c] ? b : x[a] > x[c] ? c : a));
    }

    final static void sort(int[] ref, short[] x, int off, int len) { 
	if (len < 7) {
	    for (int i=off; i<len+off; i++)
		for (int j=i; j>off && x[j-1]>x[j]; j--)
		    swap(ref, x, j, j-1);
	    return;
	}

	// Choose a partition element, v
	int m = off + len/2;       // Small arrays, middle element
	if (len > 7) {
	    int l = off;
	    int n = off + len - 1;
	    if (len > 40) {        // Big arrays, pseudomedian of 9
		int s = len/8;
		l = med3(x, l,     l+s, l+2*s);
		m = med3(x, m-s,   m,   m+s);
		n = med3(x, n-2*s, n-s, n);
	    }
	    m = med3(x, l, m, n); // Mid-size, med of 3
	}
	short v = x[m];

	// Establish Invariant: v* (<v)* (>v)* v*
	int a = off, b = a, c = off + len - 1, d = c;
	while(true) {
	    while (b <= c && x[b] <= v) {
		if (x[b] == v)
		    swap(ref, x, a++, b);
		b++;
	    }
	    while (c >= b && x[c] >= v) {
		if (x[c] == v)
		    swap(ref, x, c, d--);
		c--;
	    }
	    if (b > c)
		break;
	    swap(ref, x, b++, c--);
	}

	// Swap partition elements back to middle
	int s, n = off + len;
	s = Math.min(a-off, b-a  );  vecswap(ref, x, off, b-s, s);
	s = Math.min(d-c,   n-d-1);  vecswap(ref, x, b,   n-s, s);

	// Recursively sort non-partition-elements
	if ((s = b-a) > 1)
	    sort(ref, x, off, s);
	if ((s = d-c) > 1)
	    sort(ref, x, n-s, s);
    }


    /**
     * Swaps x[a] with x[b].
     */
    final static void swap(int ref[], byte x[], int a, int b) {
	byte t = x[a];
	x[a] = x[b];
	x[b] = t;
	int r = ref[a];
	ref[a] = ref[b];
	ref[b] = r;
    }

    /**
     * Swaps x[a .. (a+n-1)] with x[b .. (b+n-1)].
     */
    final static void vecswap(int ref[], byte x[], int a, int b, int n) {
	for (int i=0; i<n; i++, a++, b++)
	    swap(ref, x, a, b);
    }

    /**
     * Returns the index of the median of the three indexed integers.
     */
    final static int med3(byte x[], int a, int b, int c) {
	return (x[a] < x[b] ?
		(x[b] < x[c] ? b : x[a] < x[c] ? c : a) :
		(x[b] > x[c] ? b : x[a] > x[c] ? c : a));
    }

    final static void sort(int[] ref, byte[] x, int off, int len) { 
	if (len < 7) {
	    for (int i=off; i<len+off; i++)
		for (int j=i; j>off && x[j-1]>x[j]; j--)
		    swap(ref, x, j, j-1);
	    return;
	}

	// Choose a partition element, v
	int m = off + len/2;       // Small arrays, middle element
	if (len > 7) {
	    int l = off;
	    int n = off + len - 1;
	    if (len > 40) {        // Big arrays, pseudomedian of 9
		int s = len/8;
		l = med3(x, l,     l+s, l+2*s);
		m = med3(x, m-s,   m,   m+s);
		n = med3(x, n-2*s, n-s, n);
	    }
	    m = med3(x, l, m, n); // Mid-size, med of 3
	}
	byte v = x[m];

	// Establish Invariant: v* (<v)* (>v)* v*
	int a = off, b = a, c = off + len - 1, d = c;
	while(true) {
	    while (b <= c && x[b] <= v) {
		if (x[b] == v)
		    swap(ref, x, a++, b);
		b++;
	    }
	    while (c >= b && x[c] >= v) {
		if (x[c] == v)
		    swap(ref, x, c, d--);
		c--;
	    }
	    if (b > c)
		break;
	    swap(ref, x, b++, c--);
	}

	// Swap partition elements back to middle
	int s, n = off + len;
	s = Math.min(a-off, b-a  );  vecswap(ref, x, off, b-s, s);
	s = Math.min(d-c,   n-d-1);  vecswap(ref, x, b,   n-s, s);

	// Recursively sort non-partition-elements
	if ((s = b-a) > 1)
	    sort(ref, x, off, s);
	if ((s = d-c) > 1)
	    sort(ref, x, n-s, s);
    }


    /**
     * Swaps x[a] with x[b].
     */
    final static void swap(int ref[], char x[], int size, int a, int b) {
	for (int ia = a*size, ib = b*size; --size >= 0; ia++, ib++) { 
	    char t = x[ia];
	    x[ia] = x[ib];
	    x[ib] = t;
	}
	int r = ref[a];
	ref[a] = ref[b];
	ref[b] = r;
    }

    /**
     * Swaps x[a .. (a+n-1)] with x[b .. (b+n-1)].
     */
    final static void vecswap(int ref[], char x[], int size, int a, int b, int n) {
	for (int len = n*size, ia = a*size, ib = b*size; --len >= 0; ia++, ib++) { 
	    char t = x[ia];
	    x[ia] = x[ib];
	    x[ib] = t;
	}
	while (--n >= 0) { 
	    int r = ref[a];
	    ref[a] = ref[b];
	    ref[b] = r;
	    a += 1;
	    b += 1;
	}
    }

    final static int compare(char x[], int size, int a, int b) { 
	for (int ia = a*size, ib = b*size; --size >= 0; ia++, ib++) { 
	    int diff = x[ia] - x[ib];
	    if (diff != 0) { 
		return diff;
	    }
	}
	return 0;
    }

    /**
     * Returns the index of the median of the three indexed integers.
     */
    final static int med3(char x[], int size, int a, int b, int c) {
	int diff_ab = compare(x, size, a, b);
	int diff_bc = compare(x, size, b, c);
	int diff_ac = compare(x, size, a, c);
	return (diff_ab < 0 ?
		(diff_bc < 0 ? b : diff_ac < 0 ? c : a) :
		(diff_bc > 0 ? b : diff_ac > 0 ? c : a));
    }

    final static void sort(int[] ref, char[] x, int size, int off, int len) { 
	if (len < 7) {
	    for (int i=off; i<len+off; i++)
		for (int j=i; j>off && compare(x, size, j-1, j) > 0; j--)
		    swap(ref, x, size, j, j-1);
	    return;
	}

	// Choose a partition element, v
	int m = off + len/2;       // Small arrays, middle element
	if (len > 7) {
	    int l = off;
	    int n = off + len - 1;
	    if (len > 40) {        // Big arrays, pseudomedian of 9
		int s = len/8;
		l = med3(x, size, l,     l+s, l+2*s);
		m = med3(x, size, m-s,   m,   m+s);
		n = med3(x, size, n-2*s, n-s, n);
	    }
	    m = med3(x, size, l, m, n); // Mid-size, med of 3
	}

	// Establish Invariant: v* (<v)* (>v)* v*
	int a = off, b = a, c = off + len - 1, d = c;
	int diff;
	while(true) {
	    while (b <= c && (diff = compare(x, size, b, m)) <= 0) {
		if (diff == 0)
		    swap(ref, x, size, a++, b);
		b++;
	    }
	    while (c >= b && (diff = compare(x, size, c, m)) >= 0) {
		if (diff == 0)
		    swap(ref, x, size, c, d--);
		c--;
	    }
	    if (b > c)
		break;
	    swap(ref, x, size, b++, c--);
	}

	// Swap partition elements back to middle
	int s, n = off + len;
	s = Math.min(a-off, b-a  );  vecswap(ref, x, size, off, b-s, s);
	s = Math.min(d-c,   n-d-1);  vecswap(ref, x, size, b,   n-s, s);

	// Recursively sort non-partition-elements
	if ((s = b-a) > 1)
	    sort(ref, x, size, off, s);
	if ((s = d-c) > 1)
	    sort(ref, x, size, n-s, s);
    }
}






