/* Consus - Pure Java Object Relational Database Management System
 * Node for operations with three operands
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import  java.sql.SQLException;

class TripleOpNode extends ExprNode {
    ExprNode o1, o2, o3;

    String dump() { 
	return "(" + opMnem[tag] + " " + o1.dump() + " " + o2.dump() + " " 
	    + (o3 != null ? (o3.dump() + ")") : ")");
    }
    boolean equalsNode(ExprNode node) { 
	return node.tag == tag 
            && o1.equalsNode(((TripleOpNode)node).o1) 
            && o2.equalsNode(((TripleOpNode)node).o2)
	    && (o3 == ((TripleOpNode)node).o3 
                || (o3 != null && o3.equalsNode(((TripleOpNode)node).o3)));
    }

    int dependsOn() { 
	int max = Math.max(o1.dependsOn(), o2.dependsOn());
	if (o3 != null) { 
	    max =  Math.max(o3.dependsOn(), max);
	}
	return max;
    }

    String evaluateStr(ConsusStatement stmt) throws SQLException { 
	Assert.that(tag == opSubStr);
	String str = o1.evaluateStr(stmt);
	int from = (int)o2.evaluateInt(stmt);
	return (o3 != null) 
	    ? str.substring(from, from + (int)o3.evaluateInt(stmt))
	    : str.substring(from);
    }

    boolean evaluateBool(ConsusStatement stmt) throws SQLException {
	switch(tag) { 
	  case opIntBetween:
	  {
	    long val = o1.evaluateInt(stmt);
	    return val >= o2.evaluateInt(stmt) && val <= o3.evaluateInt(stmt);
	  }
	  
	  case opRealBetween:
	  {
	    double val = o1.evaluateReal(stmt);
	    return val >= o2.evaluateReal(stmt) && val <= o3.evaluateReal(stmt);
	  }
	  
	  case opStrBetween:
	  {
	    String val = o1.evaluateStr(stmt);
	    return val.compareTo(o2.evaluateStr(stmt)) >= 0 
		&& val.compareTo(o3.evaluateStr(stmt)) <= 0;
	  }
	  
	  case opStrLike:
	  {
	    String str = o1.evaluateStr(stmt);
	    String pat = o2.evaluateStr(stmt);
	    return Regexp.matchString(str, pat, '\\');
	  }
	  case opStrLikeEsc:
	  {
	    String str = o1.evaluateStr(stmt);
	    String pat = o2.evaluateStr(stmt);
	    char escape = o3.evaluateStr(stmt).charAt(0);
	    return Regexp.matchString(str, pat, escape);
	  }
	  default:
	    Assert.failed("Invalid tag " + tag);
	}
	return false;
    }

    TripleOpNode(int tag, ExprNode a, ExprNode b, ExprNode c) { 
	super(tpBool, tag);
	o1 = a;
	o2 = b;
	o3 = c;
    }

    TripleOpNode(int type, int tag, ExprNode a, ExprNode b, ExprNode c) { 
	super(type, tag);
	o1 = a;
	o2 = b;
	o3 = c;
    }
}
