/* Consus - Pure Java Object Relational Database Management System
 * Node for UPDATE statement
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

import  java.sql.SQLException;
    
class UpdateNode extends SelectNode { 
    ColumnValue     values;

    final boolean update(ConsusStatement stmt, int oid) throws SQLException { 
	int tableId = Record.getType(stmt.rec, 0);
	TableDescriptor selfTable = 
	    (tableId != table.tableId) ? stmt.db.getTableByOid(tableId) : table;
	java.util.Arrays.fill(stmt.updated, false);
	for (ColumnValue cv = values; cv != null; cv = cv.next) { 
	    cv.evaluate(stmt);
	}
	table.removeFromIndices(oid, stmt.updated);
	byte[] newRec = selfTable.updateRecord(stmt, stmt.rec);
	db.put(oid, newRec);
	if (!table.updateIndices(oid, stmt.updated)) { 
	    db.put(oid, stmt.rec);
	    table.updateIndices(oid, stmt.updated);
	    return false;
	}
	return true;
    }
    
    int executeStmt(ConsusStatement stmt) throws SQLException { 
	int saveLimit = stmt.cursor.selectionLimit;
	stmt.cursor.selectionLimit = Integer.MAX_VALUE;
	int nSelected = super.executeStmt(stmt);
	stmt.cursor.selectionLimit = saveLimit;
	stmt.updated = new boolean[table.nFields];
	stmt.columnValue = new Object[table.nFields];
	int nUpdated = 0;
	for (int i = 0; i < nSelected; i++) { 
	    stmt.oid = stmt.cursor.selection[i];
	    stmt.rec = db.get(stmt.oid);
	    if (update(stmt, stmt.oid)) {
		nUpdated += 1;
	    }
	} 
	stmt.cursor.close();
	return nUpdated;
    }

    UpdateNode(TableDescriptor desc, ExprNode criteria, ColumnValue values) { 
	super(opUpdate, desc, criteria);
	this.values = values;
    }
}
