/* Consus - Pure Java Object Relational Database Management System
 * Fixed header for varying data types (string and arrays)
 * Created by Konstantin Knizhnik, Feb 2000.
 */
package org.garret.internal.consus;
import org.garret.consus.*;

//
// Fixed header for varying component (arrays or strings)
// 
class Varying { 
    static int getSize(byte[] arr, int offs) { 
	return Bytes.unpack4(arr, offs);
    }
    static int getOffs(byte[] arr, int offs) { 
	return Bytes.unpack4(arr, offs+4);
    }
    static void setSize(byte[] arr, int offs, int size) { 
	Bytes.pack4(arr, offs, size);
    }
    static void setOffs(byte[] arr, int offs, int offs_val) { 
	Bytes.pack4(arr, offs+4, offs_val);
    }

    static String getString(byte[] rec, int offs) { 
	int len = getSize(rec, offs);
	offs = getOffs(rec, offs);
	char[] chars = new char[len];
	for (int i = 0; i < len; i++) { 
	    chars[i] = (char)Bytes.unpack2(rec, offs);
	    offs += 2;
	}
	return new String(chars);
    }

    static String getStringOrNull(byte[] rec, int offs) { 
	int len = getSize(rec, offs);
	offs = getOffs(rec, offs);
	if (offs == 0) { 
	    return null;
	}
	char[] chars = new char[len];
	for (int i = 0; i < len; i++) { 
	    chars[i] = (char)Bytes.unpack2(rec, offs);
	    offs += 2;
	}
	return new String(chars);
    }

    static void setString(byte[] rec, int offs, String str) { 
	int len = str.length();
	for (int i = 0; i < len; i++) { 
	    Bytes.pack2(rec, offs, (short)str.charAt(i));
	    offs += 2;
	}
    }

    static int setString(byte[] rec, int offs, String str, int varyingOffs) { 
	int len = str.length();
	setSize(rec, offs, len);
	setOffs(rec, offs, varyingOffs);
	for (int i = 0; i < len; i++) { 
	    Bytes.pack2(rec, varyingOffs, (short)str.charAt(i));
	    varyingOffs += 2;
	}
	return varyingOffs;
    }	

    static final int sizeof = 8;

}

