/* Consus - Pure Java Object Relational Database Management System
 * Example of mobile application
 * Created by Konstantin Knizhnik, May 2000.
 */
import java.io.*;
import java.sql.*;
import org.garret.consus.*;

public class Contract {
    public String  contractID;
    public String  company;
    public Date    delivery;
    public double  price;
    public long    quantity;
    public boolean canceled;

    public long    recordTimestamp;
    public Ref     masterOID;

    static public String indices[] = {"contractID", "delivery", "recordTimestamp", "masterOID"};
    static public String primaryKeys[] = {"contractID"};

    static BufferedReader in = new BufferedReader(new InputStreamReader(System.in)); 

    static String input(String prompt) {
	while (true) { 
	    try { 
		System.out.print(prompt);
		String answer = in.readLine().trim();
		if (answer.length() != 0) {
		    return answer;
		}
	    } catch (IOException x) {}
	}
    }
    static long inputLong(String prompt) { 
	while (true) { 
	    try { 
		return Long.parseLong(input(prompt));	    
	    } catch(NumberFormatException x) {
		System.out.println("Invalid number");
	    }
	}
    }
    static double inputDouble(String prompt) { 
	while (true) { 
	    try { 
		return Double.parseDouble(input(prompt));	    
	    } catch(NumberFormatException x) {
		System.out.println("Invalid number");
	    }
	}
    }	
    static Date inputDate(String prompt) { 
	while (true) { 
	    try { 
		return Date.valueOf(input(prompt));	    
	    } catch(Exception x) {
		System.out.println("Invalid date, use format (YEAR-MONTH-DAY)");
	    }
	}
    }	

    void print() { 
	System.out.println(contractID + "\t" + canceled + "\t" + company + "\t" + delivery 
			   + "\t" + price + "\t" + quantity);
    }

    static public void main(String[] args) throws Exception {
	if (args.length != 2) { 
	    System.out.println("Usage: java Contract <local-database-URL> <remote-database-URL>");
	    return;
	}
	Class.forName("org.garret.consus.ConsusDriver");
	Class.forName("org.garret.consus.ClientDriver");	
	ConsusConnection localConnection = 
	    (ConsusConnection)DriverManager.getConnection(args[0]);
	if (localConnection == null) { 
	    System.err.println("No connection with local database");
	    return;
	}
	Connection masterConnection = DriverManager.getConnection(args[1]);	
	if (masterConnection == null) { 
	    System.err.println("No connection with master database");
	    return;
	}
	masterConnection.setAutoCommit(false);
	localConnection.checkOutTable("Contract", masterConnection);
	masterConnection.close();
	PreparedStatement selectAll = localConnection.prepareStatement("select * from Contract order by delivery");
	PreparedStatement select = localConnection.prepareStatement("select * from Contract where contractID=? for update");
	Contract contract = new Contract();
	ResultSet cursor;

	while (true) { 
	    System.out.println("Menu:\n1. Show contracts\n2. Add contract\n3. Change contract\n4. Cancel contract\n5. Exit\n");
	    switch ((int)inputLong(">>")) { 
	      case 1:
		cursor = selectAll.executeQuery();
		System.out.println("ID\tDeleted\tCompany\tDelivery\tPrice\tQuantity");
		while (cursor.next()) { 
		    ((Contract)cursor.getObject(0)).print();
		}
		System.out.println("--------------------------------------------------------");
		break;
	      case 2:
		contract.contractID = input("ContractID: ");
		contract.company    = input("Company: ");
		contract.delivery   = inputDate("Date: ");
		contract.price      = inputDouble("Price: ");
		contract.quantity   = inputLong("Quantity: ");
		localConnection.insert(contract);
		break;
	      case 3:
		select.setString(1, input("ContractID: "));
		cursor = select.executeQuery();
		if (cursor.next()) { 
		    cursor.updateString("company", input("Company: "));
		    cursor.updateDate("delivery", inputDate("Delivery: "));
		    cursor.updateDouble("price", inputDouble("Price: "));
		    cursor.updateLong("quantity", inputLong("Quantity: "));
		    cursor.updateRow();
		} else { 
		    System.err.println("No such contract");
		}
		break;
	      case 4:
		select.setString(1, input("ContractID: "));
		cursor = select.executeQuery();
		if (cursor.next()) { 
		    cursor.updateBoolean("canceled", true);
		    cursor.updateRow();
		} else { 
		    System.err.println("No such contract");
		}
		break;
	      case 5:
		masterConnection = DriverManager.getConnection(args[1]);	
		if (masterConnection == null) { 
		    System.err.println("No connection with master database");
		    return;
		}
		masterConnection.setAutoCommit(false);
		try { 
		    localConnection.checkInTable("Contract", masterConnection);
		    masterConnection.commit();
		} catch(MergeConflictException x) { 
		    System.err.println("Conflict for contract " + 
				       ((Contract)localConnection.get((Ref)x.key)).contractID);
		    localConnection.rollback();
		}
		localConnection.close();
		masterConnection.close();
		return;
	      default:
		System.err.println("Please specify 1-5");
	    }
	}
    }
}









