/* Consus - Pure Java Object Relational Database Management System
 * Example of direct access to the objects by references: game "Guess an animal"
 * Created by Konstantin Knizhnik, Feb 2000.
 */
import java.io.*;
import java.sql.*;
import org.garret.consus.*;

public class Guess {
    public Ref  yes;
    public Ref  no;
    public String question;

    static ConsusStatement stmt;

    public Guess(Ref no, String question, Ref yes) { 
	this.yes = yes;
	this.question = question;
	this.no = no;
    }
    
    public Guess() { /* constructor for Consus */ }

    static BufferedReader in = new BufferedReader(new InputStreamReader(System.in)); 

    static String input(String prompt) {
	while (true) { 
	    try { 
		System.out.print(prompt);
		String answer = in.readLine().trim();
		if (answer.length() != 0) {
		    return answer;
		}
	    } catch (IOException x) {}
	}
    }

    static boolean askQuestion(String question) { 
	String answer = input(question);
	return answer.equalsIgnoreCase("y") || answer.equalsIgnoreCase("yes");
    }

    static Ref whoIsIt(Ref parent) throws SQLException { 
	String animal = input("What is it ? ");
	String difference = input("What is a difference from other ? ");
	return stmt.insert(new Guess(parent, difference, stmt.insert(new Guess(null, animal, null))));
    }

    static Ref dialog(Ref ref) throws SQLException {  
	Guess node = (Guess)stmt.get(ref);
	if (askQuestion("May be, " + node.question + " (y/n) ? ")) {
	    if (node.yes.equals(null)) { 
		System.out.println("It was very simple question for me...");
	    } else { 
		Ref clarify = dialog(node.yes);
		if (clarify != null) { 
		    node.yes = clarify;
		    stmt.update(ref, node);
		}
	    }
	} else { 
	    if (node.no.equals(null)) { 
		if (node.yes.equals(null)) { 
		    return whoIsIt(ref);
		} else {
		    node.no = whoIsIt(null);
		    stmt.update(ref, node);
		} 
	    } else { 
	        Ref clarify = dialog(node.no);
		if (clarify != null) { 
		    node.no = clarify;
		    stmt.update(ref, node);
		}
	    }
	}
	return null; 
    }
    
    static public void main(String[] args) throws Exception {
	Class.forName("org.garret.consus.ConsusDriver");
	Connection connection = DriverManager.getConnection("jdbc:consus:guess.dbs");	
	//Connection connection = DriverManager.getConnection("jdbc:consus:guess.mfd");	
	if (connection == null) { 
	    System.err.println("No connection with database GUESS");
	    return;
	}
	stmt = (ConsusStatement)connection.createStatement();
	Ref root = null;
	while (askQuestion("Think of an animal. Ready (y/n) ? ")) { 
	    if (root == null) { 
		ResultSet cursor;
		stmt.setMaxRows(2);
		cursor = stmt.executeQuery("select * from Guess for update");
		if (cursor.absolute(2)) { 
		    root = cursor.getRef(0);
		    cursor.close();
		} else { 
		    cursor.close();
		    root = whoIsIt(null);
		    continue;
		}
	    }
	    dialog(root);
	    connection.commit();
	}
	connection.close();
    }
}

