/* Consus - Pure Java Object Relational Database Management System
 * Example of direct access to the objects by references: game "Guess an animal"
 * Created by Konstantin Knizhnik, Feb 2000.
 */

import java.io.*;
import java.sql.*;
import org.garret.consus.*;

public class GuessPO extends Persistent {
    public GuessPO yes;
    public GuessPO no;
    public String  question;

    static PersistentObjectStorage storage;

    public GuessPO(GuessPO no, String question, GuessPO yes) { 
	this.yes = yes;
	this.question = question;
	this.no = no;
    }
    
    public GuessPO() { /* constructor for Consus */ }

    static BufferedReader in = new BufferedReader(new InputStreamReader(System.in)); 

    static String input(String prompt) {
	while (true) { 
	    try { 
		System.out.print(prompt);
		String answer = in.readLine().trim();
		if (answer.length() != 0) {
		    return answer;
		}
	    } catch (IOException x) {}
	}
    }

    static boolean askQuestion(String question) { 
	String answer = input(question);
	return answer.equalsIgnoreCase("y") || answer.equalsIgnoreCase("yes");
    }

    static GuessPO whoIsIt(GuessPO parent) throws SQLException { 
	String animal = input("What is it ? ");
	String difference = input("What is a difference from other ? ");
	return new GuessPO(parent, difference, new GuessPO(null, animal, null));
    }

    GuessPO dialog() throws SQLException {  
	if (askQuestion("May be, " + question + " (y/n) ? ")) {
	    if (yes == null) { 
		System.out.println("It was very simple question for me...");
	    } else { 
		GuessPO clarify = yes.dialog();
		if (clarify != null) { 
		    yes = clarify;
		    store();
		}
	    }
	} else { 
	    if (no == null) { 
		if (yes == null) { 
		    return whoIsIt(this);
		} else {
		    no = whoIsIt(null);
		    store();
		} 
	    } else { 
	        GuessPO clarify = no.dialog();
		if (clarify != null) { 
		    no = clarify;
		    store();
		}
	    }
	}
	return null; 
    }
    
    static public void main(String[] args) throws Exception {
	Class.forName("org.garret.consus.ConsusDriver");
	Connection con = DriverManager.getConnection("jdbc:consus:guesspo.dbs");	
	if (con == null) { 
	    System.err.println("No connection with database GUESS");
	    return;
	}
	PersistentObjectStorage storage;
	if (args.length != 0) { 
	    storage = (ConsusConnection)con;
	} else { 
	    storage = new JDBCObjectRelationalAdapter(con);
	    if (!storage.isInitialized()) { 
		Statement stmt = con.createStatement();
		// database was not initialized
		stmt.executeUpdate("create table GuessPO (RID int, TID int, yes longint, no longint, question varchar)");
		stmt.executeUpdate("create table TypeInfo (TID int, TableName varchar, ClassName varchar)");
		stmt.executeUpdate("create table OidTable (ID int)");
		stmt.executeUpdate("insert into TypeInfo values (1, 'GuessPO', 'GuessPO')");
		stmt.executeUpdate("insert into OidTable values (1)");
	    }
	    storage.open();
	    con.commit();
	}
	GuessPO root = null;
	TableIterator iterator = storage.getTableIterator("GuessPO");
	while (askQuestion("Think of an animal. Ready (y/n) ? ")) { 
	    if (root == null) { 
		if (iterator.absolute(2)) { 
		    root = (GuessPO)iterator.get();
		} else { 
		    root = whoIsIt(null);
		    root.store(storage);
		    continue;
		}
	    }
	    root.dialog();
	    con.commit();
	} 
	con.close();
    }
}
