/* Consus - Pure Java Object Relational Database Management System
 * Test of standard JDBC features
 * Created by Konstantin Knizhnik, Feb 2000.
 */
import java.sql.*;

public class Test { 
    static public void main(String args[]) throws Exception { 
	int rc;
	Class.forName("org.garret.consus.ConsusDriver");
	Connection connection = DriverManager.getConnection("jdbc:consus:test.dbs");	
	if (connection == null) { 
	    System.err.println("No connection with database TEST");
	    return;
	}
	Statement stmt = connection.createStatement();
	ResultSet cursor = stmt.executeQuery("select * from MetaTable where name = 'Person'");
	if (!cursor.next()) { 
	    stmt.execute("create table Person(name varchar primary key, salary bigint, " 
			 + "address varchar, weight double, nsubs integer, boss varchar)");
	}
	cursor.close();

	PreparedStatement pstmt = 
	    connection.prepareStatement("insert into Person (name, salary, address, weight, nsubs, boss)"
					+ "values (?,?,?,?,?,?)");
	pstmt.setString(1, "John Smith");
	pstmt.setInt(2, 75000);
	pstmt.setString(3, "1 Guildhall St., Cambridge CB2 3NH, UK");
	pstmt.setDouble(4, 80.3);
	pstmt.setByte(5, (byte)0);
	pstmt.setString(6, "Joe Cooker");

	if ((rc = pstmt.executeUpdate()) != 1) { 
	     System.err.println("insert-1 returns " + rc + " instead of 1");
	}

 	pstmt.setString(1, "Joe Cooker");
	pstmt.setInt(2, 100000);
	pstmt.setString(3, "Outlook drive, 15/3");
	pstmt.setFloat(4, (float)75.5);
	pstmt.setByte(5, (byte)1);
	pstmt.setString(6, "");

	if ((rc = pstmt.executeUpdate()) != 1) { 
	    System.err.println("insert-2 returns " + rc + " instead of 1");
	}
	pstmt.close();
	pstmt = connection.prepareStatement("select * from Person where nsubs < ? and salary > ?");
	pstmt.setInt(1, 2);
	pstmt.setLong(2, 90000);
	
	cursor = pstmt.executeQuery();
	rc = 0;
	while (cursor.next()) { 
	    System.out.println(cursor.getString(1));
	    rc += 1;
	}
	if (rc != 1) { 
	    System.err.println("select-1 fetches " + rc + " records instead of 1");
	}
	pstmt.setInt(1, 10);
	pstmt.setInt(2, 50000);
	cursor = pstmt.executeQuery();
	rc = 0;
	PreparedStatement pstmt2 = connection.prepareStatement("select * from Person where boss = ?");
	while (cursor.next()) { 
	    System.out.println(cursor.getString("name") + "\t" + cursor.getLong("salary") + "\t"
			       + cursor.getFloat("weight") + "\t" + cursor.getString("address"));
	    pstmt2.setString(1, cursor.getString(1));
	    ResultSet cursor2 = pstmt2.executeQuery();
	    while (cursor2.next()) { 
		System.out.println("\tManages " + cursor2.getString("name"));
	    }
	    rc += 1;
	}
	if (rc != 2) { 
	    System.err.println("select-2 fetches " + rc + " records instead of 2");
	}
	pstmt2.close();
	pstmt.close();
	pstmt = connection.prepareStatement("select * from Person where salary between 50000 and 100000"
					    + " for update");
	cursor = pstmt.executeQuery();
	rc = 0;
	cursor.afterLast();
	while (cursor.previous()) {
	    System.out.println("Update " + cursor.getString(1) + "'s salary");
	    cursor.updateLong("salary", cursor.getLong("salary")*90/100);
	    cursor.updateRow();
	    rc += 1;
	}
	if (rc != 2) { 
	    System.err.println("select-3 fetches " + rc + " records instead of 2");
	}
	pstmt.close();
	
	connection.commit();
	pstmt = connection.prepareStatement("select * from Person order by salary desc for update");
	rc = 0;
	cursor = pstmt.executeQuery();
	while (cursor.next()) {
	    System.out.println(cursor.getString("name") + "\t" 
			       + cursor.getInt("salary"));
	    cursor.deleteRow();
	    rc += 2;
	}
	if (rc != 4) { 
	    System.err.println("select-4 fetches " + rc + " records instead of 2");
	}
	pstmt.close();
	connection.close();
	System.out.println("Test passed!");
    }
}
	

