/* Consus - Pure Java Object Relational Database Management System
 * Test of JSQL operations with arrays
 * Created by Konstantin Knizhnik, Mar 2000.
 */

import java.sql.*;
import org.garret.consus.*;

class Elem { 
    int    i;
    String s;
    
    Elem(int x) { 
	i = x;
	s = "String-" + x;
    }
    Elem() {}
};

class TestArr { 
    Elem[] arr;
    Ref[]  refs$Elem;
    String strs[];
    int    ints[];

    static final int nRecords = 100;

    static public void main(String[] args) throws Exception {
	int i, j, k;
	Class.forName("org.garret.consus.ConsusDriver");
	Connection connection = DriverManager.getConnection("jdbc:consus:TestArr.dbs");	
	if (connection == null) { 
	    System.err.println("No connection with database TestArr");
	    return;
	}
	ConsusStatement stmt = (ConsusStatement)connection.createStatement();
	
	for (j = 0; j < nRecords; j++) { 
	    TestArr r = new TestArr();
	    r.arr = new Elem[10];
	    for (i = 0; i < 10; i++) { 
		r.arr[i] = new Elem(1 << i);
	    }
	    r.refs$Elem = new Ref[j];
	    for (i = 0; i < j; i++) { 
		Elem elem = new Elem(i*3);
		r.refs$Elem[i] = stmt.insert(elem);
	    }
	    r.strs = new String[64];
	    for (i = 0; i < 64; i++) { 
		r.strs[i] = "String-" + j + "-" + (1L << i);
	    }
	    r.ints = new int[50];
	    for (i = 0; i < 50; i++) { 
		r.ints[i] = i;
	    }
	    stmt.insert(r);
	}
	System.out.print(".");
	System.out.flush();
	ConsusPreparedStatement pstmt = (ConsusPreparedStatement)connection.prepareStatement
	    ("select from TestArr where exists i : (refs$Elem[i].i = ?*3)");
	ConsusResultSet cursor;
	for (j = 1; j < nRecords; j++) { 
	    pstmt.setInt(1, j-1);
	    cursor = (ConsusResultSet)pstmt.executeQuery();
	    Assert.that(cursor.getSelectionSize() == nRecords - j);
	}
	pstmt.close();
	System.out.print(".");
	System.out.flush();
	pstmt = (ConsusPreparedStatement)connection.prepareStatement
	    ("select from TestArr where exists i : (arr[i].s = ?)");
	for (j = 0; j < 10; j++) { 
	    pstmt.setString(1, "String-" + (1 << j));
	    cursor = (ConsusResultSet)pstmt.executeQuery();
	    Assert.that(cursor.getSelectionSize() == nRecords);
	}
	pstmt.close();
	System.out.print(".");
	System.out.flush();
	pstmt = (ConsusPreparedStatement)connection.prepareStatement
	    ("select from TestArr where ? in strs");
	for (j = 0; j < nRecords; j++) { 
	    for (i = 0; i < 64; i++) { 
		pstmt.setString(1, "String-" + j + "-" + (1L << i));
		cursor = (ConsusResultSet)pstmt.executeQuery();
		Assert.that(cursor.getSelectionSize() == 1);
	    }
	}
	pstmt.close();
	System.out.print(".");
	System.out.flush();
	for (k = 1; k <= 10; k++) { 
	    cursor = (ConsusResultSet)stmt.executeQuery("select * from TestArr where " + 
		"length(strs) >= length(ints) AND exists i:(exists j:(strs[i][j] = 45)) for update");
	    for (j = 0; j < nRecords; j++) { 
		Assert.that(cursor.next());
		TestArr r = (TestArr)cursor.getSelfObject();
		for (i = 0; i < 10; i++) { 
		    Assert.that(r.arr[i].i == (1 << i) && r.arr[i].s.equals("String-" + (1 << i)));
		}
		Ref newRefs[] = new Ref[j + 10*k];
		for (i = 0; i < j + 10*(k-1); i++) { 
		    Elem elem = (Elem)stmt.get(r.refs$Elem[i]);
		    newRefs[i] = r.refs$Elem[i];
		    Assert.that(elem.i == i*3 && elem.s.equals("String-" + i*3));
		}
		for (; i < j + 10*k; i++) { 
		    newRefs[i] = stmt.insert(new Elem(i*3));
		}
		r.refs$Elem = newRefs;
		for (i = 0; i < 64; i++) { 
		    Assert.that(r.strs[i].equals("String-" + j + "-" + (1L << i)));
		}
		for (i = 0; i < 49+k; i++) { 
		    if (r.ints[i] != i) System.out.println("j = " + j + ": " + r.ints[i] + " != " + i);
		    Assert.that(r.ints[i] == i);
		}
		r.ints = new int[++i];
		while (--i >= 0) { 
		    r.ints[i] = i;
		}
		cursor.updateRow();
	    }
	    Assert.that(!cursor.next());
	    System.out.print(".");
	    System.out.flush();
	}
	j = stmt.executeUpdate("delete from TestArr where 1 in ints AND " + 
			       "exists i : (length(strs[i]) > 10 AND '-1024' in strs[i])");
	Assert.that(j == nRecords);
	connection.close();
	System.out.println("\nTest passed");
    }
}
								    














