/* Consus - Pure Java Object Relational Database Management System
 * Test of Binary Large OBjects
 * Created by Konstantin Knizhnik, Mar 2000.
 */

import java.sql.*;
import java.io.*;

public class TestBlob { 
    static BufferedReader in = new BufferedReader(new InputStreamReader(System.in)); 

    static String input(String prompt) {
	while (true) { 
	    try { 
		System.out.print(prompt);
		String answer = in.readLine().trim();
		if (answer.length() != 0) {
		    return answer;
		}
	    } catch (IOException x) {}
	}
    }    

    static public void main(String[] args) throws Exception {
	if (args.length != 1) { 
	    System.out.println("Usage: java TestBlob (database-file-name | @host:port)");
	}
	String URL, driver;
	if (args[0].charAt(0) == '@') { 
	    driver = "org.garret.consus.ClientDriver";
	    URL = "jdbc:consus-client:" + args[0].substring(1);	
	} else { 
	    driver = "org.garret.consus.ConsusDriver";	
	    URL = "jdbc:consus:" + args[0];	
	}
	Class.forName(driver);
	Connection connection = DriverManager.getConnection(URL);
	if (connection == null) { 
	    System.err.println("No connection with database");
	    return;
	}
	Statement stmt = connection.createStatement();
	try { 
	    stmt.execute("create table File (name VARCHAR primary key,data BLOB)");
	} catch (SQLException x) { /* table already exists */ }
	PreparedStatement selectByName = 
	    connection.prepareStatement("select * from File where name=?");
	PreparedStatement updateStmt = connection.prepareStatement("update File set name=? where name=?");
	PreparedStatement deleteStmt = connection.prepareStatement("delete from File where name=?");
	PreparedStatement selectAll = connection.prepareStatement("select * from File");
	PreparedStatement insert = connection.prepareStatement("insert into File (name,data) values (?,?)");

	byte buf[] = new byte[4096];

	while (true) { 
	    int i;
	    String cmd = input(">> ");
	    int len = cmd.length();
	    for (i = 0; i < len && cmd.charAt(i) != ' '; i++);
	    String s = cmd.substring(0, i);
	    if (s.equalsIgnoreCase("exit")) { 
		connection.close();		
		return;
	    } else if (s.equalsIgnoreCase("put")) { 
		s = cmd.substring(i+1).trim();
		FileInputStream in = new FileInputStream(s);
		insert.setString(1, s);
		insert.setBinaryStream(2, in, in.available());
		insert.execute();
	    } else if (s.equalsIgnoreCase("get")) { 
		s = cmd.substring(i+1).trim();
		selectByName.setString(1, s);
		ResultSet cursor = selectByName.executeQuery();
		if (cursor.next()) { 
		    FileOutputStream out = new FileOutputStream(s);
		    Blob blob = cursor.getBlob("data");
		    InputStream in = blob.getBinaryStream();
		    int rc;
		    while ((rc = in.read(buf)) > 0) { 
			out.write(buf, 0, rc);
		    }
		    in.close();
		    out.close();
		} else { 
		    System.err.println("No such file");
		}
		cursor.close();
	    } else if (s.equalsIgnoreCase("find")) { 
		byte[] pattern = cmd.substring(i+1).getBytes();
		ResultSet cursor = selectAll.executeQuery();
		while (cursor.next()) { 
		    Blob blob = cursor.getBlob("data");
		    long start = 1;
		    long pos;
		    while ((pos = blob.position(pattern, start)) > 0) { 
			System.out.println(cursor.getString("name") + ": " + pos);
			start = pos + 1;
		    }
		}
		cursor.close();
	    } else if (s.equalsIgnoreCase("dir")) { 
		ResultSet cursor = selectAll.executeQuery();
		while (cursor.next()) { 
		    System.out.println(cursor.getString("name"));
		}
		cursor.close();
	    } else if (s.equalsIgnoreCase("rename")) { 
		int j;
		while (++i < len && cmd.charAt(i) == ' ');
		for (j = i; j < len && cmd.charAt(j) != ' '; j++); 
		if (j == len) { 
		    System.err.println("Syntax: rename old-name new-name");
		} else { 
		    updateStmt.setString(1, cmd.substring(j+1).trim());
		    updateStmt.setString(2, cmd.substring(i, j));
		    if (updateStmt.executeUpdate() != 1) { 
			System.err.println("No such file");
		    }
		}
	    } else if (s.equalsIgnoreCase("del")) { 
		deleteStmt.setString(1, cmd.substring(i+1).trim());
		if (deleteStmt.executeUpdate() != 1) { 
		    System.err.println("No such file");
		}
	    } else { 
		System.out.println("Valid command: get, put, dir, find, rename, del, exit");
	    }
	    connection.commit();
	}
    }
}



