/* Consus - Pure Java Object Relational Database Management System
 * Test OO interface to Binary Large OBjects
 * Created by Konstantin Knizhnik, Mar 2000.
 */

import org.garret.consus.*;
import java.sql.*;
import java.io.*;

class FileEntry { 
    String name;
    Blob   data;
}

public class TestBlobOO { 
    static BufferedReader in = new BufferedReader(new InputStreamReader(System.in)); 

    static String input(String prompt) {
	while (true) { 
	    try { 
		System.out.print(prompt);
		String answer = in.readLine().trim();
		if (answer.length() != 0) {
		    return answer;
		}
	    } catch (IOException x) {}
	}
    }    

    static public void main(String[] args) throws Exception {
	Class.forName("org.garret.consus.ConsusDriver");
	ConsusConnection connection = 
	    (ConsusConnection)DriverManager.getConnection("jdbc:consus:TestBlobOO.dbs");	
	if (connection == null) { 
	    System.err.println("No connection with database TestBlobOO");
	    return;
	}
	ConsusStatement stmt = (ConsusStatement)connection.createStatement();
	PreparedStatement selectByName = 
	    connection.prepareStatement("select * from FileEntry where name = ?");
	PreparedStatement selectByNameForUpdate = 
	    connection.prepareStatement("select * from FileEntry where name = ? for update");
	PreparedStatement selectAll = connection.prepareStatement("select * from FileEntry");
	byte buf[] = new byte[4096];
	FileEntry f;

	while (true) { 
	    String cmd = input(">> ");
	    int len = cmd.length();
	    int i;
	    for (i = 0; i < len && cmd.charAt(i) != ' '; i++);
	    String s = cmd.substring(0, i);
	    if (s.equalsIgnoreCase("exit")) { 
		connection.close();		
		return;
	    } else if (s.equalsIgnoreCase("put")) { 
		f = new FileEntry();
		f.name = cmd.substring(i+1).trim();
		FileInputStream in = new FileInputStream(f.name);
		f.data = connection.createBlob(in.available(), in);
		stmt.insert(f);
	    } else if (s.equalsIgnoreCase("get")) { 
		s = cmd.substring(i+1).trim();
		selectByName.setString(1, s);
		ResultSet cursor = selectByName.executeQuery();
		if (cursor.next()) { 
		    FileOutputStream out = new FileOutputStream(s);
		    f = (FileEntry)cursor.getObject("this");
		    InputStream in = f.data.getBinaryStream();
		    int rc;
		    while ((rc = in.read(buf)) > 0) { 
			out.write(buf, 0, rc);
		    }
		    in.close();
		    out.close();
		} else { 
		    System.err.println("No such file");
		}
		cursor.close();
	    } else if (s.equalsIgnoreCase("find")) { 
		byte[] pattern = cmd.substring(i+1).getBytes();
		ResultSet cursor = selectAll.executeQuery();
		while (cursor.next()) { 
		    f = (FileEntry)cursor.getObject("this");
		    long start = 1;
		    long pos;
		    while ((pos = f.data.position(pattern, start)) > 0) { 
			System.out.println(f.name + ": " + pos);
			start = pos + 1;
		    }
		}
		cursor.close();
	    } else if (s.equalsIgnoreCase("dir")) { 
		ResultSet cursor = selectAll.executeQuery();
		while (cursor.next()) { 
		    System.out.println(((FileEntry)cursor.getObject("this")).name);
		}
		cursor.close();
	    } else if (s.equalsIgnoreCase("rename")) { 
		int j;
		while (++i < len && cmd.charAt(i) == ' ');
		for (j = i; j < len && cmd.charAt(j) != ' '; j++); 
		if (j == len) { 
		    System.err.println("Syntax: rename old-name new-name");
		} else { 
		    selectByNameForUpdate.setString(1, cmd.substring(i, j));
		    ResultSet cursor = selectByNameForUpdate.executeQuery();
		    if (cursor.next()) { 
			f = (FileEntry)((ConsusResultSet)cursor).getSelfObject();
			f.name = cmd.substring(j+1).trim();
			cursor.updateRow();		    
		    } else { 
			System.err.println("No such file");
		    }
		    cursor.close();
		}
	    } else if (s.equalsIgnoreCase("del")) { 
		selectByNameForUpdate.setString(1, cmd.substring(i+1).trim());
		ResultSet cursor = selectByNameForUpdate.executeQuery();
		if (cursor.next()) { 
		    cursor.deleteRow();
		} else { 
		    System.err.println("No such file");
		}
		cursor.close();
	    } else { 
		System.out.println("Valid command: get, put, dir, find, rename, del, exit");
	    }
	    connection.commit();
	}
    }
}


