/* Consus - Pure Java Object Relational Database Management System
 * Test of B-Tree operations and measure of Consus performance
 * Created by Konstantin Knizhnik, Feb 2000.
 */
import java.sql.*;
import org.garret.consus.*;

class Record { 
    public String rkey;
    public String value;

    public final static String indices[] = {"rkey"};
};

public class TestBtree { 
    final static int nRecords = 200000;
    final static int nSequentialSearches = 10;
    final static int maxRepeatCount = 10000;

    static public void main(String args[]) throws Exception { 
	Class.forName("org.garret.consus.ConsusDriver");
	ConsusConnection connection = 
	    (ConsusConnection)DriverManager.getConnection("jdbc:consus:testbtree.dbs");
	if (connection == null) { 
	    System.err.println("No connection with database TestPerf");
	    return;
	}
	connection.createObjectCache();
	ConsusStatement stmt = (ConsusStatement)connection.createStatement();
	ResultSet cursor;
	long start = System.currentTimeMillis();
	long key = 1999;
	int i, j, n;
	Record arr[] = new Record[nRecords / 1000];
	for (i = 0; i < nRecords; i++) { 
	    Record rec = new Record();
	    key = (3141592621L*key + 2718281829L) % 1000000007L;
	    rec.rkey = rec.value = Long.toString(key);
	    stmt.insert(rec);
	    if (i % 1000 == 0) {
		arr[i/1000] = rec;
		j = (int)(key % maxRepeatCount);
		i += j;
		while (--j >= 0) { 
		    stmt.insert(rec);
		}
	    }
	}
	n = i + maxRepeatCount;
	Record rec = new Record();
	rec.rkey = rec.value = "";
	for (j = maxRepeatCount; --j >= 0;) { 
	    stmt.insert(rec);
	}
	System.out.println("Elapsed time for inserting " + n + " records: " 
			   + (System.currentTimeMillis() - start) + " milliseconds");
        start = System.currentTimeMillis();
	connection.commit();
	System.out.println("Commit time: " + (System.currentTimeMillis() - start) + " milliseconds");
	
	PreparedStatement pstmt = connection.prepareStatement("select * from Record where rkey = ?");
	start = System.currentTimeMillis();
	key = 1999;
	for (i = 0; i < nRecords; i++) { 
	    key = (3141592621L*key + 2718281829L) % 1000000007L;
	    pstmt.setLong(1, key);
	    cursor = pstmt.executeQuery();	    
	    cursor.last();
	    if (i % 1000 == 0) {
		j = (int)(key % maxRepeatCount);
		Assert.that(cursor.getRow() == j + 1);
		Assert.that(((ConsusResultSet)cursor).getSelfObject() == arr[i/1000]);
		i += j;
	    } else { 
		Assert.that(cursor.getRow() == 1);
	    }
	}
	pstmt.setString(1, "");
	cursor = pstmt.executeQuery();	    
	cursor.last();
	Assert.that(cursor.getRow() == maxRepeatCount);
	
	System.out.println("Elapsed time for " + n + " index searches: " + 
			   (System.currentTimeMillis() - start) + " milliseconds"); 
	pstmt.close();
	pstmt = connection.prepareStatement("select * from Record where rkey = value");
	start = System.currentTimeMillis();
	for (i = 0; i < nSequentialSearches; i++) { 
	    cursor = pstmt.executeQuery();
	    cursor.last();
	    Assert.that(cursor.getRow() == n);
	}
	System.out.println("Elapsed time for " + nSequentialSearches + 
			   " sequential search through " + n + " records: " +  
			   (System.currentTimeMillis() - start) + " milliseconds"); 
	pstmt.close();
	start = System.currentTimeMillis();
 	cursor = stmt.executeQuery("select * from Record where rkey = value order by value for update");
	cursor.last();
	Assert.that(cursor.getRow() == n);
	System.out.println("Elapsed time for sorting " + n + " records: " +  
			   (System.currentTimeMillis() - start) + " milliseconds"); 
	start = System.currentTimeMillis();
	i = 0;
	do { 
	    cursor.deleteRow();
	    i += 1;
        } while (cursor.previous()); 
	Assert.that(i == n);
	System.out.println("Elapsed time for deleting " + n + " records: " +  
			   (System.currentTimeMillis() - start) + " milliseconds"); 
	connection.close();
    }
}
	













