/* Consus - Pure Java Object Relational Database Management System
 * Example of table joins
 * Created by Konstantin Knizhnik, Feb 2000.
 */
import java.io.*;
import java.sql.*;
import org.garret.consus.*;

class TcDetail { 
    String name;
    double weight;
    
    final static String constrains = "primary key(name)";
}

class TcSupplier { 
    String company;
    String address;

    final static String constraints = "unique(company)";
}

class TcOrder1 { 
    Ref  $TcDetail;
    Ref  $TcSupplier;
    long price;
    long quantity;
    Date delivery;

    final static String constraints = "foreign key($TcDetail) references TcDetail(oid) on delete cascade,"
                                    + "foreign key($TcSupplier) references TcSupplier(oid) on delete cascade";
}

class TcOrder2 { 
    String detailName;
    String company;
    long   price;
    long   quantity;
    Date   delivery;

    final static String constraints = "foreign key(detailName) references TcDetail(name) on delete cascade,"
                                    + "foreign key(company) references TcSupplier on delete cascade";
}

public class TestConstraints { 
    static BufferedReader in = new BufferedReader(new InputStreamReader(System.in)); 

    static String input(String prompt) {
	while (true) { 
	    try { 
		System.out.print(prompt);
		String answer = in.readLine().trim();
		if (answer.length() != 0) {
		    return answer;
		}
	    } catch (IOException x) {}
	}
    }
    static long inputLong(String prompt) { 
	while (true) { 
	    try { 
		return Long.parseLong(input(prompt));	    
	    } catch(NumberFormatException x) {
		System.out.println("Invalid number");
	    }
	}
    }
    static double inputDouble(String prompt) { 
	while (true) { 
	    try { 
		return Double.parseDouble(input(prompt));	    
	    } catch(NumberFormatException x) {
		System.out.println("Invalid number");
	    }
	}
    }	
    static Date inputDate(String prompt) { 
	while (true) { 
	    try { 
		return Date.valueOf(input(prompt));	    
	    } catch(Exception x) {
		System.out.println("Invalid date");
	    }
	}
    }	

    static public void main(String args[]) throws Exception { 
        int rc;
	Class.forName("org.garret.consus.ConsusDriver");
	Connection connection = DriverManager.getConnection("jdbc:consus:testconst.dbs");	
	if (connection == null) { 
	    System.err.println("No connection with database TEST");
	    return;
	}
	ConsusStatement stmt = (ConsusStatement)connection.createStatement();
	PreparedStatement stmt1 
	    = connection.prepareStatement("select from TcSupplier where company = ?"); 
	PreparedStatement stmt2 
	    = connection.prepareStatement("select from TcDetail where name = ?"); 
	PreparedStatement stmt3 
	    = connection.prepareStatement("select from TcOrder1 where " 
					  + "$TcDetail.name like ? and "
					  + "$TcSupplier.company like ? " 
					  + "and $TcSupplier.address like ? order by price");
	PreparedStatement stmt4
	    = connection.prepareStatement("select from TcOrder1 where " 
					  + "$TcSupplier.company like ? " 
					  + "and $TcSupplier.address like ? " 
					  + "and $TcDetail.name like ? order by delivery");
	PreparedStatement stmt5 
	    = connection.prepareStatement("select * from TcOrder2,TcSupplier where " 
					  + "detailName like ?"
	                                  + "and TcOrder2.company like ?"
					  + "and TcOrder2.company=TcSupplier.company " 
					  + "and TcSupplier.address like ? order by price");
	PreparedStatement stmt6
	    = connection.prepareStatement("select * from TcSupplier,TcOrder2,TcDetail where " 
					  + "TcSupplier.company like ? " 
					  + "and TcSupplier.address like ? " 
					  + "and TcSupplier.company=TcOrder2.company " 
					  + "and TcOrder2.detailName like ? " 
					  + "and TcOrder2.detailName = TcDetail.name " 
					  + "order by delivery");
	PreparedStatement stmt7 
 	    = connection.prepareStatement("delete from TcSupplier where company = ?");
	PreparedStatement stmt8 
	    = connection.prepareStatement("select from TcDetail where name like ? for update"); 

	System.out.println("Supplier-Order-Detail database");
	TcSupplier  supplier = new TcSupplier();
	TcDetail    detail = new TcDetail();
	TcOrder1    order1 = new TcOrder1();
	TcOrder2    order2 = new TcOrder2();
	ConsusResultSet cursor;

	while (true) { 
	    System.out.println("Menu:\n" + 
			       "1.  Add supplier\n" + 
			       "2.  Add detail\n" + 
			       "3.  New order 1\n" + 
			       "4.  New order 2\n" + 
			       "5.  Remove supplier\n" + 
			       "6.  Remove detail\n" + 
			       "7.  Find suppliers of the detail (order-1)\n" + 
			       "8.  Find details shipped by supplier (order-1)\n" + 
			       "9.  Find suppliers of the detail (order-2)\n" + 
			       "10. Find details shipped by supplier (order-2)\n" + 
			       "11. Exit\n");
	    switch((int)inputLong("> ")) { 
	      case 1:
		supplier.company = input("Suppiler company: ");
		supplier.address = input("Suppiler address: ");
		stmt.insert(supplier);
		System.out.println("Supplier added");
		break;
	      case 2:
		detail.name = input("Detail name: ");
		detail.weight = inputDouble("Detail weight: ");
		stmt.insert(detail);
		System.out.println("Detail added");
		break;
	      case 3: 
		stmt1.setString(1, input("Company name: "));
		cursor = (ConsusResultSet)stmt1.executeQuery();
		if (!cursor.next()) { 
		    System.out.println("No such supplier");
		    break;
		}
		order1.$TcSupplier = cursor.getSelfRef();

		stmt2.setString(1, input("Detail name: "));
		cursor = (ConsusResultSet)stmt2.executeQuery();
		if (!cursor.next()) { 
		    System.out.println("No such detail");
		    break;
		}
		order1.$TcDetail = cursor.getSelfRef();
		order1.price = inputLong("Price: ");
		order1.quantity = inputLong("Quantity: ");
		order1.delivery = inputDate("Delivery: ");
		stmt.insert(order1);
		System.out.println("Order 1 added");
		break;		
	      case 4: 
		order2.company = input("Company name: ");
                order2.detailName = input("Detail name: ");
		order2.price = inputLong("Price: ");
		order2.quantity = inputLong("Quantity: ");
		order2.delivery = inputDate("Delivery: ");
		stmt.insert(order2);
		System.out.println("Order 2 added");
		break;		
              case 5:
	        stmt7.setString(1, input("Company name: "));
                stmt7.executeUpdate();
                break;
              case 6:
	        stmt8.setString(1, input("Detail name: "));
                cursor = (ConsusResultSet)stmt8.executeQuery();
                while (cursor.next()) { 
                    cursor.deleteRow();
                }
                break;
	      case 7:
		stmt3.setString(1, input("Detail like: "));
		stmt3.setString(2, input("Company like: "));
		stmt3.setString(3, input("Address like: "));
		cursor = (ConsusResultSet)stmt3.executeQuery();
		System.out.println("Detail Company Address Price Quantity Delivery");
		while (cursor.next()) { 
		    order1 = (TcOrder1)cursor.getSelfObject();
		    supplier = (TcSupplier)stmt.get(order1.$TcSupplier);
		    detail = (TcDetail)stmt.get(order1.$TcDetail);
		    System.out.println(detail.name + "\t" + supplier.company + "\t" + 
				       supplier.address + "\t" + order1.price + "\t" + 
				       order1.quantity + "\t" + order1.delivery);
		}
		System.out.println("--------------------------------------------");
		break;
	      case 8:
		stmt4.setString(1, input("Company like: "));
		stmt4.setString(2, input("Address like: "));
		stmt4.setString(3, input("Detail like: "));
		cursor = (ConsusResultSet)stmt4.executeQuery();
		System.out.println("Company Address Detail Price Quantity Delivery");
		while (cursor.next()) { 
		    order1 = (TcOrder1)cursor.getSelfObject();
		    supplier = (TcSupplier)stmt.get(order1.$TcSupplier);
		    detail = (TcDetail)stmt.get(order1.$TcDetail);
		    System.out.println(supplier.company + "\t" + supplier.address + "\t" +
				       detail.name + "\t" + order1.price + "\t" + 
				       order1.quantity + "\t" + order1.delivery);
		}
		System.out.println("--------------------------------------------");
		break;
	      case 9:
		stmt5.setString(1, input("Detail like: "));
		stmt5.setString(2, input("Company like: "));
		stmt5.setString(3, input("Address like: "));
		cursor = (ConsusResultSet)stmt5.executeQuery();
		System.out.println("Detail Company Address Price Quantity Delivery");
		while (cursor.next()) { 
		    System.out.println(cursor.getString("detailName") + "\t"  +
		                       cursor.getString("TcSupplier.company") + "\t" + 
				       cursor.getString("address") + "\t" + 
		                       cursor.getInt("price") + "\t" + 
		                       cursor.getLong("quantity") + "\t" + 
		                       cursor.getDate("delivery"));
		}
		System.out.println("--------------------------------------------");
		break;
	      case 10:
		stmt6.setString(1, input("Company like: "));
		stmt6.setString(2, input("Address like: "));
		stmt6.setString(3, input("Detail like: "));
		cursor = (ConsusResultSet)stmt6.executeQuery();
		System.out.println("Company Address Detail Price Quantity Delivery");
		while (cursor.next()) { 
		    System.out.println(cursor.getString("TcOrder2.company") + "\t" + 
		                       cursor.getString("address") + "\t" +
		                       cursor.getString("TcDetail.name") + "\t" + 
		                       cursor.getLong("price") + "\t" + 
				       cursor.getLong("quantity") + "\t" +
		                       cursor.getDate("delivery"));
		}
		System.out.println("--------------------------------------------");
		break;
	      case 11:
		connection.close();
		return;
	      default:
		System.out.println("Please specify 1..11");
	    }
	    connection.commit();
	}
    }
}




