/* Consus - Pure Java Object Relational Database Management System
 * Test of extracting information about cross references
 * Created by Konstantin Knizhnik, Jan 2001.
 */
import java.sql.*;

public class TestCrossRef { 
    static void printDependencies(ResultSet cursor) throws java.sql.SQLException  {
	while (cursor.next()) { 
	    System.out.println("PK=" + cursor.getString("PKTABLE_NAME") 
			       + "(" + cursor.getString("PKCOLUMN_NAME") + "), FK=" 
			       + cursor.getString("FKTABLE_NAME") 
			       + "(" + cursor.getString("FKCOLUMN_NAME") + ")");
	}
	cursor.close();
    }

    static void dumpCrossRefInfo(DatabaseMetaData md) throws java.sql.SQLException { 
	System.out.println("Imported keys for table SP");
	printDependencies(md.getImportedKeys("", "", "SP"));	
	System.out.println("Exported keys for table P");
	printDependencies(md.getExportedKeys("", "", "P"));
	System.out.println("Cross references between tables SP and S");
	printDependencies(md.getCrossReference("", "", "S", "", "", "SP"));
    }	
	
    static public void main(String args[]) throws Exception { 
	int rc;
	Class.forName("org.garret.consus.ConsusDriver");
	Connection connection = DriverManager.getConnection("jdbc:consus:testcross.dbs");	
	if (connection == null) { 
	    System.err.println("No connection with database TEST");
	    return;
	}
	Statement stmt = connection.createStatement();
	stmt.execute("create table S (id integer primary key, address string unique)");
	stmt.execute("create table P (id integer, name string, primary key (id), unique (name))");
	stmt.execute("create table SP (s integer references S(id), p integer, quantity integer, foreign key (p) references P(id), t integer)");
;
	DatabaseMetaData md = connection.getMetaData();
	dumpCrossRefInfo(md);
	stmt.execute("alter table SP ADD foreign key (t) references S(id)");
	dumpCrossRefInfo(md);
	connection.close();
	System.out.println("Test passed!");
    }
}
	

