/* Consus - Pure Java Object Relational Database Management System
 * Test of Consus garbage collector
 * Created by Konstantin Knizhnik, May 2000.
 */

import org.garret.consus.*;
import java.sql.*;

class Vertex extends Persistent { 
    Vertex[] incoming;     
    Vertex[] outgoing;
}

class Graph extends Persistent { 
    String name;
    Vertex root;

    public static String indices[] = {"name"};
}

public class TestGC { 
    static final int N_GRAPHS = 128;
    static final int N_VERTICES = 10000000;
    static final int MAX_HEIGHT = 10;
    
    static public void main(String[] args) throws Exception { 
	int i;
	java.util.Random rand = new java.util.Random(2000);	
	
	Class.forName("org.garret.consus.ConsusDriver");
	ConsusConnection db = 
	    (ConsusConnection)DriverManager.getConnection("jdbc:consus:TestGC.dbs");	
	if (db == null) { 
	    System.err.println("No connection with database TestGC");
	    return;
	}
	for (i = 0; i < N_GRAPHS; i++) { 
	    Graph graph = new Graph();
	    graph.name = "Graph-" + i;
	    graph.store(db);
	}
	Statement stmt = db.createStatement();
	ConsusResultSet graphs = null;
	PreparedStatement pstmt = 
	    db.prepareStatement("select distinct from Vertex start from ? following by outgoing");
	String rootTables[] = {"Graph"};
	db.setGCRoots(rootTables);
	db.setGCThreshold(1024*1024);
	db.enableTracing(true);
	for (i = 0; i < N_VERTICES; i++) { 
	    if (graphs == null) { 
		graphs = (ConsusResultSet)stmt.executeQuery("select * from Graph for update");
	    }
	    int nGraphs = graphs.getSelectionSize();		
	    Vertex vertex = new Vertex();
	    vertex.incoming = new Vertex[0];
	    graphs.absolute(1 + i % nGraphs);
	    pstmt.setRef(1, graphs.getRef(2));
	    ConsusResultSet vertices = (ConsusResultSet)pstmt.executeQuery(); 
	    int n = vertices.getSelectionSize();
	    if (n == 0 || i / nGraphs % MAX_HEIGHT == 0) { 
		vertex.outgoing = new Vertex[1];
		vertex.outgoing[0] = vertex;
	    } else { 
		vertex.outgoing = new Vertex[rand.nextInt(10)];
		for (int j = vertex.outgoing.length; --j >= 0;) { 
		    vertices.absolute(1 + rand.nextInt(n));
		    Vertex v = (Vertex)vertices.getSelfObject();
		    vertex.outgoing[j] = v;
		    v.incoming = (Vertex[])AnyArray.append(v.incoming, vertex);
		    v.store();
		}
	    }
	    vertex.store(db);
	    graphs.updateObject(2, vertex.getRef());
	    graphs.updateRow();
	    if (i % 100 == 0) { 
		db.commitSubtransaction();
		graphs = null;
	    }
	}
	db.close();
    }
}


