/* Consus - Pure Java Object Relational Database Management System
 * Test of indexed merge 
 * Created by Konstantin Knizhnik, May 2000.
 */
import java.sql.*;
import org.garret.consus.*;

class Pair { 
    public int    key1;
    public String key2;

    // public final static String indices[] = {"key1"};
    public final static String indices[] = {"key1", "key2"};
};

public class TestIndexedMerge { 
    final static int nRecords = 100000;
    final static int maxKeys = 1024;

    static public void main(String args[]) throws Exception { 
	Class.forName("org.garret.consus.ConsusDriver");
	ConsusConnection connection =
	    (ConsusConnection)DriverManager.getConnection("jdbc:consus:testimrg.dbs");	
	if (connection == null) { 
	    System.err.println("No connection with database");
	    return;
	}
	connection.createObjectCache();
	ConsusStatement stmt = (ConsusStatement)connection.createStatement();
	ResultSet cursor;
	long start = System.currentTimeMillis();
	long key = 1999;
	int i, j, n;
	int arr[] = new int[maxKeys*maxKeys];
	for (i = 0; i < nRecords; i++) { 
	    Pair rec = new Pair();
	    key = (3141592621L*key + 2718281829L) % 1000000007L;
	    rec.key1 = (int)(key % maxKeys);
	    key = (3141592621L*key + 2718281829L) % 1000000007L;
	    rec.key2 = Integer.toString((int)(key % maxKeys));
	    arr[rec.key1*maxKeys + (int)(key % maxKeys)] += 1;
	    stmt.insert(rec);
	}
	System.out.println("Elapsed time for inserting " + nRecords + " records: " 
			   + (System.currentTimeMillis() - start) + " milliseconds");
        start = System.currentTimeMillis();
	connection.commit();
	System.out.println("Commit time: " + (System.currentTimeMillis() - start) + " milliseconds");
	
	PreparedStatement pstmt = connection.prepareStatement("select * from Pair where key1 = ? and key2 = ?");
	start = System.currentTimeMillis();
	key = 1999;
	for (i = 0; i < nRecords; i++) { 
	    key = (3141592621L*key + 2718281829L) % 1000000007L;
	    int key1 = (int)(key  % maxKeys);
	    pstmt.setInt(1, key1);
	    key = (3141592621L*key + 2718281829L) % 1000000007L;
	    int key2 = (int)(key  % maxKeys);
	    pstmt.setInt(2, key2);
	    cursor = pstmt.executeQuery();	
	    Assert.that(((ConsusResultSet)cursor).getSelectionSize() == arr[key1*maxKeys+key2]);
	}
	System.out.println("Elapsed time for " + nRecords + " index object searches: " + 
			   (System.currentTimeMillis() - start) + " milliseconds"); 
	pstmt.close();
	pstmt = connection.prepareStatement("select key1,key2 from Pair where key1 = ? and key2 = ?");
	start = System.currentTimeMillis();
	key = 1999;
	for (i = 0; i < nRecords; i++) { 
	    key = (3141592621L*key + 2718281829L) % 1000000007L;
	    int key1 = (int)(key  % maxKeys);
	    pstmt.setInt(1, key1);
	    key = (3141592621L*key + 2718281829L) % 1000000007L;
	    int key2 = (int)(key  % maxKeys);
	    pstmt.setInt(2, key2);
	    cursor = pstmt.executeQuery();	
	    Assert.that(((ConsusResultSet)cursor).getSelectionSize() == arr[key1*maxKeys+key2]);
	}
	System.out.println("Elapsed time for " + nRecords + " index tupple searches: " + 
			   (System.currentTimeMillis() - start) + " milliseconds"); 
	pstmt.close();
	connection.createStatement().executeUpdate("drop table Pair");
	connection.close();
    }
}
	













