/* Consus - Pure Java Object Relational Database Management System
 * Example of table joins
 * Created by Konstantin Knizhnik, Feb 2000.
 */
import java.io.*;
import java.sql.*;
import org.garret.consus.*;

class Detail { 
    String name;
    double weight;
    
    final static String indices[] = {"name"};
}

class Supplier { 
    String company;
    String address;

    final static String indices[] = {"company"};
}

class Shipment { 
    Ref  $Detail;
    Ref  $Supplier;
    long price;
    long quantity;
    Date delivery;

    final static String indices[] = {"$Detail", "$Supplier"};
}

public class TestJoin { 
    static BufferedReader in = new BufferedReader(new InputStreamReader(System.in)); 

    static String input(String prompt) {
	while (true) { 
	    try { 
		System.out.print(prompt);
		String answer = in.readLine().trim();
		if (answer.length() != 0) {
		    return answer;
		}
	    } catch (IOException x) {}
	}
    }
    static long inputLong(String prompt) { 
	while (true) { 
	    try { 
		return Long.parseLong(input(prompt));	    
	    } catch(NumberFormatException x) {
		System.out.println("Invalid number");
	    }
	}
    }
    static double inputDouble(String prompt) { 
	while (true) { 
	    try { 
		return Double.parseDouble(input(prompt));	    
	    } catch(NumberFormatException x) {
		System.out.println("Invalid number");
	    }
	}
    }	
    static Date inputDate(String prompt) { 
	while (true) { 
	    try { 
		return Date.valueOf(input(prompt));	    
	    } catch(Exception x) {
		System.out.println("Invalid date");
	    }
	}
    }	

    static public void main(String args[]) throws Exception { 
	int rc;
	Class.forName("org.garret.consus.ConsusDriver");
	Connection connection = DriverManager.getConnection("jdbc:consus:testjoin.dbs");	
	if (connection == null) { 
	    System.err.println("No connection with database TEST");
	    return;
	}
	ConsusStatement stmt = (ConsusStatement)connection.createStatement();
	PreparedStatement stmt1 
	    = connection.prepareStatement("select from Supplier where company = ?"); 
	PreparedStatement stmt2 
	    = connection.prepareStatement("select from Detail where name = ?"); 
	PreparedStatement stmt3 
	    = connection.prepareStatement("select from Shipment where " 
					  + "$Detail.name like ? and "
					  + "$Supplier.company like ? " 
					  + "and $Supplier.address like ? order by price");
	PreparedStatement stmt4
	    = connection.prepareStatement("select from Shipment where " 
					  + "$Supplier.company like ? " 
					  + "and $Supplier.address like ? " 
					  + "and $Detail.name like ? order by delivery");
   
	System.out.println("Supplier-Shipment-Detail database");
	Supplier  supplier = new Supplier();
	Shipment  shipment = new Shipment();
	Detail    detail   = new Detail();
	ConsusResultSet cursor;

	while (true) { 
	    System.out.println("Menu:\n" + 
			       "1. Add supplier\n" + 
			       "2. Add detail\n" + 
			       "3. New shipment\n" + 
			       "4. Find suppliers of the detail\n" + 
			       "5. Find details shipped by supplier\n" + 
			       "6. Exit\n");
	    switch((int)inputLong("> ")) { 
	      case 1:
		supplier.company = input("Suppiler company: ");
		supplier.address = input("Suppiler address: ");
		stmt.insert(supplier);
		System.out.println("Supplier added");
		break;
	      case 2:
		detail.name = input("Detail name: ");
		detail.weight = inputDouble("Detail weight: ");
		stmt.insert(detail);
		System.out.println("Detail added");
		break;
	      case 3: 
		stmt1.setString(1, input("Company name: "));
		cursor = (ConsusResultSet)stmt1.executeQuery();
		if (!cursor.next()) { 
		    System.out.println("No such supplier");
		    break;
		}
		shipment.$Supplier = cursor.getSelfRef();

		stmt2.setString(1, input("Detail name: "));
		cursor = (ConsusResultSet)stmt2.executeQuery();
		if (!cursor.next()) { 
		    System.out.println("No such detail");
		    break;
		}
		shipment.$Detail = cursor.getSelfRef();
		shipment.price = inputLong("Price: ");
		shipment.quantity = inputLong("Quantity: ");
		shipment.delivery = inputDate("Delivery: ");
		stmt.insert(shipment);
		System.out.println("Shipment added");
		break;		
	      case 4:
		stmt3.setString(1, input("Detail like: "));
		stmt3.setString(2, input("Company like: "));
		stmt3.setString(3, input("Address like: "));
		cursor = (ConsusResultSet)stmt3.executeQuery();
		System.out.println("Detail Company Address Price Quantity Delivery");
		while (cursor.next()) { 
		    shipment = (Shipment)cursor.getSelfObject();
		    supplier = (Supplier)stmt.get(shipment.$Supplier);
		    detail = (Detail)stmt.get(shipment.$Detail);
		    System.out.println(detail.name + "\t" + supplier.company + "\t" + 
				       supplier.address + "\t" + shipment.price + "\t" + 
				       shipment.quantity + "\t" + shipment.delivery);
		}
		System.out.println("--------------------------------------------");
		break;
	      case 5:
		stmt4.setString(1, input("Company like: "));
		stmt4.setString(2, input("Address like: "));
		stmt4.setString(3, input("Detail like: "));
		cursor = (ConsusResultSet)stmt4.executeQuery();
		System.out.println("Company Address Detail Price Quantity Delivery");
		while (cursor.next()) { 
		    shipment = (Shipment)cursor.getSelfObject();
		    supplier = (Supplier)stmt.get(shipment.$Supplier);
		    detail = (Detail)stmt.get(shipment.$Detail);
		    System.out.println(supplier.company + "\t" + supplier.address + "\t" +
				       detail.name + "\t" + shipment.price + "\t" + 
				       shipment.quantity + "\t" + shipment.delivery);
		}
		System.out.println("--------------------------------------------");
		break;
	      case 6:
		connection.close();
		return;
	      default:
		System.out.println("Please specify 1..6");
	    }
	    connection.commit();
	}
    }
}



