/* Consus - Pure Java Object Relational Database Management System
 * Test of polymorphic Consus queries
 * Created by Konstantin Knizhnik, Feb 2000.
 */
import java.sql.*; 
import org.garret.consus.*;

class BaseClass { 
    public String skey;
    
    public String toString() { 
	return "key=" + skey;
    }

    public static final String indices[] = {"skey"};
};

class DerivedClass extends BaseClass { 
    public String value;
    public String toString() { 
	return "key=" + skey + ", value=" + value;
    }
};

public class TestOO { 
    public static void main(String args[]) throws Exception { 
	Class.forName("org.garret.consus.ConsusDriver");
	Connection connection = DriverManager.getConnection("jdbc:consus:testOO.dbs");	
	if (connection == null) { 
	    System.err.println("No connection with database TestOO");
	    return;
	}
	ConsusStatement stmt = (ConsusStatement)connection.createStatement();
	BaseClass b = new BaseClass();
	b.skey = "key-1";
	stmt.insert(b);
	DerivedClass d = new DerivedClass();
	d.skey = "key-2";
	d.value = "value-1";
	stmt.insert(d);
	b.skey = "key-3";
	stmt.insert(b);
	d.skey = "key-4";
	d.value = "value-2";
	Ref ref = stmt.insert(d);
	b = (BaseClass)stmt.get(ref);
	System.out.println(b);
	System.out.println("Selecting from BaseClass...");
	ResultSet cursor = stmt.executeQuery("select * from BaseClass where skey like 'key-_'");
	while (cursor.next()) { 
	    System.out.println(cursor.getObject("this"));
	}
	System.out.println("Selecting from DerivedClass...");
	cursor = stmt.executeQuery("select * from DerivedClass where skey between 'key-0' and 'key-9'");
	while (cursor.next()) { 
	    System.out.println(cursor.getObject("this"));
	}
	System.out.println("Updating DerivedClass...");
	cursor = stmt.executeQuery("select * from DerivedClass where " + 
				   "value in ('value-1', 'value-2', 'value-3') for update");
	while (cursor.next()) {
	    d = (DerivedClass)cursor.getObject("this");
	    System.out.println("Update object " + ((BaseClass)cursor.getObject("this")).skey);
	    d.value = "new-" + d.value;
	    cursor.updateRow();
	}
	stmt.execute("delete from DerivedClass where value='new-value-1'");
	System.out.println("Delete " + stmt.getUpdateCount() + " record from DerivedClass");
	cursor = stmt.executeQuery("select * from BaseClass");
	while (cursor.next()) { 
	    System.out.println(cursor.getObject(0));
	}
	cursor = stmt.executeQuery("select * from DerivedClass for update");
	while (cursor.next()) {
	    cursor.deleteRow();
	}
	System.out.println("Delete " + stmt.executeUpdate("delete from BaseClass") 
			   + " records from BaseClass");
	stmt.execute("drop table DerivedClass");
	connection.close();
    }
}





