/* Consus - Pure Java Object Relational Database Management System
 * Test of B-Tree operations and measure of Consus performance
 * Created by Konstantin Knizhnik, Feb 2000.
 */
import java.sql.*;

public class TestPerf { 
    final static int nRecords = 100000;
    final static int nSequentialSearches = 10;

    static public void main(String args[]) throws Exception { 
	// Parameter for Consus
	String DRIVER =   "org.garret.consus.ConsusDriver";
	String URL    =   "jdbc:consus:";
	String FILE   =   "TestPerf.dbs";
	
	// Parameter for Cloudscape
	//String DRIVER =   "COM.cloudscape.core.JDBCDriver";
	//String URL    =   "jdbc:cloudscape:";
	//String FILE   =   "CloudscapeDB;create=true";

	// Parameter for PointBase
	//String DRIVER =   "com.pointbase.jdbc.jdbcDriver";
	//String URL    =   "jdbc:pointbase:";
	//String FILE   =   "testperf.pbs,new";

	// Parameters for JDatastore
	//String DRIVER =   "com.borland.datastore.jdbc.DataStoreDriver";
	//String URL    =   "jdbc:borland:dslocal:";
	//String FILE   =   "TEMPORARY.jds";

	// Parameters for InstantDB
	//String DRIVER =   "jdbc.idbDriver";
	//String URL    =   "jdbc:idb:";
	//String FILE   =   "TestPerf.prp";

	DRIVER = System.getProperty("driver", DRIVER);
	URL    = System.getProperty("url", URL);
	FILE   = System.getProperty("file", FILE);

	Class.forName(DRIVER);
	Connection connection = DriverManager.getConnection(URL + FILE, System.getProperties());
	// PoinBase requires additional two arguments 
	//Connection connection = DriverManager.getConnection(URL + FILE, "public", "public");
	if (connection == null) { 
	    System.err.println("No connection with database " + FILE);
	    return;
	}
	connection.setAutoCommit(false);
	Statement stmt = connection.createStatement();
	ResultSet cursor;
	long start = System.currentTimeMillis();
	long key = 1999;
	int i, n;
	stmt.execute("create table Record(rkey varchar(16), rvalue varchar(16))");
	PreparedStatement pstmt = connection.prepareStatement("insert into Record values(?,?)");
	for (i = 0; i < nRecords; i++) { 
	    key = (3141592621L*key + 2718281829L) % 1000000007L;
	    String str = Long.toString(key);
	    pstmt.setString(1, str);
	    pstmt.setString(2, str);
	    pstmt.execute();
	}
	stmt.execute("create index rkeyIndex on Record(rkey)");
	System.out.println("Elapsed time for inserting " + nRecords + " records: " 
			   + (System.currentTimeMillis() - start) + " milliseconds");
        start = System.currentTimeMillis();
	connection.commit();
	pstmt.close();
	System.out.println("Commit time: " + (System.currentTimeMillis() - start) + " milliseconds");
	
	pstmt = connection.prepareStatement("select * from Record where rkey = ?");
	start = System.currentTimeMillis();
	key = 1999;
	for (i = 0; i < nRecords; i++) { 
	    key = (3141592621L*key + 2718281829L) % 1000000007L;
	    pstmt.setString(1, Long.toString(key));
	    cursor = pstmt.executeQuery();	    
	    if (!(cursor.next() && !cursor.next())) { 
		System.err.println("Not single record fetched");
	    }
	}
	System.out.println("Elapsed time for " + nRecords + " index searches: " + 
			   (System.currentTimeMillis() - start) + " milliseconds"); 
	pstmt.close();
	pstmt = connection.prepareStatement("select * from Record where rkey = rvalue");
	start = System.currentTimeMillis();
	for (i = 0; i < nSequentialSearches; i++) { 
	    cursor = pstmt.executeQuery();
	    for (n = 0; cursor.next(); n++);
	    if (n != nRecords) { 
		System.err.println("Fetch " + n + " records intead of " + nRecords);
	    }
	}
	System.out.println("Elapsed time for " + nSequentialSearches + 
			   " sequential search through " + nRecords + " records: " +  
			   (System.currentTimeMillis() - start) + " milliseconds"); 
	pstmt.close();
	start = System.currentTimeMillis();
	cursor = stmt.executeQuery("select * from Record where rkey = rvalue order by rvalue");
	System.out.println("Elapsed time for sorting " + nRecords + " records: " +  
			   (System.currentTimeMillis() - start) + " milliseconds"); 
	start = System.currentTimeMillis();
	pstmt = connection.prepareStatement("delete from Record where rkey = ?");
	for (n = 0; cursor.next(); n++) { 
	    pstmt.setString(1, cursor.getString(1));
	    pstmt.execute();
	}
	if (n != nRecords) { 
	    System.err.println("Fetch " + n + " records intead of " + nRecords);
	}
	System.out.println("Elapsed time for deleting " + nRecords + " records: " +  
			   (System.currentTimeMillis() - start) + " milliseconds"); 
	stmt.execute("drop table Record");
	connection.commit();
	connection.close();
    }
}
	













