/* Consus - Pure Java Object Relational Database Management System
 * Test of locking and transaction mechanisms, interthread sycnhronization.
 * Created by Konstantin Knizhnik, Feb 2000.
 */
import java.sql.*;
import org.garret.consus.*;

class RecordInt { 
    public long rkey;
    public long value;
    public int  id;
};

class ConsusThread extends Thread { 
    ConsusConnection connection;
    int              id;
    boolean          cooperative;

    ConsusThread(ConsusConnection conxn, int threadId, boolean coopMode) { 
	connection = conxn;
	id = threadId;
	cooperative = coopMode;
	start();
    }

    public void run() { 
	try { 
	    ResultSet cursor;
	
	    long start = System.currentTimeMillis();
	    long key = 1999;
	    int i, n;

	    RecordInt rec = new RecordInt();
	    ConsusStatement stmt = (ConsusStatement)connection.createStatement();
	    boolean delayedTransaction = System.getProperty("consus.commit.delay") != null;

	    rec.id = id;
	    for (i = 0; i < TestSync.nRecords; i++) { 
		key = (3141592621L*key + 2718281829L) % 1000000007L;
		rec.rkey = rec.value = key;
		stmt.insert(rec);
		if (cooperative) { 
		    connection.commitSubtransaction();
		} else if (delayedTransaction) { 
		    connection.commit();
		}
	    }
	    stmt.execute("create index on RecordInt(rkey)");
	    System.out.println("Thread " + id + ": Elapsed time for inserting " + TestSync.nRecords + " records: " 
			       + (System.currentTimeMillis() - start) + " milliseconds");
	    start = System.currentTimeMillis();
	    connection.commit();
	    System.out.println("Thread " + id + ": Commit time: " 
			       + (System.currentTimeMillis() - start) + " milliseconds");
	
	    PreparedStatement pstmt = connection.prepareStatement("select * from RecordInt where rkey = ? and id = ?");
	    start = System.currentTimeMillis();
	    key = 1999;
	    pstmt.setShort(2, (short)id);
	    connection.setThreadAutoCommit(true);
	    for (i = 0; i < TestSync.nRecords; i++) { 
		key = (3141592621L*key + 2718281829L) % 1000000007L;
		pstmt.setLong(1, key);
		cursor = pstmt.executeQuery();
		cursor.last();
		if (cursor.getRow() != 1) { 
		    System.out.println("cursor.getRow() = " + cursor.getRow());
		}
		Assert.that(cursor.getRow() == 1);
	    }
	    System.out.println("Thread " + id + ": Elapsed time for " 
			       + TestSync.nRecords + " index searches: " 
			       + (System.currentTimeMillis() - start) + " milliseconds"); 
	    pstmt.close();
	    pstmt = connection.prepareStatement("select * from RecordInt where id = ?");
	    pstmt.setShort(1, (short)id);	
	    start = System.currentTimeMillis();
	    for (i = 0; i < TestSync.nSequentialSearches; i++) { 
		cursor = pstmt.executeQuery();
		cursor.last();
		Assert.that(cursor.getRow() == TestSync.nRecords);
	    }
	    System.out.println("Thread " + id + ": Elapsed time for " + TestSync.nSequentialSearches
			       + " sequential search through " + TestSync.nRecords + " records: " 
			       + (System.currentTimeMillis() - start) + " milliseconds"); 
	    pstmt.close();
	    start = System.currentTimeMillis();
	    cursor = stmt.executeQuery("select * from RecordInt where id = " + id + 
				       " order by value for update");
	    cursor.last();
	    Assert.that(cursor.getRow() == TestSync.nRecords);
	    System.out.println("Thread " + id + ": Elapsed time for sorting " 
			       + TestSync.nRecords + " records: " 
			       + (System.currentTimeMillis() - start) + " milliseconds"); 
	    start = System.currentTimeMillis();
	    i = 0;
	    do { 
		cursor.deleteRow();
		i += 1;
	    } while (cursor.previous()); 
	    cursor.close();
	    Assert.that(i == TestSync.nRecords);
	    System.out.println("Thread " + id + ": Elapsed time for deleting " 
			       + TestSync.nRecords + " records: " 
			       + (System.currentTimeMillis() - start) + " milliseconds"); 
	} catch(SQLException x) { 
	    System.out.println("SQLException: " + x.getMessage());
	    x.printStackTrace();
	}
    }

}
    

public class TestSync { 
    final static int nRecords = 100000;
    final static int nSequentialSearches = 10;

    static public void main(String args[]) throws Exception { 
	int i;
	boolean cooperative = false;
	if (args.length < 1) { 
	    System.err.println("Usage: java TestSync <number-of-threads> [cooperative]");
	    return;
	}
	if (args.length == 2) { 
	    cooperative = true;
	}
	int nThreads = Integer.parseInt(args[0], 10);
	Class.forName("org.garret.consus.ConsusDriver");
	String fileName = System.getProperty("file", "testsync.dbs");
	ConsusConnection connection = 
	    (ConsusConnection)DriverManager.getConnection("jdbc:consus:" + fileName);
	if (connection == null) { 
	    System.err.println("No connection with database TestSync");
	    return;
	}
	ConsusThread threads[] = new ConsusThread[nThreads];
	for (i = 0; i < nThreads; i++) { 
	    threads[i] = new ConsusThread(connection, i, cooperative);
	}
	for (i = 0; i < nThreads; i++) { 
	    threads[i].join();
	}
	connection.close();
    }
}













