/* Consus - Pure Java Object Relational Database Management System
 * Hash table with weak reference toi the object. Weak reference is replaces with null
 *    by GC when there are no hard references to the object.
 * Created by Konstantin Knizhnik, May 2000.
 */
package consus;

class SunWeakHashTable extends WeakHashTable { 
    sun.misc.Cache cache;

    SunWeakHashTable(int initialCapacity) { 
	super(0);
	cache = new sun.misc.Cache();
    }

    boolean remove(int oid) {
	if (cache.remove(new ConsusRef(oid)) != null) { 
	    count -= 1;
	    return true;
	}
	return false;
    }

    void    put(int oid, Object obj) { 
	if (cache.put(new ConsusRef(oid), obj) == obj) { 
	    count += 1;
	}
    }

    Object  get(int oid) {
	return cache.get(new ConsusRef(oid));
    }

    synchronized void clear() { 
	java.util.Enumeration enum = cache.elements();
	while (enum.hasMoreElements()) { 
	    Object obj = enum.nextElement();
	    if (obj instanceof Persistent) { 
		((Persistent)obj).unget();
	    }
	}
	cache = null;
	count = 0;
    }

    synchronized void mark(int bitmap[]) {
	java.util.Enumeration enum = cache.elements();
	while (enum.hasMoreElements()) { 
	    Object obj = enum.nextElement();
	    if (obj instanceof Persistent) { 
		int oid = ((Persistent)obj).oid;
		bitmap[oid >>> 5] |= 1 << (oid & 31);
	    }
	}
	cache = null;
	count = 0;
    }

    int size() { 
	return count;
    }
}


